unit HrastUnit;

// version: 13
// Copyright  2014 by HrastProgrammer. All rights reserved.
// This unit contains WREN synthesizer modules developed by HrastProgrammer.
// Can be distributed only as part of the complete WREN source package.
//
// version  8: Blue Hell : changed module registration procedure
// version  9: Blue Hell : layout changes
// version 10: Blue Hell : some code changes
// version 11: Blue Hell : added constructors for the Fast/Slow selection
// version 12: Blue Hell : removal of outcommented code and object renaming
// version 13: Blue Hell : added un-attenuated FM to HrastSVF

interface

uses

  KnobsUtils, KnobsConversions, Module_Defs, Module_Implementations;

type

  THrastOsc = class( TMod)
  strict private
  const
    i_freq      =  0;
    i_morph     =  1;
    i_shape     =  2;
    i_bandlimit =  3;
    i_pw        =  4;
    i_pwm       =  5;
    i_pwminput  =  6;
    i_fm        =  7;
    i_fminput   =  8;
    i_pm        =  9;
    i_pminput   = 10;
    i_vintage   = 11;
    i_syncinput = 12;
    i_level     = 13;
    i_mute      = 14;
    i_cents     = 15;

    o_output    =  0;
  private
    FZ        : TSignal;
    FPhase    : TSignal;
    FPrevSync : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                       override;
    procedure   SetDefaults;                                                              override;
    procedure   DoTick;                                                                   override;
    procedure   GatherLights( const Prefix: string; const Callback: TLightsHandler);      override;
    function    InputName ( Index: Integer): string;                                      override;
    function    OutputName( Index: Integer): string;                                      override;
  end;

  THrastSVF = class( TMod)
  strict private
  const
    i_input       =  0;
    i_cutoff      =  1;
    i_reso        =  2;
    i_cutoffinput =  3;
    i_cutoffmod   =  4;
    i_resoinput   =  5;
    i_resomod     =  6;
    i_gain        =  7;
    i_drive       =  8;
    i_lowpass     =  9;
    i_bandpass    = 10;
    i_highpass    = 11;
    i_level12     = 12;
    i_level24     = 13;
    i_freq        = 14;

    o_outputlp    =  0;
    o_outputbp    =  1;
    o_outputhp    =  2;
    o_outputbr    =  3;
    o_output12    =  4;
    o_output24    =  5;
  private
    FBand12 : TSignal;
    FLow12  : TSignal;
    FBand24 : TSignal;
    FLow24  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                       override;
    procedure   SetDefaults;                                                              override;
    procedure   DoTick;                                                                   override;
    function    InputName ( Index: Integer): string;                                      override;
    function    OutputName( Index: Integer): string;                                      override;
  end;


  procedure CreateTables; // BLue Hell : renamed inintialize to CreateTables and made it public; table re-init is needed
                          //             after changing the system rate



implementation



uses

  Globals, Math, SysUtils;

const

  TableSize = 8192;

var

  Pow2Table : array [ 0 .. TableSize + 1] of TSignal;
  SineTable : array [ 0 .. TableSize + 1] of TSignal;


  function Pow2Freq( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;
    if f > TableSize - 1
    then f := TableSize - 1;
    n := Trunc( f);
    f := f - n;
    Result := Pow2Table[ n];
    Result := Result + f * ( Pow2Table[ n + 1] - Result);
  end;

  function Sine( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;
    if f > TableSize - 1
    then f := TableSize - 1;
    n := Trunc( f);
    f := f - n;
    Result := SineTable[ n];
    Result := Result + f * ( SineTable[ n + 1] - Result);
  end;


{ ========
  THrastOsc = class( TMod)
  private
    FZ        : TSignal;
    FPhase    : TSignal;
    FPrevSync : Boolean;
  public
}

    constructor THrastOsc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   THrastOsc.SetDefaults;
    begin
      FInputs[ i_fminput ] := 0;
      FInputs[ i_pminput ] := 0;
      FInputs[ i_pwminput] := 0;
      FDezipperMap := [
        i_freq,
        i_morph,
        i_shape,
        i_bandlimit,
        i_pw,
        i_pwm,
        i_fm,
        i_pm,
        i_level,
        i_mute,
        i_cents
      ];
    end;

    procedure THrastOsc.DoTick;
    var
      DeltaPhase: TSignal;
    var
      n   : Integer;
      t   : TSignal;
      y   : TSignal;
      ypw : TSignal;
    begin
      if not FPrevSync and SignalToLogic( FInputs[i_syncinput])
      then FPhase := 0;

      FPrevSync  := SignalToLogic( FInputs[ i_syncinput]);
      DeltaPhase := Pow2Freq( FInputs[ i_freq] + FInputs[ i_cents] + FInputs[ i_fm] * FInputs[ i_fminput]) * System_Rate_Rec;

      t := FPhase + FInputs[ i_pm] * FInputs[ i_pminput];
      n := Trunc( t);
      t := t - n;
      if t < 0
      then t := t + 1;

      FPhase := FPhase + DeltaPhase;

      n      := Trunc( FPhase);
      FPhase := FPhase - n;
      if FPhase < 0
      then FPhase := FPhase + 1;

      y := 2 * t - 1 - PolyBlep( t, FInputs[ i_bandlimit], DeltaPhase);

      t := t + FInputs[ i_pw] + FInputs[ i_pwm] * FInputs[ i_pwminput];

      n := Trunc( t);
      t := t - n;
      if t < 0
      then t := t + 1;

      ypw := 2 * t - 1 - PolyBlep( t, FInputs[ i_bandlimit], DeltaPhase);

      t := FInputs[ i_shape];

      if FInputs[ i_vintage] <> 0
      then begin                            // Vintage
        if t > 0.999
        then t := 0.999;
        y := ( y + 1) * 0.5;
        y := y * y;
        y := 2 * y - 1;

        ypw := ( ypw + 1) * 0.5;
        ypw := ypw * ypw;
        ypw := 2 * ypw - 1;
      end;
      FZ := y - ypw * FInputs[ i_morph] + t * FZ;

      if t > 0.99
      then t := 0.99;

      y := FZ * ( t - 1);
      FOutputs[ o_output] := y * FInputs[ i_level] * SignalToMute( FInputs[ i_mute]);
    end;

    procedure THrastOsc.GatherLights;
    begin
      Callback( Self, Format( '%s/%s/SyncInput/%f', [ Prefix, Name, FInputs[ i_syncinput]], AppLocale));
    end;

    function THrastOsc.InputName;
    begin
      case Index of
        i_freq      : Result := 'freq';
        i_morph     : Result := 'morph';
        i_shape     : Result := 'shape';
        i_bandlimit : Result := 'bandlimit';
        i_pw        : Result := 'pw';
        i_pwm       : Result := 'pwm';
        i_pwminput  : Result := 'pwminput';
        i_fm        : Result := 'fm';
        i_fminput   : Result := 'fminput';
        i_pm        : Result := 'pm';
        i_pminput   : Result := 'pminput';
        i_vintage   : Result := 'vintage';
        i_syncinput : Result := 'syncinput';
        i_level     : Result := 'level';
        i_mute      : Result := 'mute';
        i_cents     : Result := 'cents';
        else          Result := NO_INPUT;
      end
    end;

    function THrastOsc.OutputName;
    begin
      case Index of
        o_output : Result := 'output';
        else       Result := NO_OUTPUT;
      end
    end;


{ ========
  THrastSVF = class( TMod)
  private
    FBand12 : TSignal;
    FLow12  : TSignal;
    FBand24 : TSignal;
    FLow24  : TSignal;
  public
}

    constructor THrastSVF.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   THrastSVF.SetDefaults;
    begin
      FInputs[ i_input      ] := 0;
      FInputs[ i_cutoffinput] := 0;
      FInputs[ i_resoinput  ] := 0;
      FInputs[ i_freq       ] := 0;
      FDezipperMap:=[
        i_cutoff,
        i_reso,
        i_cutoffmod,
        i_resomod,
        i_gain,
        i_drive,
        i_lowpass,
        i_bandpass,
        i_highpass,
        i_level12,
        i_level24
      ]
    end;

    procedure THrastSVF.DoTick;
    var
      Oversample : Integer;
      f          : TSignal;
      fb         : TSignal;
      y          : TSignal;
      High12     : TSignal;
      High24     : TSignal;
      Output12   : TSignal;
      Output24   : TSignal;
    begin
      f  := Pow2Freq( FInputs[ i_cutoff] + FInputs[ i_freq] + FInputs[ i_cutoffmod] * FInputs[ i_cutoffinput]) / System_Rate_2;
      f  := 2 * Sine( f);
      fb := Clip( 1 - ( FInputs[ i_reso] + FInputs[ i_resomod] * FInputs[ i_resoinput]), 0, 1);
      y  := FInputs[ i_input] * FInputs[ i_gain];
      y  := y / ( Abs( y) + ( 1 - Clip( FInputs[ i_drive], 0, 0.999)));

      for Oversample := 1 to 2
      do begin
        High12   := Normalize( y - fb * FBand12 - FLow12);
        FBand12  := Normalize( f * High12  + FBand12);
        FLow12   := Normalize( f * FBand12 + FLow12);
        FLow12   := Normalize( FLow12 * 1.0001 / ( 0.0001 * Abs( FLow12) + 1));
        Output12 := Normalize( FLow12 * FInputs[ i_lowpass] + FBand12 * FInputs[ i_bandpass] + High12 * FInputs[ i_highpass]);

        High24   := Output12 - fb * FBand24 - FLow24;
        FBand24  := f * High24  + FBand24;
        FLow24   := f * FBand24 + FLow24;
        FLow24   := FLow24 * 1.0001 / ( 0.01 * Abs( FLow24) + 1);
      end;

      Output24 := FLow24 * FInputs[ i_lowpass] + FBand24 * FInputs[ i_bandpass] + High24 * FInputs[ i_highpass];
      FOutputs[ o_outputlp] := FLow12;
      FOutputs[ o_outputbp] := FBand12;
      FOutputs[ o_outputhp] := High12;
      FOutputs[ o_outputbr] := FLow12 + High12;
      FOutputs[ o_output12] := Output12 * FInputs[ i_level12];
      FOutputs[ o_output24] := Output24 * FInputs[ i_level24];
    end;

    function THrastSVF.InputName;
    begin
      case Index of
        i_input       : Result := 'input';
        i_cutoff      : Result := 'cutoff';
        i_reso        : Result := 'reso';
        i_cutoffinput : Result := 'cutoffinput';
        i_cutoffmod   : Result := 'cutoffmod';
        i_resoinput   : Result := 'resoinput';
        i_resomod     : Result := 'resomod';
        i_gain        : Result := 'gain';
        i_drive       : Result := 'drive';
        i_lowpass     : Result := 'lowpass';
        i_bandpass    : Result := 'bandpass';
        i_highpass    : Result := 'highpass';
        i_level12     : Result := 'level12';
        i_level24     : Result := 'level24';
        i_freq        : Result := 'freq';
        else            Result := NO_INPUT;
      end
    end;

    function THrastSVF.OutputName;
    begin
      case Index of
        o_outputlp : Result := 'outputlp';
        o_outputbp : Result := 'outputbp';
        o_outputhp : Result := 'outputhp';
        o_outputbr : Result := 'outputbr';
        o_output12 : Result := 'output12';
        o_output24 : Result := 'output24';
        else         Result := NO_OUTPUT;
      end
    end;



  procedure CreateTables;
  var
    i: Integer;
    x: TSignal;
  begin
    for i := 0 to TableSize
    do Pow2Table[ i] := UnitsToFrequency( i / TableSize);
    Pow2Table[ TableSize + 1] := Pow2Table[ TableSize];

    x := Pi / TableSize;
    for i := 0 to TableSize - 1
    do SineTable[ i] := Sin( x * i);
    SineTable[ TableSize div 2] := 0;
    SineTable[ TableSize      ] := SineTable[ 0];
    SineTable[ TableSize + 1  ] := SineTable[ 1];
  end;

initialization

  CreateTables;

end.
