unit KnobsUtils;

{

  (C) COPYRIGHT 1999 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Types, System.Classes, System.DateUtils, WinApi.Windows, System.Math,
  Vcl.Controls, Vcl.Forms, Vcl.Graphics, System.TypInfo;


type

  TSignal          = Double;
  TBinary          = Int32;
  TBitCount        = 1 .. 32;
  TSignalArray     = array of TSignal;
  TOscTime         = UInt64;
  TKnobsModuleType = Integer;
  TByteSet         = set of Byte;


  TDistanceMode = (
    dmCircular,
    dmHorizontal,
    dmVertical
  );

  TSignalType = (
    stUnknown,        // treat as stAudio
    stControl,        // control rate
    stAudio,          // audio rate
    stLogic,          // audio rate logic levels
    stContrLogic      // control rate logic signals
  );


var

  AppLocale : TFormatSettings;


  procedure Nop;
  procedure KilledException( const anException: Exception);
  function  GetCpuClockCycleCount: Int64;                                                           assembler; register;
  function  Clip( const aValue: Integer; const aLow, aHigh: Integer): Integer;                         inline; overload;
  function  Clip( const aValue: TSignal; const aLow, aHigh: TSignal): TSignal;                         inline; overload;
  procedure PClip( var aValue: TSignal; const aLow, aHigh: TSignal);                                             inline;
  function  FloatMod( const X, Y: TSignal): TSignal;                                                assembler; register;
  function  MathFloatMod( const X, Y: TSignal): TSignal;                                                         inline;
  function  MathIntMod( X, Y: Integer): Integer;                                                                 inline;
  function  MathIntDiv( X, Y: Integer): Integer;                                                                 inline;
  function  RandomGaus( aSigma : TSignal; AbsVal: Boolean = False): TSignal;
  function  RandomExpo( aLambda: TSignal; AbsVal: Boolean = False): TSignal;

  function  ApplicationPath: String;
  procedure Exchange( var A, B: Integer);
  function  RectanglesOverlap( A, B: TRect): Boolean;
  function  Distance( aPoint: TPoint): Integer;
  function  MultPoint( A: TPoint; M: Integer): TPoint;
  function  DivPoint( A: TPoint; D: Integer): TPoint;
  function  ValuesToAngle( MinValue, Value, MaxValue: TSignal): TSignal;
  function  PolarToPoint( aCenter: TPoint; anAngle, aLength: TSignal): TPoint;
  function  DistanceToValue( MinValue, MaxValue: TSignal; aPoint: TPoint; aDistanceMode: TDistanceMode): TSignal;
  function  Rotate( A: TPoint; anAngle: TSignal): TPoint;

  function  Indent( anInd: Integer): string;                                                                   overload;
  function  Indent( anInd: Integer; const S: string): string;                                                  overload;
  function  Indent( anInd: Integer; const aFmt: string; const anArgs: array of const): string;                 overload;

  procedure WriteInd  ( var aFile: TextFile; anInd: Integer; S: string);
  procedure WriteLnInd( var aFile: TextFile; anInd: Integer; S: string);

  function  TrimStr  ( const S: string; aLength: Integer): string;
  function  ExtendStr( const S: string; aLength: Integer): string;

  function  MakePath( const aParts: array of string): string;
  function  Explode( const aString: string; const aChar: Char): TStringList;                                   overload;
  function  Explode( aString: string; aSeparator: string): TStringList;                                        overload;
  function  Implode( const aStrings: TStringList    ; const aSeparator: string): string;                       overload;
  function  Implode( const aStrings: array of string; const aSeparator: string): string;                       overload;
  function  Implode( const aStrings: TStringList    ): string;                                                 overload;
  function  Implode( const aStrings: array of string): string;                                                 overload;
  function  StripQuotes( const aValue, aLeftQuote, aRightQuote: string): string;
  function  TabsToSpaces( const aValue: string): string;

  function  OscTimeToDateTime( aTime: TOscTime): TDateTime;
  function  DateTimeToOscTime( aTime: TDateTime): TOscTime;
  function  OscNow: TOsctime;
  function  GetTimeZoneBias: Integer;
  function  IsDaylightSavingOn: Boolean;
  function  UTC: TDateTime;

  function  InterpolateColors( aColor, bColor: TColor; aMix: Byte): TColor;
  function  IsPrime( aValue: Cardinal): Boolean;
  function  NextPrime( aValue: Cardinal): Cardinal;

  function  SignalTypeToStr( const aValue: TSignalType): string;
  function  SignalIsFast( const aValue: TSignalType): Boolean;
  function  ByteSetToStr( const aValue: TByteSet): string;


var

  ExceptionCounter : Int64;


implementation


var

  strIndent : string;


  procedure Nop;
  begin
    // a Nop ... allows for setting breakpoints in places
  end;

  procedure KilledException( const anException: Exception);
  begin
    {$Q-R-}
    Inc( ExceptionCounter);
    {$Q+R+}
  end;

  function GetCpuClockCycleCount: Int64;
  asm
    rdtsc
  end;

  function KnobClip( const aValue: TSignal; const aLow, aHigh: TSignal): TSignal;
  begin
    Result := aValue;

    if aLow < aHigh
    then begin
      if aValue < aLow
      then Result := aLow
      else If aValue > aHigh
      then Result := aHigh;
    end
    else begin
      if aValue < aHigh
      then Result := aHigh
      else If aValue > aLow
      then Result := aLow;
    end;
  end;

  function  Clip( const aValue: Integer; const aLow, aHigh: Integer): Integer; inline; overload;
  begin
    if aValue < aLow
    then Result := aLow
    else if aValue > aHigh
    then Result := aHigh
    else Result := aValue;
  end;

  function Clip( const aValue: TSignal; const aLow, aHigh: TSignal): TSignal; inline; overload;
  begin
    if aValue < aLow
    then Result := aLow
    else if aValue > aHigh
    then Result := aHigh
    else Result := aValue;
  end;

  procedure PClip( var aValue: TSignal; const aLow, aHigh: TSignal); inline;
  begin
    if aValue < aLow
    then aValue := aLow
    else if aValue > aHigh
    then aValue := aHigh;
  end;


  function  FLoatMod( const X, Y: TSignal): TSignal; assembler; register;
  asm
    fld qword ptr[ Y]
    fld qword ptr[ X]
@r:
    fprem
    fstsw ax
    sahf
    jp    @r
    fstp  st( 1)
  end;


//  function  FloatMod( X, Y: Double): Double; inline;
//  begin
//    Result := Frac( X / Y) * Y;
//  end;

  function  MathFloatMod( const X, Y: TSignal): TSignal; inline;
  begin
    Result := FloatMod( FloatMod( X, Y) + Y, Y);
  end;

  function  MathIntMod( X, Y: Integer): Integer; inline;
  begin
    Result := ( X mod Y + Y) mod Y;
  end;

  function  MathIntDiv( X, Y: Integer): Integer; inline;
  begin
    Result := Floor( X / Y);
  end;

  function  RandomGaus( aSigma: TSignal; AbsVal: Boolean = False): TSignal;
  begin
    if AbsVal
    then Result := Abs( RandG( 0, aSigma))
    else Result := RandG( 0, aSigma);
  end;

  function  RandomExpo( aLambda: TSignal; AbsVal: Boolean = False): TSignal;
  begin
    Result := - Ln( 1.0 - Random) / aLambda;
    if not AbsVal
    then begin
      if Random( 2) = 0
      then Result := - Result;
    end;
  end;

  function  ApplicationPath: string;
  begin
    Result := ExtractFilePath( Application.ExeName);
  end;

  procedure Exchange( var A, B: Integer);
  var
    T : Integer;
  begin
    T := A;
    A := B;
    B := T;
  end;

  function  RectanglesOverlap( A, B: TRect): Boolean;
  var
    R : TRect;
  begin
    Result := IntersectRect( R, A, B);
  end;

  function  Distance( aPoint: TPoint): Integer;
  begin
    with aPoint
    do Result := Round( Sqrt( 1.0 * x * x + 1.0 * y * y));
  end;

  function   MultPoint( A: TPoint; M: Integer): TPoint;
  begin
    Result := Point( A.x * M, A.y * M);
  end;

  function   DivPoint( A: TPoint; D: Integer): TPoint;
  begin
    Result := Point( A.x div D, A.y div D);
  end;

  function  ValuesToAngle( MinValue, Value, MaxValue: TSignal): TSignal;
  begin
    if MinValue <> MaxValue
    then Result := 270 * ( Value - MinValue) / ( MinValue - MaxValue) + 225
    else Result := 0; // arbitrary ...
  end;

  function  PolarToPoint( aCenter: TPoint; anAngle, aLength: TSignal): TPoint;
  begin
    with Result
    do begin
      x := Round( aCenter.x + aLength * Cos( DegToRad( anAngle)));
      y := Round( aCenter.y - aLength * Sin( DegToRad( anAngle)));
    end;
  end;

  function  DistanceToValue( MinValue, MaxValue: TSignal; aPoint: TPoint; aDistanceMode: TDistanceMode): TSignal;
  var
    anAngle : Extended;
    anAlpha : Extended;
  begin
    case aDistanceMode of

      dmCircular :
        begin
          anAngle := RadToDeg( ArcTan2( - aPoint.y, aPoint.x));

          while anAngle < 0
          do anAngle := anAngle + 360;

          while anAngle > 270
          do anAngle := anAngle - 360;

          Result := KnobClip( MinValue + (( MinValue - MaxValue) * ( anAngle - 225) / 270), MinValue, MaxValue);
        end;

      dmHorizontal :
        begin
          anAlpha := 0.5 + aPoint.X / 200;
          Result  := KnobClip( MinValue + anAlpha * ( MaxValue - MinValue), MinValue, MaxValue);
        end;

      dmVertical :
        begin
          anAlpha := 0.5 - aPoint.Y / 200;
          Result  := KnobClip( MinValue + anAlpha * ( MaxValue - MinValue), MinValue, MaxValue);
        end;

      else begin
        Result := 0;
        Assert( False, 'unhandled knob control mode');
      end;
    end;
  end;

  function   Rotate( A: TPoint; anAngle: TSignal): TPoint;
  begin
    anAngle := DegToRad( anAngle);

    with Result
    do begin
      x := Round( Cos( anAngle) * A.x - Sin( anAngle) * A.y);
      y := Round( Sin( anAngle) * A.x + Cos( anAngle) * A.y);
    end;
  end;

  function  Indent( anInd: Integer): string; // overload;
  begin
    Result := Copy( strIndent, 1, 2 * anInd);
  end;

  function  Indent( anInd: Integer; const S: string): string; // overload;
  begin
    Result := Format( '%s%s', [ Indent( anInd), S]);
  end;

  function  Indent( anInd: Integer; const aFmt: string; const anArgs: array of const): string; // overload;
  begin
    Result := Indent( anInd, Format( aFmt, anArgs));
  end;

  procedure WriteInd( var aFile: TextFile; anInd: Integer; S: string);
  begin
    Write( aFile, Indent( anInd), S);
  end;

  procedure WriteLnInd( var aFile: TextFile; anInd: Integer; S: string);
  begin
    WriteLn( aFile, Indent( anInd), S);
  end;

  function  BuildStr( const aParts: array of string; const aSeparator: string): string;
  var
    i : Integer;
  begin
    Result := '';
    for i := 0 to Length( aParts) - 1
    do begin
      if i = Length( aParts) - 1
      then Result := Result + aParts[ i]
      else Result := Result + aParts[ i] + aSeparator;
    end;
  end;

  function  TrimStr( const S: string; aLength: Integer): string;
  begin
    Result := Copy( ExtendStr( S, aLength), 1, aLength); // Procustrus trimming.
  end;

  function  ExtendStr( const S: string; aLength: Integer): string;
  begin
    Result := S;
    while Length( Result) < aLength
    do Result := Result + ' ';
  end;

  function  MakePath( const aParts: array of string): string;
  begin
    Result := BuildStr( aParts, '/');
  end;

  function  Explode( const aString: string; const aChar: Char): TStringList; // overload;
  var
    p : Integer;
    i : integer;
    S : string;
  begin
    Result := TStringList.Create;
    p := 1;
    i := 1;
    while i <= Length( aString)
    do begin
      if aString[ i] = aChar
      then begin
        S := Copy( aString, p, i - p);
        Result.Add( S);
        if CharInSet( aChar, [ ' ', ^I])
        then begin
          while ( i <= Length( aString)) and ( aString[ i] = aChar)
          do Inc( i);
        end
        else Inc( i);
        p := i;
      end
      else Inc( i);
    end;
    if p <= i
    then begin
      S := Copy( aString, p, Length( aString));
      Result.Add( S);
    end;
  end;

  function Explode( aString: string; aSeparator: string): TStringList; // overload;
  var
    c : word;
    p : Integer;
  begin
    Result := TStringList.Create;
    c := 0;
    while aString <> ''
    do begin
      p := Pos( aSeparator, aString);
      if p > 0
      then begin
        Result.Add( Copy( aString, 1, p - 1));
        Delete( aString, 1, Length( Result[ c]) + Length( aSeparator));
      end
      else begin
        Result.Add( aString);
        aString := '';
      end;
      inc( c);
    end;
  end;

  function  Implode( const aStrings: TStringList; const aSeparator: string): string; // overload;
  var
    i : Integer;
  begin
    Result := '';
    if Assigned( aStrings)
    then begin
      for i := 0 to aStrings.Count - 1
      do begin
        if i = aStrings.Count - 1
        then Result := Result + aStrings[ i]
        else Result := Result + aStrings[ i] + aSeparator;
      end;
    end;
  end;

  function  Implode( const aStrings: array of string; const aSeparator: string): string; // overload;
  var
    i : Integer;
  begin
    Result := '';
    for i := Low( aStrings) to High( aStrings)
    do begin
      if i = High( aStrings)
      then Result := Result + aStrings[ i]
      else Result := Result + aStrings[ i] + aSeparator;
    end;
  end;

  function  Implode( const aStrings: TStringList): string; // overload;
  begin
    Result := Implode( aStrings, ' ');
  end;

  function  Implode( const aStrings: array of string): string; // overload;
  begin
    Result := Implode( aStrings, ' ');
  end;

  function  StripQuotes( const aValue, aLeftQuote, aRightQuote: string): string;
  begin
    if ( Pos( aLeftQuote, aValue) = 1) and ( Pos( aRightQuote, aValue, Length( aValue) - Length( aRightQuote) + 1) = Length( aValue) - Length( aRightQuote) + 1)
    then Result := Copy( aValue, Length( aLeftQuote) + 1, Length( aValue) - Length( aLeftQuote) - Length( aRightQuote))
    else Result := aValue;
  end;

  function  TabsToSpaces( const aValue: string): string;
  var
    C : Char;
  begin
    Result := '';
    for C in aValue
    do begin
      if C = ^H
      then Result := Result + ' '
      else Result := Result + C;
    end;
  end;



const

  SecsPerDay = 86400.0;          // seconds per day
  C1shl32    = 4294967296.0;     // 1 << 32 as a float
  DayShift   = 2.0;              // to move from dec 30 1899 to jan 1 1900

  // OSC time is a 64 bit integer, where the leftmost 32 bits specify seconds
  // and the right most 32 bits specify fractional seconds, the time is relative
  // to 1900-01-01 00:00:00 UTC - this is the same as an NTP timestamp.
  // TDateTime is a floating point time, the integer part holds days, the
  // fractional bit holds the fraction of a day. This time is relative
  // to December 30, 1899; 12:00 A.M, in the local time zone ...

  procedure DateTimeToNTP( aTime: TDateTime; var aSecs, aFracs: Cardinal);
  var
    Value1 : Double;
    Value2 : Double;
  begin
    Value1 := ( aTime - DayShift) * SecsPerDay;
    Value2 := Value1;

    if Value2 > C1shl32
    then Value2 := Value2 - C1shl32;

    aSecs := Cardinal( Trunc( Value2));
    Value2 := (( Frac( Value1) * 1000) / 1000) * C1shl32;

    if Value2 > C1shl32
    then Value2 := Value2 - C1shl32;

    aFracs := Cardinal( Trunc( Value2));
  end;

  function NTPToDateTime( aSecs, aFrac: Cardinal): TDateTime;
  var
    Value1 : Double;
    Value2 : Double;
  begin
    Value1 := aSecs;

    if Value1 < 0
    then Value1 := C1shl32 + Value1 - 1;

    Value2 := aFrac;

    if Value2 < 0
    then Value2 := C1shl32 + Value2 - 1;

    Value2 := Trunc(( Value2 / C1shl32) * 1000) / 1000;
    Result := (( Value1 + Value2) / SecsPerDay) + DayShift;
  end ;

  function  OscTimeToDateTime( aTime: TOscTime): TDateTime;
  var
    aSecs : Cardinal;
    aFrac : Cardinal;
  begin
    aSecs  := ( aTime shr 32) and $ffffffff;
    aFrac  := ( aTime shr  0) and $ffffffff;
    Result := NTPToDateTime( aSecs, aFrac);
  end;

  function  DateTimeToOscTime( aTime: TDateTime): TOscTime;
  var
    aSecs : Cardinal;
    aFrac : Cardinal;
  begin
    DateTimeToNTP( aTime, aSecs, aFrac);
    Result := ( UInt64( aSecs) shl 32) + UInt64( aFrac);
  end;

  function  OscNow: TOsctime;
  begin
    Result := DateTimeToOscTime( UTC);
  end;

  function  GetTimeZoneBias: Integer;
  var
    tz: TTimeZoneInformation;
  begin
    case GetTimeZoneInformation( tz) of
      TIME_ZONE_ID_STANDARD: Result := - ( tz.StandardBias + tz.Bias) div 60;
      TIME_ZONE_ID_DAYLIGHT: Result := - ( tz.DaylightBias + tz.Bias) div 60;
    else
      Result := 0;
    end;
  end;

  function  IsDaylightSavingOn: Boolean;
  var
    tz: TTimeZoneInformation;
  begin
    Result := GetTimeZoneInformation( tz) = TIME_ZONE_ID_DAYLIGHT;
  end;

  function  UTC: TDateTime;
  begin
    Result := Now - GetTimeZoneBias * OneHour;
  end;

  function  InterpolateColors( aColor, bColor: TColor; aMix: Byte): TColor;
  begin
    Result := RGB(
      (( Word( 255 - aMix) * Word( GetRValue( aColor)) + Word( aMix) * Word( GetRValue( bColor))) div 256) and $ff,
      (( Word( 255 - aMix) * Word( GetGValue( aColor)) + Word( aMix) * Word( GetGValue( bColor))) div 256) and $ff,
      (( Word( 255 - aMix) * Word( GetBValue( aColor)) + Word( aMix) * Word( GetBValue( bColor))) div 256) and $ff
    );
  end;

  function  IsPrime( aValue: Cardinal): Boolean;
  var
    X       : Cardinal;
    Largest : Cardinal;
  begin
    if aValue < 4                                  // 1, 2 and 3 are primes
    then Result := True
    else begin
      if ( aValue mod 2) = 0                       // skip all even values to make it a tad faster
      then Result := False
      else begin
        Largest := Trunc( Sqrt( 1.0 * aValue));    // Only check up to Sqrt( aValue)
        Result  := True;                           // Assume prime
        X       := 3;                              // First divisor to try is 3
        while ( X <= Largest) and Result
        do begin
          Inc( X, 2);                              // Go to next odd value
          Result := ( aValue mod X) <> 0;          // Get false when a divisor found, so not prime then.
        end;
      end;
    end;
  end;

  function  NextPrime( aValue: Cardinal): Cardinal;
  begin
    Result := aValue;
    if not IsPrime( Result)                        // when not a prime yet
    then begin
      Result := Result or 1;                       // make sure its not a multiple of two ..
      while not IsPrime( Result)                   // then repeatedly add two till its a prime
      do Inc( Result, 2);
    end;
  end;

  function  SignalTypeToStr( const aValue: TSignalType): string;
  begin
    Result := GetEnumName( TypeInfo( TSignalType), Ord( aValue));
  end;

  function  SignalIsFast( const aValue: TSignalType): Boolean;
  begin
    Result := aValue in [ stAudio, stLogic];
  end;

  function  ByteSetToStr( const aValue: TByteSet): string;
  var
    Curr    : Byte;
    Prev    : Integer;
    InRange : Boolean;
  begin
    Result  := '';
    InRange := False;
    Prev    := -2;                    // To make '[ 0' not result in '[ .. 0'

    for Curr in aValue
    do begin
      if Curr = Prev + 1
      then begin
        if not InRange
        then begin
          InRange := True;
          Result  := Result + ' .. ';
        end;
      end
      else begin
        if InRange
        then begin
          InRange := False;
          Result  := Result + IntToStr( Prev) + ', ' + IntToStr( Curr);
        end
        else begin
          if Result = ''
          then Result := Result +        IntToStr( Curr)
          else Result := Result + ', ' + IntToStr( Curr);
        end;
      end;
      Prev := Curr;
    end;

    if InRange
    then Result := Result + IntToStr( Prev);

    if Result = ''
    then Result := '[]'
    else Result := '[ ' + Result + ']';
  end;


  // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure InitializeUnit;
  var
    i : Integer;
  begin
    strIndent := ' ';

    while Length( strIndent) < 256
    do strIndent := strIndent + strIndent;

    AppLocale.CurrencyFormat    := FormatSettings.CurrencyFormat;
    AppLocale.NegCurrFormat     := FormatSettings.NegCurrFormat;
    AppLocale.ThousandSeparator := ',';
    AppLocale.DecimalSeparator  := '.';
    AppLocale.CurrencyDecimals  := FormatSettings.CurrencyDecimals;
    AppLocale.DateSeparator     := '-';
    AppLocale.TimeSeparator     := ':';
    AppLocale.ListSeparator     := ',';
    AppLocale.CurrencyString    := FormatSettings.CurrencyString;
    AppLocale.ShortDateFormat   := 'yyyy-MM-dd';
    AppLocale.LongDateFormat    := 'yyyy-MM-dd';
    AppLocale.TimeAMString      := FormatSettings.TimeAMString;
    AppLocale.TimePMString      := FormatSettings.TimePMString;
    AppLocale.ShortTimeFormat   := 'hh:mm';
    AppLocale.LongTimeFormat    := 'hh:mm:ss';
    for i := Low( FormatSettings.ShortMonthNames) to High( FormatSettings.ShortMonthNames) do AppLocale.ShortMonthNames[ i] := FormatSettings.ShortMonthNames[ i];
    for i := Low( FormatSettings.LongMonthNames ) to High( FormatSettings.LongMonthNames ) do AppLocale.LongMonthNames [ i] := FormatSettings.LongMonthNames [ i];
    for i := Low( FormatSettings.ShortDayNames  ) to High( FormatSettings.ShortDayNames  ) do AppLocale.ShortDayNames  [ i] := FormatSettings.ShortDayNames  [ i];
    for i := Low( FormatSettings.LongDayNames   ) to High( FormatSettings.LongDayNames   ) do AppLocale.LongDayNames   [ i] := FormatSettings.LongDayNames   [ i];
    AppLocale.TwoDigitYearCenturyWindow := FormatSettings.TwoDigitYearCenturyWindow ;

    FormatSettings.ThousandSeparator := AppLocale.ThousandSeparator;
    FormatSettings.DecimalSeparator  := AppLocale.DecimalSeparator;
    FormatSettings.DateSeparator     := AppLocale.DateSeparator;
    FormatSettings.TimeSeparator     := AppLocale.TimeSeparator;
    FormatSettings.ListSeparator     := AppLocale.ListSeparator;
    FormatSettings.ShortDateFormat   := AppLocale.ShortDateFormat;
    FormatSettings.LongDateFormat    := AppLocale.LongDateFormat;
    FormatSettings.ShortTimeFormat   := AppLocale.ShortTimeFormat;
    FormatSettings.LongTimeFormat    := AppLocale.LongTimeFormat;
  end;


initialization

  InitializeUnit;

finalization


end.
