unit OldOSCif;

(*

  project name
  ------------

    OSCUtils


  description
  -----------

    Utility library to encode/decode osc-packets.

    Inspired by original OSC reference implementation ( OSC-Kit)
    and OSC.Net library as shipped with the TUIO-CSharp sample
    from http://reactable.iua.upf.edu/?software


  licence
  -------

    GNU Lesser General Public License ( LGPL)
    english: http://www.gnu.org/licenses/lgpl.html
    german: http://www.gnu.de/lgpl-ger.html


  language/ide
  ------------

    delphi


  initial author
  --------------

    joreg -> joreg@vvvv.org


  additions for FreePascal
  ------------------------

    simon -> simonmoscrop@googlemail.com


  addition of 'm' type message, and a reordering of the code.
  -----------------------------------------------------------

    jan -> http://bluehell.electro-music.com [2012-04-14]


  instructions for use with FreePascal
  ------------------------------------

    Define: FPC


  [jan] rationale for adding 'm' type message
  -------------------------------------------

    I needed to control my caparana with MIDI messages.


  [jan] rationale for reordering the code
  ---------------------------------------

    I got bored ;-)


  encoding a single message:
  --------------------------

    First create a message: msg := TOSCMessage.Create( address)
    then call msg.AddAsAnsiString( typetag, value) to add any number of arguments
    with msg.AsOSCAnsiString you get its osc-string representation

    [jan] a Note for MIDI messages :

    [jan] MIDI messages are compatible with what Kyma uses, that is they need a
    Port parameter that is currently ignored (set it to zero); then they need a
    MIDI status byte which is composed of a MIDI channel and a MIDI message and
    then zero, one or two arguments (depending on the actual MIDI message) where
    unused arguments should be set to $ff (all $ff bytes are ignored by Kyma).

    [jan] This is not available according to the OSC 1.1 standard.


  encoding a bundle:
  ------------------

    First create a bundle: bundle := TOSCBundle.Create
    then add any number of packets ( i.e. message, bundle) via bundle.Add( packet)
    with bundle.AsOSCAnsiString you get its osc-string representation


  decoding a string:
  ------------------

    Use TOSCPacket.Unpack( PByte( osc-string), Length( osc-string)) to create
    TOSCPackets of your osc-strings ( those can be either bundles or single
    messages.

    If you want to decode several packets at once you can create a container
    bundle first and add the packets you create like this and then use
    msg := FPacket.MatchAddress( address) to find a message with the according
    address in your packet-structure.

    Before you now can access the arguments and typetags of a message you have
    to call msg.Decode

    ~ voila ~

*)

interface

uses

  Classes, Contnrs;

type

  TOnLog      = procedure( aSender: TObject; const aMsg: string) of object;
  TOSCPacket  = class;
  TOSCMessage = class;


  TOSCPacket = class( TObject)
  protected
    FBytes : PByte;
    FOnLog : TOnLog;
  protected
    function    MatchBrackets( aMessage, anAddress: PAnsiChar): Boolean;
    function    MatchList    ( aMessage, anAddress: PAnsiChar): Boolean;
    function    MatchPattern ( aMessage, anAddress: PAnsiChar): Boolean;
  protected
    procedure   Log( const aMsg: string);
    procedure   LogFmt( const aFmt: string; const anArgs: array of const);
  public
    constructor Create( const aBytes: PByte);
    function    MatchAddress( const anAddress: AnsiString): TOSCMessage;                              virtual; abstract;
    function    AsOSCString: AnsiString;                                                              virtual; abstract;
    function    AsOScLog: string;                                                                     virtual; abstract;
    procedure   Unmatch;                                                                              virtual; abstract;
  public
    class function Unpack( aBytes: PByte; aCount: Integer): TOSCPacket;                                        overload;
    class function Unpack( aBytes: PByte; anOffset, aCount: Integer; aTimeTag: Extended = 0): TOSCPacket;  overload; virtual;
  end;


  TOSCMessage = class( TOSCPacket)
  private
    FAddress       : AnsiString;
    FTypeTags      : AnsiString;
    FArguments     : TStringList;
    FIsDecoded     : Boolean;
    FMatched       : Boolean;
    FTimeTag       : Extended;
    FTypeTagOffset : Integer;
  private
    function    GetArgument( anIndex: Integer): AnsiString;
    function    GetArgumentAsFloat(Index: Integer): Single;
    function    GetArgumentAsInt(Index: Integer): Integer;
    function    GetArgumentCount: Integer;
    function    GetTypeTag( anIndex: Integer): AnsiChar;
  public
    constructor Create( const anAddress: AnsiString);                                                          overload;
    constructor Create( const aBytes: PByte);                                                                  overload;
    destructor  Destroy;                                                                                       override;
    function    AddAsAnsiString( aTypeTag: AnsiChar; const aValue: AnsiString): Integer;
    procedure   AddFloat     ( aValue: Single);
    procedure   AddInteger   ( aValue: Integer);
    procedure   AddMIDI      ( aPort, aCh, aMsg: Byte; aV1: Byte = $ff; aV2: Byte = $ff);
    procedure   AddAnsiString( const aValue: AnsiString);
    procedure   Decode;
    function    MatchAddress( const anAddress: AnsiString): TOSCMessage;                                       override;
    function    AsOSCString: AnsiString;                                                                       override;
    function    AsOScLog: string;                                                                              override;
    procedure   Unmatch;                                                                                       override;
  public
    class function Unpack( aBytes: PByte; aPacketOffset, aCount: Integer; aTimeTag: Extended = 0): TOSCPacket; overload; override;
  public
    property    Address                           : AnsiString read FAddress           write FAddress;
    property    Argument       [ anIndex: Integer]: AnsiString read GetArgument;
    property    ArgumentAsFloat[ anIndex: Integer]: Single     read GetArgumentAsFloat;
    property    ArgumentAsInt  [ anIndex: Integer]: Integer    read GetArgumentAsInt;
    property    ArgumentCount                     : Integer    read GetArgumentCount;
    property    IsDecoded                         : Boolean    read FIsDecoded         write FIsDecoded;
    property    Matched                           : Boolean    read FMatched           write FMatched;
    property    TimeTag                           : Extended   read FTimeTag           write FTimeTag;
    property    TypeTag[ anIndex: Integer]        : AnsiChar   read GetTypeTag;
    property    TypeTagOffset                     : Integer    read FTypeTagOffset     write FTypeTagOffset;
  end;


  TOSCBundle = class( TOSCPacket)
  private
    FPackets : TObjectList;
  public
    constructor Create( const aBytes: PByte);
    destructor  Destroy;                                                                                       override;
    procedure   Addpacket( const aPacket: TOSCPacket);
    function    MatchAddress( const anAddress: AnsiString): TOSCMessage;                                       override;
    function    AsOSCString: AnsiString;                                                                       override;
    function    AsOScLog: string;                                                                              override;
    procedure   Unmatch;                                                                                       override;
  public
    class function Unpack( aBytes: PByte; aPacketOffset, aCount: Integer; aTimeTag: Extended = 0): TOSCPacket; overload; override;
  end;


  function MarshallToNetwork( aValue: LongInt): LongInt; inline;

const

  OSC_OK                   = 0;
  OSC_UNRECOGNIZED_TYPETAG = 1;
  OSC_CONVERT_ERROR        = 2;



implementation

uses

  SysUtils, Math {$IFNDEF FPC}, WinSock {$ENDIF};


  function MarshallToNetwork( aValue: LongInt): LongInt; inline;
  begin
    Result := {$IFDEF FPC} BEtoN( aValue) {$ELSE} htonl( aValue) {$ENDIF} ;
  end;

  function MarshallToLocal( aValue: LongInt): LongInt; inline;
  begin
    Result := {$IFDEF FPC} NtoBE( aValue) {$ELSE} ntohl( aValue); {$ENDIF};
  end;


  function  MakeOSCFloat( aValue: Single): AnsiString;
  var
    aTmp : Byte;
    aVal : Integer;
    i    : Integer;
  begin
    Result := '';
    aVal := MarshallToNetwork( Integer( Pointer( aValue)));
    for i := 0 to 3
    do begin
      aTmp   := aVal and $ff;
      Result := Result + AnsiChar( aTmp);
      aVal   := aVal shr 8;
    end;
  end;


  function  MakeOSCInt( aValue: Integer): AnsiString;
  var
    aTmp  : Byte;
    i    : Integer;
    aVal : Integer;
  begin
    Result := '';
    aVal   := MarshallToNetwork( aValue);
    for i := 0 to 3
    do begin
      aTmp   := aVal and $ff;
      Result := Result + AnsiChar( aTmp);
      aVal   := aVal shr 8;
    end;
  end;


  function  MakeOSCMIDI( aPort, aCh, aMsg, aV1, aV2: Byte): AnsiString;
  // [jan] aV1 and aV2 are optional, when not used for the particular
  // type of aMsg they should be set to $ff.
  var
    aStatus : Byte;
  begin
    aStatus := (( aMsg and $0f) shl 4) or ( aCh and $0f);
    Result  := AnsiChar( aPort) + AnsiChar( aStatus) + AnsiChar( aV1) + AnsiChar( aV2);
  end;


  function  MakeOSCAnsiString( const aValue: AnsiString): AnsiString;
  var
    i    : Integer;
    aLen : Integer;
  begin
    Result := aValue;

    aLen := 4 - ( length( aValue)) mod 4;
    for i := 0 to aLen - 1
    do Result := Result + #0;
  end;


  function  UnpackFloat( aBytes: PByte; var anOffset: Integer): Single;
  var
    i    : Integer;
    aVal : Integer;
    aTmp : PByte;
  begin
    if Assigned( aBytes)
    then begin
      aVal := 0;
      aTmp := aBytes;
      Inc( aTmp, anOffset);

      for i := 0 to 3
      do begin
        aVal := aVal + aTmp^ shl ( i * 8);
        Inc( aTmp);
      end;

      Inc( anOffset, 4);
      Result := Single( Pointer( MarshallToLocal( aVal)));
    end
    else Result := 0.0;
  end;


  function  UnpackInt( aBytes: PByte; var anOffset: Integer): Integer;
  var
    i    : Integer;
    aVal : Integer;
    aTmp : PByte;
  begin
    if Assigned( aBytes)
    then begin
      aVal := 0;
      aTmp := aBytes;
      Inc( aTmp, anOffset);

      for i := 0 to 3
      do begin
        aVal := aVal + aTmp^ shl ( i * 8);
        Inc( aTmp);
      end;

      Inc( anOffset, 4);
      Result := MarshallToLocal( aVal);
    end
    else Result := 0;
  end;


  function  UnpackString( aBytes: PByte; var anOffset: Integer): AnsiString;
  var
    aTmp : PByte;
    aLen : Integer;
  begin
    if Assigned( aBytes)
    then begin
      aTmp := aBytes;
      Inc( aTmp, anOffset);

      Result := PAnsiChar( aTmp);
      aLen   := Length( PAnsiChar( aTmp));
      aLen   := aLen + ( 4 - aLen mod 4);
      Inc( anOffset, aLen)
    end
    else Result := AnsiString( '');
  end;


  function   UnpackMIDI( aBytes: PByte; var anOffset: Integer): AnsiString;
  // [jan]
  var
    aPort   : Byte;
    aStatus : Byte;
    aCh     : Byte;
    aMsg    : Byte;
    aV1     : Byte;
    aV2     : Byte;
    aTmp    : PByte;
  begin
    if Assigned( aBytes)
    then begin
      aTmp    := aBytes;           Inc( aTmp, anOffset);
      aPort   := aTmp^ and $ff;    Inc( aTmp);
      aStatus := aTmp^ and $ff;    Inc( aTmp);
      aV1     := aTmp^ and $ff;    Inc( aTmp);
      aV2     := aTmp^ and $ff;
      aCh     := aStatus and $0f;
      aMsg    := ( aStatus shr 4) and $0f;
      Result  := AnsiChar( aPort) + AnsiChar( aCh) + AnsiChar( aMsg);
      if aV1 <> $ff then Result  := Result + AnsiChar( aV1);
      if aV2 <> $ff then Result  := Result + AnsiChar( aV2);
      Inc( anOffset, 4);
    end
    else Result := AnsiString( #0 + #0 + #0 + #0);
  end;



{ ========
  TOSCPacket = class( TObject)
  protected
    FBytes : PByte;
  protected
}

    // we know that pattern[ 0] == '[' and test[ 0] != 0 */
    function    TOSCPacket.MatchBrackets( aMessage, anAddress: PAnsiChar): Boolean;
    var
      Negated  : Boolean;
      p        : PAnsiChar;
      p1       : PAnsiChar;
      p2       : PAnsiChar;
    begin
      p       := aMessage;
      Result  := False;
      Negated := False;

      Inc( aMessage);
      if aMessage^ = #0
      then begin
        LogFmt( 'Unterminated [ in message: %s', [ FBytes[ 0]]);
        exit;
      end;

      if aMessage^ = '!'
      then begin
        Negated := True;
        Inc( p);
      end;

      Result := Negated;

      while p^ <> ']'
      do begin
        if p^ = #0
        then begin
          LogFmt( 'Unterminated [ in message: %s', [ FBytes[ 0]]);
          exit;
        end;

        p1 := p  + 1; // sizeOf( PAnsiChar);
        p2 := p1 + 1; // sizeOf( PAnsiChar);

        if ( p1^ = '-') and ( p2^ <> #0)
        then begin
          if
            ( Ord( anAddress^) >= Ord( p^ )) and
            ( Ord( anAddress^) <= Ord( p2^))
          then begin
            Result := not Negated;
            break;
          end;
        end;

        if p^ = anAddress^
        then begin
          Result := not Negated;
          break;
        end;

        Inc( p);
      end;

      if not Result
      then exit;

      while p^ <> ']'
      do begin
        if p^ = #0
        then begin
          LogFmt( 'Unterminated [ in message: %s', [ FBytes[ 0]]);
          exit;
        end;

        Inc( p);
      end;

      Inc( p);
      Inc( anAddress);
      Result := MatchPattern( p, anAddress);
    end;

    function    TOSCPacket.MatchList( aMessage, anAddress: PAnsiChar): Boolean;
    var
      p  : PAnsiChar;
      tp : PAnsiChar;
    begin
      Result := False;
      p      := aMessage;
      tp     := anAddress;

      while p^ <> '}'
      do begin
        if p^ = #0
        then begin
          LogFmt( 'Unterminated { in message: %s', [ FBytes[ 0]]);
          exit;
        end;

        Inc( p);
      end;

      Inc( p       ); // skip close curly brace
      Inc( aMessage); // skip open  curly brace

      while True
      do begin
        if aMessage^ = ','
        then begin
          if MatchPattern( p, tp)
          then begin
            Result := True;
            exit;
          end
          else begin
            tp := anAddress;
            Inc( aMessage);
          end;
        end
        else if aMessage^ = '}'
        then begin
          Result := MatchPattern( p, tp);
          exit;
        end
        else if aMessage^ = tp^
        then begin
          Inc( aMessage);
          Inc( tp);
        end
        else begin
          tp := anAddress;
          while ( aMessage^ <> ',') and ( aMessage^ <> '}')
          do Inc( aMessage);

          if aMessage^ = ','
          then Inc( aMessage);
        end;
      end;
    end;

    function    TOSCPacket.MatchPattern( aMessage, anAddress: PAnsiChar): Boolean;
    begin
      if anAddress^ = '*'
      then begin
        Result := True;
        Exit;
      end;

      if ( aMessage = nil) or ( aMessage^ = #0)
      then begin
        Result := anAddress^ = #0;
        exit;
      end;

      if anAddress^ = #0
      then begin
        if aMessage^ = '*'
        then begin
          Result := MatchPattern( aMessage + 1, anAddress);
          exit;
        end
        else begin
          Result := False;
          exit;
        end;
      end;

      case aMessage^ of
        #0 : Result := anAddress^ = #0;
        '?': Result := MatchPattern( aMessage + 1, anAddress + 1);
        '*':
          begin
            if MatchPattern( aMessage + 1, anAddress)
            then Result := True
            else Result := MatchPattern( aMessage, anAddress + 1);
          end;
        ']','}':
          begin
            LogFmt( 'Spurious %s in message: %s', [ aMessage^, FBytes[ 0]]);
            Result := False;
          end;
      '[': Result := MatchBrackets( aMessage, anAddress);
      '{': Result := MatchList( aMessage, anAddress);
      else
        if aMessage^ = anAddress^
        then Result := MatchPattern( aMessage + 1,anAddress + 1)
        else Result := False;
      end;
    end;


//  protected

    procedure   TOSCPacket.Log( const aMsg: string);
    begin
      if Assigned( FOnLog)
      then FOnLog( Self, aMsg);
    end;

    procedure   TOSCPacket.LogFmt( const aFmt: string; const anArgs: array of const);
    begin
      Log( Format( aFmt, anArgs));
    end;

//  public

    constructor TOSCPacket.Create( const aBytes: PByte);
    begin
      FBytes := aBytes;
    end;

//  public

    class function TOSCPacket.Unpack( aBytes: PByte; aCount: Integer): TOSCPacket;
    begin
      Result := UnPack( aBytes, 0, aCount);
    end;

    class function TOSCPacket.Unpack( aBytes: PByte; anOffset, aCount: Integer; aTimeTag: Extended = 0): TOSCPacket;
    var
      aTmp: PByte;
    begin
      aTmp := aBytes;
      Inc( aTmp, anOffset);

      if AnsiChar( aTmp^) = '#'
      then Result := TOSCBundle .UnPack( aBytes, anOffset, aCount)
      else Result := TOSCMessage.UnPack( aBytes, anOffset, aCount, aTimeTag);
    end;


{ ========
  TOSCMessage = class( TOSCPacket)
  private
    FAddress       : AnsiString;
    FArguments     : TStringList;
    FIsDecoded     : Boolean;
    FMatched       : Boolean;
    FTimeTag       : Extended;
    FTypeTagOffset : Integer;
    FTypeTags      : string;
  public
    property    Address                           : AnsiString read FAddress           write FAddress;
    property    Argument       [ anIndex: Integer]: AnsiString read GetArgument;
    property    ArgumentAsFloat[ anIndex: Integer]: Single     read GetArgumentAsFloat;
    property    ArgumentAsInt  [ anIndex: Integer]: Single     read GetArgumentAsInt;
    property    ArgumentCount                     : Integer    read GetArgumentCount;
    property    IsDecoded                         : Boolean    read FIsDecoded         write FIsDecoded;
    property    Matched                           : Boolean    read FMatched           write FMatched;
    property    TimeTag                           : Extended   read FTimeTag           write FTimeTag;
    property    TypeTag[ anIndex: Integer]        : AnsiChar   read GetTypeTag;
    property    TypeTagOffset                     : Integer    read FTypeTagOffset     write FTypeTagOffset;
  private
}

    function    TOSCMessage.GetArgument( anIndex: Integer): AnsiString;
    begin
      Result := AnsiString( FArguments[ anIndex]);
    end;

    function    TOSCMessage.GetArgumentAsFloat(Index: Integer): Single;
    begin
      Result := PSingle( Pointer( FArguments[ Index]))^;
    end;

    function    TOSCMessage.GetArgumentAsInt(Index: Integer): Integer;
    begin
      Result := PInteger( Pointer( FArguments[ Index]))^;
    end;

    function    TOSCMessage.GetArgumentCount: Integer;
    begin
      Result := FArguments.Count;
    end;

    function    TOSCMessage.GetTypeTag( anIndex: Integer): AnsiChar;
    begin
      Result := FTypeTags[ anIndex + 2];
    end;

//  public

    constructor TOSCMessage.Create( const anAddress: AnsiString);
    begin
      Create( nil);
      FAddress := anAddress;
    end;

    constructor TOSCMessage.Create( const aBytes: PByte);
    begin
      inherited;

      FTypeTags  := ',';
      FArguments := TStringList.Create;
      FIsDecoded := False;
    end;

    destructor  TOSCMessage.Destroy;
    begin
      FArguments.Free;
      inherited;
    end;

    function    TOSCMessage.AddAsAnsiString( aTypeTag: AnsiChar; const aValue: AnsiString): Integer;
    begin
      Result := OSC_OK;

      try
        if      aTypeTag = 'f' then FArguments.Add( string( MakeOSCFloat( StrToFloat( string( aValue)))))
        else if aTypeTag = 'i' then FArguments.Add( string( MakeOSCInt  ( StrToInt  ( string( aValue)))))
        else if aTypeTag = 'm' then FArguments.Add( string( MakeOSCInt  ( StrToInt  ( string( aValue)))))
        else if aTypeTag = 's' then FArguments.Add( string( MakeOSCAnsiString( aValue)))
        else Result := OSC_UNRECOGNIZED_TYPETAG;
      except
        on EConvertError
        do Result := OSC_CONVERT_ERROR;
      end;

      if Result = OSC_OK
      then FTypeTags := FTypeTags + aTypeTag;
    end;

    procedure   TOSCMessage.AddFloat( aValue: Single);
    begin
      FTypeTags := FTypeTags + 'f';
      FArguments.Add( string( MakeOSCFloat( aValue)));
    end;

    procedure   TOSCMessage.AddInteger( aValue: Integer);
    begin
      FTypeTags := FTypeTags + 'i';
      FArguments.Add( string( MakeOSCInt( aValue)));
    end;

    procedure   TOSCMessage.AddMIDI( aPort, aCh, aMsg: Byte; aV1: Byte = $ff; aV2: Byte = $ff);
    begin
      FTypeTags := FTypeTags + 'm';
      FArguments.Add( string( MakeOSCMIDI( aPort, aCh, aMsg, aV1, aV2)));
    end;

    procedure   TOSCMessage.AddAnsiString( const aValue: AnsiString);
    begin
      FTypeTags := FTypeTags + 's';
      FArguments.Add( string( MakeOSCAnsiString( aValue)));
    end;

    procedure   TOSCMessage.Decode;
    var
      i        : Integer;
      anOffset : Integer;
    begin
      if FIsDecoded
      then exit;

      anOffset  := FTypeTagOffset;
      FTypeTags := UnpackString( FBytes, anOffset);

      for i := 1 to Length( FTypeTags) - 1
      do begin
        if      FTypeTags[ i + 1] = 'f' then FArguments.Add( FloatToStr( UnpackFloat ( FBytes, anOffset)))
        else if FTypeTags[ i + 1] = 'i' then FArguments.Add( IntToStr  ( UnpackInt   ( FBytes, anOffset)))
        else if FTypeTags[ i + 1] = 'm' then FArguments.Add( IntToStr  ( UnpackInt   ( FBytes, anOffset)))
        else if FTypeTags[ i + 1] = 's' then FArguments.Add( string    ( UnpackString( FBytes, anOffset)));
      end;

      FIsDecoded := True;
    end;



    function    TOSCMessage.MatchAddress( const anAddress: AnsiString): TOSCMessage;
    begin
      if not FMatched and MatchPattern( PAnsiChar( FAddress), PAnsiChar( anAddress))
      then begin
        FMatched := True;
        Result   := Self
      end
      else Result := nil;
    end;

    function    TOSCMessage.AsOSCString: AnsiString;
    var
      i: Integer;
    begin
      Result := MakeOSCAnsiString( FAddress) + MakeOSCAnsiString( FTypeTags);

      for i := 0 to FArguments.Count - 1
      do Result := Result + Argument[ i];
    end;

    function    TOSCMessage.AsOScLog: string; // override;
    var
      i         : Integer;
      aTypeTag  : AnsiChar;
      aValue    : string;
      anOffset  : Integer;
    begin
      Result := string( FAddress);

      for i := 0 to FArguments.Count - 1
      do begin
        Result := Result + ' ';

        aTypeTag  := TypeTag [ i];
        aValue    := '<undefined>';
        anOffset  := 0;

        if      aTypeTag = 'f' then aValue := FloatToStr( UnpackFloat ( PByte( @ Argument[ i][ 1]), anOffset))
        else if aTypeTag = 'i' then aValue := IntToStr  ( UnpackInt   ( PByte( @ Argument[ i][ 1]), anOffset))
        else if aTypeTag = 'm' then aValue := string    ( UnpackMIDI  ( PByte( @ Argument[ i][ 1]), anOffset))
        else if aTypeTag = 's' then aValue := string    ( UnpackString( PByte( @ Argument[ i][ 1]), anOffset));

        Result := Result + format( '%s(%s)', [ aTypeTag, aValue])
      end;
    end;

    procedure   TOSCMessage.Unmatch;
    begin
      FMatched := False;
    end;

    class function TOSCMessage.Unpack( aBytes: PByte; aPacketOffset, aCount: Integer; aTimeTag: Extended = 0): TOSCPacket;
    begin
      Result := TOSCMessage.Create( aBytes);
      with Result as TOSCMessage
      do begin
        Address       := UnpackString( aBytes, aPacketOffset);  // for now decode address only
        TimeTag       := aTimeTag;
        TypeTagOffset := aPacketOffset;                         // save offset for later decoding on demand
        IsDecoded     := False;
      end;
    end;


{ ========
  TOSCBundle = class( TOSCPacket)
  private
    FPackets : TObjectList;
  public
}

    constructor TOSCBundle.Create( const aBytes: PByte);
    begin
      inherited;
      FPackets := TObjectList.Create;
      FPackets.OwnsObjects := True;
    end;

    destructor  TOSCBundle.Destroy;
    begin
      FPackets.Free;
      inherited;
    end;

    procedure   TOSCBundle.Addpacket( const aPacket: TOSCPacket);
    begin
      FPackets.Add( aPacket);
    end;

    function    TOSCBundle.MatchAddress( const anAddress: AnsiString): TOSCMessage;
    var
      i: Integer;
    begin
      Result := nil;

      for i := 0 to FPackets.Count - 1
      do begin
        Result := ( FPackets[ i] as TOSCPacket).MatchAddress( anAddress);
        if Assigned( Result)
        then break;
      end;
    end;

    function    TOSCBundle.AsOSCString: AnsiString;
    var
      aTimeTag : AnsiString;
      i        : Integer;
      S        : AnsiString;
    begin
      aTimeTag := #0#0#0#0#0#0#0#1;                           // immediately
      Result   := MakeOSCAnsiString( '#bundle') + aTimeTag;

      for i := 0 to FPackets.Count - 1
      do begin
        S      := ( FPackets[ i] as TOSCPacket).AsOSCString;
        Result := Result + MakeOSCInt( Length( S)) + S;
      end;
    end;

    function    TOSCBundle.AsOScLog: string; // override;
    begin
      // todo : make osc's AsOscLog
      Result := 'TOSCBundle.AsOScLog ... to be invented ...';
    end;

    procedure   TOSCBundle.Unmatch;
    var
      i: Integer;
    begin
      for i := 0 to FPackets.Count - 1
      do ( FPackets[ i] as TOSCPacket).UnMatch;
    end;

//  public

    class function TOSCBundle.Unpack( aBytes: PByte; aPacketOffset, aCount: Integer; aTimeTag: Extended = 0): TOSCPacket;
    var
      aPacketLen : Integer;
      tt1        : Cardinal;
      tt2        : Cardinal;
    begin
      Result := TOSCBundle.Create( aBytes);

      // advance the '#bundle' string
      UnpackString( aBytes, aPacketOffset);

      // advance the timestamp
      tt1 := Cardinal( UnpackInt( aBytes, aPacketOffset));
      tt2 := Cardinal( UnpackInt( aBytes, aPacketOffset));

      aTimeTag := tt1 + tt2 / power( 2, 32);

      while aPacketOffset < aCount
      do begin
        aPacketLen := UnpackInt( aBytes, aPacketOffset);
        // NOTE: aPacketOffset is always from the very beginning of aBytes, and not the beginning of the current packet.
        // todo : this looks like a bug ???? (aPacketOffset + aPacketLen should be aPacketLen I think) - jan.
        ( Result as TOSCBundle).Addpacket( TOSCPacket.Unpack( aBytes, aPacketOffset, aPacketOffset + aPacketLen, aTimeTag));
        Inc( aPacketOffset, aPacketLen);
      end;
    end;


end.
