unit module_implementations;

{

   COPYRIGHT 2013 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Classes, System.Math, System.UITypes, Vcl.Dialogs, WinApi.WinSock,

  Globals, module_defs, ComplexMath, KnobsUtils, KnobsConversions, midi_defs, midi_reader, OSCif, windowing,

  WaveUtils, WaveStorage, scala, freereverb, gverb, fftwgen, fftw_, FANN, FannNetwork, Talkie, forth, module_scales,
  loro_sc, poetry_generator, grains;

const

  DELAY_BUF_SIZE  = 65536;   // main delay buffer size - 65536 samples on 44100 Hz means 1.49 s.
  MAX_XTAB        = 20480;   // crossfade lookup table size for pitch change

type

  TARState         = ( arIdle, arA, arR, arWait);
  TAHDState        = ( ahdIdle, ahdA, ahdH, ahdD, ahdWait);
  TADSRState       = ( adsrIdle, adsrA, adsrD, adsrS, adsrR, adsrWait);
  TEnvShape        = ( esLog, esLin, esExp, esS);
  TLfoRange        = ( lsSlow, lsMedium, lsFast, lsBPM, lsLin);
  TDelayRange      = ( dsShort, dsMedium, dsLong);
  TRWState         = ( trwIdle, trwFired);
  TTrigOScState    = ( toIdle, toRunning, toWaiting, toDone);
  TSeqInState      = ( seqClkLow, seqClkHigh);
  TSeqStepState    = ( sssIdle, sssActive, sssWaiting);
  TPulsesState     = ( psIdle, psPulsing, psWaiting);
  TMidiPlayerState = ( msActive, msDone);

  TLongWord = record
    Lo, Hi : Word;
  end;

  TFadeTable = array[ 0 .. MAX_XTAB - 1] of TSignal;
  PFadeTable = ^TFadeTable;


  TModConstant = class( TMod)
  strict private
  const
    i_value = 0;
    i_mute  = 1;
    i_mode  = 2;
  const
    o_out   = 0;
  const
    m_lin44 = 0;
    m_db4   = 1;
    m_lin11 = 2;
    m_db0   = 3;
    m_lin04 = 4;
    m_lin01 = 5;
    m_note  = 6;
  private
    FMode : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Mode: Integer read FMode write FMode;
  end;

  TModPad = class( TMod)
  strict private
  const
    i_padx       = 0;
    i_pady       = 1;
    i_outputtype = 2;
  const
    o_x          = 0;
    o_y          = 1;
  private
    FMode : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Mode: Integer read FMode write FMode;
  end;

  TModInverter = class( TMod)
  strict private
  const
    i_in1  = 0; // Keep these consecutive
    i_in2  = 1;
    i_in3  = 2;
    i_in4  = 3;
  const
    o_out1 = 0; // Keep these consecutive
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModAdder = class( TMod)
  strict private
  const
    i_mute = 0;
    i_in1  = 1;
    i_in2  = 2;
    i_in3  = 3;
    i_in4  = 4;
  const
    o_out  = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMultiplier = class( TMod)
  strict private
  const
    i_mute = 0;
    i_in1  = 1;
    i_in2  = 2;
    i_in3  = 3;
    i_in4  = 4;
  const
    o_out  = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMultiMult = class( TMod)
  strict private
  const
    i_mult = 0;
    i_in1  = 1;
    i_in2  = 2;
    i_in3  = 3;
    i_in4  = 4;
    i_in5  = 5;
    i_in6  = 6;
    i_in7  = 7;
    i_in8  = 8;
  const
    o_out1 = 0;
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
    o_out5 = 4;
    o_out6 = 5;
    o_out7 = 6;
    o_out8 = 7;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModNoteToDelay = class( TMod)
  strict private
  const
    i_in  = 0;
  const
    o_out = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModZeroCounter = class( TMod)
  strict private
  const
    i_in        = 0;
    i_threshold = 1;
  const
    o_period    = 0;
    o_gate      = 1;
    o_note      = 2;
  private
    FOldSign : Boolean;
    FTime    : Integer;
    FFreq    : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMedian = class( TMod)
  strict private
  const
    i_in   = 0;
    i_trig = 1;
  const
    o_out  = 0;
    o_high = 1;
    o_low  = 2;
  private
    FData      : TMedianCalculator;
    FOldTrig   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModCompare = class( TMod)
  strict private
  const
    i_in1      = 0;
    i_in2      = 1;
    i_range    = 2;
  const
    o_outless  = 0;
    o_outequal = 1;
    o_outmore  = 2;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModGate = class( TMod)
  strict private
  const
    i_function = 0;
    i_width    = 1;
    i_in1      = 2;
    i_in2      = 3;
    i_in3      = 4;
    i_in4      = 5;
  const
    o_out      = 0;
  const
    m_and      = 0;
    m_nand     = 1;
    m_or       = 2;
    m_nor      = 3;
    m_xor      = 4; // odd parity, true when an odd number of inputs is active
    m_xnor     = 5; // even pariry, true when an even number of inputs is active
    m_one      = 6; // true when exactly one input is active
    m_none     = 7; // false when exactly one input is activee
  private
    FFunction : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDivider = class( TMod)
  strict private
  const
    i_count = 0;
    i_trig  = 1;
    i_reset = 2;
  const
    o_out   = 0;
  const
    m_states : array[ Boolean] of TSignal = ( -1, 1);
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FOldCount  : Integer;
    FState     : Boolean;
    FOldReset  : Boolean;
  private
    procedure   PerformReset;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModProgDivider = class( TMod)
  strict private
  const
    i_trig      = 0;
    i_reset     = 1;
    i_select    = 2;
    i_selectlev = 3;
    i_steps     = 4;
    i_lowcount  = 5;
    i_highcount = 6;
  const
    o_out       = 0;
  const
    m_states : array[ Boolean] of TSignal = ( -1, 1);
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FCount     : Integer;
    FState     : Boolean;
    FOldReset  : Boolean;
  private
    procedure   PerformReset;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDFlipFlop = class( TMod)
  strict private
  const
    i_d    = 0;
    i_clk  = 1;
  const
    o_q    = 0;
    o_notq = 1;
  private
    FClockState : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModRSFlipFlop = class( TMod)
  strict private
  const
    i_s        = 0;
    i_r        = 1;
    i_trigmode = 2;
  const
    o_q        = 0;
    o_notq     = 1;
  const
    m_edge     = 0;
    m_level    = 1;
  private
    FSetState   : Boolean;
    FResetState : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModCounter = class( TMod)
  strict private
  const
    i_trig  =  0;
    i_reset =  1;
    i_dir   =  2;
    i_mode  =  3;
  const
    o_out0  =  0;
    o_out1  =  1;
    o_out2  =  2;
    o_out3  =  3;
    o_out4  =  4;
    o_out5  =  5;
    o_out6  =  6;
    o_out7  =  7;
    o_out8  =  8;
    o_out9  =  9;
    o_out10 = 10;
    o_out11 = 11;
    o_out12 = 12;
    o_out13 = 13;
    o_out14 = 14;
    o_out15 = 15;
  const
    m_bin   =  0;
    m_seq   =  1;
  private
    FState   : Word;
    FMode    : Integer;
    FOldRes  : Boolean;
    FOldTrig : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAdc = class( TMod)
  strict private
  const
    i_in    =  0;
  const
    o_out0  =  0;
    o_out1  =  1;
    o_out2  =  2;
    o_out3  =  3;
    o_out4  =  4;
    o_out5  =  5;
    o_out6  =  6;
    o_out7  =  7;
    o_out8  =  8;
    o_out9  =  9;
    o_out10 = 10;
    o_out11 = 11;
    o_out12 = 12;
    o_out13 = 13;
    o_out14 = 14;
    o_out15 = 15;
  const
    n_bits  = 16;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDac = class( TMod)
  strict private
  const
    i_in0  =  0;
    i_in1  =  1;
    i_in2  =  2;
    i_in3  =  3;
    i_in4  =  4;
    i_in5  =  5;
    i_in6  =  6;
    i_in7  =  7;
    i_in8  =  8;
    i_in9  =  9;
    i_in10 = 10;
    i_in11 = 11;
    i_in12 = 12;
    i_in13 = 13;
    i_in14 = 14;
    i_in15 = 15;
  const
    o_out  =  0;
  const
    n_bits = 16;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModPhaseDetect = class( TMod)
  strict private
  type
    t_state = 1 .. 12;
  const
    i_in1         = 0;
    i_in2         = 1;
    i_alpha       = 2;
    i_alphamod    = 3;
    i_alphamodamt = 4;
    i_type        = 5;
    i_typesel     = 6;
  const
    o_out         = 0;
    o_lock        = 1;
  private
    FOldIn1 : Boolean;
    FOldIn2 : Boolean;
    FState  : t_state;
    FAlpha  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModNot = class( TMod)
  strict private
  const
    i_in1  = 0;
    i_in2  = 1;
    i_in3  = 2;
    i_in4  = 3;
  const
    o_out1 = 0;
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModLogicSelector = class( TMod)
  strict private
  const
    i_sel  = 0;
    i_ina  = 1;
    i_inb  = 2;
  const
    o_outa = 0;
    o_outb = 1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModAmplifier = class( TMod)
  strict private
  const
    i_gain = 0;
    i_mute = 1;
    i_in1  = 2; // keep in1 .. in4 consecutive
    i_in2  = 3;
    i_in3  = 4;
    i_in4  = 5;
  const
    o_out1 = 0; // keep out1 .. out4 consecutive
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModScaler = class( TMod)
  strict private
  const
    i_gain   = 0;
    i_offset = 1;
    i_mode   = 2;
    i_mute   = 3;
    i_in1    = 4;
    i_in2    = 5;
    i_in3    = 6;
    i_in4    = 7;
  const
    o_out1   = 0;
    o_out2   = 1;
    o_out3   = 2;
    o_out4   = 3;
    o_sum    = 4;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRatio = class( TMod)
  strict private
  const
    i_multiplier = 0;
    i_divider    = 1;
    i_mute       = 2;
    i_in1        = 3;
    i_in2        = 4;
    i_in3        = 5;
    i_in4        = 6;
  const
    o_out1       = 0;
    o_out2       = 1;
    o_out3       = 2;
    o_out4       = 3;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRangeConverter = class( TMod)
  strict private
  const
    i_lowvalue   = 0;
    i_highvalue  = 1;
    i_mode       = 2;
    i_in1        = 3;
    i_in2        = 4;
    i_in3        = 5;
    i_in4        = 6;
  const
    o_out1       = 0;
    o_out2       = 1;
    o_out3       = 2;
    o_out4       = 3;
  const
    s_normal            = 0;
    s_inverted          = 1;
    s_positive          = 2;
    s_positive_inverted = 3;
    s_negative          = 4;
    s_negative_inverted = 5;
  private
    FInLow  : TSignal;
    FInHigh : TSignal;
    FMode   : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModLogisticMap = class( TMod)
  strict private
  const
    i_clock   =  0;
    i_start   =  1;
    i_smod    =  2;
    i_smodamt =  3;
    i_svalue  =  4;
    i_rmod    =  5;
    i_rmodamt =  6;
    i_lowr    =  7;
    i_highr   =  8;
    i_xfade   =  9;
    i_outtype = 10;
    i_mute    = 11;
  const
    o_out     =  0;
  private
    FOldStart   : Boolean;
    FOldClock   : Boolean;
    FTicks      : Integer;
    FPeriodTime : Integer;
    FOutValue   : TSignal;
    FNextValue  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRotator = class( TMod)
  strict private
  const
    i_angle = 0;
    i_in1   = 1;
    i_in2   = 2;
  const
    o_out1  = 0;
    o_out2  = 1;
  private
    FAngle: TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRectifier = class( TMod)
  strict private
  const
    i_mode   = 0;
    i_in     = 1;
  const
    o_out    = 0;
    o_outinv = 1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMinMax = class( TMod)
  strict private
  const
    i_in1 = 0;
    i_in2 = 1;
  const
    o_min = 0;
    o_max = 1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModWaveWiper = class( TMod)
  strict private
  const
    i_break  = 0;
    i_mute   = 1;
    i_in1    = 2;
    i_in2    = 3;
  const
    o_out1   = 0;
    o_out2   = 1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModWaveWrapper = class( TMod)
  strict private
  const
    i_in        = 0;
    i_mute      = 1;
    i_lowlevel  = 2;
    i_highlevel = 3;
    i_mod       = 4;
    i_modamt    = 5;
    i_type      = 6;
  const
    o_out       = 0;
  const
    m_type1     = 0;
    m_type2     = 1;
    m_type3     = 2;
  private
    FLastValue  : TSignal;
    FOutValue   : TSignal;
    FStateValue : TSignal;
    FGoingUp    : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModIntDif = class( TMod)
  strict private
  const
    i_in   = 0;
    i_mode = 1;
  const
    o_int  = 0;
    o_dif  = 1;
  private
    FSum        : TSignal;
    FlastSample : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModBaseOsc = class( TMod)
  strict private
  const
    i_frequency  = 0;
    i_fm         = 1;
    i_phase      = 2;
    i_sync       = 3;
    i_mute       = 4;
    i_shape      = 5;
    i_pmlevel    = 6;
    i_fmlevel    = 7;
    i_cents      = 8;
    i_bandlimit  = 9;
    i_freq       = 10;
  const
    o_out        = 0;
  private
    FPhase      : TSignal;
    FPhaseDelta : TSignal;
    FPosition   : TSignal;
    FOldSync    : Boolean;
    FShape      : Integer;
    FBandLimit  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModBaseTrigOsc = class( TModBaseOsc)
  strict private
  const
    i_frequency =  0;
    i_fm        =  1;
    i_trig      =  2;
    i_count     =  3;
    i_mute      =  4;
    i_shape     =  5;
    i_pmlevel   =  6;
    i_fmlevel   =  7;
    i_cents     =  8;
    i_bandlimit =  9;
    i_freq      = 10;
  const
    o_out       =  0;
  private
    FState   : TTrigOScState;
    FCounter : Integer;
  public
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModOsc = class( TModBaseOsc)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_phase      =  2;
    i_sync       =  3;
    i_mute       =  4;
    i_shape      =  5;
    i_pmlevel    =  6;
    i_fmlevel    =  7;
    i_cents      =  8;
    i_bandlimit  =  9;
    i_freq       = 10;
  const
    o_out        =  0;
  public
    procedure   DoTick;                                                                          override;
  end;

  TModTrigOsc = class( TModBaseTrigOsc)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_trig       =  2;
    i_count      =  3;
    i_mute       =  4;
    i_shape      =  5;
    i_pmlevel    =  6;
    i_fmlevel    =  7;
    i_cents      =  8;
    i_bandlimit  =  9;
    i_freq       = 10;
  const
    o_out        = 0;
  public
    procedure   DoTick;                                                                          override;
  end;

  TModSquare = class( TModBaseOsc)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_phase      =  2;
    i_sync       =  3;
    i_mute       =  4;
    i_shape      =  5;
    i_pmlevel    =  6;
    i_fmlevel    =  7;
    i_cents      =  8;
    i_bandlimit  =  9;
    i_freq       = 10;
    i_pwm        = 11;
    i_pwmlevel   = 12;
  const
    i_out        =  0;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareTrig = class( TModBaseTrigOsc)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_phase      =  2;
    i_sync       =  3;
    i_mute       =  4;
    i_shape      =  5;
    i_pmlevel    =  6;
    i_fmlevel    =  7;
    i_cents      =  8;
    i_bandlimit  =  9;
    i_freq       = 10;
    i_pwm        = 11;
    i_pwmlevel   = 12;
  const
    o_out        = 0;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModBaseLfo = class( TMod)
  strict private
  const
    i_frequency  = 0;
    i_fm         = 1;
    i_phase      = 2;
    i_res        = 3;
    i_mute       = 4;
    i_fmlevel    = 5;
    i_pmlevel    = 6;
    i_shape      = 7;
    i_range      = 8;
    i_outputtype = 9;
  const
    o_out        = 0;
    o_sync       = 1;
  private
    FPhase       : TSignal;
    FPosition    : TSignal;
    FOldRes      : Boolean;
    FShape       : Integer;
    FWrapped     : Boolean;
    FRandomValue : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModBaseTrigLfo = class( TModBaseLfo)
  strict private
  const
    i_frequency  = 0;
    i_fm         = 1;
    i_count      = 2;
    i_trig       = 3;
    i_mute       = 4;
    i_fmlevel    = 5;
    i_pmlevel    = 6;
    i_shape      = 7;
    i_range      = 8;
    i_outputtype = 9;
  const
    o_out        = 0;
    o_ready      = 1;
  private
    FState   : TTrigOScState;
    FCounter : Integer;
    FDidRun  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
  end;

  TModCounterLfo = class( TMod)
  strict private
  const
    i_frequency  = 0;
    i_fm         = 1;
    i_fmlevel    = 2;
    i_range      = 3;
    i_mute       = 4;
    i_outputtype = 5;
  private
    FPhase : TSignal;
  protected
    procedure   Wrapped;                                                                          virtual;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModLfo = class( TModBaseLfo)
  strict private
  const
    i_frequency  = 0;
    i_fm         = 1;
    i_phase      = 2;
    i_res        = 3;
    i_mute       = 4;
    i_fmlevel    = 5;
    i_pmlevel    = 6;
    i_shape      = 7;
    i_range      = 8;
    i_outputtype = 9;
  const
    o_out        = 0;
    o_sync       = 1;
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModLfoTrig = class( TModBaseTrigLfo)
  strict private
  const
    i_frequency  = 0;
    i_fm         = 1;
    i_count      = 2;
    i_trig       = 3;
    i_mute       = 4;
    i_fmlevel    = 5;
    i_pmlevel    = 6;
    i_shape      = 7;
    i_range      = 8;
    i_outputtype = 9;
  const
    o_out        = 0;
    o_ready      = 1;
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModSquareLfo = class( TModBaseLfo)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_phase      =  2;
    i_res        =  3;
    i_mute       =  4;
    i_fmlevel    =  5;
    i_pmlevel    =  6;
    i_shape      =  7;
    i_range      =  8;
    i_outputtype =  9;
    i_pwm        = 10;
    i_pwmlevel   = 11;
  const
    o_out        =  0;
    o_sync       =  1;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareLfoTrig = class( TModBaseTrigLfo)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_count      =  2;
    i_trig       =  3;
    i_mute       =  4;
    i_fmlevel    =  5;
    i_pmlevel    =  6;
    i_shape      =  7;
    i_range      =  8;
    i_outputtype =  9;
    i_pwm        = 10;
    i_pwmlevel   = 11;
  const
    o_out        =  0;
    o_ready      =  1;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModRandSig = class( TModCounterLfo)
  strict private
  const
    i_frequency  =  0;
    i_fm         =  1;
    i_fmlevel    =  2;
    i_range      =  3;
    i_mute       =  4;
    i_outputtype =  5;
    i_dist       =  6;
    i_distmod    =  7;
    i_distmodamt =  8;
    i_lambda     =  9;
    i_mode       = 10;
  const
    o_out        =  0;
  const
    m_lin        =  0;
    m_exp        =  1;
  private
    FStart  : TSignal;
    FTarget : TSignal;
  protected
    procedure   Wrapped;                                                                         override;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPulses = class( TMod)
  strict private
  const
    i_frequency      = 0;
    i_range          = 1;
    i_dist           = 2;
    i_distmod        = 3;
    i_distmodamt     = 4;
    i_lambda         = 5;
    i_rndmode        = 6;
    i_syncmode       = 7;
    i_outputtype     = 8;
  const
    o_out            = 0;
  const
    m_rnd_lin        = 0;
    m_rnd_exp        = 1;
  const
    m_sync_async     = 0;
    m_sync_sync      = 1;
  private
    FPhase         : TSignal;
    FState         : TPulsesState;
    FPulseCounter  : Integer;
    FPeriodCounter : Integer;
    FLedCounter    : Integer;
    FOldFrequency  : TSignal;
    FOldLambda     : TSignal;
    FOldDist       : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModRandomWalkLfo = class( TModBaseLfo)
  strict private
  const
    i_range       =  0;
    i_frequency   =  1;
    i_fm          =  2;
    i_fmlevel     =  3;
    i_distance    =  4;
    i_mute        =  5;
    i_color       =  6;
    i_colormod    =  7;
    i_colormodamt =  8;
    i_outputtype  =  9;
    i_lambda      = 10;
    i_mode        = 11;
  const
    o_outx        =  0;
    o_outy        =  1;
    o_outz        =  2;
  const
    m_lin         =  0;
    m_exp         =  0;
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAttractorLfo = class( TModBaseLfo)
  strict private
  const
    i_range       =  0;
    i_frequency   =  1;
    i_fm          =  2;
    i_fmlevel     =  3;
    i_mute        =  4;
    i_outputtype  =  5;
    i_mode        =  6;
  const
    o_outx        =  0;
    o_outy        =  1;
    o_outz        =  2;
  const
    m_lorenz      = 0;
    m_rossler     = 1;
  private
    FLorenz  : TLorenzOsc;
    FRossler : TRosslerOsc;
    FX       : TSignal;
    FY       : TSignal;
    FZ       : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   SampleRateChanged;                                                               override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoise = class( TMod)
  strict private
  const
    i_color       = 0;
    i_colormod    = 1;
    i_colormodamt = 2;
    i_mute        = 3;
  const
    o_out         = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseTrig = class( TMod)
  strict private
  const
    i_color       = 0;
    i_colormod    = 1;
    i_colormodamt = 2;
    i_trig        = 3;
    i_mute        = 4;
  const
    o_out         = 0;
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfo = class( TMod)
  strict private
  const
    i_color       = 0;
    i_colormod    = 1;
    i_colormodamt = 2;
    i_mute        = 3;
    i_outputtype  = 4;
  const
    o_out         = 0;
  private
    FOutValue : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfoTrig = class( TMod)
  strict private
  const
    i_color       = 0;
    i_colormod    = 1;
    i_colormodamt = 2;
    i_trig        = 3;
    i_mute        = 4;
    i_outputtype  = 5;
  const
    o_out         = 0;
  private
    FLastTrig : Boolean;
    FOutValue : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAverage = class( TMod)
  strict private
  const
    i_in          = 0;
    i_alpha       = 1;
    i_alphamod    = 2;
    i_alphamodamt = 3;
  const
    o_lp          = 0;
    o_hp          = 1;
  private
    FAlpha : TSignal;  // Alpha = System_Rate / ( 2 * PI * Freq + System_Rate) ... Fc := (log(b1))/(-2.0 * M_PI) * Fs; {b1 < 0.0432139182637722579772}
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSVF = class( TMod)
  strict private
  const
    i_in         = 0;
    i_frequency  = 1;
    i_q          = 2;
    i_distortion = 3;
    i_fm         = 4;
    i_fmlevel    = 5;
    i_inlevel    = 6;
    i_freq       = 7;
  const
    o_lp         = 0;
    o_bp         = 1;
    o_hp         = 2;
    o_br         = 3;
  private
    FFreq  : TSignal;
    FDamp  : TSignal;
    FDrive : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPinkFilter = class( TMod)
  strict private
  const
    i_in   = 0;
    i_mode = 1;
    i_gain = 2;
  const
    o_out  = 0;
  private
    FB0 : TSignal;
    FB1 : TSignal;
    FB2 : TSignal;
    FB3 : TSignal;
    FB4 : TSignal;
    FB5 : TSignal;
    FB6 : TSignal;
    FG  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModResonator = class( TMod)
  // http://www.katjaas.nl/complexintegrator/complexresonator.html
  strict private
  const
    i_freq      =  0;
    i_decay     =  1;
    i_in        =  2;
    i_fm        =  3;
    i_fmlevel   =  4;
    i_frequency =  5;
    i_cents     =  6;
    i_pm        =  7;
    i_pmlevel   =  8;
    i_dm        =  9;
    i_dmlevel   = 10;
    i_click     = 11;
    i_inlevel   = 12;
  const
    o_out       =  0;
  private
    FOldRe  : TSignal;
    FOldIm  : TSignal;
    FRadius : TSignal;
    FSin    : TSignal;
    FCos    : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFreqShifter = class( TMod)
  // https://github.com/swh/ladspa/blob/master/bode_shifter_1431.xml
  strict private
  const
    i_in         = 0;
    i_shift      = 1;
    i_shiftmod   = 2;
    i_shiftdepth = 3;
  const
    o_up    = 0;
    o_down  = 1;
  const
    c_coefficients : array[ 0 .. 99] of TSignal = (
      +0.0008103736, +0.0008457886, +0.0009017196, +0.0009793364,
      +0.0010798341, +0.0012044365, +0.0013544008, +0.0015310235,
      +0.0017356466, +0.0019696659, +0.0022345404, +0.0025318040,
      +0.0028630784, +0.0032300896, +0.0036346867, +0.0040788644,
      +0.0045647903, +0.0050948365, +0.0056716186, +0.0062980419,
      +0.0069773575, +0.0077132300, +0.0085098208, +0.0093718901,
      +0.0103049226, +0.0113152847, +0.0124104218, +0.0135991079,
      +0.0148917649, +0.0163008758, +0.0178415242, +0.0195321089,
      +0.0213953037, +0.0234593652, +0.0257599469, +0.0283426636,
      +0.0312667947, +0.0346107648, +0.0384804823, +0.0430224431,
      +0.0484451086, +0.0550553725, +0.0633242001, +0.0740128560,
      +0.0884368322, +0.1090816773, +0.1412745301, +0.1988673273,
      +0.3326528346, +0.9997730178, -0.9997730178, -0.3326528346,
      -0.1988673273, -0.1412745301, -0.1090816773, -0.0884368322,
      -0.0740128560, -0.0633242001, -0.0550553725, -0.0484451086,
      -0.0430224431, -0.0384804823, -0.0346107648, -0.0312667947,
      -0.0283426636, -0.0257599469, -0.0234593652, -0.0213953037,
      -0.0195321089, -0.0178415242, -0.0163008758, -0.0148917649,
      -0.0135991079, -0.0124104218, -0.0113152847, -0.0103049226,
      -0.0093718901, -0.0085098208, -0.0077132300, -0.0069773575,
      -0.0062980419, -0.0056716186, -0.0050948365, -0.0045647903,
      -0.0040788644, -0.0036346867, -0.0032300896, -0.0028630784,
      -0.0025318040, -0.0022345404, -0.0019696659, -0.0017356466,
      -0.0015310235, -0.0013544008, -0.0012044365, -0.0010798341,
      -0.0009793364, -0.0009017196, -0.0008457886, -0.0008103736
    );
  private
    FDelay : array[ 0 .. 255] of TSignal;
    FDPtr  : Integer;
    FPhase : TSignal;
    FDelta : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAR = class( TMod)
  strict private
  const
    i_trig    = 0;
    i_attack  = 1;
    i_release = 2;
    i_gain    = 3;
    i_sigin   = 4;
    i_mute    = 5;
  const
    o_out     = 0;
    o_sigout  = 1;
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FEnvValue     : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvARRetrig = class( TMod)
  strict private
  const
    i_trig    = 0;
    i_attack  = 1;
    i_release = 2;
    i_gain    = 3;
    i_sigin   = 4;
    i_mute    = 5;
  const
    o_out     = 0;
    o_sigout  = 1;
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
    FEnvValue     : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAHD = class( TMod)
  strict private
  const
    i_trig       =  0;
    i_mode       =  1;
    i_ashape     =  2;
    i_dshape     =  3;
    i_attack     =  4;
    i_hold       =  5;
    i_decay      =  6;
    i_gain       =  7;
    i_sigin      =  8;
    i_mute       =  9;
    i_arange     = 10;
    i_hrange     = 11;
    i_drange     = 12;
    i_amod       = 13;
    i_hmod       = 14;
    i_dmod       = 15;
    i_amodneg    = 16;
    i_hmodneg    = 17;
    i_dmodneg    = 18;
    i_amodlevel  = 19;
    i_hmodlevel  = 20;
    i_dmodlevel  = 21;
    i_outputtype = 22;
  const
    o_out        =  0;
    o_sigout     =  1;
  private
    FData      : TSignalArray;
    FState     : TAHDState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvADSR = class( TMod)
  strict private
  const
    i_trig       =  0;
    i_mode       =  1;
    i_ashape     =  2;
    i_dshape     =  3;
    i_rshape     =  4;
    i_attack     =  5;
    i_decay      =  6;
    i_sustain    =  7;
    i_release    =  8;
    i_gain       =  9;
    i_sigin      = 10;
    i_mute       = 11;
    i_arange     = 12;
    i_drange     = 13;
    i_rrange     = 14;
    i_amod       = 15;
    i_dmod       = 16;
    i_smod       = 17;
    i_rmod       = 18;
    i_amodneg    = 19;
    i_dmodneg    = 20;
    i_smodneg    = 21;
    i_rmodneg    = 22;
    i_amodlevel  = 23;
    i_dmodlevel  = 24;
    i_smodlevel  = 25;
    i_rmodlevel  = 26;
    i_outputtype = 27;
  const
    o_out        =  0;
    o_sigout     =  1;
  private
    FData      : TSignalArray;
    FState     : TADSRState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSequencer = class( TMod)
  strict private
  const
    i_trig   =  0;
    i_reset  =  1;
    i_dir    =  2;
    i_chain  =  3;
    i_rnd    =  4;
    i_active =  5;
    i_step1  =  6;   // Keep steps consecutive
    i_step2  =  7;
    i_step3  =  8;
    i_step4  =  9;
    i_step5  = 10;
    i_step6  = 11;
    i_step7  = 12;
    i_step8  = 13;
    i_step9  = 14;
    i_step10 = 15;
    i_step11 = 16;
  const
    o_out    =  0;   // Keep trigs comsecutive
    o_trig1  =  1;
    o_trig2  =  2;
    o_trig3  =  3;
    o_trig4  =  4;
    o_trig5  =  5;
    o_trig6  =  6;
    o_trig7  =  7;
    o_trig8  =  8;
    o_trig9  =  9;
    o_trig10 = 10;
    o_trig11 = 11;
  private
    FOutValue : TSignal;
    FSteps    : Integer;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModSeqSeq = class( TMod)
  strict private
  const
    i_res      =  0;
    i_trig     =  1;
    i_dir      =  2;
    i_resmod   =  3;
  const
    o_outres   =  0;
    o_outtrig  =  1;
    o_outdir   =  2;
    o_out1     =  3;
    o_out2     =  4;
    o_out3     =  5;
    o_out4     =  6;
    o_out5     =  7;
    o_out6     =  8;
    o_out7     =  9;
    o_out8     = 10;
    o_out9     = 11;
    o_out10    = 12;
    o_out11    = 13;
    o_out12    = 14;
    o_out13    = 15;
    o_out14    = 16;
    o_out15    = 17;
    o_out16    = 18;
  const
    c_clkdelay = 17;
  const
    m_res1     =  0;
    m_res2     =  1;
    m_res3     =  2;
  private
    FState           : Word;
    FSteps           : Integer;
    FSequenceCounter : Integer;
    FPeriodCounter   : Integer;
    FOldRes          : Boolean;
    FOldTrig         : Boolean;
    FClkCnt          : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModSeq16 = class( TMod)
  strict private
  const
    i_trig       =  0;
    i_res        =  1;
    i_dir        =  2;
    i_chain      =  3;
    i_rnd        =  4;
    i_active     =  5;
    i_mode       =  6;
    i_curveup    =  7;
    i_curvedown  =  8;
    i_xfade      =  9;
    i_step1      = 10;   // Keep steps consecutive
    i_step2      = 11;
    i_step3      = 12;
    i_step4      = 13;
    i_step5      = 14;
    i_step6      = 15;
    i_step7      = 16;
    i_step8      = 17;
    i_step9      = 18;
    i_step10     = 19;
    i_step11     = 20;
    i_step12     = 21;
    i_step13     = 22;
    i_step14     = 23;
    i_step15     = 24;
    i_step16     = 25;
  const
    o_out        =  0;
    o_outres     =  1;
    o_outtrig    =  2;
    o_outdir     =  3;
  const
    m_lin44      =  0;   // linear [-4, 4]
    m_db4        =  1;   // exp --140, 12] dB
    m_lin11      =  2;   // linear [-1, 1]
    m_db0        =  3;   // exp --140, 0] dB
    m_lin04      =  4;   // linear [0, 4]
    m_lin01      =  5;   // linear [0, 1]
    m_note       =  6;   // note values ([0,127] nn - MiddleNote) / NOTE_SCALING
  private
    FOutValue   : TSignal;
    FReset      : Boolean;
    FNextValue  : TSignal;
    FForwards   : Boolean;
    FActive     : Boolean;
    FChainValue : TSignal;
    FSteps      : Integer;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FOldRes     : Boolean;
    FOldActive  : Boolean;
    FTicks      : Integer;
    FPeriodTime : Integer;
  private
    procedure   SampleInputs;
    procedure   MakeNextValue;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModTextSequencer = class( TMod)
  strict private
  const
    i_trig    = 0;
    i_reset   = 1;
    i_dir     = 2;
    i_mode    = 3;
    i_chain   = 4;
  const
    o_out     = 0;
    o_gateout = 1;
  private
    FOutValue   : TSignal;
    FStepCount  : Integer;
    FStepValues : TSignalArray;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FOldRes     : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    procedure   SetStepValues( const aValues: TSignalArray);
  end;

  TModSeqStep = class( TMod)
  strict private
  const
    i_triginfwd  = 0;
    i_triginbwd  = 1;
    i_chain      = 2;
    i_reset      = 3;
    i_dir        = 4;
    i_note       = 5;
    i_time       = 6;
  const
    o_trigoutfwd = 0;
    o_trigoutbwd = 1;
    o_out        = 2;
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModSeqClockStep = class( TMod)
  strict private
  const
    i_triginfwd  = 0;
    i_triginbwd  = 1;
    i_chain      = 2;
    i_reset      = 3;
    i_dir        = 4;
    i_clock      = 5;
    i_note       = 6;
    i_counts     = 7;
  const
    o_trigoutfwd = 0;
    o_trigoutbwd = 1;
    o_out        = 2;
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModPattern = class( TMod)
  strict private
  const
    i_res   = 0;
    i_trig  = 1;
    i_step1 = 2;
    i_step2 = 3;
    i_step3 = 4;
    i_step4 = 5;
    i_step5 = 6;
    i_step6 = 7;
    i_step7 = 8;
    i_step8 = 9;
  const
    i_last  = 9;
  const
    o_out   = 0;
  private
    FPosition : Integer;
    FCounter  : Integer;
    FOldTrig  : Boolean;
    FOldRes   : Boolean;
    FOutVal   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFriemels = class( TMod)
  strict private
  const
    i_res   = 0;
  const
    i_last  = 9;
  const
    o_out   = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
  end;

  TModCsvData = class( TMod)
  strict private
  const
    i_reset =  0;
    i_trig  =  1;
    i_dir   =  2;
    i_mode  =  3;
  const
    o_out1  =  0;
    o_out2  =  1;
    o_out3  =  2;
    o_out4  =  3;
    o_out5  =  4;
    o_out6  =  5;
    o_out7  =  6;
    o_out8  =  7;
    o_out9  =  8;
    o_out10 =  9;
    o_out11 = 10;
    o_out12 = 11;
    o_out13 = 12;
    o_out14 = 13;
    o_out15 = 14;
    o_out16 = 15;
  private
    FMemory   : array[ 0 .. 15] of TSignalArray;
    FComments : TStringList;
    FFileName : string;
    FChannels : Integer;
    FCounter  : Integer;
    FSteps    : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
    FMode     : Integer;
    FStopped  : Boolean;
  private
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoTick;                                                                          override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModKleeBit = class( TMod)
  strict private
  const
    i_instep       =  0;
    i_inload       =  1;
    i_loopmode     =  2;
    i_inanalog     =  3;
    i_inbus1       =  4;
    i_inbus2       =  5;
    i_inbus3       =  6;
    i_pre1         =  7;
    i_pre2         =  8;
    i_pre3         =  9;
    i_pre4         = 10;
    i_pre5         = 11;
    i_pre6         = 12;
    i_pre7         = 13;
    i_pre8         = 14;
    i_level1       = 15;
    i_level2       = 16;
    i_level3       = 17;
    i_level4       = 18;
    i_level5       = 19;
    i_level6       = 20;
    i_level7       = 21;
    i_level8       = 22;
    i_bussel1      = 23;
    i_bussel2      = 24;
    i_bussel3      = 25;
    i_bussel4      = 26;
    i_bussel5      = 27;
    i_bussel6      = 28;
    i_bussel7      = 29;
    i_bussel8      = 30;
    i_inbit        = 31;
    i_invertmode   = 32;
    i_mainlevel    = 33;
    i_loopinvert   = 34;
    i_invertinvert = 35;
    i_mainlevelmod = 36;
  const
    o_outanalog    =  0;
    o_outbus1      =  1;
    o_outbus2      =  2;
    o_outbus3      =  3;
    o_outbit       =  4;
    o_outloop      =  5;
    o_outinvert    =  6;
  private
    FPattern    : Byte;
    FOldStep    : Boolean;
    FOldLoad    : Boolean;
    FBus1       : Boolean;
    FBus2       : Boolean;
    FBus3       : Boolean;
    FAnalog     : TSignal;
    FLoopMode   : Boolean;
    FInvertMode : Boolean;
  private
    procedure   LoadPreset;
    procedure   Step;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModSeqRandom = class( TMod)
  strict private
  const
    i_trig       = 0;
    i_changeone  = 1;
    i_scale      = 2;
    i_basenote   = 3;
    i_lownote    = 4;
    i_highnote   = 5;
    i_distmod    = 6;
    i_distmodamt = 7;
    i_dist       = 8;
    i_chain      = 9;
  const
    o_out        = 0;
  const
    s_count      = 0;
  private
    FValues       : array[ Byte] of TScaleNote;
    FPattern      : array[ Byte] of TScaleNote;
    FCount        : Integer;
    FScale        : Integer;
    FBaseNote     : TScaleNote;
    FLowNote      : TScaleNote;
    FHighNote     : TScaleNote;
    FNoteCount    : Integer;
    FScaleNotes   : array[ TScaleNote] of TScaleNote;
    FOldTrig      : Boolean;
    FOldOne       : Boolean;
    FPosition     : Byte;
    FDistribution : TSignal;
  private
    procedure   MakeScale;
    function    RandomNote( aPosition: Byte; aDistribution: TSignal): TScaleNote;
    procedure   ChangeScale( aScale: Integer);
    procedure   ChangeOneNote( aPosition: Byte);
    procedure   ChangeAllNotes;
    procedure   Step;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModMorse = class( TMod)
  strict private
  const
    i_reset = 0;
    i_clock = 1;
  const
    o_out   = 0;
  private
    FMessage     : string;
    FSymbols     : string;
    FSymbolIndex : Integer;
    FOldReset    : Boolean;
    FOldClock    : Boolean;
  private
    function    CharToSymbols( aChar: Char): string;
    procedure   MessageToSymbols;
    procedure   SetMessage( const aValue: string);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    Message: string read FMessage write SetMessage;
  end;

  TModGateSeq = class( TMod)
  strict private
  const
    i_trig    =  0;
    i_step1   =  1;
    i_step2   =  2;
    i_step3   =  3;
    i_step4   =  4;
    i_step5   =  5;
    i_step6   =  6;
    i_step7   =  7;
    i_step8   =  8;
    i_step9   =  9;
    i_step10  = 10;
    i_step11  = 11;
    i_step12  = 12;
    i_step13  = 13;
    i_step14  = 14;
    i_step15  = 15;
    i_step16  = 16;
    i_chain   = 17;
    i_res     = 18;
    i_active  = 19;
    i_dir     = 20;
  const
    o_out     =  0;
    o_outres  =  1;
    o_outtrig =  2;
    o_outdir  =  3;
  const
    s_steps  =  0;
  private
    FOutValue   : TSignal;
    FSteps      : Integer;
    FOldRes     : Boolean;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FForwards   : Boolean;
    FReset      : Boolean;
  private
    procedure   SampleInputs;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModMemory = class( TMod)
  private
    FMemoryL : TSignalArrayWrapper;
    FMemoryR : TSignalArrayWrapper;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
  end;

  TModPitchShift = class( TMod)
  strict private
  const
    i_in          = 0;
    i_shift       = 1;
    i_shiftfine   = 2;
    i_shiftmod    = 3;
    i_shiftmodamt = 4;
    i_mute        = 5;
  const
    o_out         = 0;
  const
    s_delay       = 0;
  private
    FDelayLine : TSignalArrayWrapper;
    FDelay     : TSignal;
    FLength    : Integer;
    FInPointer : Integer;
    FDelta     : TSignal;
    FPhase     : TSignal;
  private
    procedure   SetDelay( aValue: TSignal);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModDelay = class( TModMemory)
  strict private
  const
    i_in       = 0;
    i_delay    = 1;
  const
    o_out      = 0;
    o_outfixed = 1;
  const
    s_length   = 0;
  private
    FMemLength : TSignal;
    FInPointer : Integer;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModDelay8 = class( TModMemory)
  strict private
  const
    i_in       = 0;
    i_delay1   = 1;
    i_delay2   = 2;
    i_delay3   = 3;
    i_delay4   = 4;
    i_delay5   = 5;
    i_delay6   = 6;
    i_delay7   = 7;
    i_delay8   = 8;
  const
    o_outfixed = 0;
    o_out1     = 1;
    o_out2     = 2;
    o_out3     = 3;
    o_out4     = 4;
    o_out5     = 5;
    o_out6     = 6;
    o_out7     = 7;
    o_out8     = 8;
  const
    s_length   = 0;
  private
    FMemLength : TSignal;
    FInPointer : Integer;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModClockedDelay = class( TModMemory)
  strict private
  const
    i_in       = 0;
    i_clock    = 1;
  const
    o_out      = 0;
  const
    s_length   = 0;
  private
    FMemLength : TSignal;
    FInPointer : Integer;
    FOldClock  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModWave = class( TModMemory)
  strict private
  const
    o_outl = 0;
    o_outr = 1;
  private
    FFileName    : string;
    FDuration    : string;
    FIsStereo    : Boolean;
  private
    procedure   SetDuration( aValue: Cardinal);
    procedure   SwapMemoryFor( aMemory: TSignalArray);
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
  public
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModWavePlayer = class( TModWave)
  strict private
  const
    i_speed     = 0;
    i_position  = 1;
    i_res       = 2;
    i_mode      = 3;
    i_fm        = 4;
  const
    o_outl      = 0;
    o_outr      = 1;
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModRndGranulator = class( TMod)
  strict private
  const
    i_in                  =  0;
    i_mute                =  1;
    i_envtype             =  2;
    i_interonset          =  3;
    i_interonsetmod       =  4;
    i_interonsetmodamt    =  5;
    i_grainduration       =  6;
    i_graindurationmod    =  7;
    i_graindurationmodamt =  8;
    i_volume              =  9;
    i_volumemod           = 10;
    i_volumemodamt        = 11;
    i_pan                 = 12;
    i_panmod              = 13;
    i_panmodamt           = 14;
    i_record              = 15;
    i_rec                 = 16;
    i_delay               = 17;
    i_grainoffset         = 18;
  const
    o_outl                =  0;
    o_outr                =  1;
  const
    s_delaytime           =  0;
    s_graincount          =  1;
  private
    FGranulator       : TRndGranulator;
    FLocked           : Boolean;
    FDelayTime        : TSignal;
    FGrainCount       : Integer;
    FInterOnset       : TSignal;
    FInterOnsetMod    : TSignal;
    FGrainDuration    : TSignal;
    FGrainDurationMod : TSignal;
    FVolume           : TSignal;
    FVolumeMod        : TSignal;
    FPan              : TSignal;
    FPanMod           : TSignal;
    FRecording        : TSignal;
  private
    procedure   Recreate;
    procedure   SetDelayTime ( aValue: TSignal);
    procedure   SetGrainCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    DelayTime  : TSignal read FDelayTime  write SetDelayTime;
    property    GrainCount : Integer read FGrainCount write SetGrainCount;
  end;

  TModGranulator = class( TMod)
  strict private
  const
    i_in                  =  0;
    i_mute                =  1;
    i_envtype             =  2;
    i_interonset          =  3;
    i_interonsetmod       =  4;
    i_interonsetmodamt    =  5;
    i_grainduration       =  6;
    i_graindurationmod    =  7;
    i_graindurationmodamt =  8;
    i_volume              =  9;
    i_volumemod           = 10;
    i_volumemodamt        = 11;
    i_pan                 = 12;
    i_panmod              = 13;
    i_panmodamt           = 14;
    i_record              = 15;
    i_rec                 = 16;
    i_delay               = 17;
    i_grainoffset         = 18;
  const
    o_outl                =  0;
    o_outr                =  1;
  const
    s_delaytime           =  0;
    s_graincount          =  1;
  private
    FGranulator    : TGranulator;
    FLocked        : Boolean;
    FDelayTime     : TSignal;
    FGrainCount    : Integer;
  private
    procedure   Recreate;
    procedure   SetDelayTime ( aValue: TSignal);
    procedure   SetGrainCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    DelayTime  : TSignal read FDelayTime  write SetDelayTime;
    property    GrainCount : Integer read FGrainCount write SetGrainCount;
  end;

  TModSimpleGranulator = class( TMod)
  strict private
  const
    i_in                  =  0;
    i_mute                =  1;
    i_envtype             =  2;
    i_grainduration       =  3;
    i_graindurationmod    =  4;
    i_graindurationmodamt =  5;
    i_record              =  6;
    i_rec                 =  7;
    i_delay               =  8;
    i_grainoffset         =  9;
    i_trig                = 10;
  const
    o_out                 =  0;
  const
    s_delaytime           =  0;
    s_graincount          =  1;
  private
    FGranulator : TSimpleGranulator;
    FLocked     : Boolean;
    FDelayTime  : TSignal;
    FGrainCount : Integer;
    FOldTrig    : Boolean;
  private
    procedure   Recreate;
    procedure   SetDelayTime ( aValue: TSignal);
    procedure   SetGrainCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    DelayTime  : TSignal read FDelayTime  write SetDelayTime;
    property    GrainCount : Integer read FGrainCount write SetGrainCount;
  end;

  TModSampleAndHold = class( TMod)
  strict private
  const
    i_trig = 0;
    i_in   = 1;
  const
    o_out  = 0;
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMultiSAndH = class( TMod)
  strict private
  const
    i_trig  =  0;
    i_in1   =  1;
    i_in2   =  2;
    i_in3   =  3;
    i_in4   =  4;
    i_in5   =  5;
    i_in6   =  6;
    i_in7   =  7;
    i_in8   =  8;
    i_in9   =  9;
    i_in10  = 10;
    i_in11  = 11;
    i_in12  = 12;
  const
    o_out1  =  0;
    o_out2  =  1;
    o_out3  =  2;
    o_out4  =  3;
    o_out5  =  4;
    o_out6  =  5;
    o_out7  =  6;
    o_out8  =  7;
    o_out9  =  8;
    o_out10 =  9;
    o_out11 = 10;
    o_out12 = 11;
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMux = class( TMod)
  strict private
  const
    i_in0     =  0;
    i_in1     =  1;
    i_in2     =  2;
    i_in3     =  3;
    i_in4     =  4;
    i_in5     =  5;
    i_in6     =  6;
    i_in7     =  7;
    i_in8     =  8;
    i_in9     =  9;
    i_in10    = 10;
    i_in11    = 11;
    i_control = 12;
    i_xfade   = 13;
  const
    o_out     =  0;
  private
    FState : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDemux = class( TMod)
  strict private
  const
    i_in      =  0;
    i_control =  1;
    i_xfade   =  2;
  const
    o_out0    =  0;
    o_out1    =  1;
    o_out2    =  2;
    o_out3    =  3;
    o_out4    =  4;
    o_out5    =  5;
    o_out6    =  6;
    o_out7    =  7;
    o_out8    =  8;
    o_out9    =  9;
    o_out10   = 10;
    o_out11   = 11;
  private
    FState : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDigiMux = class( TMod)
  strict private
  const
    i_in0     =  0;
    i_in1     =  1;
    i_in2     =  2;
    i_in3     =  3;
    i_in4     =  4;
    i_in5     =  5;
    i_in6     =  6;
    i_in7     =  7;
    i_sel0    =  8;
    i_sel1    =  9;
    i_sel2    = 10;
  const
    o_out     =  0;
  private
    FState : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDigiDemux = class( TMod)
  strict private
  const
    i_in   = 0;
    i_sel0 = 1;
    i_sel1 = 2;
    i_sel2 = 3;
    i_hold = 4;
  const
    o_out0 = 0;
    o_out1 = 1;
    o_out2 = 2;
    o_out3 = 3;
    o_out4 = 4;
    o_out5 = 5;
    o_out6 = 6;
    o_out7 = 7;
  private
    FState : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTrackAndHold = class( TMod)
  strict private
  const
    i_trig = 0;
    i_in   = 1;
  const
    o_out  = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSwitch2to1 = class( TMod)
  strict private
  const
    i_control = 0;
    i_in1     = 1;
    i_in2     = 2;
  const
    o_out     = 0;
    o_out2    = 1;
  private
    FOn : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModXFade = class( TMod)
  strict private
  const
    i_control  = 0;
    i_in1      = 1;
    i_in2      = 2;
    i_position = 3;
    i_posmod   = 4;
    i_intype   = 5;
  const
    o_out      = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPan = class( TMod)
  strict private
  const
    i_control  = 0;
    i_in       = 1;
    i_position = 2;
    i_posmod   = 3;
    i_intype   = 4;
  const
    o_out1     = 0;
    o_out2     = 1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM4to1 = class( TMod)
  strict private
  const
    i_in      =  0;
    i_in1     =  1;
    i_in2     =  2;
    i_in3     =  3;
    i_in4     =  4;
    i_mute1   =  5;
    i_mute2   =  6;
    i_mute3   =  7;
    i_mute4   =  8;
    i_level1  =  9;
    i_level2  = 10;
    i_level3  = 11;
    i_level4  = 12;
    i_levmod1 = 13;
    i_levmod2 = 14;
    i_levmod3 = 15;
    i_levmod4 = 16;
  const
    o_out     =  0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM16to1 = class( TMod)
  strict private
  const
    i_in      =  0;
    i_in1     =  1;
    i_in2     =  2;
    i_in3     =  3;
    i_in4     =  4;
    i_in5     =  5;
    i_in6     =  6;
    i_in7     =  7;
    i_in8     =  8;
    i_in9     =  9;
    i_in10    = 10;
    i_in11    = 11;
    i_in12    = 12;
    i_in13    = 13;
    i_in14    = 14;
    i_in15    = 15;
    i_in16    = 16;
    i_lev1    = 17;
    i_lev2    = 18;
    i_lev3    = 19;
    i_lev4    = 20;
    i_lev5    = 21;
    i_lev6    = 22;
    i_lev7    = 23;
    i_lev8    = 24;
    i_lev9    = 25;
    i_lev10   = 26;
    i_lev11   = 27;
    i_lev12   = 28;
    i_lev13   = 29;
    i_lev14   = 30;
    i_lev15   = 31;
    i_lev16   = 32;
    i_mute    = 33;
  const
    o_out     =  0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixS2to1 = class( TMod)
  strict private
  const
    i_inl     =  0;
    i_inr     =  1;
    i_inl1    =  2;
    i_inr1    =  3;
    i_inl2    =  4;
    i_inr2    =  5;
    i_mute1   =  6;
    i_mute2   =  7;
    i_level1  =  8;
    i_level2  =  9;
    i_levmod1 = 10;
    i_levmod2 = 11;
  const
    o_outl    =  0;
    o_outr    =  1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixT1to1 = class( TMod)
  strict private
  const
    i_inl     = 0;
    i_inm     = 1;
    i_inr     = 2;
    i_inl1    = 3;
    i_inm1    = 4;
    i_inr1    = 5;
    i_mute1   = 6;
    i_level1  = 7;
    i_levmod1 = 8;
  const
    o_outl    = 0;
    o_outm    = 1;
    o_outr    = 2;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixQ1to1 = class( TMod)
  strict private
  const
    i_inll    =  0;
    i_inlr    =  1;
    i_inrl    =  2;
    i_inrr    =  3;
    i_inll1   =  4;
    i_inlr1   =  5;
    i_inrl1   =  6;
    i_inrr1   =  7;
    i_mute1   =  8;
    i_level1  =  9;
    i_levmod1 = 10;
  const
    o_outll    = 0;
    o_outlr    = 1;
    o_outrl    = 2;
    o_outrr    = 3;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM6to6 = class( TMod)
  strict private
  const
    i_in1       =  0;
    i_in2       =  1;
    i_in3       =  2;
    i_in4       =  3;
    i_in5       =  4;
    i_in6       =  5;
    i_mute1     =  6;
    i_mute2     =  7;
    i_mute3     =  8;
    i_mute4     =  9;
    i_mute5     = 10;
    i_mute6     = 11;
    i_inchcol1  = 12;
    i_inchcol2  = 13;
    i_inchcol3  = 14;
    i_inchcol4  = 15;
    i_inchcol5  = 16;
    i_inchcol6  = 17;
    i_levelr1c1 = 18;
    i_levelr1c2 = 19;
    i_levelr1c3 = 20;
    i_levelr1c4 = 21;
    i_levelr1c5 = 22;
    i_levelr1c6 = 23;
    i_levelr2c1 = 24;
    i_levelr2c2 = 25;
    i_levelr2c3 = 26;
    i_levelr2c4 = 27;
    i_levelr2c5 = 28;
    i_levelr2c6 = 29;
    i_levelr3c1 = 30;
    i_levelr3c2 = 31;
    i_levelr3c3 = 32;
    i_levelr3c4 = 33;
    i_levelr3c5 = 34;
    i_levelr3c6 = 35;
    i_levelr4c1 = 36;
    i_levelr4c2 = 37;
    i_levelr4c3 = 38;
    i_levelr4c4 = 39;
    i_levelr4c5 = 40;
    i_levelr4c6 = 41;
    i_levelr5c1 = 42;
    i_levelr5c2 = 43;
    i_levelr5c3 = 44;
    i_levelr5c4 = 45;
    i_levelr5c5 = 46;
    i_levelr5c6 = 47;
    i_levelr6c1 = 48;
    i_levelr6c2 = 49;
    i_levelr6c3 = 50;
    i_levelr6c4 = 51;
    i_levelr6c5 = 52;
    i_levelr6c6 = 53;
    i_inr1c1    = 54;
    i_inr1c2    = 55;
    i_inr1c3    = 56;
    i_inr1c4    = 57;
    i_inr1c5    = 58;
    i_inr1c6    = 59;
    i_inr2c1    = 60;
    i_inr2c2    = 61;
    i_inr2c3    = 62;
    i_inr2c4    = 63;
    i_inr2c5    = 64;
    i_inr2c6    = 65;
    i_inr3c1    = 66;
    i_inr3c2    = 67;
    i_inr3c3    = 68;
    i_inr3c4    = 69;
    i_inr3c5    = 70;
    i_inr3c6    = 71;
    i_inr4c1    = 72;
    i_inr4c2    = 73;
    i_inr4c3    = 74;
    i_inr4c4    = 75;
    i_inr4c5    = 76;
    i_inr4c6    = 77;
    i_inr5c1    = 78;
    i_inr5c2    = 79;
    i_inr5c3    = 80;
    i_inr5c4    = 81;
    i_inr5c5    = 82;
    i_inr5c6    = 83;
    i_inr6c1    = 84;
    i_inr6c2    = 85;
    i_inr6c3    = 86;
    i_inr6c4    = 87;
    i_inr6c5    = 88;
    i_inr6c6    = 89;
  const
    o_out1      =  0;
    o_out2      =  1;
    o_out3      =  2;
    o_out4      =  3;
    o_out5      =  4;
    o_out6      =  5;
    o_outchrow1 =  6;
    o_outchrow2 =  7;
    o_outchrow3 =  8;
    o_outchrow4 =  9;
    o_outchrow5 = 10;
    o_outchrow6 = 11;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixAudio = class( TMod)
  strict private
  const
    i_insolo    =  0;
    i_inchain1l =  1;
    i_inchain1r =  2;
    i_inchain2l =  3;
    i_inchain2r =  4;
    i_in1       =  5;
    i_in2       =  6;
    i_in3       =  7;
    i_in4       =  8;
    i_in5       =  9;
    i_in6       = 10;
    i_in7       = 11;
    i_in8       = 12;
    i_pan1      = 13;
    i_pan2      = 14;
    i_pan3      = 15;
    i_pan4      = 16;
    i_pan5      = 17;
    i_pan6      = 18;
    i_pan7      = 19;
    i_pan8      = 20;
    i_vol11     = 21;
    i_vol12     = 22;
    i_vol13     = 23;
    i_vol14     = 24;
    i_vol15     = 25;
    i_vol16     = 26;
    i_vol17     = 27;
    i_vol18     = 28;
    i_vol21     = 29;
    i_vol22     = 30;
    i_vol23     = 31;
    i_vol24     = 32;
    i_vol25     = 33;
    i_vol26     = 34;
    i_vol27     = 35;
    i_vol28     = 36;
    i_mute1     = 37;
    i_mute2     = 38;
    i_mute3     = 39;
    i_mute4     = 40;
    i_mute5     = 41;
    i_mute6     = 42;
    i_mute7     = 43;
    i_mute8     = 44;
    i_solo1     = 45;
    i_solo2     = 46;
    i_solo3     = 47;
    i_solo4     = 48;
    i_solo5     = 49;
    i_solo6     = 50;
    i_solo7     = 51;
    i_solo8     = 52;
  const
    o_outsolo   =  0;
    o_out1l     =  1;
    o_out1r     =  2;
    o_out2l     =  3;
    o_out2r     =  4;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModQuantize = class( TMod)
  strict private
  const
    i_qsteps = 0;
    i_in     = 1;
    i_mute   = 2;
    i_level  = 3;
  const
    o_out    = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModHardClip = class( TMod)
  strict private
  const
    i_in        = 0;
    i_mute      = 1;
    i_lowlevel  = 2;
    i_highlevel = 3;
    i_outamp    = 4;
  const
    o_out       = 0;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTransInv = class( TMod)
  strict private
  const
    i_in1       = 0;
    i_in2       = 1;
    i_in3       = 2;
    i_in4       = 3;
    i_amtpivot  = 4;
    i_modulo    = 5;
    i_mode      = 6;
  const
    o_out1      = 0;
    o_out2      = 1;
    o_out3      = 2;
    o_out4      = 3;
  const
    m_transpose = 0;
    m_invert    = 1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModScala = class( TMod)
  strict private
  const
    i_in  = 0;
  const
    i_out = 0;
  private
    FMappingFileName : string;
    FScaleFileName   : string;
    FScale           : TScalaScale;
    FLastNote        : Integer;
    FlastOut         : TSignal;
  private
    procedure   SetMappingFileName( const aValue: string);
    procedure   SetScaleFileName  ( const aValue: string);
    function    Lookup( aValue: TSignal): TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    MappingFileName : string read FMappingFileName write SetMappingFileName;
    property    ScaleFileName   : string read FScaleFileName   write SetScaleFileName;
  end;

  TModTerraGen = class( TMod)
  strict private
  const
    i_x              =  0;
    i_y              =  1;
    i_mute           =  2;
    i_algo           =  3;
  const
    o_out            =  0;
  const
    m_off            =  0; // Output is zero
    m_inx            =  1; // Just x input
    m_iny            =  2; // Just y input
    m_xy             =  3; // Just y input
    m_curtis_roads   =  4; // Curtis roads standdard example
    m_bass           =  5; // all others from http://mymbs.mobeard.org/~pfisher/FOV2-0010016C/FOV2-0010016E/FOV2-001001A3/chapters/26mikelson/index%20.html
    m_lead1          =  6;
    m_hollow_lead    =  7;
    m_pulsar         =  8;
    m_flat_growl     =  9;
    m_atan           = 10; // except this one, standard arc tangent function
    m_softclip       = 11; // except these, these are from http://music.columbia.edu/cmc/musicandcomputers/chapter4/04_06.php
    m_third          = 12;
    m_cheb2          = 13;
    m_cheb3          = 14;
    m_cheb4          = 15;
    m_cheb5          = 16;
    m_cheb6          = 17;
    m_cheb7          = 18;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModChebGen = class( TMod)
  strict private
  const
    i_in    =  0;
    i_wmode =  1;
  const
    o_out2  =  0;
    o_out3  =  1;
    o_out4  =  2;
    o_out5  =  3;
    o_out6  =  4;
    o_out7  =  5;
    o_out8  =  6;
    o_out9  =  7;
    o_out10 =  8;
    o_out11 =  9;
    o_out12 = 10;
    o_out13 = 11;
  const
    m_sine  =  0;
    m_tri   =  1;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPoetry = class( TMod)
  strict private
  const
    i_in         = 0;
    i_newinput   = 1;
    i_newoutput  = 2;
  const
    o_out        = 0;
  const
    s_history    = 0;
    s_seedlen    = 1;
  private
    FLocked       : Boolean;
    FGenerator    : TPoetryGenerator;
    FHistory      : Integer;
    FSeedLen      : Integer;
    FEpsilon      : TSignal;
    FOldNewInput  : Boolean;
    FOldNewOutput : Boolean;
  private
    procedure   Recreate;
    procedure   SetHistory( aValue: Integer);
    procedure   SetSeedLen( aValue: Integer);
  private
    property    History : Integer read FHistory write SetHistory;
    property    SeedLen : Integer read FSeedLen write SetSeedLen;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPluck = class( TMod)
  strict private
  const
    i_pluck       =  0;
    i_plucklevel  =  1;
    i_feedback    =  2;
    i_brightness  =  3;
    i_brighctrl   =  4;
    i_bclevel     =  5;
    i_fm          =  6;
    i_fmlevel     =  7;
    i_frequency   =  8;
    i_cents       =  9;
    i_mute        = 10;
    i_noise       = 11;
    i_noisemod    = 12;
    i_noisedepth  = 13;
    i_freq        = 14;
  const
    o_out         =  0;
  private
    FMemory      : TSignalArrayWrapper;
    FDelayTime   : TSignal;
    FInPointer   : Integer;
    FOutValue    : TSignal;
    FOldFeedback : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDisplay = class( TMod)
  strict private
  const
    i_in1   = 0;
    i_in2   = 1;
    i_mode  = 2;
    i_speed = 3;
  private
    FMax1 : TSignal;
    FMax2 : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TMod_O_S_C = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModSync = class( TMod_O_S_C)
  strict private
  const
    i_sync = 0;
  const
    o_sync = 0;
  private
    FOscBytes    : TBytes;
    FSync        : Boolean;
    FOldSync     : Boolean;
    FLedCounter  : Integer;
    FSyncCounter : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket);                    override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModOscMessage = class( TMod_O_S_C)
  strict private
  const
    i_trigger = 0;
    i_input   = 1;
  const
    o_out     = 0;
  private
    FAddress    : string;
    FValue      : Single;
    FValuePtr   : PSingle;
    FOscBytes   : TBytes;
    FOldTrigger : Boolean;
    FLedCounter : Integer;
  private
    procedure   Recreate;
    procedure   SetAddress( const aValue: string);
    procedure   SetValue( const aValue: Single);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket);                    override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Address : string read FAddress write SetAddress;
    property    Value   : Single read FValue   write SetValue;
  end;

  TModMidiIn = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModMidiNoteIn = class( TModMidiIn)
  strict private
  const
    i_ch     = 0;
  const
    o_gate   = 0;
    o_note   = 1;
    o_velon  = 2;
    o_veloff = 3;
  private
    FChannel     : Byte;
    FLastNote    : Byte;
    FLastChannel : Byte;
    FOldChannel  : Byte;
    FGate        : Boolean;
    FNote        : TSignal;
    FVelocityOn  : TSignal;
    FVelocityOff : TSignal;
  public
    procedure   NotesOff;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                                         override;
    procedure   DoSlowTick;                                                                                    override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);                         override;
    function    InputName ( anIndex: Integer): string;                                                         override;
    function    OutputName( anIndex: Integer): string;                                                         override;
  end;

  TModMidiCCIn = class( TModMidiIn)
  strict private
  const
    i_ch    = 0;
    i_cc    = 1;
  const
    o_gate  = 0;
    o_value = 1;
  private
    FChannel     : Byte;
    FCC          : Byte;
    FGate        : Boolean;
    FValue       : TSignal;
    FCounter     : Integer;
    FLedCounter  : Integer;
  public
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                                         override;
    procedure   DoSlowTick;                                                                                    override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);                         override;
    function    InputName ( anIndex: Integer): string;                                                         override;
    function    OutputName( anIndex: Integer): string;                                                         override;
  end;

  TModMidiOut = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModMidiNoteOut = class( TModMidiOut)
  strict private
  const
    i_ch     = 0;
    i_gate   = 1;
    i_note   = 2;
    i_velon  = 3;
    i_veloff = 4;
  private
    FLastNote    : Byte;
    FLastChannel : Byte;
    FGate        : Boolean;
    FDidSend     : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModMidiCCOut = class( TModMidiOut)
  strict private
  const
    i_ch    = 0;
    i_cc    = 1;
    i_gate  = 2;
    i_value = 3;
  private
    FGate  : Boolean;
  public
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModMidiPlayer = class( TMod)
  strict private
  const
    i_trig      = 0;
    i_ch        = 1;
    i_reset     = 2;
    i_mode      = 3;
  const
    o_note      = 0;
    o_velon     = 1;
    o_veloff    = 2;
    o_gate      = 3;
    o_done      = 4;
  const
    m_once      = 0;
    m_loop      = 1;
  private
    FState           : TMidiPlayerState;
    FReader          : TWrenMidiReader;
    FEvent           : TWrenMidiTimedEvent;
    FFileName        : string;
    FReading         : Boolean;
    FCounter         : Integer;
    FChannel         : Byte;
    FOldTrig         : Boolean;
    FOldReset        : Boolean;
    FGateCount       : Integer;
    FNote            : TSignal;
    FVelocityOn      : TSignal;
    FVelocityOff     : TSignal;
    FDoneCounter     : Integer;
    FDoneLedCounter  : Integer;
    FResetLedCounter : Integer;
    FChOffset        : Integer;
  private
    procedure   SetFileName( const aValue: string);
    function    GetNextEvent: TWrenMidiTimedEvent;
  public
    procedure   NotesOff;
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   DoSlowTick;                                                                      override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModReverb = class( TMod)
  // http://members.chello.nl/t.koning8/kolindex.htm
  strict private
  const
    i_inl      = 0;
    i_inr      = 1;
    i_roomsize = 2;
    i_damp     = 3;
    i_wet      = 4;
    i_dry      = 5;
    i_width    = 6;
    i_mode     = 7;
    i_mute     = 8;
    i_freezein = 9;
  const
    o_outl     = 0;
    o_outr     = 1;
  private
    FReverb : TReverb;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModGVerb = class( TMod)
  // https://github.com/swh/lv2/tree/master/gverb
  strict private
  const
    i_inl             =  0;
    i_inr             =  1;
    i_roomsize        =  2;
    i_roomsizemod     =  3;
    i_roomsizemodlev  =  4;
    i_revtime         =  5;
    i_damping         =  6;
    i_inputbw         =  7;
    i_earlylevel      =  8;
    i_taillevel       =  9;
    i_drylevel        = 10;
    i_inputlevel      = 11;
    i_mute            = 12;
    i_revtimemod      = 13;
    i_dampingmod      = 14;
    i_inputbwmod      = 15;
    i_earlylevelmod   = 16;
    i_taillevelmod    = 17;
    i_drylevelmod     = 18;
  const
    o_outl            =  0;
    o_outr            =  1;
  const
    gverb_maxroomsize =  0;
    gverb_spread      =  1;
    gverb_mode        =  2;
  private
    FReverb       : TGVerb;
    FRecreateFlag : Boolean;
    FMaxRoomSize  : TSignal;
    FRoomSize     : TSignal;
    FRevTime      : TSignal;
    FDamping      : TSignal;
    FSpread       : TSignal;
    FInputBW      : TSignal;
    FEarlyLevel   : TSignal;
    FTailLevel    : TSignal;
    FDryLevel     : TSignal;
    FInputLevel   : TSignal;
    FMode         : Integer;
  private
    procedure   ReCreateReverb;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModEnvelope = class( TMod)
  // http://www.musicdsp.org/showone.php?id=97
  // Modified to have an inverted output
  strict private
  const
    i_in          = 0;
    i_attack      = 1;
    i_release     = 2;
    i_peakdecay   = 3;
    i_peakdecayon = 4;
  const
    o_out         = 0;
    o_invout      = 1;
  private
    FValue     : TSignal;
    FAttack    : TSignal;
    FRelease   : TSignal;
    FPeakValue : TSignal;
    FPeakDecay : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModCompressor = class( TMod)
  strict private
  const
    i_inl            = 0;
    i_inr            = 1;
    i_reflevel       = 2;
    i_gain           = 3;
    i_attack         = 4;
    i_release        = 5;
    i_bypass         = 6;
    i_preamp         = 7;
  const
    o_outl           = 0;
    o_outr           = 1;
    o_env            = 2;
  const
    m_bypass_off     = 0; // normal working mode
    m_bypass_byp     = 1; // bypass, outputs are same as inputs
  const
    COMP_MAX_GAIN =  64.0;
    COMP_MIN_GAIN = -32.0;
  private
    FEnvelope    : TSignal;
    FAttack      : TSignal;
    FRelease     : TSignal;
    FRefLevel    : TSignal;
    FPreGain     : TSignal;
    FGain        : TSignal;
    FBypass      : Integer;
    FOutL        : TSignal;
    FOutR        : TSignal;
    FLoopAmp     : TSignal;
    FMaxLoopAmp  : TSignal;
    FOldAttack   : TSignal;
    FOldRelease  : TSignal;
    FMinDbVal    : TSignal;
    FMaxDbVal    : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFormant = class( TMod)
  // http://www.musicdsp.org/showone.php?id=110
  strict private
  const
    i_in         = 0;
    i_vowel      = 1;
    i_vowelin    = 2;
    i_controlmod = 3;
  const
    o_out        = 0;
  private
    FCoeffStore    : array[ 0 .. 10] of TSignalArrayWrapper;
    FCoeffiecients : array[ 0 .. 10] of TSignal;
    FMemory        : TSignalArray;
    FVowel         : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFlangeChorus = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  strict private
  const
    i_in       = 0;
    i_mute     = 1;
    i_freeze   = 2;
    i_dry      = 3;
    i_wet      = 4;
    i_feedback = 5;
    i_rate     = 6;
    i_sweep    = 7;
    i_delay    = 8;
    i_freezein = 9;
  const
    o_out      = 0;
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedBack : TSignal;
    FRate     : TSignal;
    FSweep    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp        : Integer;
    ep1       : Integer;
    ep2       : Integer;
    step      : Integer;
    depth     : Integer;
    idelay    : Integer;
    min_sweep : Integer;
    max_sweep : Integer;
    wlSweep   : TLongWord;
    Buf       : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
  private
    procedure   Setup;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;


  TModPitchChange = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  strict private
  const
    i_in       = 0;
    i_mute     = 1;
    i_freeze   = 2;
    i_dry      = 3;
    i_wet      = 4;
    i_feedback = 5;
    i_rate     = 6;
    i_depth    = 7;
    i_delay    = 8;
    i_freezein = 9;
  const
    o_out      = 0;
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedback : TSignal;
    FRate     : TSignal;
    FDepth    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp         : Integer;
    ep1        : Integer;
    ep2        : Integer;
    ep3        : Integer;
    ep4        : Integer;
    idepth     : Integer;
    idelay     : Integer;
    min_sweep  : Integer;
    max_sweep  : Integer;
    step       : Word;
    xfade      : Integer;
    xfade_cnt  : Integer;
    active     : Integer;
    active_cnt : Integer;
    sweep_up   : Boolean;
    chanA      : Boolean;
    blendA     : TSignal;
    blendB     : TSignal;
    fadeA      : PFadeTable;
    fadeB      : PFadeTable;
    wlSweep    : TLongWord;
    Buf        : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
    Fade_Out   : TFadeTable;
    Fade_In    : TFadeTable;
  private
    procedure   Setup;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEuclideanRhythm = class( TMod)
  strict private
  const
    i_trig       = 0;
    i_pattern    = 1;
    i_patternmod = 2;
    i_pulses     = 3;
    i_steps      = 4;
  const
    o_out        = 0;
  private
    FBits     : TScaleNoteSet;
    FSteps    : Integer;
    FPattern  : Integer;
    FOldTrig  : Boolean;
    FPosition : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEuclideanScale = class( TMod)
  strict private
  const
    i_in            = 0;
    i_pattern       = 1;
    i_patternmod    = 2;
    i_patternmodamt = 3;
    i_steps         = 4;
  const
    o_out           = 0;
  private
    FBits       : TScaleNoteSet;
    FCount      : Integer;
    FSteps      : Integer;
    FPattern    : Integer;
    FScaleNotes : array[ Byte] of Byte;
    FNoteCount  : Integer;
  private
    procedure   MakeScale;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModXYScope = class( TMod)
  strict private
  const
    i_inx     = 0;
    i_iny     = 1;
    i_levx    = 2;
    i_levy    = 3;
    i_lockxy  = 5;
  const
    s_history = 0;
  private
    FHistory   : Integer;
    FDataQueue : TSignalPairFifo;
  private
    procedure   SetHistory( aValue: Integer);
    procedure   AddPoint( anX, anY: TSignal);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    History: Integer read FHistory write SetHistory;
  end;

  TFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FIndex        : Integer;
    FFTWInBuffer  : Pfftw_real;
    FFTWOutBuffer : Pfftw_complex;
    FFTWPlan      : fftw_plan;
  public
    procedure   Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    procedure   Finalize;
    procedure   Tick( anInValue: TSignal; var anOutValue: TComplex);
  end;
  TFFTDataArray = array of TFFTData;

  TIFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FIndex        : Integer;
    FFTWInBuffer  : Pfftw_complex;
    FFTWOutBuffer : Pfftw_real;
    FFTWPlan      : fftw_plan;
  public
    procedure   Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    procedure   Finalize;
    procedure   Tick( anInValue: TComplex; var anOutValue: TSignal);
  end;
  TIFFTDataArray = array of TIFFTData;

  TFFTHandler = class
  private
    FDestroying : Boolean;
    FBinCount   : Integer;
    FWindow     : TFFTWindow;
    FData       : TFFTDataArray;
  private
    function    GetWindowCount: Integer;
  public
    constructor Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    destructor  Destroy;                                                                         override;
    procedure   Tick( anInValue: TSignal; var anOutValue: TComplex);
  public
    property    WindowCount: Integer read GetWindowCount;
  end;

  TIFFTHandler = class
  private
    FDestroying : Boolean;
    FBinCount   : Integer;
    FWindow     : TFFTWindow;
    FData       : TIFFTDataArray;
  private
    function    GetWindowCount: Integer;
  public
    constructor Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    destructor  Destroy;                                                                         override;
    procedure   Tick( anInValue: TComplex; var anOutValue: TSignal);
  public
    property    WindowCount: Integer read GetWindowCount;
  end;

  TModVocoder = class( TMod)
  strict private
  const
    i_modulator = 0;
    i_carrier   = 1;
    i_mute      = 2;
  const
    o_out       = 0;
  private
    FBinCount         : Integer;
    FWindowCount      : Integer;
    FModulatorHandler : TFFTHandler;
    FCarrierHandler   : TFFTHandler;
    FOutHandler       : TIFFTHandler;
  private
    procedure   SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModConvoder = class( TMod)
  strict private
  const
    i_modulator    = 0;
    i_carrier      = 1;
    i_mute         = 2;
    i_modulatorlev = 3;
    i_carrierlev   = 4;
  const
    o_out          = 0;
  const
    s_size         = 0;
  const
    c_sizes : array[ 0 .. 11] of Integer = (
      64    ,
      128   ,
      256   ,
      512   ,
      1024  ,
      2048  ,
      4096  ,
      8192  ,
      16384 ,
      32768 ,
      65536 ,
      131072
    );
  private
    FSize             : Integer;
    FBinCount         : Integer;
    FWindowCount      : Integer;
    FAttenuation      : TSignal;
    FModulatorHandler : TFFTHandler;
    FCarrierHandler   : TFFTHandler;
    FOutHandler       : TIFFTHandler;
  private
    procedure   SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModFilterBank = class( TMod)
  strict private
  const
    i_in     =  0;
    i_fader1 =  1;
    i_fader2 =  2;
    i_fader3 =  3;
    i_fader4 =  4;
    i_fader5 =  5;
    i_fader6 =  6;
    i_fader7 =  7;
    i_fader8 =  8;
    i_fader9 =  9;
    i_fader10 = 10;
    i_fader11 = 11;
    i_fader12 = 12;
    i_fader13 = 13;
    i_fader14 = 14;
    i_fader15 = 15;
    i_fader16 = 16;
    i_fader17 = 17;
  const
    o_out     =  0;
  private
    FBinCount     : Integer;
    FWindowCount  : Integer;
    FInHandler    : TFFTHandler;
    FOutHandler   : TIFFTHandler;
    FCoefficients : TSignalArray;
    FCurrCoeff    : Integer;
    FFrequencies  : TSignalArray;
  private
    procedure   SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModTalkie = class( TMod)
  strict private
  const
    i_start    =  0;
    i_phrase   =  1;
    i_mute     =  2;
    i_fr       =  3;
    i_frmod    =  4;
    i_frmodamt =  5;
    i_sr       =  6;
    i_srmod    =  7;
    i_srmodamt =  8;
    i_bank     =  9;
    i_banksel  = 10;
    i_fm       = 11;
  const
    o_out      =  0;
    o_rdy      =  1;
  private
    FTalkie     : TTalkie;
    FDestroying : Boolean;
    FStart      : Boolean;
    FOldBankSel : TSignal;
    FBank       : TSignal;
    FPhrase     : TSignal;
    FSampleRate : TSignal;
    FFrameRate  : TSignal;
    FSpeaking   : Boolean;
    FDidRun     : Boolean;
  private
    procedure   TalkStarted( aSender: TObject);
    procedure   TalkDone   ( aSender: TObject);
    procedure   Say;
    procedure   NewFramRate  ( aValue: TSignal);
    procedure   NewSampleRate( aValue: TSignal);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAttractor = class( TMod)
  strict private
  const
    i_frequency   = 0;
    i_fm          = 1;
    i_fmlevel     = 2;
    i_cents       = 3;
    i_freq        = 4;
    i_mute        = 5;
    i_mode        = 6;
  const
    o_outx        = 0;
    o_outy        = 1;
    o_outz        = 2;
  const
    m_lorenz      = 0;
    m_rossler     = 1;
  private
    FLorenz  : TLorenzOsc;
    FRossler : TRosslerOsc;
    FX       : TSignal;
    FY       : TSignal;
    FZ       : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNeuron = class( TMod)
  strict private
  const
    i_inlevel   =  0;
    i_outlevel  =  1;
  const
    i_in1       =  2;          // Keeps ins consecutive
    i_in2       =  3;
    i_in3       =  4;
    i_in4       =  5;
    i_in5       =  6;
    i_in6       =  7;
    i_in7       =  8;
    i_in8       =  9;
    i_in9       = 10;
    i_in10      = 11;
    i_in11      = 12;
    i_in12      = 13;
  const
    i_in_first  = i_in1;
    i_in_last   = i_in12;
  const
    o_out1      =  0;          // Keeps outs consecutive
    o_out2      =  1;
    o_out3      =  2;
    o_out4      =  3;
    o_out5      =  4;
    o_out6      =  5;
    o_out7      =  6;
    o_out8      =  7;
    o_out9      =  8;
    o_out10     =  9;
    o_out11     = 10;
    o_out12     = 11;
  const
    o_out_first = o_out1;
    o_out_last  = o_out12;
  private
    FBrain      : TFannNetwork;
    FIns        : array of fann_type;
    FOuts       : array of fann_type;
    FFileName   : string;
    FDestroying : Boolean;
    FUpdating   : Boolean;
  private
    procedure   SetFileName( const aValue: string);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    FileName : string read FFileName write SetFileName;
  end;

  TModInterpreter = class( TMod)
  strict private
  const
    i_in1  = 0;
    i_in2  = 1;
    i_in3  = 2;
    i_in4  = 3;
    i_in5  = 4;
    i_in6  = 5;
    i_in7  = 6;
    i_in8  = 7;
  const
    o_out1 = 0;
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
    o_out5 = 4;
    o_out6 = 5;
    o_out7 = 6;
    o_out8 = 7;
  private
    FForth    : TForth;
    FFileName : string;
  private
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
    procedure   HandleOutData      ( const aSender: TObject; const aData: TExternalData);
    procedure   HandleForthLog     ( const aSender: TObject; const aMsg: string);
    procedure   HandleForthResponse( const aSender: TObject; const aMsg: string);
    procedure   HandleForthError   ( const aSender: TObject; const aMsg: string);
    procedure   SetCallbacks;
    procedure   ClearCallbacks;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;


  function  SignalToEnvRange  ( aValue: TSignal): TEnvRange;
  function  SignalToEnvShape  ( aValue: TSignal): TEnvShape;
  function  SignalToLfoRange  ( aValue: TSignal): TLfoRange;
  function  SignalToDelayRange( aValue: TSignal): TDelayRange;



implementation


var

  IssuedFftwWarning : Boolean = False;

  procedure FftwWarning( const aModuleType: string);
  begin
    if not IssuedFftwWarning
    then begin
      IssuedFftwWarning := True;
      MessageDlg(
        Format(
          'Could not initialize the %s module, the required FFTW DLL may not be present or can not be found.'^M^M +
          'This is a one time warning for all FFT based module types, they will all not work.'^M^M +
          'Please check the debug log (F5) after closing this window',
          [ aModuleType],
          AppLocale
        ),
        mtWarning,
        [ mbOK],
        0
      );
    end;
  end;


  function SignalToEnvRange( aValue: TSignal): TEnvRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := erFast
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := erMedium
    else Result := erSlow;
  end;

  function SignalToEnvShape( aValue: TSignal): TEnvShape;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := esLog
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := esLin
    else if AlmostEqual( aValue, 2, 0.45)
    then Result := esExp
    else Result := esS;
  end;

  function SignalToLfoRange( aValue: TSignal): TLfoRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := lsFast
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := lsMedium
    else if AlmostEqual( aValue, 2, 0.45)
    then Result := lsSlow
    else if AlmostEqual( aValue, 3, 0.45)
    then Result := lsBPM
    else Result := lsLin;
  end;

  function SignalToDelayRange( aValue: TSignal): TDelayRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := dsShort
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := dsMedium
    else Result := dsLong;
  end;

  function EuclideanPattern( aPattern, aSteps: Integer; out aCount: Integer): TScaleNoteSet;
  //
  // Idea from the wiki Bresenham and Eclidean RHythm pages,
  // Changed to let the pattern start with a one always,
  // the original is:
  //
  // function EuclideanRhythm( aPulses, aSteps: Integer): string;
  // var
  //   Error : Single;
  //   Slope : Single;
  //   i     : Integer;
  // begin
  //   Result := '';
  //   if aSteps > 0
  //   then begin
  //     Error := 0;
  //     Slope := aPulses / aSteps;
  //     for i := 0 to aSteps - 1
  //     do begin
  //       Error := Error + Slope;
  //       if Error >= 0.5
  //       then begin
  //         Result := Result + '1';
  //         Error  := Error  -  1;
  //       end
  //       else Result := Result + '0';
  //     end;
  //   end;
  // end;
  //
  // Then the integer version became:
  //
  // function EuclideanRhythm2( aPulses, aSteps: Integer): string;
  // var
  //   Error : Integer;
  //   i     : Integer;
  // begin
  //   Result := '';
  //   if aSteps > 0
  //   then begin
  //     Error := 0;
  //     for i := 0 to aSteps - 1
  //     do begin
  //       Error := Error + aPulses;
  //       if Error >= aSteps div 2
  //       then begin
  //         Result := Result + '1';
  //         Error  := Error - aSteps;
  //       end
  //       else Result := Result + '0';
  //     end;
  //   end;
  // end;
  //
  // Then finally the modification to make the pattern start with a one
  // always (and uses a result 'set of Byte' instead of a 'string' to avoid memory allocation issues):
   //
  // done : test a more compact one from : http://electro-music.com/forum/viewtopic.php?p=406480#406480
  // also see : http://www.pdpatchrepo.info/hurleur/euclid.pd (do not test for now)
  //
  /// * generates euclideian patterns
  //    c: current step number
  //    k: hits per bar
  //    n: bar length
  //    r: rotation
  //    returns 1 or 0 according to the euclidean pattern*/
  //
  // fun int euclide( int c, int k, int n, int r ) {
  //   return (((c + r) * k) % n) < k;
  //}
  //
  // 0 => int i;
  // while(i < 12) {
  //   euclide(i, 7, 12, 4) => int e;
  //   //declaring b just to print e
  //   <<<e>>> => int b;
  //   1 +=> i;
  //   100::ms => now;
  //}
  var
    Error : Integer;
    i     : Integer;
  begin
    Result := [];
    aCount := 0;
    if aSteps > 0
    then begin
      Error  := 0;
      for i := 0 to aSteps - 1
      do begin
        Error := Error + aPattern;
        if Error > 0
        then begin
          Result := Result + [ i];
          Error  := Error - aSteps;
          Inc( aCount);
        end;
      end;
    end;
  end;


{ ========
  TModConstant = class( TMod)
  public
}

    constructor TModConstant.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModConstant.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_value
      ];
    end;

    procedure   TModConstant.DoSlowTick; // override;
    begin
      if Round( FInputs[ i_mode]) = m_note
      then FOutputs[ o_out] := ( FInputs[ i_value]) - NoteNumberToUnits( MiddleNote)
      else FOutputs[ o_out] :=   FInputs[ i_value] * SignalToMute( FInputs[ i_mute]);
    end;

    function    TModConstant.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_value :  Result := 'value';
        i_mute  :  Result := 'mute';
        i_mode  :  Result := 'mode';
        else       Result := NO_INPUT;
      end;
    end;

    function    TModConstant.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out :  Result := 'out';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPad = class( TMod)
  public
}

    constructor TModPad.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPad.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_padx,
        i_pady
      ];
    end;

    procedure   TModPad.DoSlowTick; // override;
    var
      aType : Integer;
    begin
      aType := Round( FInputs[ i_outputtype]);
      FOutputs[ o_x] := SignalForOutputType( FInputs[ i_padx], aType);
      FOutputs[ o_y] := SignalForOutputType( FInputs[ i_pady], aType);
    end;

    function    TModPad.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_padx       :  Result := 'padx';
        i_pady       :  Result := 'pady';
        i_outputtype :  Result := 'outputtype';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModPad.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_x :  Result := 'x';
        o_y :  Result := 'y';
        else   Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModInverter = class( TMod)
  public
}

    constructor TModInverter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModInverter.DoSlowTick; // override;
    var
      i : Integer;
    begin
      for i := i_in1 to i_in4
      do FOutputs[ i - i_in1 + o_out1] := - FInputs[ i];
    end;

    function    TModInverter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1 : Result := 'in1';
        i_in2 : Result := 'in2';
        i_in3 : Result := 'in3';
        i_in4 : Result := 'in4';
        else    Result := NO_INPUT;
      end;
    end;

    function    TModInverter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAdder = class( TMod)
  public
}

    constructor TModAdder.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModAdder.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := i_in1 to i_in4          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInputs[ i] := 0;
      FDezipperMap := [
        i_mute
      ];
    end;

    procedure   TModAdder.DoSlowTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := 0.0;
      for i := i_in1 to i_in4
      do R := R + FInputs[ i];
      FOutputs[ o_out] := R * SignalToMute( FInputs[ i_mute]);
    end;

    function    TModAdder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_mute : Result := 'mute';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModAdder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMultiplier = class( TMod)
  public
}

    constructor TModMultiplier.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMultiplier.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := i_in1 to i_in4          // Make sure unused inputs are set to 1, so it'll 'work'
      do FInputs[ i] := 1;
      FDezipperMap := [
        i_mute
      ];
    end;

    procedure   TModMultiplier.DoSlowTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := SignalToMute( FInputs[ i_mute]);
      for i := i_in1 to i_in4
      do R := R * FInputs[ i];
      FOutputs[ o_out] := R;
    end;

    function    TModMultiplier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_mute : Result := 'mute';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModMultiplier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMultiMult = class( TMod)
  public
}

    constructor TModMultiMult.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMultiMult.SetDefaults; // override;
    begin
      FInputs[ i_mult] := 1; // set module transparent with unconnected multiplier
    end;

    procedure   TModMultiMult.DoSlowTick; // override;
    var
      Mult : TSignal;
    begin
      Mult              := FInputs[ i_mult];
      FOutputs[ o_out1] := Mult * FInputs[ i_in1];
      FOutputs[ o_out2] := Mult * FInputs[ i_in2];
      FOutputs[ o_out3] := Mult * FInputs[ i_in3];
      FOutputs[ o_out4] := Mult * FInputs[ i_in4];
      FOutputs[ o_out5] := Mult * FInputs[ i_in5];
      FOutputs[ o_out6] := Mult * FInputs[ i_in6];
      FOutputs[ o_out7] := Mult * FInputs[ i_in7];
      FOutputs[ o_out8] := Mult * FInputs[ i_in8];
    end;

    function    TModMultiMult.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_mult : Result := 'mult';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        i_in5  : Result := 'in5';
        i_in6  : Result := 'in6';
        i_in7  : Result := 'in7';
        i_in8  : Result := 'in8';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModMultiMult.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        o_out5 : Result := 'out5';
        o_out6 : Result := 'out6';
        o_out7 : Result := 'out7';
        o_out8 : Result := 'out8';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoteToDelay = class( TMod)
  public
}

    constructor TModNoteToDelay.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNoteToDelay.DoSlowTick; // override;
    begin
      FOutputs[ o_out] := 1 / LookupFrequency( FInputs[ i_in]);
    end;

    function    TModNoteToDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in : Result := 'in';
        else   Result := NO_INPUT;
      end;
    end;

    function    TModNoteToDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModZeroCounter = class( TMod)
  private
    FOldSign : Boolean;
    FTime    : Integer;
    FFreq    : TSignal;
  public
}

    constructor TModZeroCounter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModZeroCounter.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_threshold
      ];
    end;

    procedure   TModZeroCounter.DoTick; // override;
    var
      aValue : TSignal;
      aLevel : TSignal;
      aSign  : Boolean;
      aTime  : TSignal;
    begin
      Inc( FTime);
      aValue := FInputs[ i_in];
      aLevel := Abs( aValue);
      if aLevel > FInputs[ i_threshold]
      then begin
        FOutputs[ o_gate] := 1;
        aSign := SignalToLogic( aValue);
        if aSign <> FOldSign
        then begin
          FOldSign := aSign;
          FOutputs[ o_period] := LogicToSignal( aSign);
          if not aSign
          then begin
            aTime := FTime * System_Rate_Rec;
            if aTime > 0
            then FFreq := 1 / aTime;
            FTime := 0;
          end;
        end;
      end
      else FOutputs[ o_gate] := -1;
    end;

    procedure   TModZeroCounter.DoSlowTick; // override;
    begin
      FOutputs[ o_note] := FrequencyToUnits( FFreq) - NoteNumberToUnits( MiddleNote);
    end;

    procedure   TModZeroCounter.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, FOutputs[ o_gate]], AppLocale));
    end;

    function    TModZeroCounter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in        : Result := 'in';
        i_threshold : Result := 'threshold';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModZeroCounter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_period : Result := 'period';
        o_gate   : Result := 'gate';
        o_note   : Result := 'note';
        else       Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModMedian = class( TMod)
  private
    FData      : TMedianCalculator;
    FOldTrig   : Boolean;
  public
}

    constructor TModMedian.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FData   := TMedianCalculator.Create( 0);
      FIsFast := False;
      FIsSlow := True;
    end;

    destructor  TModMedian.Destroy; // override;
    var
      OldData  : TMedianCalculator;
    begin
      OldData := AtomicExchange( Pointer( FData), Nil);
      FreeAndNil( OldData);
      inherited;
    end;

    procedure   TModMedian.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      aBinSize : Integer;
      NewData  : TMedianCalculator;
      OldData  : TMedianCalculator;
      i        : Integer;
    begin
      if SameText( aName, 'binsize')
      then begin
        aBinSize := Round( aValue);
        if not Assigned( FData) or ( aBinSize <> FData.Count)
        then begin
          NewData := TMedianCalculator.Create( aBinSize);
          for i := 0 to FData.Count - 1
          do NewData.Append( FData.GetData);
          OldData := AtomicExchange( Pointer( FData), Pointer( NewData));
          FreeAndnil( OldData);
        end;
      end
      else inherited;
    end;

    function    TModMedian.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'binsize')
      then begin
        if Assigned( FData)
        then Result := FData.Count
        else Result := 0;
      end
      else Result := inherited;
    end;

    procedure   TModMedian.DoSlowTick; // override;
    var
      aTrig : Boolean;
    begin
      aTrig := SignalToLogic( FInputs[ i_trig]);
      if ( aTrig <> FOldTrig) and aTrig and Assigned( FData) and ( FData.Count > 0)
      then begin
        FData.Append( FInputs[ i_in]);
        FOutputs[ o_out ] := FData.Median;
        FOutputs[ o_high] := FData.Highest;
        FOutputs[ o_low ] := FData.Lowest;
      end;
      FOldTrig := aTrig;
    end;

    procedure   TModMedian.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModMedian.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in   : Result := 'in';
        i_trig : Result := 'trig';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModMedian.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out  : Result := 'out';
        o_high : Result := 'high';
        o_low  : Result := 'low';
        else     Result := NO_OUTPUT;
      end;
    end;

    function    TModMedian.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'binsize';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModCompare = class( TMod)
  public
}

    constructor TModCompare.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModCompare.DoSlowTick; // override;
    var
      in1      : TSignal;
      in2      : TSignal;
      outless  : TSignal;
      outequal : TSignal;
      outmore  : TSignal;
    begin
      in1      := FInputs[ i_in1];
      in2      := FInputs[ i_in2];

      outless  := LogicToSignal( False);
      outequal := LogicToSignal( False);
      outmore  := LogicToSignal( False);

      if AlmostEqual( in1, in2, FInputs[ i_range])
      then outequal := LogicToSignal( True)
      else if in1 < in2
      then outless := LogicToSignal( True)
      else outmore  := LogicToSignal( True);

      FOutputs[ o_outless ] := outless;
      FOutputs[ o_outequal] := outequal;
      FOutputs[ o_outmore ] := outmore;
    end;

    procedure   TModCompare.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/less/%f' , [ aPrefix, Name, FOutputs[ o_outless ]], AppLocale));
      aCallback( Self, Format( '%s/%s/equal/%f', [ aPrefix, Name, FOutputs[ o_outequal]], AppLocale));
      aCallback( Self, Format( '%s/%s/more/%f' , [ aPrefix, Name, FOutputs[ o_outmore ]], AppLocale));
    end;

    function    TModCompare.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1   : Result := 'in1';
        i_in2   : Result := 'in2';
        i_range : Result := 'range';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModCompare.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outless  : Result := 'outless';
        o_outequal : Result := 'outequal';
        o_outmore  : Result := 'outmore';
        else         Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModLogic = class( TMod)
  public
}

    constructor TModGate.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModGate.SetDefaults; // override;
    var
      i : Integer;
    begin
      case Round( FInputs[ i_function]) of
        m_and :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 1, so it'll 'work'
            do FInputs[ i] := 1;
          end;
        m_nand :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 1, so it'll 'work'
            do FInputs[ i] := 1;
          end;
        m_or :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        m_nor :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        m_xor :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        m_xnor :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        m_one :
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        else // none
          begin
            for i := i_in1 to i_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
      end;
    end;

    procedure   TModGate.DoSlowTick; // override;
    var
      i         : Integer;
      R         : TBinary;
      OneCounts : Integer;
      BitCount  : Integer;
      aFunction : Integer;
    begin
      aFunction := Round( FInputs[ i_function]);
      if aFunction <> FFunction
      then begin
        FFunction := aFunction;
        SetDefaults;
      end;
      BitCount := Round( FInputs[ i_width]) + 1;

      case aFunction of
        m_and :
          begin
            if BitCount = 1
            then begin
              FOutputs[ o_out] :=
                LogicToSignal(
                  SignalToLogic( FInputs[ i_in1]) and
                  SignalToLogic( FInputs[ i_in2]) and
                  SignalToLogic( FInputs[ i_in3]) and
                  SignalToLogic( FInputs[ i_in4])
                );
            end
            else begin
              R := -1;
              for i := i_in1 to i_in4
              do R := R and SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ o_out] := BinaryToSignal( R, BitCount);
            end;
          end;
        m_nand :
          begin
            if BitCount = 1
            then begin
              FOutputs[ o_out] :=
                LogicToSignal(
                  Not(
                    SignalToLogic( FInputs[ i_in1]) and
                    SignalToLogic( FInputs[ i_in2]) and
                    SignalToLogic( FInputs[ i_in3]) and
                    SignalToLogic( FInputs[ i_in4])
                  )
                );
            end
            else begin
              R := -1;
              for i := i_in1 to i_in4
              do R := R and SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ o_out] := BinaryToSignal( not R, BitCount);
            end;
          end;
        m_or :
          begin
            if BitCount = 1
            then begin
              FOutputs[ o_out] :=
                LogicToSignal(
                  SignalToLogic( FInputs[ i_in1]) or
                  SignalToLogic( FInputs[ i_in2]) or
                  SignalToLogic( FInputs[ i_in3]) or
                  SignalToLogic( FInputs[ i_in4])
                );
            end
            else begin
              R := 0;
              for i := i_in1 to i_in4
              do R := R or SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ o_out] := BinaryToSignal( R, BitCount);
            end;
          end;
        m_nor :
          begin
            if BitCount = 1
            then begin
              FOutputs[ o_out] :=
                LogicToSignal(
                  Not(
                    SignalToLogic( FInputs[ i_in1]) or
                    SignalToLogic( FInputs[ i_in2]) or
                    SignalToLogic( FInputs[ i_in3]) or
                    SignalToLogic( FInputs[ i_in4])
                  )
                );
            end
            else begin
              R := 0;
              for i := i_in1 to i_in4
              do R := R or SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ o_out] := BinaryToSignal( not R, BitCount);
            end;
          end;
        m_xor :
          begin
            if BitCount = 1
            then begin
              FOutputs[ o_out] :=
                LogicToSignal(
                  SignalToLogic( FInputs[ i_in1]) xor
                  SignalToLogic( FInputs[ i_in2]) xor
                  SignalToLogic( FInputs[ i_in3]) xor
                  SignalToLogic( FInputs[ i_in4])
                );
            end
            else begin
              R := 0;
              for i := i_in1 to i_in4
              do R := R xor SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ o_out] := BinaryToSignal( R, BitCount);
            end;
          end;
        m_xnor :
          begin
            if BitCount = 1
            then begin
              FOutputs[ o_out] :=
                LogicToSignal(
                  Not(
                    SignalToLogic( FInputs[ i_in1]) xor
                    SignalToLogic( FInputs[ i_in2]) xor
                    SignalToLogic( FInputs[ i_in3]) xor
                    SignalToLogic( FInputs[ i_in4])
                  )
                );
            end
            else begin
              R := 0;
              for i := i_in1 to i_in4
              do R := R xor SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ o_out] := BinaryToSignal( not R, BitCount);
            end;
          end;
        m_one :
          begin
            OneCounts := 0;
            for i := i_in1 to i_in4
            do begin
              if SignalToLogic( FInputs[ i])
              then Inc( OneCounts);
            end;
            FOutputs[ o_out] := LogicToSignal( OneCounts = 1);
          end;
        else // none
          begin
            OneCounts := 0;
            for i := i_in1 to i_in4
            do begin
              if SignalToLogic( FInputs[ i])
              then Inc( OneCounts);
            end;
            FOutputs[ o_out] := LogicToSignal( OneCounts <> 1);
          end;
      end;
    end;

    procedure   TModGate.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutputs[ o_out]], AppLocale));
    end;

    function    TModGate.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_function : Result := 'function';
        i_width    : Result := 'width';
        i_in1      : Result := 'in1';
        i_in2      : Result := 'in2';
        i_in3      : Result := 'in3';
        i_in4      : Result := 'in4';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModGate.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDivider = class( TMod)
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FOldCount  : Integer;
    FState     : Boolean;
    FOldReset  : Boolean;
  private
}

    procedure   TModDivider.PerformReset;
    begin
      FCounter   := Max( 0, FOldCount - 1);
      FState     := True;
      FLastInput := True;
      FOutputs[ o_out] := m_states[ FState];
    end;

//  public

    constructor TModDivider.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModDivider.DoSlowTick; // override;
    var
      Inp      : Boolean;
      Res      : Boolean;
      NewCount : Integer;
    begin
      Res      := SignalToLogic( FInputs[ i_reset]);
      NewCount := Round        ( FInputs[ i_count]);

      if ResetFlag or ( Res and not FOldReset) or ( FOldCount <> NewCount)
      then begin
        FOldCount := NewCount;
        PerformReset
      end
      else begin
        Inp := SignalToLogic( FInputs[ i_trig]);
        if FLastInput xor Inp
        then begin
          if FCounter <= 0
          then begin
            FCounter := Max( 0, FOldCount - 1);
            FState   := not FState;
            FOutputs[ o_out] := m_states[ FState];
          end
          else Dec( FCounter);
        end;
        FLastInput := Inp;
      end;
      ResetFlag := False;
      FOldReset := Res;
    end;

    procedure   TModDivider.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutputs[ o_out]], AppLocale));
    end;

    function    TModDivider.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_count : Result := 'count';
        i_trig  : Result := 'trig';
        i_reset : Result := 'reset';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModDivider.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModProgDivider = class( TMod)
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FCount     : Integer;
    FState     : Boolean;
    FOldReset  : Boolean;
  private
}

    procedure   TModProgDivider.PerformReset;
    begin
      FCounter   := Max( 0, FCount - 1);
      FState     := True;
      FLastInput := True;
      FOutputs[ o_out] := m_states[ FState];
    end;

//  public

    constructor TModProgDivider.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModProgDivider.DoSlowTick; // override;
    var
      Inp       : Boolean;
      Res       : Boolean;
      Steps     : Integer;
      LowCount  : Integer;
      HighCount : Integer;
      Selection : TSignal;
    begin
      // What to do when highcount < lowcount ... ok then it is just treated as equal to lowcount
      Res       := SignalToLogic( FInputs[ i_reset    ]);
      LowCount  := Round        ( FInputs[ i_lowcount ]);
      HighCount := Round        ( FInputs[ i_highcount]);
      Selection := FInputs[ i_select] * FInputs[ i_selectlev];
      Steps     := Max( 1, Round      ( FInputs[ i_steps    ]));
      FCount    := (( Round(( LowCount + Max( 0, HighCount - LowCount)) * Selection) div Steps) + 1) * Steps;

      if ResetFlag or ( Res and not FOldReset)
      then PerformReset
      else begin
        Inp := SignalToLogic( FInputs[ i_trig]);
        if FLastInput xor Inp
        then begin
          if FCounter <= 0
          then begin
            FCounter := Max( 0, FCount - 1);
            FState   := not FState;
            FOutputs[ o_out] := m_states[ FState];
          end
          else Dec( FCounter);
        end;
        FLastInput := Inp;
      end;
      ResetFlag := False;
      FOldReset := Res;
    end;

    procedure   TModProgDivider.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutputs[ o_out]], AppLocale));
    end;

    function    TModProgDivider.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig      : Result := 'trig';
        i_reset     : Result := 'reset';
        i_select    : Result := 'select';
        i_selectlev : Result := 'selectlev';
        i_steps     : Result := 'steps';
        i_lowcount  : Result := 'lowcount';
        i_highcount : Result := 'highcount';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModProgDivider.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDFlipFlop = class( TMod)
  private
    FClockState : Boolean;
  public
}

    constructor TModDFlipFlop.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModDFlipFlop.DoSlowTick; // override;
    var
      CLK : Boolean;
      Q   : Boolean;
    begin
      if ResetFlag
      then begin
        ResetFlag      := False;
        FOutputs[ o_q] := LogicToSignal( False);
        Q              := False;
      end
      else begin
        CLK := SignalToLogic( FInputs[ i_clk]);
        if not FClockState and CLK
        then begin
          Q              := SignalToLogic( FInputs[ i_d]);
          FOutputs[ o_q] := LogicToSignal( Q);
        end
        else Q := SignalToLogic( FOutputs[ o_q]);
        FClockState := CLK;
      end;
      FOutputs[ o_notq] := LogicToSignal( not Q);
    end;

    procedure   TModDFlipFlop.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/q/%f'   , [ aPrefix, Name, FOutputs[ o_q   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/notq/%f', [ aPrefix, Name, FOutputs[ o_notq]], AppLocale));
    end;

    function    TModDFlipFlop.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_d   : Result := 'd';
        i_clk : Result := 'clk';
        else    Result := NO_INPUT;
      end;
    end;

    function    TModDFlipFlop.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_q    : Result := 'q';
        o_notq : Result := 'notq';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRSFlipFlop = class( TMod)
  private
    FSetState   : Boolean;
    FResetState : Boolean;
  public
}

    constructor TModRSFlipFlop.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRSFlipFlop.DoSlowTick; // override;
    var
      S    : Boolean;
      R    : Boolean;
      Q    : Boolean;
      Mode : Integer;
    begin
      Mode := Round( FInputs[ i_trigmode]);
      S    := SignalToLogic( FInputs [ i_s]);
      R    := SignalToLogic( FInputs [ i_r]) or ResetFlag;
      Q    := SignalToLogic( FOutputs[ o_q]);
      if (( not FResetState) and R) or (( Mode = m_level) and R)
      then Q := False
      else if (( not FSetState) and S) or (( Mode = m_level) and S)
      then Q := True;
      FOutputs[ o_q   ] := LogicToSignal(     Q);
      FOutputs[ o_notq] := LogicToSignal( not Q);
      FSetState   := S;
      FResetState := R;
      ResetFlag   := False;
    end;

    procedure   TModRSFlipFlop.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/q/%f'   , [ aPrefix, Name, FOutputs[ o_q   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/notq/%f', [ aPrefix, Name, FOutputs[ o_notq]], AppLocale));
    end;

    function    TModRSFlipFlop.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_s        : Result := 's';
        i_r        : Result := 'r';
        i_trigmode : Result := 'trigmode';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModRSFlipFlop.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_q    : Result := 'q';
        o_notq : Result := 'notq';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModCounter = class( TMod)
  private
    FState   : Word;
    FMode    : Integer;
    FOldRes  : Boolean;
    FOldTrig : Boolean;
  public
}

    constructor TModCounter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModCounter.SetDefaults; // override;
    begin
      FInputs[ i_dir] := 1; // Default direction is up
    end;

    procedure   TModCounter.DoSlowTick; // override;
    const
      MaxCount = 16;
    var
      Direction : Boolean;
      Trig      : Boolean;
      Res       : Boolean;
      Mode      : Integer;
      i         : Integer;
      p         : Integer;
    begin
      Direction := SignalToLogic( FInputs[ i_dir  ]);
      Mode      := Round        ( FInputs[ i_mode ]);
      Trig      := SignalToLogic( FInputs[ i_trig ]);
      Res       := SignalToLogic( FInputs[ i_reset]);

      if (( not FOldRes) and Res) or ResetFlag or ( Mode <> FMode)
      then begin
        FMode    := Mode;
        FOldTrig := True;
        if Direction
        then FState := 0               // Forwards, reset to first state
        else begin
          if FMode = m_bin       // Backwards, reset to last state
          then FState := $ffff         // binary
          else FState := MaxCount - 1; // sequential (or ring)
        end;
      end
      else begin
        if Trig and not FOldTrig
        then begin
          if Direction
          then begin                     // Forwards, go to nexxt state
          {$Q-R-}
            Inc( FState);                // binary counter will oveflow properly
          {$Q+R+}
            if FMode = m_seq       // ring counter needs a check
            then FState := FState mod MaxCount;
          end
          else begin                     // Backwards, go to previous state
            if FState > 0
            then Dec( FState)
            else begin
              if FMode = m_bin
              then FState := $ffff
              else FState := MaxCount - 1;
            end;
          end;
        end;
        FOldTrig := Trig;
      end;
      FOldRes   := Res;
      ResetFlag := False;


      if FMode = m_bin
      then begin
        p := 1;
        for i := 0 to MaxCount - 1
        do begin
          if ( FState and p) <> 0
          then FOutputs[ o_out0 + i] := LogicToSignal( True )
          else FOutputs[ o_out0 + i] := LogicToSignal( False);
          p := p shl 1;
        end;
      end
      else begin
        for i := 0 to MaxCount - 1
        do begin
          if i = FState
          then FOutputs[ o_out0 + i] := LogicToSignal( True )
          else FOutputs[ o_out0 + i] := LogicToSignal( False);
        end;
      end;
    end;

    procedure   TModCounter.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := o_out0 to o_out15
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, 1 + i - o_out1, FOutputs[ i]], AppLocale));

      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs[ i_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInputs[ i_reset]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInputs[ i_dir  ]], AppLocale));
    end;

    function    TModCounter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig  : Result := 'trig';
        i_reset : Result := 'reset';
        i_dir   : Result := 'dir';
        i_mode  : Result := 'mode';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModCounter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out0  : Result := 'out0';
        o_out1  : Result := 'out1';
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        o_out12 : Result := 'out12';
        o_out13 : Result := 'out13';
        o_out14 : Result := 'out14';
        o_out15 : Result := 'out15';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAdc = class( TMod)
  public
}

    constructor TModAdc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModAdc.DoSlowTick; // override;
    var
      aValue : Integer;
      i      : Integer;
    begin
      aValue := SignalToBinary( FInputs[ i_in], o_out15 - o_out0 + 1);
      for i := o_out0 to o_out15
      do begin
        FOutputs[ i] := LogicToSignal(( aValue and 1) = 1);
        aValue := aValue shr 1;
      end;
    end;

    procedure   TModAdc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := o_out0 to o_out15
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, i - o_out0, FOutputs[ i]], AppLocale));
    end;

    function    TModAdc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in : Result := 'in';
        else   Result := NO_INPUT;
      end;
    end;

    function    TModAdc.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out0  : Result := 'out0';
        o_out1  : Result := 'out1';
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        o_out12 : Result := 'out12';
        o_out13 : Result := 'out13';
        o_out14 : Result := 'out14';
        o_out15 : Result := 'out15';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDac = class( TMod)
  public
}

    constructor TModDac.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModDac.DoSlowTick; // override;
    var
      aValue : Integer;
      i      : Integer;
    begin
      aValue := 0;
      for i := i_in0 to i_in15
      do begin
        if SignalToLogic( FInputs[ i])
        then aValue := ( aValue shr 1) or ( 1 shl ( n_bits - 1))
        else aValue := aValue shr 1;
      end;
      FOutputs[ o_out] := BinaryToSignal( aValue, n_bits);
    end;

    procedure   TModDac.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := i_in0 to i_in15
      do aCallback( Self, Format( '%s/%s/in%d/%f', [ aPrefix, Name, i - i_in0, FInputs[ i]], AppLocale));
    end;

    function    TModDac.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in0  : Result := 'in0';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        i_in5  : Result := 'in5';
        i_in6  : Result := 'in6';
        i_in7  : Result := 'in7';
        i_in8  : Result := 'in8';
        i_in9  : Result := 'in9';
        i_in10 : Result := 'in10';
        i_in11 : Result := 'in11';
        i_in12 : Result := 'in12';
        i_in13 : Result := 'in13';
        i_in14 : Result := 'in14';
        i_in15 : Result := 'in15';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModDac.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out  : Result := 'out';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TPDState = 1 .. 12;

  TModPhaseDetect = class( TMod)
  private
    FOldIn1 : Boolean;
    FOldIn2 : Boolean;
    FState  : TPDState;
    FAlpha  : TSignal;
  public
}

    constructor TModPhaseDetect.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPhaseDetect.SetDefaults; // override;
    begin
      FState := 1;
      FInputs[ i_typesel] := -1.0;
      FDezipperMap := [
        i_alpha,
        i_alphamodamt
      ];
    end;

    procedure   TModPhaseDetect.DoSlowTick; // override;
    var
      NewIn1   : Boolean;
      NewIn2   : Boolean;
      In1Up    : Boolean;
      In1Down  : Boolean;
      In2Up    : Boolean;
      In2Down  : Boolean;
      Inp      : TSignal;
      Outp     : TSignal;
      aType    : Integer;
      aTypeSel : Boolean;
    begin
      FAlpha   := Normalize( ExpAttack( FInputs[ i_alpha] + FInputs[ i_alphamod] * FInputs[ i_alphamodamt], 1, 0, 1));
      NewIn1   := SignalToLogic( FInputs[ i_in1    ]);
      NewIn2   := SignalToLogic( FInputs[ i_in2    ]);
      aType    := Round        ( FInputs[ i_type   ]);
      aTypeSel := SignalToLogic( FInputs[ i_typesel]);

      if aTypeSel
      then aType := 1 - aType;

      In1Up   := not FOldIn1 and     NewIn1;
      In1Down :=     FOldIn1 and not NewIn1;
      In2Up   := not FOldIn2 and     NewIn2;
      In2Down :=     FOldIn2 and not NewIn2;

      case FState of
         1 : if in1up   then FState :=  3 else if in2up   then FState :=  2;
         2 : if In1Up   then FState :=  6 else if In2Down then FState :=  4;
         3 : if In1Down then FState :=  5 else if In2Up   then FState :=  6;
         4 : if In1Up   then FState :=  8 else if In2Up   then FState :=  2;
         5 : if In1Up   then FState :=  3 else if In2Up   then FState :=  7;
         6 : if In1Down then FState :=  7 else if in2down then FState :=  8;
         7 : if In1Up   then FState :=  9 else if In2Down then FState :=  1;
         8 : if In1Down then FState :=  1 else if In2Up   then FState := 10;
         9 : if In1Down then FState := 11 else if In2Down then FState :=  3;
        10 : if In1Down then FState :=  2 else if In2Down then FState := 12;
        11 : if In1Up   then FState :=  9 else if In2Down then FState :=  5;
        12 : if In1Down then FState :=  4 else if In2Up   then FState := 10;
      end;

      Outp := FOutputs[ o_out];

      if aType = 0
      then Inp := LogicToSignal( NewIn1 xor NewIn2)
      else begin
        if FState in [ 3, 5, 9, 11]
        then Inp := 1.0                                      // Input frequency is higher
        else if FState in [ 2, 4, 10, 12]
        then Inp := - 1.0                                    // Input frequency is higher
        else Inp := Outp;                                    // Input frequency is the same
      end;

      Outp  := Outp + FAlpha * ( Inp - Outp);

      FOutputs[ o_out ] := Normalize( Outp);
      FOutputs[ o_lock] := LogicToSignal( FState in [ 1, 6, 7, 8]);

      FOldIn1 := NewIn1;
      FOldIn2 := NewIn2;
    end;

    procedure   TModPhaseDetect.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/up/%f'  , [ aPrefix, Name, LogicToSignal( FState in [ 3, 5,  9, 11])], AppLocale));
      aCallback( Self, Format( '%s/%s/lock/%f', [ aPrefix, Name, LogicToSignal( FState in [ 1, 6,  7,  8])], AppLocale));
      aCallback( Self, Format( '%s/%s/down/%f', [ aPrefix, Name, LogicToSignal( FState in [ 2, 4, 10, 12])], AppLocale));
    end;

    function    TModPhaseDetect.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1         : Result := 'in1';
        i_in2         : Result := 'in2';
        i_alpha       : Result := 'alpha';
        i_alphamod    : Result := 'alphamod';
        i_alphamodamt : Result := 'alphamodamt';
        i_type        : Result := 'type';
        i_typesel     : Result := 'typesel';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModPhaseDetect.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out  : Result := 'out';
        o_lock : Result := 'lock';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNot = class( TMod)
  public
}

    constructor TModNot.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNot.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := i_in1 to i_in4          // Set unused inputs to 1 - just to keep the lights off
      do FInputs[ i] := 1;
    end;

    procedure   TModNot.DoSlowTick; // override;
    var
      i : Integer;
    begin
      for i := i_in1 to i_in4
      do FOutputs[ i] := LogicToSignal( not SignalToLogic( FInputs[ i]));
    end;

    procedure   TModNot.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := o_out1 to o_out4
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, i + 1, FOutputs[ i]], AppLocale));
    end;

    function    TModNot.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1 : Result := 'in1';
        i_in2 : Result := 'in2';
        i_in3 : Result := 'in3';
        i_in4 : Result := 'in4';
        else    Result := NO_INPUT;
      end;
    end;

    function    TModNot.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModLogicSelector = class( TMod)
  public
}

    constructor TModLogicSelector.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModLogicSelector.SetDefaults; // override;
    begin
      FInputs[ i_sel] := LogicToSignal( False);
      FInputs[ i_ina] := LogicToSignal( False);
      FInputs[ i_inb] := LogicToSignal( False);
    end;

    procedure   TModLogicSelector.DoSlowTick; // override;
    begin
      if SignalToLogic( FInputs[ i_sel])
      then begin
        FOutputs[ o_outa] := LogicToSignal( SignalToLogic( FInputs[ i_inb]));
        FOutputs[ o_outb] := LogicToSignal( SignalToLogic( FInputs[ i_ina]));
      end
      else begin
        FOutputs[ o_outa] := LogicToSignal( SignalToLogic( FInputs[ i_ina]));
        FOutputs[ o_outb] := LogicToSignal( SignalToLogic( FInputs[ i_inb]));
      end;
    end;

    procedure   TModLogicSelector.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/outa/%f', [ aPrefix, Name, FOutputs[ o_outa]], AppLocale));
      aCallback( Self, Format( '%s/%s/outb/%f', [ aPrefix, Name, FOutputs[ o_outb]], AppLocale));
    end;

    function    TModLogicSelector.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_sel : Result := 'sel';
        i_ina : Result := 'ina';
        i_inb : Result := 'inb';
        else    Result := NO_INPUT;
      end;
    end;

    function    TModLogicSelector.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outa : Result := 'outa';
        o_outb : Result := 'outb';
        else     Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModAmplifier = class( TMod)
  public
}

    constructor TModAmplifier.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModAmplifier.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_gain,
        i_mute
      ];
    end;

    procedure   TModAmplifier.DoSlowTick; // override;
    var
      i : Integer;
      G : TSignal;
    begin
      G := FInputs[ i_gain] * SignalToMute( FInputs[ i_mute]);
      for i := i_in1 to i_in4
      do FOutputs[ i - i_in1 + o_out1] := FInputs[ i] * G;
    end;

    function    TModAmplifier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_gain : Result := 'gain';
        i_mute : Result := 'mute';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModAmplifier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1: Result := 'out1';
        o_out2: Result := 'out2';
        o_out3: Result := 'out3';
        o_out4: Result := 'out4';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModScaler = class( TMod)
  public
}

    constructor TModScaler.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModScaler.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_gain,
        i_offset,
        i_mute
      ];
    end;

    procedure   TModScaler.DoSlowTick; // override;
    var
      i : Integer;
      G : TSignal;
      O : TSignal;
      S : TSignal;
      T : TSignal;
    begin
      G := FInputs[ i_gain  ] * SignalToMute( FInputs[ i_mute]);
      O := FInputs[ i_offset] * SignalToMute( FInputs[ i_mute]);
      S := 0.0;
      for i := i_in1 to i_in4
      do begin
        T :=  G * FInputs[ i] + O;
        FOutputs[ i - i_in1 + o_out1] := T;
        S := S + T;
      end;
      FOutputs[ o_sum] := S;
    end;

    function    TModScaler.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_gain   : Result := 'gain';
        i_offset : Result := 'offset';
        i_mode   : Result := 'mode';
        i_mute   : Result := 'mute';
        i_in1    : Result := 'in1';
        i_in2    : Result := 'in2';
        i_in3    : Result := 'in3';
        i_in4    : Result := 'in4';
        else       Result := NO_INPUT;
      end;
    end;

    function    TModScaler.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        o_sum  : Result := 'sum';   // Sum output
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRatio = class( TMod)
  public
}

    constructor TModRatio.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRatio.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_multiplier,
        i_divider,
        i_mute
      ];
    end;

    procedure   TModRatio.DoSlowTick; // override;
    var
      i      : Integer;
      MulDiv : TSignal;
      RDiv   : Integer;
    begin
      RDiv := Round( FInputs[ i_divider]);
      if RDiv = 0
      then Muldiv := 1
      else Muldiv := Round( SignalToMute( FInputs[ i_mute]) * FInputs[ i_multiplier]) div RDiv;
      for i := i_in1 to i_in4
      do FOutputs[ i - i_in1 + o_out1] := MulDiv * FInputs[ i];
    end;

    function    TModRatio.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_multiplier : Result := 'multiplier';
        i_divider    : Result := 'divider';
        i_mute       : Result := 'mute';
        i_in1        : Result := 'in1';
        i_in2        : Result := 'in2';
        i_in3        : Result := 'in3';
        i_in4        : Result := 'in4';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModRatio.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRange = class( TMod)
  private
    FInLow   : TSignal;
    FInHigh  : TSignal;
    FSigType : Integer;
  public
}

    constructor TModRangeConverter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRangeConverter.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_lowvalue,
        i_highvalue
      ];
      FMode := s_positive;
    end;

    procedure   TModRangeConverter.DoSlowTick; // override;
    // Called from DoSlowTick when in slow mode, otherwise called directly from the TSynthPatch
    begin
      FInLow  :=              FInputs[ i_lowvalue ];
      FInHigh :=              FInputs[ i_highvalue];
      FMode   := Round( FInputs[ i_mode     ]);

      case FMode of

        s_inverted : begin
          FOutputs[ o_out1] := RangeMap( FInputs[ i_in1],  1, -1, FInLow, FInHigh);
          FOutputs[ o_out2] := RangeMap( FInputs[ i_in2],  1, -1, FInLow, FInHigh);
          FOutputs[ o_out3] := RangeMap( FInputs[ i_in3],  1, -1, FInLow, FInHigh);
          FOutputs[ o_out4] := RangeMap( FInputs[ i_in4],  1, -1, FInLow, FInHigh);
        end;

        s_positive : begin
          FOutputs[ o_out1] := RangeMap( FInputs[ i_in1],  0,  1, FInLow, FInHigh);
          FOutputs[ o_out2] := RangeMap( FInputs[ i_in2],  0,  1, FInLow, FInHigh);
          FOutputs[ o_out3] := RangeMap( FInputs[ i_in3],  0,  1, FInLow, FInHigh);
          FOutputs[ o_out4] := RangeMap( FInputs[ i_in4],  0,  1, FInLow, FInHigh);
        end;

        s_positive_inverted : begin
          FOutputs[ o_out1] := RangeMap( FInputs[ i_in1],  1,  0, FInLow, FInHigh);
          FOutputs[ o_out2] := RangeMap( FInputs[ i_in2],  1,  0, FInLow, FInHigh);
          FOutputs[ o_out3] := RangeMap( FInputs[ i_in3],  1,  0, FInLow, FInHigh);
          FOutputs[ o_out4] := RangeMap( FInputs[ i_in4],  1,  0, FInLow, FInHigh);
        end;

        s_negative : begin
          FOutputs[ o_out1] := RangeMap( FInputs[ i_in1], -1,  0, FInLow, FInHigh);
          FOutputs[ o_out2] := RangeMap( FInputs[ i_in2], -1,  0, FInLow, FInHigh);
          FOutputs[ o_out3] := RangeMap( FInputs[ i_in3], -1,  0, FInLow, FInHigh);
          FOutputs[ o_out4] := RangeMap( FInputs[ i_in4], -1,  0, FInLow, FInHigh);
        end;

        s_negative_inverted : begin
          FOutputs[ o_out1] := RangeMap( FInputs[ i_in1],  0, -1, FInLow, FInHigh);
          FOutputs[ o_out2] := RangeMap( FInputs[ i_in2],  0, -1, FInLow, FInHigh);
          FOutputs[ o_out3] := RangeMap( FInputs[ i_in3],  0, -1, FInLow, FInHigh);
          FOutputs[ o_out4] := RangeMap( FInputs[ i_in4],  0, -1, FInLow, FInHigh);
        end;

        else begin // s_normal
          FOutputs[ o_out1] := RangeMap( FInputs[ i_in1], -1,  1, FInLow, FInHigh);
          FOutputs[ o_out2] := RangeMap( FInputs[ i_in2], -1,  1, FInLow, FInHigh);
          FOutputs[ o_out3] := RangeMap( FInputs[ i_in3], -1,  1, FInLow, FInHigh);
          FOutputs[ o_out4] := RangeMap( FInputs[ i_in4], -1,  1, FInLow, FInHigh);
        end;
      end;
    end;


    function    TModRangeConverter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_lowvalue  : Result := 'lowvalue';
        i_highvalue : Result := 'highvalue';
        i_mode      : Result := 'mode';
        i_in1       : Result := 'in1';
        i_in2       : Result := 'in2';
        i_in3       : Result := 'in3';
        i_in4       : Result := 'in4';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModRangeConverter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModLogisticMap = class( TMod)
  private
    FOldStart   : Boolean;
    FOldClock   : Boolean;
    FTicks      : Integer;
    FPeriodTime : Integer;
    FOutValue   : TSignal;
    FNextValue  : TSignal;
  public
 }

    constructor TModLogisticMap.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModLogisticMap.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_smodamt,
        i_svalue ,
        i_rmodamt,
        i_lowr   ,
        i_highr  ,
        i_xfade  ,
        i_mute
      ];
      FOldStart := False;
      FOldClock := False;
    end;

    procedure   TModLogisticMap.DoSlowTick;  // override;
    var
      Clock  : Boolean;
      Start  : Boolean;
      XFade  : TSignal;
      R      : TSignal;
      OutVal : TSignal;
    begin
      Start := SignalToLogic( FInputs[ i_start]);

      if ( not FOldStart and Start) or ResetFlag
      then begin
        FNextValue  := Clip( FInputs[ i_svalue] + FInputs[ i_smodamt] * FInputs[ i_smod], 0.0, 1.0);
        FPeriodTime := Max( 2, FTicks);
        FTicks      := 0;
      end;

      ResetFlag := False;
      FOldStart := Start;

      Clock := SignalToLogic( FInputs[ i_clock]);

      if not FOldClock and Clock
      then begin
        FPeriodTime := Max( 2, FTicks);
        FTicks      := 0;
        XFade       := Clip( FInputs[ i_rmod] * FInputs[ i_rmodamt], 0.0, 1.0);
        R           := FInputs[ i_lowr] * ( 1 - XFade) + FInputs[ i_highr] * XFade;
        FOutValue   := FNextValue;
        FNextValue  := Normalize( R * FOutValue * ( 1.0 - FOutValue));
        Nop;
      end
      else if FOldClock and not Clock
      then FPeriodTime := FTicks * 2;

      XFade := Clip((( FPeriodTime - FTicks) / FPeriodTime) * FInputs[ i_xfade], -1.0, 1.0);

      if FOutValue < FNextValue
      then OutVal := 2 * LinDecay ( XFade, 1, FNextValue, FOutValue) - 1  // going up, note that as final and current value are reversed this must call decay ...
      else OutVal := 2 * LinAttack( XFade, 1, FNextValue, FOutValue) - 1; // going down

      if FTicks < MAXINT shr 2    // Cap FTicks value in case the module gets not clocked prolly MAXINT shr 1 would be OK too
      then Inc( FTicks);

      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( OutVal, Round( FInputs[ i_outtype]));

      FOldClock := Clock;
    end;

    function    TModLogisticMap.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_clock   : Result := 'clock';
        i_start   : Result := 'start';
        i_smod    : Result := 'smod';
        i_smodamt : Result := 'smodamt';
        i_svalue  : Result := 'svalue';
        i_rmod    : Result := 'rmod';
        i_rmodamt : Result := 'rmodamt';
        i_lowr    : Result := 'lowr';
        i_highr   : Result := 'highr';
        i_xfade   : Result := 'xfade';
        i_outtype : Result := 'outtype';
        i_mute    : Result := 'mute';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModLogisticMap.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRotator = class( TMod)
  public
}

    constructor TModRotator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRotator.SetDefaults; // override;
    begin
      FInputs[ i_angle] := 0;
      FInputs[ i_in1  ] := 0;
      FInputs[ i_in2  ] := 0;
    end;

    procedure   TModRotator.DoSlowTick; // override;
    var
      aCos    : TSignal;
      aSin    : TSignal;
      in1     : TSignal;
      in2     : TSignal;
    begin
      FAngle := FInputs[ i_angle] * TWO_PI;
      SinCos( FAngle, aSin, aCos);
      in1 := FInputs[ i_in1];
      in2 := FInputs[ i_in2];
      FOutputs[ o_out1] := aCos * in1 - aSin * in2;
      FOutputs[ o_out2] := aSin * in1 + aCos * in2;
    end;

    function    TModRotator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_angle : Result := 'angle';
        i_in1   : Result := 'in1';
        i_in2   : Result := 'in2';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModRotator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRectifier = class( TMod)
  public
}

    constructor TModRectifier.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRectifier.SetDefaults; // override;
    begin
      FInputs[ i_in] := 0;
    end;

    procedure   TModRectifier.DoSlowTick; // override;
    var
      Mode : Integer;
      Inp  : TSignal;
    begin
      Mode := Round( FInputs[ i_mode]);
      Inp  :=        FInputs[ i_in  ] ;

      if Mode = 0                                      // Half wave
      then begin
        if SignalToLogic( Inp)
        then FOutputs[ o_out] := Inp
        else FOutputs[ o_out] := 0;
      end
      else if Mode = 1                                 // Full wave
      then FOutputs[ o_out] := Abs( Inp)
      else FOutputs[ o_out] :=      Inp;      // transparent

      FOutputs[ o_outinv] := - FOutputs[ o_out];
    end;

    function    TModRectifier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_mode : Result := 'mode';
        i_in   : Result := 'in';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModRectifier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_outinv : Result := 'outinv';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMinMax = class( TMod)
  public
}

    constructor TModMinMax.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMinMax.SetDefaults; // override;
    begin
      FInputs[ i_in1] := 0;
      FInputs[ i_in2] := 0;
    end;

    procedure   TModMinMax.DoSlowTick; // override;
    var
      in1 : TSignal;
      in2 : TSignal;
    begin
      in1              := FInputs[ i_in1];
      in2              := FInputs[ i_in2];
      FOutputs[ o_min] := Min( in1, in2);
      FOutputs[ o_max] := Max( in1, in2);
    end;

    function    TModMinMax.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1 : Result := 'in1';
        i_in2 : Result := 'in2';
        else    Result := NO_INPUT;
      end;
    end;

    function    TModMinMax.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_min : Result := 'min';
        o_max : Result := 'max';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModWaveWiper = class( TMod)
  public
}

    constructor TModWaveWiper.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
      FDezipperMap := [
        i_mute
      ];
    end;

    procedure   TModWaveWiper.SetDefaults; // override;
    begin
      FInputs[ i_in1] := 0;
      FInputs[ i_in2] := 0;
    end;

    procedure   TModWaveWiper.DoSlowTick; // override;
    var
      Insig1   : TSignal;
      Insig2   : TSignal;
      BreakSig : TSignal;
      Mute     : TSignal;
    begin
      InSig1   := FInputs[ i_in1  ];
      InSig2   := FInputs[ i_in2  ];
      BreakSig := FInputs[ i_break];
      Mute     := SignalToMute( FInputs[ i_mute]);
      FOutPuts[ o_out1] := ( Max( Insig1, BreakSig) + Min( InSig2, BreakSig) - BreakSig) * Mute;
      FOutPuts[ o_out2] := ( Max( Insig2, BreakSig) + Min( Insig1, BreakSig) - BreakSig) * Mute;
    end;

    function    TModWaveWiper.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_break : Result := 'break';
        i_mute  : Result := 'mute';
        i_in1   : Result := 'in1';
        i_in2   : Result := 'in2';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModWaveWiper.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModWaveWrapper = class( TMod)
  private
    FLastValue  : TSignal;
    FOutValue   : TSignal;
    FStateValue : TSignal;
    FGoingUp    : Boolean;
  public
}

    constructor TModWaveWrapper.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    procedure   TModWaveWrapper.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute     ,
        i_lowlevel ,
        i_highlevel,
        i_modamt
      ];
    end;

    procedure   TModWaveWrapper.DoSlowTick; // override;
    var
      Modulation  : TSignal;
      LoMirror    : TSignal;
      MirrorDelta : TSignal;
      Amp         : TSignal;
      Mute        : TSignal;
      Signal      : TSignal;
    var
      Difference  : TSignal;
      Invalue     : TSignal;
      LowMirror   : TSignal;
      HighMirror  : TSignal;
    begin
      case Round( FInputs[ i_type]) of

        m_type2 : begin
          // Source :: https://www.muffwiggler.com/forum/viewtopic.php?p=1630760
          InValue      := FInputs[ i_in];
          Modulation   := Normalize( FInputs[ i_mod] * FInputs[ i_modamt]);         // Modulation amount
          LowMirror    := FInputs[ i_lowlevel ] - Modulation;                       // Get low  mirror and subtract modulation
          HighMirror   := FInputs[ i_highlevel] + Modulation;                       // Get high mirror and add      modulation
          Difference   := InValue - FLastValue;
          FLastValue   := InValue;

          if FGoingUp
          then FOutValue := FOutValue + Difference
          else FOutValue := FOutValue - Difference;

          if FOutValue > HighMirror
          then begin
            FGoingUp  := not FGoingUp;
            FOutValue := 2.0 * HighMirror - FOutValue;
          end
          else if FOutValue < LowMirror
          then begin
            FGoingUp  := not FGoingUp;
            FOutValue := 2.0 * LowMirror - FOutValue;
          end;
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * FOutValue;
        end;

        m_type3 : begin
          // Source :: https://www.muffwiggler.com/forum/viewtopic.php?p=1630760 - my take at the smrl variation.
          InValue      := FInputs[ i_in];
          Modulation   := Normalize( FInputs[ i_mod] + FInputs[ i_modamt]);         // Modulation amount
          LowMirror    := FInputs[ i_lowlevel ];                                    // Get low  mirror
          HighMirror   := FInputs[ i_highlevel];                                    // Get high mirror
          Difference   := InValue - FLastValue;
          FLastValue   := InValue;

          if FGoingUp
          then FStateValue := FStateValue + Difference
          else FStateValue := FStateValue - Difference;

          if FStateValue > HighMirror
          then begin
            FGoingUp    := not FGoingUp;
            FStateValue := 2.0 * HighMirror - FStateValue;
            FOutValue   := FOutValue + Modulation;
            if FOutValue > 1
            then FOutValue := -1;
          end
          else if FStateValue < LowMirror
          then begin
            FGoingUp    := not FGoingUp;
            FStateValue := 2.0 * LowMirror - FStateValue;
            FOutValue   := FOutValue + Modulation;
            if FOutValue > 1
            then FOutValue := -1;
          end;
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * FOutValue;
        end;

        else begin
          Modulation  := Normalize( FInputs[ i_mod] * FInputs[ i_modamt]);         // Modulation amount
          LoMirror    := FInputs[ i_lowlevel ] - Modulation;                       // Get low mirror and subtract modulatio
          MirrorDelta := 2.0 * ( FInputs[ i_Highlevel] + Modulation - LoMirror);   // High mirror gets modulation added. calc mirror distance, times 2
          Mute        := SignalToMute( FInputs[ i_mute]);                          // Get Mute multiplier

          if Abs( MirrorDelta) < 1e-6                                              // When mirrors are too close to each other
          then FOutputs[ o_out] := 0                                               // make an exception ...
          else begin
            Amp    := 4.0 / MirrorDelta;                                           // Amplitude reduction factor due to wrapping - times two
            Signal := MathFloatMod( FInputs[ i_in] + LoMirror, MirrorDelta);       // Lift signal by LoMirror, and take signal modulo 2 * Delta
            if Signal > 0.5 * MirrorDelta                                          // Reflect bit above Delta back
            then Signal := MirrorDelta - Signal;
            FOutputs[ o_out] := Normalize(( Amp * Signal - 1.0) * Mute);           // Bring signal into -1, 1 range and apply Mute signal
          end;
        end;

      end;
    end;

    function    TModWaveWrapper.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in        : Result := 'in';
        i_mute      : Result := 'mute';
        i_lowlevel  : Result := 'lowlevel';
        i_highlevel : Result := 'highlevel';
        i_mod       : Result := 'mod';
        i_modamt    : Result := 'modamt';
        i_type      : Result := 'type';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModWaveWrapper.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModIntDif = class( TMod)
  private
    FSum        : TSignal;
    FlastSample : TSignal;
  public
}

    constructor TModIntDif.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModIntDif.SetDefaults; // override;
    begin
      FInputs[ i_in] := 0;
    end;

    procedure   TModIntDif.DoTick; // override;
    var
      Inp  : TSignal;
      Dif  : TSignal;
      Mode : Integer;
    begin
      Inp := FInputs[ i_in];

      if ResetFlag
      then begin
        ResetFlag   := False;
        FSum        := Inp;
        FlastSample := Inp;
      end
      else FSum := Clip( FSum + Inp, -10, 10);

      Mode := Round( FInputs[ i_mode]);
      Dif  := Inp - FlastSample;

      if Mode = 0                                      // Half wave rectification of dif signal
      then begin
        if SignalToLogic( Dif)
        then FOutputs[ o_dif] := Dif
        else FOutputs[ o_dif] := 0;
      end
      else if Mode = 1                                 // Full wave rectification of dif signal
      then FOutputs[ o_dif] := Abs( Dif)
      else FOutputs[ o_dif] :=      Dif;         // Transparent

      FOutputs[ o_int] := FSum;
      FlastSample      := Inp;
    end;

    function    TModIntDif.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in   : Result := 'in';
        i_mode : Result := 'mode';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModIntDif.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_int : Result := 'int';
        o_dif : Result := 'dif';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseOsc = class( TMod)
  private
    FPhase      : TSignal;
    FPhaseDelta : TSignal;
    FPosition   : TSignal;
    FOldSync    : Boolean;
    FShape      : Integer;
    FBandLimit  : TSignal;
  public
}

    constructor TModBaseOsc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModBaseOsc.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_frequency,
        i_mute,
        i_pmlevel,
        i_fmlevel,
        i_cents
      ];
      FInputs[ i_bandlimit] := 3;
    end;

    procedure   TModBaseOsc.DoTick; // override;
    var
      Trunced: Integer;
    begin
      if ( not FOldSync) and SignalToLogic( FInputs[ i_sync])
      then FPhase  := 0.0;

      FOldSync    := SignalToLogic( FInputs[ i_sync]);
      FPhaseDelta := LookupPhaseDelta( FInputs[ i_frequency] + FInputs[ i_freq] + FInputs[ i_fm] * FInputs[ i_fmlevel] + FInputs[ i_cents]);

      FPhase      := FPhase + FPhaseDelta;
      Trunced     := Trunc( FPhase);
      FPhase      := FPhase - Trunced;
      if FPhase < 0
      then FPhase := FPhase + 1;

      FPosition   := FPhase + Clip( FInputs[ i_phase] * FInputs[ i_pmlevel], -1, 1);
      Trunced     := Trunc( FPosition);
      FPosition   := FPosition - Trunced;
      if FPosition < 0
      then FPosition := FPosition + 1;

      FShape      := Round( FInputs[ i_shape    ]);
      FBandLimit  :=              FInputs[ i_bandlimit];
    end;

    function    TModBaseOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_frequency : Result := 'frequency';
        i_fm        : Result := 'fm';
        i_phase     : Result := 'phase';
        i_sync      : Result := 'sync';
        i_mute      : Result := 'mute';
        i_shape     : Result := 'shape';
        i_pmlevel   : Result := 'pmlevel';
        i_fmlevel   : Result := 'fmlevel';
        i_cents     : Result := 'cents';
        i_bandlimit : Result := 'bandlimit';
        i_freq      : Result := 'freq';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModBaseOsc.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseTrigOsc = class( TModBaseOsc)
  private
    FState   : TTrigOscState;
    FCounter : TSignal;
  public
}

    procedure   TModbaseTrigOsc.DoTick; // override;
    var
      FOldCounter : Integer;
      Trunced     : Integer;
    begin

      case FState of

        toIdle :
          begin
            FPosition := 0.0;
            FCounter  := 0;
            if SignalToLogic( FInputs[ i_trig])
            then FState := toRunning;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FPhaseDelta := LookupPhaseDelta( FInputs[ i_frequency] + FInputs[ i_freq] + FInputs[ i_fm] * FInputs[ i_fmlevel] + FInputs[ i_cents]);
            FPosition   := FPosition + FPhaseDelta;
            Trunced     := Trunc( FPosition);
            FPosition   := FPosition - Trunced;
            if FPosition < 0
            then FPosition := FPosition + 1;
            FShape := Round( FInputs[ i_shape]);

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPosition, 0.5, FPhaseDelta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPosition, 1.0, FPhaseDelta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInputs[ i_count]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPosition, 0.5, FPhaseDelta) and
                not AlmostEqual( FPosition, 1.0, FPhaseDelta)
              then FState := toRunning;
            end;

          end;

        toDone :
          begin
            FPosition := 0.0;
            if FInputs[ i_trig] <= 0.0
            then FState := toIdle;
          end;
      end;
    end;

    function    TModBaseTrigOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig  : Result := 'trig';
        i_count : Result := 'count';
        else      Result := inherited;
      end;
    end;


{ ========
  TModOsc = class( TModBaseOsc)
  public
}

    procedure   TModOSc.DoTick; // override;
    var
      aValue : TSignal;
    begin
      inherited;
      case FShape of
        0  : // sine
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * LookupSine( FPosition);
        1  : // tri
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            aValue := 4 * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FloatMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta));
            FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * ( FPhaseDelta * aValue + ( 1 - FPhaseDelta) * FOutputs[ o_out]);
          end;
        2  : // saw
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * (( 2.0 * FPosition - 1) - PolyBLEP( FPosition, FBandLimit, FPhaseDelta));
        3  : // square
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FloatMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta));
          end;
      end;
    end;


{ ========
  TModTrigOsc = class( TModBaseTrigOsc)
  public
}

    procedure   TModTrigOsc.DoTick; // override;
    var
      aValue : TSignal;
    begin
      inherited;
      case FShape of
        0  : // sine
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * LookupSine( FPosition);
        1  : // tri
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            aValue := 4 * aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FloatMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta);
            FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * ( FPhaseDelta * aValue + ( 1 - FPhaseDelta) * FOutputs[ o_out]);
          end;
        2  : // saw
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * (( 2.0 * FPosition - 1) - PolyBLEP( FPosition, FBandLimit, FPhaseDelta));
        3  : // square
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FloatMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta));
          end;
      end;
    end;


{ ========
  TModSquare = class( TModBaseOsc)
  public
}

    procedure   TModSquare.SetDefaults;
    begin
      inherited;
      FDezipperMap := FDezipperMap + [
        i_pwmlevel
      ];
    end;

    procedure   TModSquare.DoTick; // override;
    var
      pwmp   : TSignal;
      aValue : TSignal;
    begin
      inherited;                                // For phase calculations
      pwmp := Clip( 0.5 * ( 1.0 - FInputs[ i_pwm] * FInputs[ i_pwmlevel]), 0.01, 0.99);
      if FPosition < pwmp
      then aValue :=   1
      else aValue := - 1;
      FOutputs[ i_out] := SignalToMute( FInputs[ i_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FloatMod( FPosition + ( 1 - pwmp), 1), FBandLimit, FPhaseDelta));
    end;

    function    TModSquare.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         i_pwm      : Result := 'pwm';
         i_pwmlevel : Result := 'pwmlevel';
        else          Result := inherited;
      end;
    end;


{ ========
  TModSquareTrig = class( TModBaseTrigOsc)
  public
}

    procedure   TModSquareTrig.SetDefaults;
    begin
      inherited;
      FDezipperMap := FDezipperMap + [
        i_pwmlevel
      ];
    end;

    procedure   TModSquareTrig.DoTick; // override;
    var
      pwmp   : TSignal;
      aValue : TSignal;
    begin
      inherited;                                // For phase calculations
      pwmp := Clip( 0.5 * ( 1.0 - FInputs[ i_pwm] * FInputs[ i_pwmlevel]), 0.01, 0.99);
      if FPosition < pwmp
      then aValue :=   1
      else aValue := - 1;
      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FloatMod( FPosition + ( 1 - pwmp), 1), FBandLimit, FPhaseDelta));
    end;

    function    TModSquareTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         i_pwm      : Result := 'pwm';
         i_pwmlevel : Result := 'pwmlevel';
        else          Result := inherited;
      end;
    end;


{ ========
  TModBaseLfo = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldRes   : Boolean;
    FShape    : Integer;
  public
}

    constructor TModBaseLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModBaseLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_frequency,
        i_mute,
        i_fmlevel,
        i_pmlevel
      ];
    end;

    procedure   TModBaseLfo.DoSlowTick; // override;
    var
      Delta   : TSignal;
      Trunced : Integer;
    begin
      if ( not FOldRes) and SignalToLogic( FInputs[ i_res]) or ResetFlag
      then FPhase := 0.0;
      ResetFlag := False;
      FOldRes   := SignalToLogic( FInputs[ i_res  ]);
      FShape    := Round        ( FInputs[ i_shape]);
      FWrapped  := False;

      case Round( FInputs[ i_range]) of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        else Delta := LookupLfoLinPhaseDelta   ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
      end;

      FPhase      := FPhase + Delta;
      Trunced     := Trunc( FPhase);
      FWrapped    := Trunced > 0;
      FPhase      := FPhase - Trunced;
      if FPhase < 0
      then FPhase := FPhase + 1;

      FPosition   := FPhase + Clip( FInputs[ i_phase] * FInputs[ i_pmlevel], -1, 1);
      Trunced     := Trunc( FPosition);
      FPosition   := FPosition - Trunced;
      if FPosition < 0
      then FPosition := FPosition + 1;

      if FWrapped
      then FRandomValue := 1 - 2 * Random;

      FOutputs[ o_sync] := LogicToSignal( FWrapped);
//      FOutputs[ o_sync] := LogicToSignal( FPosition < 0.01);
    end;

    procedure   TModBaseLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f' , [ aPrefix, Name, FInputs [ i_res]               ], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, LogicToSignal( FPosition < 0.5)], AppLocale));
    end;

    function    TModBaseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_frequency  : Result := 'frequency';
        i_fm         : Result := 'fm';
        i_phase      : Result := 'phase';
        i_res        : Result := 'res';
        i_mute       : Result := 'mute';
        i_fmlevel    : Result := 'fmlevel';
        i_pmlevel    : Result := 'pmlevel';
        i_shape      : Result := 'shape';
        i_range      : Result := 'range';
        i_outputtype : Result := 'outputtype';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModBaseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out  : Result := 'out';
        o_sync : Result := 'sync';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseTrigLfo = class( TModLfo)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
    FDidRun  : Boolean;
  public
}

    constructor TModBaseTrigLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModBaseTrigLfo.DoSlowTick; // override;
    var
      FOldCounter : Integer;
      Delta       : TSignal;
      Trunced     : Integer;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := toIdle;
        FDidRun   := False;
      end;

      case FState of

        toIdle :
          begin
            FPosition := 0.0;
            FCounter  := 0;
            if SignalToLogic( FInputs[ i_trig])
            then begin
              FState  := toRunning;
              FDidRun := True;
            end;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FShape := Round( FInputs[ i_shape]);

            case Round( FInputs[ i_range]) of
              0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
              1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
              2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
              3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
              else Delta := LookupLfoLinPhaseDelta   ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
            end;

            FPosition   := FPosition + Delta;
            Trunced     := Trunc( FPosition);
            FPosition   := FPosition - Trunced;
            if FPosition < 0
            then FPosition := FPosition + 1;

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPosition, 0.5, Delta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPosition, 1.0, Delta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInputs[ i_count]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPosition, 0.5, Delta) and
                not AlmostEqual( FPosition, 1.0, Delta)
              then FState := toRunning;
            end;
          end;

        toDone :
          begin
            FPosition := 0.0;
            if not SignalToLogic( FInputs[ i_trig])
            then FState := toIdle;
          end;
      end;
      FOutputs[ o_ready] := LogicToSignal(( FState = toIdle) and FDidRun);
    end;

    function    TModBaseTrigLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig  : Result := 'trig';
        i_count : Result := 'count';
        else      Result := inherited;
      end;
    end;

    function    TModBaseTrigLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_ready : Result := 'ready';
        else      Result := inherited;
      end;
    end;

    procedure   TModBaseTrigLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs[ i_trig  ]              ], AppLocale));
      aCallback( Self, Format( '%s/%s/ready/%f', [ aPrefix, Name, FOutputs[ o_ready]              ], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'  , [ aPrefix, Name, LogicToSignal( FPosition <= 0.5)], AppLocale));
    end;


{ ========
  TModCounterLfo = class( TMod)
  private
    FPhase : TSignal;
  protected
}

    procedure   TModCounterLfo.Wrapped; // virtual;
    begin
      // To be overridden
    end;

//  public

    constructor TModCounterLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModCounterLfo.DoSlowTick; // override;
    var
      Delta   : TSignal;
      DidWrap : Boolean;
    begin
      if ResetFlag
      then FPhase := 0;
      ResetFlag := False;

      case Round( FInputs[ i_range]) of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
        else Delta := LookupLfoLinPhaseDelta   ( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
      end;

      FPhase  := FPhase - Delta;
      DidWrap := False;

      while FPhase < 0
      do begin
        if not DidWrap
        then begin
          Wrapped;
          DidWrap := True;
        end;
        FPhase := FPhase + 1.0;
      end;
    end;

    procedure   TModCounterLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, LogicToSignal( FPhase >= 0.5)], AppLocale));
    end;

    function    TModCounterLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_frequency : Result := 'frequency';
        i_fm        : Result := 'fm';
        i_fmlevel   : Result := 'fmlevel';
        i_range     : Result := 'range';
        else          Result := NO_INPUT;
      end;
    end;


{ ========
  TModLfo = class( TModBaseLfo)
  public
}

    procedure   TModLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      case FShape of
        0  : // sine
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( LookupSine( FPosition), Round( FInputs[ i_outputtype]));
        1  : // tri
          if FPosition <= 0.5
          then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( 4.0 *         FPosition  - 1, Round( FInputs[ i_outputtype]))
          else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( 4.0 * ( 1.0 - FPosition) - 1, Round( FInputs[ i_outputtype]));
        2  : // saw
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( 2.0 * FPosition - 1, Round( FInputs[ i_outputtype]));
        3  : // square
          if FPosition < 0.5
          then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  1.0, Round( FInputs[ i_outputtype]))
          else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( -1.0, Round( FInputs[ i_outputtype]));
        4  : // random PWM square
          if FRandomValue < 0
          then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  1.0, Round( FInputs[ i_outputtype]))
          else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( -1.0, Round( FInputs[ i_outputtype]));
        5  : // random
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  FRandomValue, Round( FInputs[ i_outputtype]))
      end;
    end;


{ ========
  TModLfoTrig = class( TModbaseTrigLfo)
  public
}

    procedure   TModLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      case FShape of
        0  : // sine
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( LookupSine( FPosition), Round( FInputs[ i_outputtype]));
        1  : // tri
          if FPosition <= 0.5
          then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( 4.0 *         FPosition  - 1, Round( FInputs[ i_outputtype]))
          else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( 4.0 * ( 1.0 - FPosition) - 1, Round( FInputs[ i_outputtype]));
        2  : // saw
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( 2.0 * FPosition - 1, Round( FInputs[ i_outputtype]));
        3  : // square
          if FPosition < 0.5
          then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  1.0, Round( FInputs[ i_outputtype]))
          else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( -1.0, Round( FInputs[ i_outputtype]));
        4  : // random PWM square
          if FRandomValue < 0
          then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  1.0, Round( FInputs[ i_outputtype]))
          else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( -1.0, Round( FInputs[ i_outputtype]));
        5  : // random
          FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  FRandomValue, Round( FInputs[ i_outputtype]))
      end;
    end;


{ ========
  TModSquareLfo = class( TModLfo)
  public
}

    procedure   TModSquareLfo.SetDefaults;
    begin
      inherited;
      FDezipperMap := [
        i_frequency,
        i_fmlevel,
        i_pmlevel,
        i_pwmlevel
      ];
    end;

    procedure   TModSquareLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if FPosition < Clip( 0.5 * ( 1 - FInputs[ i_pwm] * FInputs[ i_pwmlevel]), 0.01, 0.99)
      then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  1.0, Round( FInputs[ i_outputtype]))
      else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( -1.0, Round( FInputs[ i_outputtype]));
    end;

    function    TModSquareLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         i_pwm      : Result := 'pwm';
         i_pwmlevel : Result := 'pwmlevel';
        else          Result := inherited;
      end;
    end;


{ ========
  TModTrigSquareLfo = class( TModBaseTrigLfo)
  public
}

    procedure   TModSquareLfoTrig.SetDefaults;
    begin
      inherited;
      FDezipperMap := [
        i_frequency,
        i_fmlevel,
        i_pmlevel,
        i_pwmlevel
      ];
    end;

    procedure   TModSquareLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if FPosition < Clip( 0.5 * ( 1.0 - FInputs[ i_pwm] * FInputs[ i_pwmlevel]), 0.01, 0.99)
      then FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType(  1.0, Round( FInputs[ i_outputtype]))
      else FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( -1.0, Round( FInputs[ i_outputtype]));
    end;

    function    TModSquareLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         i_pwm      : Result := 'pwm';
         i_pwmlevel : Result := 'pwmlevel';
        else          Result := inherited;
      end;
    end;


{ ========
  TModRandSig = class( TModCounterLfo)
  private
    FStart  : TSignal;
    FTarget : TSignal;
  protected
}

    procedure   TModRandSig.Wrapped; // override;
    var
      Dist   : TSignal;
      Lambda : TSignal;
      Mode   : Integer;
    begin
      Dist    := Clip( FInputs[ i_dist] + FInputs[ i_distmod] * FInputs[ i_distmodamt], 0, 1);
      Mode    := Round( FInputs[ i_mode]);
      FStart  := FTarget;
      if Mode = m_exp
      then begin
        Lambda  := Max( FInputs[ i_lambda], 1.0000001);
        FTarget := Clip( Dist * ( RandomExpo( Lambda, False)) + ( 1 - Dist) * FStart, -1, 1);
      end
      else FTarget := Clip( Dist * ( 1 - 2 * Random) + ( 1 - Dist) * FStart, -1, 1);
    end;

//  public

    procedure   TModRandSig.SetDefaults; // override;
    begin
      FTarget := 2 * Random - 1;
      Wrapped; // Start with a wrap to make it start moving from the start.
      FDezipperMap :=
        FDezipperMap +
        [
          i_mute,
          i_dist,
          i_distmodamt
        ];
    end;

    procedure   TModRandSig.DoSlowTick; // override;
    var
      Value : TSignal;
    begin
      inherited;
      Value := FTarget + FPhase * ( FStart - FTarget); // Phase goes from 1 -> 0
      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( Value, Round( FInputs[ i_outputtype]));
    end;

    function    TModRandSig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_mute       : Result := 'mute';
        i_outputtype : Result := 'outputtype';
        i_dist       : Result := 'dist';
        i_distmod    : Result := 'distmod';
        i_distmodamt : Result := 'distmodamt';
        i_lambda     : Result := 'lambda';
        i_mode       : Result := 'mode';
        else           Result := inherited;
      end;
    end;

    function    TModRandSig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         o_out : Result := 'out';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPulses = class( TMod)
  private
    FPhase         : TSignal;
    FState         : TPulsesState;
    FPulseCounter  : Integer;
    FPeriodCounter : Integer;
    FOldFrequency  : TSignal;
    FOldLambda     : TSignal;
    FOldDist       : TSignal;
  public
}

    constructor TModPulses.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPulses.SetDefaults; // override;
    begin
      FDezipperMap :=
        [
          i_dist      ,
          i_distmodamt,
          i_lambda
        ];
    end;

    procedure   TModPulses.DoSlowTick; // override;
    var
      Delta      : TSignal;
      Frequency  : TSignal;
      Range      : Integer;
      Dist       : TSignal;
      Lambda     : TSignal;
      RndMode    : Integer;
      SyncMode   : Integer;
      OutputType : Integer;
      Wrap       : Boolean;
    begin
      Frequency  := Max( 1e-6, FInputs[ i_frequency]);
      Range      := Round( FInputs[ i_range]);
      Dist       := FInputs[ i_dist];
      Lambda     := Max( FInputs[ i_lambda], 1.0000001);
      RndMode    := Round( FInputs[ i_rndmode]);
      SyncMode   := Round( FInputs[ i_syncmode]);
      OutputType := Round( FInputs[ i_outputtype]);

      if
        ResetFlag or
        ( FOldFrequency <> Frequency) or
        ( FOldLambda    <> Lambda   ) or
        ( FOldDist      <> Dist     )
      then begin
        ResetFlag := False;
        FPhase    := 0.0;
        FState    := psIdle;
      end;
      FOldFrequency := Frequency;
      FOldLambda    := Lambda;
      FOldDist      := Dist;

      Dist := Clip( Dist + FInputs[ i_distmod] * FInputs[ i_distmodamt], 0, 1);

      case Range of
        0  : Delta := LookupLfoFastPhaseDelta  ( Frequency);
        1  : Delta := LookupLfoMediumPhaseDelta( Frequency);
        2  : Delta := LookupLfoSlowPhaseDelta  ( Frequency);
        3  : Delta := LookupLfoBPMPhaseDelta   ( Frequency);
        else Delta := LookupLfoLinPhaseDelta   ( Frequency);
      end;

      FPhase := FPhase + Delta;
      Wrap   := False;
      while FPhase > 1.0
      do begin
        FPhase := FPhase - 1.0;
        Wrap   := True;
      end;

      if ( SyncMode = m_sync_async) or Wrap
      then begin
        case FState of

          psIdle : begin
            FPulseCounter := Trunc( SlowTimeToSampleCount( 1e-3)); // Pulse for    1 ms
            FLedCounter   := Trunc( SlowTimeToSampleCount( 1e-1)); // Led on for 100 ms
            if RndMode = m_rnd_exp
            then begin
              if SyncMode = m_sync_async
              then FPeriodCounter := Max( 1, Trunc(( Control_Rate / Frequency ) * Dist * RandomExpo( Lambda, True)))
              else FPeriodCounter := Trunc( 25 * Dist * RandomExpo( Lambda, True));
            end
            else begin
              if SyncMode = m_sync_async
              then FPeriodCounter := Max( 1, Trunc(( Control_Rate / Frequency ) * Dist * Random))
              else FPeriodCounter := Trunc( 25 * Dist * Random);
            end;
            FState := psPulsing;
          end;

          else begin
            if FPeriodCounter > 0
            then Dec( FPeriodCounter)
            else FState := psIdle;
          end;
        end;
      end;

      if FLedCounter > 0
      then Dec( FLedCounter);

      if FState = psPulsing
      then begin
        if FPulseCounter > 0
        then Dec( FPulseCounter)
        else FState := psWaiting;
      end;

      FOutputs[ o_out] := SignalForOutputType( LogicToSignal( FState = psPulsing), OutputType);
    end;

    procedure   TModPulses.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, LogicToSignal( FLedCounter > 0)], AppLocale));
    end;

    function    TModPulses.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_frequency  : Result := 'frequency';
        i_range      : Result := 'range';
        i_dist       : Result := 'dist';
        i_distmod    : Result := 'distmod';
        i_distmodamt : Result := 'distmodamt';
        i_lambda     : Result := 'lambda';
        i_rndmode    : Result := 'rndmode';
        i_syncmode   : Result := 'syncmode';
        i_outputtype : Result := 'outputtype';
        else           Result := inherited;
      end;
    end;

    function    TModPulses.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRandomWalkLfo = class( TModLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
}

    constructor TModRandomWalkLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRandomWalkLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_frequency,
        i_fmlevel,
        i_distance,
        i_mute,
        i_color,
        i_colormodamt
      ];
      FInputs[ i_fm      ] := 0;
      FInputs[ i_colormod] := 0;
    end;

    procedure   TModRandomWalkLfo.DoSlowTick; // override;
    var
      Distance : TSignal;
      Delta    : TSignal;
      Color    : TSignal;
      Mute     : TSignal;
      OutType  : Integer;
      Lambda   : TSignal;
      Mode     : Integer;
    begin

      case Round( FInputs[ i_range]) of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ i_frequency] + Clip( FInputs[ i_fm] * FInputs[ i_fmlevel], -1, 1));
        1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ i_frequency] + Clip( FInputs[ i_fm] * FInputs[ i_fmlevel], -1, 1));
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ i_frequency] + Clip( FInputs[ i_fm] * FInputs[ i_fmlevel], -1, 1));
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ i_frequency] + Clip( FInputs[ i_fm] * FInputs[ i_fmlevel], -1, 1));
        else Delta := LookupLfoLinPhaseDelta   ( FInputs[ i_frequency] + Clip( FInputs[ i_fm] * FInputs[ i_fmlevel], -1, 1));
      end;

      FPosition := MathFloatMod( FPosition + Delta, 1.0);

      case FState of
        trwIdle :
          begin
            if FPosition >= 0.5
            then begin
              FState   := trwFired;
              Lambda   := Max( FInputs[ i_lambda], 1.0000001);
              Distance := FInputs[ i_distance];
              Color    := Clip( FInputs[ i_color] + FInputs[ i_colormod] * FInputs[ i_colormodamt], 0, 1);
              Mute     := SignalToMute( FInputs[ i_mute]);
              OutType  := Round( FInputs[ i_outputtype]);
              Mode     := Round( FInputs[ i_mode]);

              if Mode = m_exp
              then begin
                FX := ( 1 - Color) * FX + Color * ( FX + RandomExpo( Lambda, False) * Distance);
                FY := ( 1 - Color) * FY + Color * ( FY + RandomExpo( Lambda, False) * Distance);
                FZ := ( 1 - Color) * FZ + Color * ( FZ + RandomExpo( Lambda, False) * Distance);
              end
              else begin
                FX := ( 1 - Color) * FX + Color * ( FX + ( 1.0 - 2.0 * Random) * Distance);
                FY := ( 1 - Color) * FY + Color * ( FY + ( 1.0 - 2.0 * Random) * Distance);
                FZ := ( 1 - Color) * FZ + Color * ( FZ + ( 1.0 - 2.0 * Random) * Distance);
              end;

              if FX < -1 then FX := - 2 - FX else if FX > 1 then FX := 2 - FX;
              if FY < -1 then FY := - 2 - FY else if FY > 1 then FY := 2 - FY;
              if FZ < -1 then FZ := - 2 - FZ else if FZ > 1 then FZ := 2 - FZ;

              FOutputs[ o_outx] := Mute * SignalForOutputType( FX, OutType);
              FOutputs[ o_outy] := Mute * SignalForOutputType( FY, OutType);
              FOutputs[ o_outz] := Mute * SignalForOutputType( FZ, OutType);
            end;
          end;

        trwFired :
          begin
            if FPosition < 0.5
            then FState := trwIdle;
          end;
      end;
    end;

    procedure   TModRandomWalkLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/outx/%f', [ aPrefix, Name, FX], AppLocale));
      aCallback( Self, Format( '%s/%s/outy/%f', [ aPrefix, Name, FY], AppLocale));
      aCallback( Self, Format( '%s/%s/outz/%f', [ aPrefix, Name, FZ], AppLocale));
    end;

    function    TModRandomWalkLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_range       : Result := 'range';
        i_frequency   : Result := 'frequency';
        i_fm          : Result := 'fm';
        i_fmlevel     : Result := 'fmlevel';
        i_distance    : Result := 'distance';
        i_mute        : Result := 'mute';
        i_color       : Result := 'color';
        i_colormod    : Result := 'colormod';
        i_colormodamt : Result := 'colormodamt';
        i_outputtype  : Result := 'outputtype';
        i_lambda      : Result := 'lambda';
        i_mode        : Result := 'mode';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModRandomWalkLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outx : Result := 'outx';
        o_outy : Result := 'outy';
        o_outz : Result := 'outz';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAttractor = class( TModBaseLfo)
  private
    FLorenz  : TLorenzOsc;
    FRossler : TRosslerOsc;
    FX       : TSignal;
    FY       : TSignal;
    FZ       : TSignal;
  public
}

    constructor TModAttractorLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast  := False;
      FIsSlow  := True;
      FLorenz  := TLorenzOsc.Create;
      FRossler := TRosslerOsc.Create;
    end;

    destructor  TModAttractorLfo.Destroy; // override;
    begin
      FLorenz .DisposeOf;
      FRossler.DisposeOf;
      inherited;
    end;

    procedure   TModAttractorLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_frequency,
        i_fm       ,
        i_fmlevel  ,
        i_mute
      ];
    end;

    procedure   TModAttractorLfo.DoSlowTick; // override;
    var
      Mode      : Integer;
      OutType   : Integer;
      Range     : Integer;
      Mute      : TSignal;
      Frequency : TSignal;
    begin
      if Assigned( FLorenz) and Assigned( FRossler)
      then begin
        Mode    := Round       ( FInputs[ i_mode      ]);
        OutType := Round       ( FInputs[ i_outputtype]);
        Range   := Round       ( FInputs[ i_range     ]);
        Mute    := SignalToMute( FInputs[ i_mute      ]);

        case Range of
          0  : Frequency := LookupLfoFastFrequency  ( Clip( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel], -1.0, 1.0));
          1  : Frequency := LookupLfoMediumFrequency( Clip( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel], -1.0, 1.0));
          2  : Frequency := LookupLfoSlowFrequency  ( Clip( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel], -1.0, 1.0));
          3  : Frequency := LookupLfoBPMFrequency   ( Clip( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel], -1.0, 1.0));
          else Frequency := LookupLfoLinFrequency   ( Clip( FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel], -1.0, 1.0));
        end;

        case Mode of
          m_lorenz : begin if ResetFlag then FLorenz .Reset; FLorenz .SetFreq( Frequency); FLorenz .Tick( FX, FY, FZ); end;
          else       begin if ResetFlag then FRossler.Reset; FRossler.SetFreq( Frequency); FRossler.Tick( FX, FY, FZ); end;
        end;

        FOutputs[ o_outx] := SignalForOutputType( FX * Mute, OutType);
        FOutputs[ o_outy] := SignalForOutputType( FY * Mute, OutType);
        FOutputs[ o_outz] := SignalForOutputType( FZ * Mute, OutType);
      end;
      ResetFlag := False;
    end;

    procedure   TModAttractorLfo.SampleRateChanged; // override;
    begin
      if Assigned( FLorenz ) then FLorenz .SetSampleRate( Control_Rate);
      if Assigned( FRossler) then FRossler.SetSampleRate( Control_Rate);
    end;

    procedure   TModAttractorLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/outx/%f', [ aPrefix, Name, FX], AppLocale));
      aCallback( Self, Format( '%s/%s/outy/%f', [ aPrefix, Name, FY], AppLocale));
      aCallback( Self, Format( '%s/%s/outz/%f', [ aPrefix, Name, FZ], AppLocale));
    end;

    function    TModAttractorLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_range      : Result := 'range';
        i_frequency  : Result := 'frequency';
        i_fm         : Result := 'fm';
        i_fmlevel    : Result := 'fmlevel';
        i_mute       : Result := 'mute';
        i_outputtype : Result := 'outputtype';
        i_mode       : Result := 'mode';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModAttractorLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outx : Result := 'outx';
        o_outy : Result := 'outy';
        o_outz : Result := 'outz';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoise = class( TMod)
  public
}

    constructor TModNoise.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModNoise.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_color,
        i_colormodamt,
        i_mute
      ];
    end;

    procedure   TModNoise.DoTick; // override;
    var
      Color : TSignal;
    begin
      Color := Clip( FInputs[ i_color] + FInputs[ i_colormod] * FInputs[ i_colormodamt], 0, 1);
      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * (( 1 - Color) * FOutputs[ o_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
    end;

    function    TModNoise.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_color       : Result := 'color';
        i_colormod    : Result := 'colormod';
        i_colormodamt : Result := 'colormodamt';
        i_mute        : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModNoise.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseTrig = class( TMod)
  private
    FLastInput: TSignal;
  public
}

    constructor TModNoiseTrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModNoiseTrig.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_color,
        i_colormodamt,
        i_mute
      ];
    end;

    procedure   TModNoiseTrig.DoTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInputs[ i_trig]);
      if not FLastTrig and Trig
      then begin
        Color := Clip( FInputs[ i_color] + FInputs[ i_colormod] * FInputs[ i_colormodamt], 0, 1);
        FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * (( 1 - Color) * FOutputs[ o_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModNoiseTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_color       : Result := 'color';
        i_colormod    : Result := 'colormod';
        i_colormodamt : Result := 'colormodamt';
        i_trig        : Result := 'trig';
        i_mute        : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModNoiseTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfo = class( TMod)
  private
    FOutValue : TSignal;
  public
}

    constructor TModNoiseLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNoiseLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_color,
        i_colormodamt,
        i_mute
      ];
    end;

    procedure   TModNoiseLfo.DoSlowTick; // override;
    var
      Color : TSignal;
    begin
      Color     := Clip( FInputs[ i_color] + FInputs[ i_colormod] * FInputs[ i_colormodamt], 0, 1);
      FOutValue := (( 1 - Color) * FOutputs[ o_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * SignalForOutputType( FOutValue, Round( FInputs[ i_outputtype]));
    end;

    procedure   TModNoiseLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutValue], AppLocale));
    end;

    function    TModNoiseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_color       : Result := 'color';
        i_colormod    : Result := 'colormod';
        i_colormodamt : Result := 'colormodamt';
        i_mute        : Result := 'mute';
        i_outputtype  : Result := 'outputtype';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfoTrig = class( TMod)
  private
    FLastTrig : Boolean;
    FOutValue : TSignal;
  public
}

    constructor TModNoiseLfoTrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNoiseLfoTrig.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_color,
        i_colormodamt,
        i_mute
      ];
    end;

    procedure   TModNoiseLfoTrig.DoSlowTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInputs[ i_trig]);
      if not FLastTrig and Trig
      then begin
        Color     := Clip( FInputs[ i_color] + FInputs[ i_colormod] * FInputs[ i_colormodamt], 0, 1);
        FOutValue := (( 1 - Color) * FOutputs[ o_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
        FOutputs[ o_out] :=  SignalToMute( FInputs[ i_mute]) * SignalForOutputType( FOutValue, Round( FInputs[ i_outputtype]));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseLfoTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs [ i_trig]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutValue                   ], AppLocale));
    end;

    function    TModNoiseLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_color       : Result := 'color';
        i_colormod    : Result := 'colormod';
        i_colormodamt : Result := 'colormodamt';
        i_trig        : Result := 'trig';
        i_mute        : Result := 'mute';
        i_outputtype  : Result := 'outputtype';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfoTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAverage = class( TMod)
  private
    FAlpha : TSignal;  // Alpha = System_Rate / ( 2 * PI * Freq + System_Rate) ... Fc := (log(b1))/(-2.0 * M_PI) * Fs; (*b1 < 0.0432139182637722579772*)
  public
}

    constructor TModAverage.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModAverage.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_alpha,
        i_alphamodamt
      ];
    end;

    procedure   TModAverage.DoSlowTick; // override;
    var
      Inp   : TSignal;
      Outp  : TSignal;
      Delta : TSignal;
    begin
      FAlpha := Normalize( ExpAttack( FInputs[ i_alpha] + FInputs[ i_alphamod] * FInputs[ i_alphamodamt], 1, 0, 1));
      if ResetFlag
      then begin
        Inp   := 0;
        Outp  := 0;
      end
      else begin
        Inp   := FInputs [ i_in];
        Outp  := FOutputs[ o_lp];
      end;

      Delta           := Inp  - Outp;
      Outp            := Outp + FAlpha * Delta;
      FOutputs[ o_lp] := Normalize( Outp );
      FOutputs[ o_hp] := Normalize( Delta);

      ResetFlag := False;
    end;

    function    TModAverage.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in          : Result := 'in';
        i_alpha       : Result := 'alpha';
        i_alphamod    : Result := 'alphamod';
        i_alphamodamt : Result := 'alphamodamt';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModAverage.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_lp : Result := 'lp';
        o_hp : Result := 'hp';
        else   Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSVF = class( TMod)
  private
    FFreq  : TSignal;
    FDamp  : TSignal;
    FDrive : TSignal;
  public
}

    (*
      from    : http://www.musicdsp.org/archive.php?classid=3#92
      also see: http://www.music.mcgill.ca/~ich/classes/FiltersChap2.pdf

      input  = input buffer;
      output = output buffer;
      fs     = sampling frequency;
      fc     = cutoff frequency normally something like:
               440.0*pow(2.0, (midi_note - 69.0)/12.0);                 :: We use signals rather, Frequency = Signal * SampleRate / 2
      res    = resonance 0 to 1;
      drive  = internal distortion 0 to 0.1
      freq   = 2.0*sin(PI*MIN(0.25, fc/(fs*2)));  // the fs*2 is because it's double sampled
      damp   = MIN(2.0*(1.0 - pow(res, 0.25)), MIN(2.0, 2.0/freq - freq*0.5));
      notch  = notch output
      low    = low pass output
      high   = high pass output
      band   = band pass output
      peak   = peaking output = low - high
      --
      double sampled svf loop:
      for (i=0; i<numSamples; i++)
      {
        in    = input[i];
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out   = 0.5*(notch or low or high or band or peak);
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out  += 0.5*(same out as above);
        output[i] = out;
      }
    *)

    constructor TModSVF.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModSVF.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_frequency,
        i_q,
        i_distortion,
        i_fmlevel,
        i_inlevel
      ];
    end;

    procedure   TModSVF.DoSlowTick; // override;
    begin
      // NOTE: this assumes an SVF sampled at twice the system rate

      FFreq := LookupSvfFrequency( FInputs[ i_frequency] + FInputs[ i_freq] + FInputs[ i_fm] * FInputs[ i_fmlevel]);
      FDamp := 1 - Clip( FInputs[ i_q], 0.01, 1);

      // see: http://www.earlevel.com/main/2003/03/02/the-digital-state-variable-filter/
      // instability when FFreq aproaches 1

      if FFreq > 0.99
      then FFreq := 0.99;

      FDrive := Clip( FInputs[ i_distortion] * 0.1, 0.0, 0.1);
    end;

    procedure   TModSVF.DoTick; // override;
    var
      INP : TSignal;
      HP  : TSignal;
      BP  : TSignal;
      LP  : TSignal;
      BR  : TSignal;
    begin
      INP := FInputs [ i_in] * FInputs[ i_inlevel];
      LP  := FOutputs[ o_lp];
      BP  := FOutputs[ o_bp];

      BR := INP - FDamp * BP;
      LP := LP  + FFreq * BP;
      HP := BR  - LP;
      BP := FFreq * HP + BP - FDrive * BP * BP * BP; // with 0 <= Drive <= 0.1
      BR := INP - FDamp * BP;
      LP := LP  + FFreq * BP;
      HP := BR  - LP;
      BP := FFreq * HP + BP - FDrive * BP * BP * BP;

      FOutputs[ o_lp] := Normalize( LP);
      FOutputs[ o_bp] := Normalize( BP);
      FOutputs[ o_hp] := Normalize( HP);
      FOutputs[ o_br] := Normalize( BR);
    end;

    function    TModSVF.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in         : Result := 'in';
        i_frequency  : Result := 'frequency';
        i_q          : Result := 'q';
        i_distortion : Result := 'distortion';
        i_fm         : Result := 'fm';
        i_fmlevel    : Result := 'fmlevel';
        i_inlevel    : Result := 'inlevel';
        i_freq       : Result := 'freq';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModSVF.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_lp : Result := 'lp';
        o_bp : Result := 'bp';
        o_hp : Result := 'hp';
        o_br : Result := 'br';
        else   Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPinkFilter = class( TMod)
  private
    FB0 : TSignal;
    FB1 : TSignal;
    FB2 : TSignal;
    FB3 : TSignal;
    FB4 : TSignal;
    FB5 : TSignal;
    FB6 : TSignal;
    FG  : TSignal;
  public
}

    constructor TModPinkFilter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModPinkFilter.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_gain
      ];
      FInputs[ i_in] := 0;
      FB0            := 0.0;
      FB1            := 0.0;
      FB2            := 0.0;
      FB3            := 0.0;
      FB4            := 0.0;
      FB5            := 0.0;
      FB6            := 0.0;
      SampleRateChanged;
    end;

    procedure   TModPinkFilter.DoTick; // override;
    // http://www.musicdsp.org/files/pink.txt
    var
      White : TSignal;
    begin
      White := Normalize( FInputs[ i_in]);
      case Round( FInputs[ i_mode]) of
        0 : begin
          FB0 := Normalize( 0.99886 * FB0 + White * 0.0555179);
          FB1 := Normalize( 0.99332 * FB1 + White * 0.0750759);
          FB2 := Normalize( 0.96900 * FB2 + White * 0.1538520);
          FB3 := Normalize( 0.86650 * FB3 + White * 0.3104856);
          FB4 := Normalize( 0.55000 * FB4 + White * 0.5329522);
          FB5 := Normalize( -0.7616 * FB5 - White * 0.0168980);
          FOutputs[ o_out] := FInputs[ i_gain] * FG * ( FB0 + FB1 + FB2 + FB3 + FB4 + FB5 + FB6 + White * 0.5362);
          FB6 := White * 0.115926;
        end;
        else begin
          FB0 := Normalize( 0.99765 * FB0 + White * 0.0990460);
          FB1 := Normalize( 0.96300 * FB1 + White * 0.2965164);
          FB2 := Normalize( 0.57000 * FB2 + White * 1.0526913);
          FOutputs[ o_out] := FInputs[ i_gain] * FG * ( FB0 + FB1 + FB2 + White * 0.1848);
        end
      end;
    end;

    procedure   TModPinkFilter.SampleRateChanged; // override;
    begin
      FG  := dBToSignal( -13) * ( 44100 * System_Rate_Rec);
    end;

    function    TModPinkFilter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in   : Result := 'in';
        i_mode : Result := 'mode';
        i_gain : Result := 'gain';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModPinkFilter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModResonator = class( TMod)
  // http://www.katjaas.nl/complexintegrator/complexresonator.html
  private
    FOldRe  : TSignal;
    FOldIm  : TSignal;
    FRadius : TSignal;
    FSin    : TSignal;
    FCos    : TSignal;
  public
}

    constructor TModResonator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModResonator.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_decay    ,
        i_fmlevel  ,
        i_frequency,
        i_cents    ,
        i_pmlevel  ,
        i_dmlevel  ,
        i_click
      ];
    end;

    procedure   TModResonator.DoSlowTick; // override;
    const
      PM_ATT = 1.0 / 32.0;
    var
      aFreq     : TSignal;
      anAngle   : TSignal;
      aDecayMod : TSignal;
      aDecay    : TSignal;
    begin
      aFreq     := LookupFrequency( FInputs[ i_freq] + FInputs[ i_frequency] + FInputs[ i_fm] * FInputs[ i_fmlevel] + FInputs[ i_cents]);
      anAngle   := TWO_PI * (( aFreq * System_Rate_Rec) + FInputs[ i_pm] * FInputs[ i_pmlevel] * PM_ATT);
      aDecayMod := FInputs[ i_dm] * FInputs[ i_dmlevel];
      aDecay    := FInputs[ i_decay] * ( 1 + aDecayMod);
      if aDecay <= 0
      then aDecay := 10e-6;   // decay time is seconds - so that's 10 s minimal time
      FRadius := Normalize( exp( - System_Rate_Rec / aDecay));
      SinCos( anAngle, FSin, FCos);
    end;

    procedure   TModResonator.DoTick; // override;
    var
      inRe : TSignal;
      inIm : TSignal;
    begin
      inRe             := FRadius * FOldRe + FInputs[ i_in] * FInputs[ i_inlevel];   // Use Re. for input
      inIm             := FRadius * FOldIm;
      FOldRe           := Normalize( FCos * inRe - FSin * inIm);
      FOldIm           := Normalize( FSin * inRe + FCos * inIm);
      FOutputs[ o_out] := FOldIm + FInputs[ i_click] * ( FOldRe - FOldIm);           // and then Im. for output
    end;

    function    TModResonator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_freq      : Result := 'freq';
        i_decay     : Result := 'decay';
        i_in        : Result := 'in';
        i_fm        : Result := 'fm';
        i_fmlevel   : Result := 'fmlevel';
        i_frequency : Result := 'frequency';
        i_cents     : Result := 'cents';
        i_pm        : Result := 'pm';
        i_pmlevel   : Result := 'pmlevel';
        i_dm        : Result := 'dm';
        i_dmlevel   : Result := 'dmlevel';
        i_click     : Result := 'click';
        i_inlevel   : Result := 'inlevel';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModResonator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFreqShifter = class( TMod)
  // https://github.com/swh/ladspa/blob/master/bode_shifter_1431.xml
  private
    FDelay : array[ 0 .. 255] of TSignal;
    FDPtr  : Integer;
    FPhase : TSignal;
    FDelta : TSignal;
  public
}

    constructor TModFreqShifter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModFreqShifter.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_shift
      ];
    end;

    procedure   TModFreqShifter.DoSlowTick; // override;
    begin
      FDelta := Clip( LookupFrequencyCubed( FInputs[ i_shift]) + LookupFrequencyLin( FInputs[ i_shiftmod] * FInputs[ i_shiftdepth]), 0.0, 10000.0)  * System_Rate_Rec;
    end;

    procedure   TModFreqShifter.DoTick; // override;
    var
      i       : Integer;
      d       : Word;
      hilb    : TSignal;
      rm1     : TSignal;
      rm2     : TSignal;
      Trunced : Integer;
    begin
      FDelay[ FDPtr] := Normalize( FInputs[ i_in]);

      hilb := 0;
      d    := FDPtr;

      for i := Low( c_coefficients) to High( c_coefficients)
      do begin
        hilb := hilb + c_coefficients[ i] * FDelay[ d];
        d    := ( d + 254) and 255;
      end;

      FPhase  := FPhase + FDelta;
      Trunced := Trunc( FPhase);
      FPhase  := FPhase - Trunced;

      if FPhase < 0
      then FPhase := FPhase + 1;

      rm1 := hilb * 0.63661978              * LookupSine  ( FPhase);
      rm2 := FDelay[ ( FDPtr - 99) and 255] * LookupCosine( FPhase);

      FOutputs[ o_down] := ( rm2 - rm1) * 0.5;
      FOutputs[ o_up  ] := ( rm2 + rm1) * 0.5;

      FDPtr := ( FDPtr + 1) and 255;
    end;

    function    TModFreqShifter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in         : Result := 'in';
        i_shift      : Result := 'shift';
        i_shiftmod   : Result := 'shiftmod';
        i_shiftdepth : Result := 'shiftdepth';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModFreqShifter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_up   : Result := 'up';
        o_down : Result := 'down';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FEnvValue     : TSignal;
  public
}

    constructor TModEnvAR.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModEnvAR.SetDefaults; // override;
    begin
      FInputs[ i_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        i_attack,
        i_release
      ];
    end;

    procedure   TModEnvAR.DoTick; // override;
    var
      R : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := arIdle;
        FValue    := 0;
      end;

      case FState of

        arIdle :
          if SignalToLogic( FInputs[ i_trig])
          then begin
            if FInputs[ i_attack] <= 0
            then FCoeffAttack := 1
            else FCoeffAttack := 1 / Max( 1, TimeToSampleCount( LookupEnvTime( erFast, FInputs[ i_attack]))); // Linear attack
            if FInputs[ i_release] <= 0
            then FCoeffRelease := -1
            else FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / Max( 1, TimeToSampleCount( LookupEnvTime( erFast, FInputs[ i_release]))); // Exp release
            FState := arA;
          end;

        arA :
          begin
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
          end;

        arR :
          begin
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
          end;

        arWait :
          if not SignalToLogic( FInputs[ i_trig])
          then FState := arIdle;
      end;

      case Round( FInputs[ i_mute]) of
        1  : FEnvValue := Normalize(  EnvelopeSmoothing * FEnvValue                              / ( 1 + EnvelopeSmoothing));
        2  : FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + 1)                         / ( 1 + EnvelopeSmoothing));
        else FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + FValue * FInputs[ i_gain]) / ( 1 + EnvelopeSmoothing));
      end;

      FOutputs[ o_out   ] := FEnvValue;
      FOutputs[ o_sigout] := FEnvValue * FInputs[ i_sigin];
    end;

    procedure   TModEnvAR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModEnvAR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig    : Result := 'trig';
        i_attack  : Result := 'attack';
        i_release : Result := 'release';
        i_gain    : Result := 'gain';
        i_sigin   : Result := 'sigin';
        i_mute    : Result := 'mute';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModEnvAR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_sigout : Result := 'sigout';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
    FEnvValue     : TSignal;
  public
}

    constructor TModEnvARRetrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModEnvARRetrig.SetDefaults; // override;
    begin
      FInputs[ i_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        i_attack,
        i_release
      ];
    end;

    procedure   TModEnvARRetrig.DoTick; // override;
    var
      R : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := arIdle;
        FValue    := 0;
      end;

      if FArmed and SignalToLogic( FInputs[ i_trig])
      then FState := arIdle;

      case FState of

        arIdle :
          if SignalToLogic( FInputs[ i_trig])
          then begin
            FCoeffAttack  := 1 / Max( 1, TimeToSampleCount( LookupEnvTime( erFast, FInputs[ i_attack]))); // Linear attack
            FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / Max( 1, TimeToSampleCount( LookupEnvTime( erFast, FInputs[ i_release]))); // Exp release
            FState        := arA;
            FArmed        := False;
          end;

        arA :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInputs[ i_trig]);
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
          end;

        arR :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInputs[ i_trig]);
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
          end;

        arWait :
          if not SignalToLogic( FInputs[ i_trig])
          then FState := arIdle;

      end;

      case Round( FInputs[ i_mute]) of
        1  : FEnvValue := Normalize(  EnvelopeSmoothing * FEnvValue                              / ( 1 + EnvelopeSmoothing));
        2  : FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + 1)                         / ( 1 + EnvelopeSmoothing));
        else FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + FValue * FInputs[ i_gain]) / ( 1 + EnvelopeSmoothing));
      end;

      FOutputs[ o_out   ] := FEnvValue;
      FOutputs[ o_sigout] := FEnvValue * FInputs[ i_sigin];
    end;

    procedure   TModEnvARRetrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModEnvARRetrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig    : Result := 'trig';
        i_attack  : Result := 'attack';
        i_release : Result := 'release';
        i_gain    : Result := 'gain';
        i_sigin   : Result := 'sigin';
        i_mute    : Result := 'mute';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModEnvARRetrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_sigout : Result := 'sigout';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAHD = class( TMod)
  private
    FData      : TSignalArray;
    FState     : TAHDState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
}

    constructor TModEnvAHD.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEnvAHD.SetDefaults; // override;
    begin
      FInputs[ i_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        i_attack,
        i_hold,
        i_decay,
        i_amodlevel,
        i_hmodlevel,
        i_dmodlevel
      ];
    end;

    procedure   TModEnvAHD.DoSlowTick; // override;
    var
      Trig : Boolean;
      AMod : TSignal;
      HMod : TSignal;
      DMod : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := ahdIdle;
        FValue    := 0;
      end;

      Trig      := FTrigFlag;
      FTrigFlag := False;

      if FArmed and Trig and ( Round( FInputs[ i_mode]) = 1)
      then FState := ahdIdle;

      case FState of

        ahdIdle :
          if Trig
          then begin
            FArmed     := False;
            FLastLevel := FValue;
            FTime      := 0;
            AMod       := ( FInputs[ i_amod] - FInputs[ i_amodneg]) * FInputs[ i_amodlevel];
            FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( LookupEnvTime( SignalToEnvRange( FInputs[ i_arange]), FInputs[ i_attack] + AMod))));
            FState     := ahdA;
          end;

        ahdA :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FTime     := 0;
              HMod      := ( FInputs[ i_hmod] - FInputs[ i_hmodneg]) * FInputs[ i_hmodlevel];
              FDuration := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( LookupEnvTime( SignalToEnvRange( FInputs[ i_hrange]), FInputs[ i_hold] + HMod))));
              FState    := ahdH;
            end
            else begin
              case SignalToEnvShape( FInputs[ i_ashape]) of
                esLog : FValue := LogAttack( FTime, FDuration, FLastLevel, 1);
                esLin : FValue := LinAttack( FTime, FDuration, FLastLevel, 1);
                esExp : FValue := ExpAttack( FTime, FDuration, FLastLevel, 1);
                else    FValue := SAttack  ( FTime, FDuration, FLastLevel, 1);
              end;
            end;
          end;

        ahdH :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FLastLevel := FValue;
              FTime      := 0;
              DMod       := ( FInputs[ i_dmod] - FInputs[ i_dmodneg]) * FInputs[ i_dmodlevel];
              FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( LookupEnvTime( SignalToEnvRange( FInputs[ i_drange]), FInputs[ i_decay] + DMod))));
              FState     := ahdD;
            end;
          end;

        ahdD :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then FState := ahdWait
            else begin
              case SignalToEnvShape( FInputs[ i_dshape]) of
                esLog : FValue := LogDecay( FTime, FDuration, FLastLevel, 0);
                esLin : FValue := LinDecay( FTime, FDuration, FLastLevel, 0);
                esExp : FValue := ExpDecay( FTime, FDuration, FLastLevel, 0);
                else    FValue := SDecay  ( FTime, FDuration, FLastLevel, 0);
              end;
            end;
          end;

        ahdWait :
          if not Trig or FArmed
          then FState := ahdIdle;
      end;

      case Round( FInputs[ i_mute]) of
        1  : FEnvValue := Normalize(  EnvelopeSmoothingSlow * FEnvValue                              / ( 1 + EnvelopeSmoothingSlow));
        2  : FEnvValue := Normalize(( EnvelopeSmoothingSlow * FEnvValue + 1)                         / ( 1 + EnvelopeSmoothingSlow));
        else FEnvValue := Normalize(( EnvelopeSmoothingSlow * FEnvValue + FValue * FInputs[ i_gain]) / ( 1 + EnvelopeSmoothingSlow));
      end;

      FOutputs[ o_out] := SignalForEnvOutputType( FEnvValue, Round( FInputs[ i_outputtype]));
    end;

    procedure   TModEnvAHD.DoTick; // override;
    begin
      if SignalToLogic( FInputs[ i_trig])
      then FTrigFlag := True;
      FOutputs[ o_sigout] := FInputs[ i_sigin] * FOutputs[ o_out]; // sigout := sigin * out
    end;

    procedure   TModEnvAHD.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, LogicToSignal( not ( FState in [ ahdIdle, ahdWait]))], AppLocale));
      aCallback( Self, Format( '%s/%s/trigin/%f', [ aPrefix, Name, FInputs[ i_trig]                                    ], AppLocale));
    end;

    procedure   TModEnvAHD.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime    : TSignal;
      HTime    : TSignal;
      DTime    : TSignal;
      TTime    : TSignal;
      TTimeRec : TSignal;
      i        : Integer;
      p        : Integer;
      AShape   : TEnvShape;
      DShape   : TEnvShape;
      R        : TSignal;
    begin
      ATime  := LookupEnvTime( SignalToEnvRange( FInputs[ i_arange]), FInputs[ i_attack]);
      HTime  := LookupEnvTime( SignalToEnvRange( FInputs[ i_hrange]), FInputs[ i_hold  ]);
      DTime  := LookupEnvTime( SignalToEnvRange( FInputs[ i_drange]), FInputs[ i_decay ]);
      TTime  := ATime + HTime + DTime;

      if TTIme = 0
      then Exit;

      TTimeRec := 67 / TTime;
      ATime    := ATime * TTimeRec;
      HTime    := HTime * TTimeRec;
      DTime    := DTime * TTimeRec;
      AShape   := SignalToEnvShape( FInputs[ i_ashape]);
      DShape   := SignalToEnvShape( FInputs[ i_dshape]);

      i := 0;
      while i < ATime
      do begin
        case AShape of
          esLog : R := LogAttack( i, ATime, 0, 1);
          esLin : R := LinAttack( i, ATime, 0, 1);
          esExp : R := ExpAttack( i, ATime, 0, 1);
          else    R := SAttack  ( i, ATime, 0, 1);
        end;
        FData[ i] :=  SignalForEnvOutputType( R, Round( FInputs[ i_outputtype]));
        Inc( i);
      end;
      while i < ATime + HTime
      do begin
        FData[ i] := SignalForEnvOutputType( 1, Round( FInputs[ i_outputtype]));
        Inc( i);
      end;
      p := 0;
      while i < 67
      do begin
        case DShape of
          esLog : R := LogDecay( p, DTime, 1, 0);
          esLin : R := LinDecay( p, DTime, 1, 0);
          esExp : R := ExpDecay( p, DTime, 1, 0);
          else    R := SDecay  ( p, DTime, 1, 0);
        end;
        FData[ i] := SignalForEnvOutputType( R, Round( FInputs[ i_outputtype]));
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvAHD.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig       : Result := 'trig';
        i_mode       : Result := 'mode';
        i_ashape     : Result := 'ashape';
        i_dshape     : Result := 'dshape';
        i_attack     : Result := 'attack';
        i_hold       : Result := 'hold';
        i_decay      : Result := 'decay';
        i_gain       : Result := 'gain';
        i_sigin      : Result := 'sigin';
        i_mute       : Result := 'mute';
        i_arange     : Result := 'arange';
        i_hrange     : Result := 'hrange';
        i_drange     : Result := 'drange';
        i_amod       : Result := 'amod';
        i_hmod       : Result := 'hmod';
        i_dmod       : Result := 'dmod';
        i_amodneg    : Result := 'amodneg';
        i_hmodneg    : Result := 'hmodneg';
        i_dmodneg    : Result := 'dmodneg';
        i_amodlevel  : Result := 'amodlevel';
        i_hmodlevel  : Result := 'hmodlevel';
        i_dmodlevel  : Result := 'dmodlevel';
        i_outputtype : Result := 'outputtype';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModEnvAHD.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_sigout : Result := 'sigout';
        else       Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModEnvADSR = class( TMod)
  private
    FData      : TSignalArray;
    FState     : TADSRState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
}

    constructor TModEnvADSR.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEnvADSR.SetDefaults; // override;
    begin
      FInputs[ i_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        i_attack,
        i_decay,
        i_sustain,
        i_release,
        i_amodlevel,
        i_dmodlevel,
        i_smodlevel,
        i_rmodlevel
      ];
    end;

    procedure   TModEnvADSR.DoSlowTick; // override;
    var
      Trig : Boolean;
      AMod : TSignal;
      DMod : TSignal;
      SMod : TSignal;
      RMod : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := adsrIdle;
        FValue    := 0;
      end;

      Trig      := FTrigFlag;
      FTrigFlag := False;

      if FArmed and Trig and ( Round( FInputs[ i_mode]) = 1)
      then FState := adsrIdle;

      case FState of

        adsrIdle :

          if Trig
          then begin
            FArmed     := False;
            FLastLevel := FValue;
            FTime      := 0;
            AMod       := ( FInputs[ i_amod] - FInputs[ i_amodneg]) * FInputs[ i_amodlevel];
            FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( LookupEnvTime( SignalToEnvRange( FInputs[ i_arange]), FInputs[ i_attack] + AMod))));
            FState     := adsrA;
          end;

        adsrA :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FLastLevel := FValue;
              FTime      := 0;
              DMod       := ( FInputs[ i_dmod] - FInputs[ i_dmodneg]) * FInputs[ i_dmodlevel];
              FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( LookupEnvTime( SignalToEnvRange( FInputs[ i_drange]), FInputs[ i_decay] + DMod))));
              FState     := adsrD;
            end
            else begin
              case SignalToEnvShape( FInputs[ i_ashape ]) of
                esLog : FValue := LogAttack( FTime, FDuration, FLastLevel, 1);
                esLin : FValue := LinAttack( FTime, FDuration, FLastLevel, 1);
                esExp : FValue := ExpAttack( FTime, FDuration, FLastLevel, 1);
                else    FValue := SAttack  ( FTime, FDuration, FLastLevel, 1);
              end;
            end;
          end;

        adsrD :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FLastLevel := FValue;
              FState     := adsrS;
            end
            else begin
              SMod := ( FInputs[ i_smod] - FInputs[ i_smodneg]) * FInputs[ i_smodlevel];
              case SignalToEnvShape( FInputs[ i_dshape ]) of
                esLog : FValue := LogDecay( FTime, FDuration, FLastLevel, FInputs[ i_sustain] + SMod);
                esLin : FValue := LinDecay( FTime, FDuration, FLastLevel, FInputs[ i_sustain] + SMod);
                esExp : FValue := ExpDecay( FTime, FDuration, FLastLevel, FInputs[ i_sustain] + SMod);
                else    FValue := SDecay  ( FTime, FDuration, FLastLevel, FInputs[ i_sustain] + SMod);
              end;
            end;
          end;

        adsrS :
          begin
            if not FArmed
            then FArmed := not Trig;

            if FArmed
            then begin
              FTime     := 0;
              RMod      := ( FInputs[ i_rmod] - FInputs[ i_rmodneg]) * FInputs[ i_rmodlevel];
              FDuration := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( LookupEnvTime( SignalToEnvRange( FInputs[ i_rrange]), FInputs[ i_release] + RMod))));
              FState    := adsrR;
            end;
          end;

        adsrR :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then FState := adsrWait
            else begin
              case SignalToEnvShape( FInputs[ i_rshape ]) of
                esLog : FValue := LogDecay( FTime, FDuration, FLastLevel, 0);
                esLin : FValue := LinDecay( FTime, FDuration, FLastLevel, 0);
                esExp : FValue := ExpDecay( FTime, FDuration, FLastLevel, 0);
                else    FValue := SDecay  ( FTime, FDuration, FLastLevel, 0);
              end;
            end;
          end;

        adsrWait :
          if not Trig or FArmed
          then FState := adsrIdle;
      end;

      case Round( FInputs[ i_mute]) of
        1  : FEnvValue := Normalize(  EnvelopeSmoothingSlow * FEnvValue                              / ( 1 + EnvelopeSmoothingSlow));
        2  : FEnvValue := Normalize(( EnvelopeSmoothingSlow * FEnvValue + 1 * 1)                     / ( 1 + EnvelopeSmoothingSlow));
        else FEnvValue := Normalize(( EnvelopeSmoothingSlow * FEnvValue + FValue * FInputs[ i_gain]) / ( 1 + EnvelopeSmoothingSlow));
      end;
      FOutputs[ o_out] := SignalForEnvOutputType( FEnvValue, Round( FInputs[ i_outputtype]));
    end;

    procedure   TModEnvADSR.DoTick; // override;
    begin
      if SignalToLogic( FInputs[ i_trig])
      then FTrigFlag := True;
      FOutputs[ o_sigout] := FInputs[ i_sigin] * FOutputs[ o_out]; // sigout := sigin * out
    end;

    procedure   TModEnvADSR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, LogicToSignal( not ( FState in [ adsrIdle, adsrWait]))], AppLocale));
      aCallback( Self, Format( '%s/%s/trigin/%f', [ aPrefix, Name, FInputs[ i_trig]                                      ], AppLocale));
    end;

    procedure   TModEnvADSR.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime    : TSignal;
      DTime    : TSignal;
      STime    : TSignal;
      RTime    : TSignal;
      TTime    : TSignal;
      TTimeRec : TSignal;
      SLevel   : TSignal;
      i        : Integer;
      p        : Integer;
      R        : TSignal;
    begin
      ATime  := LookupEnvTime( SignalToEnvRange( FInputs[ i_arange]), FInputs[ i_attack ]);
      DTime  := LookupEnvTime( SignalToEnvRange( FInputs[ i_drange]), FInputs[ i_decay  ]);
      SLevel := FInputs[ i_sustain];
      RTime  := LookupEnvTime( SignalToEnvRange( FInputs[ i_rrange]), FInputs[ i_release]);
      STime  := ( ATime + DTime + RTime) * 0.333333;
      TTime  := ATime + DTime + STime + RTime;

      if TTIme = 0
      then Exit;

      TTimeRec := 67 / TTime;
      ATime    := ATime * TTimeRec;
      DTime    := DTime * TTimeRec;
      RTime    := RTime * TTimeRec;
      STime    := STime * TTimeRec;

      i := 0;
      while i < ATime
      do begin
        case Round( FInputs[ i_ashape]) of
          0  : R := LogAttack( i, ATime, 0, 1);
          1  : R := LinAttack( i, ATime, 0, 1);
          else R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] := SignalForEnvOutputType( R, Round( FInputs[ i_outputtype]));
        Inc( i);
      end;

      p := 0;
      while i < ATime + DTime
      do begin
        case SignalToEnvShape( FInputs[ i_dshape]) of
          esLog : R := LogDecay( p, DTime, 1, SLevel);
          esLin : R := LinDecay( p, DTime, 1, SLevel);
          esExp : R := ExpDecay( p, DTime, 1, SLevel);
          else    R := SDecay  ( p, DTime, 1, SLevel);
        end;
        FData[ i] := SignalForEnvOutputType( R, Round( FInputs[ i_outputtype]));
        Inc( i);
        Inc( p);
      end;

      while i < ATime + DTime + STime
      do begin
        FData[ i] := SignalForEnvOutputType( SLevel, Round( FInputs[ i_outputtype]));
        Inc( i);
      end;

      p := 0;
      while i < 67
      do begin
        case SignalToEnvShape( FInputs[ i_rshape]) of
          esLog : R := LogDecay( p, RTime, SLevel, 0);
          esLin : R := LinDecay( p, RTime, SLevel, 0);
          esExp : R := ExpDecay( p, RTime, SLevel, 0);
          else    R := SDecay  ( p, RTime, SLevel, 0);
        end;
        FData[ i] := SignalForEnvOutputType( R, Round( FInputs[ i_outputtype]));
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvADSR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig       : Result := 'trig';
        i_mode       : Result := 'mode';
        i_ashape     : Result := 'ashape';
        i_dshape     : Result := 'dshape';
        i_rshape     : Result := 'rshape';
        i_attack     : Result := 'attack';
        i_decay      : Result := 'decay';
        i_sustain    : Result := 'sustain';
        i_release    : Result := 'release';
        i_gain       : Result := 'gain';
        i_sigin      : Result := 'sigin';
        i_mute       : Result := 'mute';
        i_arange     : Result := 'arange';
        i_drange     : Result := 'drange';
        i_rrange     : Result := 'rrange';
        i_amod       : Result := 'amod';
        i_dmod       : Result := 'dmod';
        i_smod       : Result := 'smod';
        i_rmod       : Result := 'rmod';
        i_amodneg    : Result := 'amodneg';
        i_dmodneg    : Result := 'dmodneg';
        i_smodneg    : Result := 'smodneg';
        i_rmodneg    : Result := 'rmodneg';
        i_amodlevel  : Result := 'amodlevel';
        i_dmodlevel  : Result := 'dmodlevel';
        i_smodlevel  : Result := 'smodlevel';
        i_rmodlevel  : Result := 'rmodlevel';
        i_outputtype : Result := 'outputtype';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModEnvADSR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_sigout : Result := 'sigout';
        else       Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps    : Integer;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
}

    constructor TModSequencer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModSequencer.SetDefaults; // override;
    begin
      FOutValue          := 0;    // Neutral output
      FInputs[ i_dir   ] := 1;    // Make seq run forwards by default.
      FInputs[ i_chain ] := 0;    // Chain input to zero
      FInputs[ i_active] := 1;    // Sequencer active when active input not connected
      FCounter           := 0;
    end;

    procedure   TModSequencer.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 11)
      else inherited;
    end;

    function    TModSequencer.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSequencer.DoTick; // override;
    begin
      if not FOldRes and SignalToLogic( FInputs[ i_reset])
      then begin
        FInState  := seqClkHigh;
        FCounter  := 0;
        FOutValue := 0;
      end;
      FOldRes := SignalToLogic( FInputs[ i_reset]);
    end;

    procedure   TModSequencer.DoSlowTick; // override;
    var
      i : Integer;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FInState  := seqClkHigh;
        FCounter  := 0;
        FOutValue := 0;
      end;

      case FInState of
        seqClkLow :
          begin
            if SignalToLogic( FInputs[ i_trig])
            then begin
              if FSteps > 0
              then begin
                FInState := seqClkHigh;
                if SignalToLogic( FInputs[ i_dir])
                then FCounter := ( FCounter + 1) mod FSteps
                else begin
                  if FCounter <= 0
                  then FCounter := FSteps - 1
                  else Dec( FCounter);
                end;
              end;
            end;
          end;
        seqClkHigh :
          begin
            if not SignalToLogic( FInputs[ i_trig])
            then FInState := seqClkLow;
          end;
      end; // case
      if SignalToLogic( FInputs[ i_active])
      then FOutValue := FInputs[ FCounter + i_step1] - MiddleNote * NOTE_SCALING_REC
      else FOutValue := 0;

      FOutputs[ o_out] := FOutValue + FInputs[ i_chain];

      for i := o_trig1 to o_trig11
      do begin
        if ( i = FCounter + o_trig1) and SignalToLogic( FInputs[ i_active])
        then FOutputs[ i] := LogicToSignal( True )
        else FOutputs[ i] := LogicToSignal( False);
      end;
    end;

    procedure   TModSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, FInputs [ i_trig  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f' , [ aPrefix, Name, FInputs [ i_reset ]], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, FInputs [ i_active]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'   , [ aPrefix, Name, FInputs [ i_dir   ]], AppLocale));
      for i := o_trig1 to o_trig11
      do begin
        if ( i = FCounter + o_trig1) and SignalToLogic( FInputs[ i_active])
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - o_trig1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - o_trig1 + 1], AppLocale));
      end;
    end;

    function    TModSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig   : Result := 'trig';
        i_reset  : Result := 'reset';
        i_dir    : Result := 'dir';
        i_chain  : Result := 'chain';
        i_rnd    : Result := 'rnd';
        i_active : Result := 'active';
        i_step1  : Result := 'step1';
        i_step2  : Result := 'step2';
        i_step3  : Result := 'step3';
        i_step4  : Result := 'step4';
        i_step5  : Result := 'step5';
        i_step6  : Result := 'step6';
        i_step7  : Result := 'step7';
        i_step8  : Result := 'step8';
        i_step9  : Result := 'step9';
        i_step10 : Result := 'step10';
        i_step11 : Result := 'step11';
        else       Result := NO_INPUT;;
      end;
    end;

    function    TModSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_trig1  : Result := 'trig1';
        o_trig2  : Result := 'trig2';
        o_trig3  : Result := 'trig3';
        o_trig4  : Result := 'trig4';
        o_trig5  : Result := 'trig5';
        o_trig6  : Result := 'trig6';
        o_trig7  : Result := 'trig7';
        o_trig8  : Result := 'trig8';
        o_trig9  : Result := 'trig9';
        o_trig10 : Result := 'trig10';
        o_trig11 : Result := 'trig11';
        else       Result := NO_OUTPUT;
      end;
    end;

    function    TModSequencer.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;

{ ========
  TModSeqSeq = class( TMod)
  private
    FState           : Word;
    FSteps           : Integer;
    FSequenceCounter : Integer;
    FPeriodCounter   : Integer;
    FOldRes          : Boolean;
    FOldTrig         : Boolean;
    FClkCnt          : Integer;
  public
}

    constructor TModSeqSeq.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqSeq.SetDefaults; // override;
    begin
      FInputs[ i_dir]  :=  1;        // Go forwards with unconnected dir input
      FInputs[ i_res]  := -1;        // No reset on unconnected input
      FSequenceCounter := 16;
      FPeriodCounter   := 24;
    end;

    procedure   TModSeqSeq.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 256)
      else inherited;
    end;

    function    TModSeqSeq.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSeqSeq.DoSlowTick; // override;
    const
      MaxCount = 16;
    var
      Direction : Boolean;
      Trig      : Boolean;
      Res       : Boolean;
      i         : Integer;
      DoChange  : Boolean;
      EndOfSeq  : Boolean;
      MustReset : Boolean;
      ResMod    : Integer;
    begin
      DoChange  := False;
      MustReset := False;

      ResMod    := Round        ( FInputs[ i_resmod]);
      Direction := SignalToLogic( FInputs[ i_dir   ]);
      Trig      := SignalToLogic( FInputs[ i_trig  ]);
      Res       := SignalToLogic( FInputs[ i_res   ]);
      EndOfSeq  := FPeriodCounter <= 0;

      if (( not FOldRes) and Res) or ResetFlag or EndOfSeq
      then begin
        if Direction
        then begin                                             // Forwards , reset to first state
          FState           :=  0;
          FSequenceCounter := 15;
        end
        else begin                                             // Backwards , reset to last state
          FState           :=   FSteps shr 4;
          FSequenceCounter := ( FSteps and 15) - 1;
        end;
        FClkCnt        := c_clkdelay;
        FPeriodCounter := FSteps;
        DoChange       := True;
        MustReset      := MustReset or ( ResMod in [ m_res1, m_res3]);   // Propagate reset on received reset or end of sequence in mode 1 and 3
        ResetFlag      := False;
        FOldTrig       := True;
      end
      else begin
        if Trig and not FOldTrig
        then begin
          if not EndOfSeq
          then Dec( FPeriodCounter);
          if FSequenceCounter > 0
          then Dec( FSequenceCounter)
          else begin
            FSequenceCounter := 15;
            if Direction
            then begin                                         // Forwards, go to nexxt state
              FState := FState + 1;
              if FState > FSteps shr 4
              then FState := 0;
            end
            else begin                                         // Backwards, go to previous state
              if FState > 0
              then Dec( FState)
              else FState := Fsteps shr 4;
            end;
            MustReset := MustReset or ( ResMod in [ m_res2, m_res3]);    // Issue reset after state change in mode 2 and 3
            DoChange  := True;
          end;
          FClkCnt := c_clkdelay;
        end
        else if FOldTrig and not Trig
        then begin
          DoChange := True;
          FClkCnt  := c_clkdelay;
        end;
        FOldTrig := Trig;
      end;
      FOldRes := Res;

      if FClkCnt > 0
      then begin
        Dec( FClkCnt);
        if FClkCnt = 0
        then FOutputs[ o_outtrig] := LogicToSignal( Trig);
      end;

      if DoChange
      then begin
        for i := 0 to MaxCount - 1
        do begin
          if i = FState
          then FOutputs[ o_out1 + i] := LogicToSignal( True )
          else FOutputs[ o_out1 + i] := LogicToSignal( False);
        end;
        FOutputs[ o_outres] := FInputs[ i_res];
        FOutputs[ o_outdir] := FInputs[ i_dir];
      end;

      if MustReset
      then begin
        FOutputs[ o_outres ] := LogicToSignal( True);
        FOutputs[ o_outtrig] := LogicToSignal( True);
      end;
    end;

    procedure   TModSeqSeq.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := o_out1 to o_out16
      do aCallback( Self, Format( '%s/%s/step%d/%f', [ aPrefix, Name, 1 + i - o_out1, FOutputs[ i]], AppLocale));

      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
      aCallback( Self, Format( '%s/%s/res/%f' , [ aPrefix, Name, FInputs[ i_res ]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f' , [ aPrefix, Name, FInputs[ i_dir ]], AppLocale));
    end;

    function    TModSeqSeq.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_res    : Result := 'res';
        i_trig   : Result := 'trig';
        i_dir    : Result := 'dir';
        i_resmod : Result := 'resmod';
        else       Result := NO_INPUT;;
      end;
    end;

    function    TModSeqSeq.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outres  : Result := 'outres';
        o_outtrig : Result := 'outtrig';
        o_outdir  : Result := 'outdir';
        o_out1    : Result := 'out1';
        o_out2    : Result := 'out2';
        o_out3    : Result := 'out3';
        o_out4    : Result := 'out4';
        o_out5    : Result := 'out5';
        o_out6    : Result := 'out6';
        o_out7    : Result := 'out7';
        o_out8    : Result := 'out8';
        o_out9    : Result := 'out9';
        o_out10   : Result := 'out10';
        o_out11   : Result := 'out11';
        o_out12   : Result := 'out12';
        o_out13   : Result := 'out13';
        o_out14   : Result := 'out14';
        o_out15   : Result := 'out15';
        o_out16   : Result := 'out16';
        else        Result := NO_OUTPUT;
      end;
    end;

    function    TModSeqSeq.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModSeq16 = class( TMod)
  private
    FOutValue   : TSignal;
    FReset      : Boolean;
    FNextValue  : TSignal;
    FForwards   : Boolean;
    FActive     : Boolean;
    FChainValue : TSignal;
    FSteps      : Integer;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FOldRes     : Boolean;
    FOldActive  : Boolean;
    FTicks      : Integer;
    FPeriodTime : Integer;
  private
}

    procedure   TModSeq16.SampleInputs;
    begin
      FInState    := seqClkHigh;
      FActive     := SignalToLogic( FInputs[ i_active]) and ( FSteps > 0);
      FChainValue := FInputs[ i_chain];
      FForwards   := SignalToLogic( FInputs[ i_dir]);
      FReset      := SignalToLogic( FInputs[ i_res]);
    end;

    procedure   TModSeq16.MakeNextValue;
    begin
      FOutValue := FNextValue;
      if FActive
      then begin
        if Round( FInputs[ i_mode]) = m_note
        then FNextValue := FInputs[ FCounter + i_step1] - MiddleNote * NOTE_SCALING_REC
        else FNextValue := FInputs[ FCounter + i_step1];
      end
      else FNextValue := FChainValue;
    end;

//  public

    constructor TModSeq16.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeq16.SetDefaults; // override;
    begin
      inherited;
      FInputs[ i_dir   ] := 1;         // Make seq run forwards by default.
      FInputs[ i_chain ] := 0;         // Chain input to zero wwhen not connected
      FInputs[ i_active] := 1;         // Sequencer active when active input not connected
      FInputs[ i_mode  ] := m_note;    // Default startup mode is notes
      FNextValue         := 0;         // Neutral output
      FPeriodTime        := 2;
      FTicks             := 0;
      SampleInputs;
      if FForwards
      then FCounter := 0
      else FCounter := FSteps - 1;
    end;

    procedure   TModSeq16.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 16)
      else inherited;
    end;

    function    TModSeq16.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSeq16.DoSlowTick; // override;
    var
      XFade : TSignal;
    begin
      if ResetFlag
      then begin
        SampleInputs;
        FPeriodTime := Max( 2, FTicks);
        FTicks      := 0;
        FNextValue  := 0;
        if FForwards
        then FCounter := 0
        else FCounter := FSteps - 1;
        MakeNextValue;
      end
      else begin
        case FInState of
          seqClkLow :
            begin
              if SignalToLogic( FInputs[ i_trig])
              then begin
                SampleInputs;
                if ResetFlag or ( not FOldRes and FReset)
                then begin
                  FPeriodTime := Max( 2, FTicks);
                  FTicks      := 0;
                  if FForwards
                  then FCounter := 0
                  else FCounter := FSteps - 1;
                end
                else begin
                  if FSteps > 0
                  then begin
                    SampleInputs;
                    FPeriodTime := Max( 2, FTicks);
                    FTicks      := 0;
                    if FForwards
                    then FCounter := ( FCounter + 1) mod FSteps
                    else begin
                      if FCounter <= 0
                      then FCounter := FSteps - 1
                      else Dec( FCounter);
                    end;
                  end;
                end;
                MakeNextValue;
              end;
            end;
          seqClkHigh :
            begin
              if not SignalToLogic( FInputs[ i_trig])
              then begin
                FInState    := seqClkLow;
                FPeriodTime := FTicks * 2;
              end;
            end;
        end; // case
      end;

      XFade := Clip((( FPeriodTime - FTicks) / FPeriodTime) * FInputs[ i_xfade], -1, 1);

      if FOutValue < FNextValue      // going up, note that as final and current value are reversed this must call decay ...
      then begin
        case Round( FInputs[ i_curveup]) of
          0  : FOutputs[ o_out] := LogDecay( XFade, 1, FNextValue, FOutValue);
          1  : FOutputs[ o_out] := LinDecay( XFade, 1, FNextValue, FOutValue);
          else FOutputs[ o_out] := ExpDecay( XFade, 1, FNextValue, FOutValue);
        end;
      end
      else begin                     // going down
        case Round( FInputs[ i_curvedown]) of
          0  : FOutputs[ o_out] := LogAttack( XFade, 1, FNextValue, FOutValue);
          1  : FOutputs[ o_out] := LinAttack( XFade, 1, FNextValue, FOutValue);
          else FOutputs[ o_out] := ExpAttack( XFade, 1, FNextValue, FOutValue);
        end;
      end;

      FOutputs[ o_outres ] := FInputs[ i_res ];
      FOutputs[ o_outtrig] := FInputs[ i_trig];
      FOutputs[ o_outdir ] := FInputs[ i_dir ];

      if FTicks < MAXINT shr 2    // Cap FTicks value in case the module gets not clocked prolly MAXINT shr 1 would be OK too
      then Inc( FTicks);

      FOldRes    := FReset;
      ResetFlag  := False;
      FOldActive := FActive;
    end;

    procedure   TModSeq16.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, FInputs [ i_trig  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f' , [ aPrefix, Name, FInputs [ i_res   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, FInputs [ i_active]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'   , [ aPrefix, Name, FInputs [ i_dir   ]], AppLocale));
      for i := i_step1 to i_step16
      do begin
        if ( i = FCounter + i_step1) and SignalToLogic( FInputs[ i_active])
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - i_step1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - i_step1 + 1], AppLocale));
      end;
    end;

    function    TModSeq16.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig      : Result := 'trig';
        i_res       : Result := 'res';
        i_dir       : Result := 'dir';
        i_chain     : Result := 'chain';
        i_rnd       : Result := 'rnd';
        i_active    : Result := 'active';
        i_mode      : Result := 'mode';
        i_xfade     : Result := 'xfade';
        i_curveup   : Result := 'curveup';
        i_curvedown : Result := 'curvedown';
        i_step1     : Result := 'step1';
        i_step2     : Result := 'step2';
        i_step3     : Result := 'step3';
        i_step4     : Result := 'step4';
        i_step5     : Result := 'step5';
        i_step6     : Result := 'step6';
        i_step7     : Result := 'step7';
        i_step8     : Result := 'step8';
        i_step9     : Result := 'step9';
        i_step10    : Result := 'step10';
        i_step11    : Result := 'step11';
        i_step12    : Result := 'step12';
        i_step13    : Result := 'step13';
        i_step14    : Result := 'step14';
        i_step15    : Result := 'step15';
        i_step16    : Result := 'step16';
        else          Result := NO_INPUT;;
      end;
    end;

    function    TModSeq16.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out     : Result := 'out';
        o_outres  : Result := 'outres';
        o_outtrig : Result := 'outtrig';
        o_outdir  : Result := 'outdir';
        else        Result := NO_OUTPUT;
      end;
    end;

    function    TModSeq16.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModTextSequencer = class( TMod)
  private
    FOutValue   : TSignal;
    FStepCount  : Integer;
    FStepValues : TSignalArray;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FOldRes     : Boolean;
  public
}

    constructor TModTextSequencer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModTextSequencer.SetDefaults; // override;
    begin
      FOutValue            := 0;
      FInputs [ i_dir    ] := 1; // Make seq run forward by default.
      FInputs [ i_chain  ] := 0; // Clear chain input
      FOutputs[ o_gateout] := LogicToSignal( False);
    end;

    procedure   TModTextSequencer.DoSlowTick; // override;
    var
      CurrentValue : TSIgnal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FInState  := seqClkLow;
        FCounter  := 0;
        FOutValue := 0;
      end;

      if not FOldRes and SignalToLogic( FInputs[ i_reset])
      then begin
        FInState := seqClkLow;
        FCounter := 0;
      end;

      case FInState of

        seqClkLow :
          begin
            if SignalToLogic( FInputs[ i_trig])
            then begin
              FInState := seqClkHigh;
              if FStepCount > 0
              then begin
                if FInputs[ i_mode] = 0
                then begin
                  if SignalToLogic( FInputs[ i_dir])
                  then FCounter := ( FCounter + 1) mod FStepCount
                  else begin
                    if FCounter <= 0
                    then FCounter := FStepCount - 1
                    else Dec( FCounter);
                  end;
                end
                else FCounter := Random( FStepCount);
              end;

              CurrentValue := FStepValues[ FCounter];
              if not AlmostEqual( CurrentValue, NO_NOTE, DELTA)
              then begin
                FOutValue := CurrentValue - MiddleNote * NOTE_SCALING_REC;
                FOutputs[ o_gateout] := LogicToSignal( True);
              end;
            end;
          end;

        seqClkHigh :
          begin
            if FInputs[ i_trig] <= 0
            then begin
              FInState := seqClkLow;
              FOutputs[ o_gateout] := LogicToSignal( False);
            end;
          end;

      end;
      FOutputs[ o_out] := FOutValue + FInputs[ i_chain];
      FOldRes := SignalToLogic( FInputs[ i_reset]);
    end;

    procedure   TModTextSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs[ i_trig    ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInputs[ i_reset   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInputs[ i_dir     ]], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%d' , [ aPrefix, Name, FCounter + 1        ], AppLocale));
      aCallback( Self, Format( '%s/%s/gate/%f' , [ aPrefix, Name, FOutputs[ o_gateout]], AppLocale));
      aCallback( Self, Format( '%s/%s/steps/%d', [ aPrefix, Name, FStepCount          ], AppLocale));
    end;

    function    TModTextSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig  : Result := 'trig';
        i_reset : Result := 'reset';
        i_dir   : Result := 'dir';
        i_mode  : Result := 'mode';
        i_chain : Result := 'chain';
      end;
    end;

    function    TModTextSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out     : Result := 'out';
        o_gateout : Result := 'gateout';
        else        Result := NO_OUTPUT;
      end;
    end;

    procedure   TModTextSequencer.SetStepValues( const aValues: TSignalArray);
    var
      i       : Integer;
      aLength : Integer;
    begin
      FStepCount := 0;
      aLength := Length( aValues);
      SetLength( FStepValues, aLength);
      for i := 0 to aLength - 1
      do FStepValues[ i] := aValues[ i];
      FStepCount := aLength;
    end;


{ ========
  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
}

    constructor TModSeqStep.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqStep.SetDefaults; // override;
    begin
      FInputs[ i_dir] := 1; // Make seq run forward by default.
      FDezipperMap := [
        i_time
      ];
    end;

    procedure   TModSeqStep.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState := sssIdle;
        FOutputs[ o_trigoutfwd] := LogicToSignal( False);                        // Do not trigger next in forward  direction
        FOutputs[ o_trigoutbwd] := LogicToSignal( False);                        // Do not trigger next in backward direction
      end;

      if not FOldRes and SignalToLogic( FInputs[ i_reset])
      then begin                                                                 // Reset from inactive to active
        FState := sssIdle;
        FOutputs[ o_trigoutfwd] := LogicToSignal( False);                        // Do not trigger next in forward  direction
        FOutputs[ o_trigoutbwd] := LogicToSignal( False);                        // Do not trigger next in backward direction
      end;

      case FState of

        sssIdle :
          begin
            if SignalToLogic( FInputs[ i_dir])
            then begin                                                           // running Forwards
              if SignalToLogic( FInputs[ i_triginfwd])                           // Forward trig in
              then begin
                FState := sssActive;
                FTime  := SlowTimeToSampleCount( FInputs[ i_time]);              // Duration
                FOutputs[ o_trigoutfwd] := LogicToSignal( False);                // Do not trigger next in forward  direction
                FOutputs[ o_trigoutbwd] := LogicToSignal( False);                // Do not trigger next in backward direction
              end;
            end
            else begin                                                           // running Backwards
              if SignalToLogic( FInputs[ i_triginbwd])                           // Backward trig in
              then begin
                FState := sssActive;
                FTime  := SlowTimeToSampleCount( FInputs[ i_time]);              // Duration
                FOutputs[ o_trigoutfwd] := LogicToSignal( False);                // Do not trigger next in forward  direction
                FOutputs[ o_trigoutbwd] := LogicToSignal( False);                // Do not trigger next in backward direction
              end;
            end;
          end;

        sssActive :
          begin
            FTime := FTime - 1;
            if FTime <= 0
            then begin
              FState := sssWaiting;
              if SignalToLogic( FInputs[ i_dir])                                 // running Forwards
              then FOutputs[ o_trigoutfwd] := LogicToSignal( True)               // Trigger next in forward  direction
              else FOutputs[ o_trigoutbwd] := LogicToSignal( True);              // Trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if SignalToLogic( FInputs[ i_dir])
            then begin                                                           // running Forwards
              if not SignalToLogic( FInputs[ i_triginfwd])                       // Forward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end
            else begin                                                           // running Backwards
              if not SignalToLogic( FInputs[ i_triginbwd])                       // Backward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end;
          end;

      end;

      if FState = sssActive
      then FOutputs[ o_out] := FInputs[ i_note] - MiddleNote * NOTE_SCALING_REC  // Own value to output
      else FOutputs[ o_out] := FInputs[ i_chain];                                // Chained value to output (was compensated there already)

      FOldRes := SignalToLogic( FInputs[ i_reset]);
    end;

    procedure   TModSeqStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInputs[ i_triginfwd]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInputs[ i_triginbwd]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInputs[ i_reset    ]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInputs[ i_dir      ]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_triginfwd : Result := 'triginfwd';
        i_triginbwd : Result := 'triginbwd';
        i_chain     : Result := 'chain';
        i_reset     : Result := 'reset';
        i_dir       : Result := 'dir';
        i_note      : Result := 'note';
        i_time      : Result := 'time';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModSeqStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_trigoutfwd : Result := 'trigoutfwd';
        o_trigoutbwd : Result := 'trigoutbwd';
        o_out        : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
}

    constructor TModSeqClockStep.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqClockStep.SetDefaults; // override;
    begin
      FInputs[ i_dir] := 1; // Make seq run forward by default.
    end;

    procedure   TModSeqClockStep.DoSlowTick; // override;
    begin

      if ResetFlag
      then begin
        ResetFlag   := False;
        FState      := sssIdle;
        FTrigOutFwd := False;
        FTrigOutBwd := False;
        FOutputs[ o_trigoutfwd] := LogicToSignal( False);                       // Do not trigger next in forward  direction
        FOutputs[ o_trigoutbwd] := LogicToSignal( False);                       // Do not trigger next in backward direction
        FOutputs[ o_out] := FInputs[ i_chain];                                  // Chained value to output
      end;

      if not FOldRes and SignalToLogic( FInputs[ i_reset])
      then begin                                                                // Reset from inactive to active - reset is asynchroneous
        FState      := sssIdle;
        FTrigOutFwd := False;
        FTrigOutBwd := False;
        FOutputs[ o_trigoutfwd] := LogicToSignal( False);                       // Do not trigger next in forward  direction
        FOutputs[ o_trigoutbwd] := LogicToSignal( False);                       // Do not trigger next in backward direction
        FOutputs[ o_out] := FInputs[ i_chain];                                  // Chained value to output
      end;

      FTrigInFwd := SignalToLogic( FInputs[ i_triginfwd]);
      FTrigInBwd := SignalToLogic( FInputs[ i_triginbwd]);
      FDir       := SignalToLogic( FInputs[ i_dir]);

      case FState of

        sssIdle :
          begin
            if ( FDir and FTrigInFwd) or (( not FDir) and FTrigInBwd)
            then begin                                                          // Trigger seen
              FState      := sssActive;
              FCount      := Ceil( FInputs[ i_counts]);                         // Duration
              FTrigOutFwd := False;                                             // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                             // Do not trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if ( FDir and ( not FTrigInFwd)) or (( not FDir) and ( not FTrigInBwd))
            then begin
              FState      := sssIdle;                                           // Trigger fell off - go idle
              FTrigOutFwd := False;                                             // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                             // Do not trigger next in backward direction
            end;
          end;
      end;

      if not FOldClock and SignalToLogic( FInputs[ i_clock])
      then begin                                                                // Clock from inactive to active - state changes should only occur here
        case FState of
          sssActive :
            begin
              FCount := FCount - 1;
              if FCount <= 1
              then begin
                FCount := 0;
                FState := sssWaiting;
                if FDir                                                         // running Forwards
                then begin
                  FTrigOutFwd := True;                                          // Trigger next in forward  direction
                  FTrigOutBwd := False;                                         // No Trigger next in backward  direction
                end
                else begin
                  FTrigOutFwd := False;                                         // No Trigger next in forward direction
                  FTrigOutBwd := True;                                          // Trigger next in backward direction
                end;
              end;
            end;
        end;
      end;

      FOutputs[ o_trigoutfwd] := LogicToSignal( FTrigOutFwd);                   // Clock out internal state to ouptputs
      FOutputs[ o_trigoutbwd] := LogicToSignal( FTrigOutBwd);

      if FState = sssActive
      then FOutputs[ o_out] := FInputs[ i_note] - MiddleNote * NOTE_SCALING_REC // Own     value to output
      else FOutputs[ o_out] := FInputs[ i_chain];                               // Chained value to output (was compensated there already)

      FOldRes   := SignalToLogic( FInputs[ i_triginbwd]);
      FOldClock := SignalToLogic( FInputs[ i_clock    ]);
    end;


    procedure   TModSeqClockStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInputs[ i_triginfwd]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInputs[ i_triginbwd]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInputs[ i_reset    ]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInputs[ i_dir      ]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/clock/%f'  , [ aPrefix, Name, FInputs[ i_clock    ]                ], AppLocale));
      aCallback( Self, Format( '%s/%s/count/%d'  , [ aPrefix, Name, FCount                               ], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqClockStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_triginfwd : Result := 'triginfwd';
        i_triginbwd : Result := 'triginbwd';
        i_chain     : Result := 'chain';
        i_reset     : Result := 'reset';
        i_dir       : Result := 'dir';
        i_clock     : Result := 'clock';
        i_note      : Result := 'note';
        i_counts    : Result := 'counts';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModSeqClockStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_trigoutfwd : Result := 'trigoutfwd';
        o_trigoutbwd : Result := 'trigoutbwd';
        o_out        : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPattern = class( TMod)
  private
    FPosition : Integer;
    FCounter  : Integer;
    FOldTrig  : Boolean;
    FOldRes   : Boolean;
    FOutVal   : Boolean;
  public
}

    constructor TModPattern.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPattern.DoSlowTick; // override;
    var
      Trig   : Boolean;
      Res    : Boolean;
      Steps  : Integer;
      i      : Integer;
      p      : Integer;
    begin
      Res   := SignalToLogic( FInputs[ i_res ]);
      Trig  := SignalToLogic( FInputs[ i_trig]);
      Steps := Round( FInputs[ i_step1 + FPosition]);

      if ( not FOldRes and Res) or ResetFlag
      then FPosition := 0;

      if ( not FOldTrig and Trig)
      then begin
        Dec( FCounter);
        FOutVal := ( FCounter = 0) and ( Steps > 0);
        if FCounter <= 0
        then begin
          p       := FPosition;

          for i := i_step1 to i_last
          do begin
            p := ( p + 1) mod ( i_last - i_step1 + 1);
            if Round( FInputs[ i_step1 + p]) > 0
            then begin
              FPosition := p;
              Break;
            end;
          end;

          FCounter := Round( FInputs[ i_step1 + FPosition]);
        end;
      end;

      FOutputs[ o_out] := LogicToSignal( FOutVal and Trig);

      ResetFlag := False;
      FOldRes   := Res;
      FOldTrig  := Trig;
    end;

    procedure   TModPattern.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs [ i_trig]], AppLocale));
      aCallback( Self, Format( '%s/%s/res/%f' , [ aPrefix, Name, FInputs [ i_res ]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutputs[ o_out ]], AppLocale));
    end;

    function    TModPattern.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_res   : Result := 'res';
        i_trig  : Result := 'trig';
        i_step1 : Result := 'step1';
        i_step2 : Result := 'step2';
        i_step3 : Result := 'step3';
        i_step4 : Result := 'step4';
        i_step5 : Result := 'step5';
        i_step6 : Result := 'step6';
        i_step7 : Result := 'step7';
        i_step8 : Result := 'step8';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModPattern.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFriemels = class( TMod)
  strict private
  const
    i_res   = 0;
  const
    i_last  = 9;
  const
    o_out   = 0;
  public
}

    constructor TModFriemels.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    procedure   TModFriemels.DoSlowTick; // override;
    begin
    end;


{ ========
  TModCsvData = class( TMod)
  private
    FMemory   : Array[ 0 .. 15] of TSignalArray;
    FComments : TStringList;
    FFileName : string;
    FChannels : Integer;
    FCounter  : Integer;
    FSteps    : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
    FMode     : Integer;
    FStopped  : Boolean;
  public
    property    FileName: string read FFileName write SetFileName;
  private
 }

    procedure   TModCsvData.LoadFile;

      procedure AddSignal( aMemoryIndex: Integer; aValue: TSignal);
      var
        aLength : Integer;
      begin
        aLength := Length( FMemory[ aMemoryIndex]);
        SetLength( FMemory[ aMemoryIndex], aLength + 1);
        FMemory[ aMemoryIndex][ aLength] := aValue;
      end;

    var
      aData  : TStringList;
      aparts : TStringList;
      aLine  : string;
      i      : Integer;
      j      : Integer;
      p      : Integer;
      S      : string;
      T      : string;
      Cmt    : string;
      First  : Boolean;
    begin
      aData := TStringList.Create;
      try
        if FileExists( FFileName)
        then begin
          aData.LoadFromFile( FFileName);
          First := True;

          for i := Low( FMemory) to High( FMemory)
          do begin
            for j := Low( FMemory[ i]) to High( FMemory[ i])
            do FMemory[ i][ j] := 0.0;
          end;
          FComments.Clear;

          for aLine in aData
          do begin
            // Strip comment, but keep end of line comment for comment label in module
            p := Pos( '#', aLine);
            if p > 0
            then begin
              Cmt := Trim( Copy( aLine, p + 1, Length( aLine)));
              S   := Trim( Copy( aLine, 1, p - 1))
            end
            else begin
              Cmt := '';
              S   := Trim( aLine);
            end;

            // if line not empty after that handle it
            if Length( S) > 0
            then begin
              aParts := Explode( S, ',');
              try
                if First                           // First line determines column count
                then begin
                  FChannels := aParts.Count;
                  First     := False;
                end;
                for i := 0 to FChannels - 1
                do begin
                  if i > aParts.Count - 1
                  then AddSignal( i, 0)            // When not enough data on this line set remaining columns to zero
                  else begin
                    T := Trim( aParts[ i]);
                    AddSignal( i, StrToFloatDef( T, 0));
                  end;
                end;
              finally
                aParts.DisposeOf;
              end;
              FComments.Add( Cmt);
            end;
          end;
          FSteps   := Length( FMemory[ 0]);
          FCounter := Min( FCounter, FSteps - 1);  // Do not address outside data area
        end;
      finally
        aData.DisposeOf;
      end;
    end;

    procedure   TModCsvData.SetFileName( const aValue: string);
    begin
      if  aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

//  public

    constructor TModCsvData.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      i : Integer;
    begin
      inherited;
      for i := Low( FMemory) to High( FMemory)
      do FMemory[ i] := nil;
      FComments := TStringList.Create;
      FIsFast   := True;
      FIsSlow   := True;
    end;

    destructor  TModCsvData.Destroy; // override;
    var
      i : Integer;
    begin
      FreeAndNil( FComments);
      for i := Low( FMemory) to High( FMemory)
      do FreeAndNil( FMemory[ i]);
      inherited;
    end;

    procedure   TModCsvData.SetDefaults; // override;
    begin
      FInputs[ i_dir] := 1; // Make seq run forward by default.
    end;

    procedure   TModCsvData.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModCsvData.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModCsvData.DoTick; // override;
    // Follow inputs at audio rate
    var
      OldCounter : Integer;
      Forwards   : Boolean;
    begin
      Forwards := SignalToLogic( FInputs[ i_dir]);
      if ResetFlag or ( not FOldRes and SignalToLogic( FInputs[ i_reset]))
      then begin
        ResetFlag := False;
        FInState  := seqClkLow;
        if Forwards
        then FCounter := 0
        else FCounter := FSteps - 1;
        FStopped   := False;
      end;
      FOldRes := SignalToLogic( FInputs[ i_reset]);

      if ( FSteps > 0) and not FStopped
      then begin
        FMode      := Round( FInputs[ i_mode]);
        OldCounter := FCounter;

        case FInState of
          seqClkLow :
            begin
              if SignalToLogic( FInputs[ i_trig])
              then begin
                FInState := seqClkHigh;
                if Forwards
                then begin
                  FCounter := ( FCounter + 1) mod FSteps;
                  if ( FCounter < OldCounter) and ( FMode = 1)
                  then FStopped := True;
                end
                else begin
                  if FCounter <= 0
                  then begin
                    FCounter := FSteps - 1;
                    if FMode = 1
                    then FStopped := True;
                  end
                  else Dec( FCounter);
                end;
              end;
            end;
          seqClkHigh :
            begin
              if FInputs[ i_trig] <= 0
              then FInState := seqClkLow;
            end;
        end;
      end;
    end;

    procedure   TModCsvData.DoSlowTick; // override;
    // generate output at control rate
    var
      i : Integer;
    begin
      if ( FSteps > 0) and ( FChannels > 0)
      then begin
        for i := o_out1 to Min( FChannels - 1, o_out16)
        do FOutputs[ i] := FMemory[ i - o_out1][ FCounter];
      end;
    end;

    procedure   TModCsvData.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'    , [ aPrefix, Name, FInputs[ i_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'   , [ aPrefix, Name, FInputs[ i_reset]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'     , [ aPrefix, Name, FInputs[ i_dir  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%d'    , [ aPrefix, Name, FCounter + 1     ], AppLocale));
      aCallback( Self, Format( '%s/%s/steps/%d'   , [ aPrefix, Name, FSteps           ], AppLocale));
      aCallback( Self, Format( '%s/%s/channels/%d', [ aPrefix, Name, FChannels        ], AppLocale));
    end;

    procedure   TModCsvData.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    var
      Cmt : string;
    begin
      if FCounter < FComments.Count
      then Cmt := FComments[ FCounter]
      else Cmt := '';
      aCallback( Self, Format( '%s/%s/comment/%s', [ aPrefix, Name, Cmt], AppLocale));
    end;

    function    TModCsvData.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_reset : Result := 'reset';
        i_trig  : Result := 'trig';
        i_dir   : Result := 'dir';
        i_mode  : Result := 'mode';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModCsvData.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1  : Result := 'out1';
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        o_out12 : Result := 'out12';
        o_out13 : Result := 'out13';
        o_out14 : Result := 'out14';
        o_out15 : Result := 'out15';
        o_out16 : Result := 'out16';
        else      Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModKleeBit = class( TMod)
  private
    FPattern    : Byte;
    FOldStep    : Boolean;
    FOldLoad    : Boolean;
    FBus1       : Boolean;
    FBus2       : Boolean;
    FBus3       : Boolean;
    FAnalog     : TSignal;
    FLoopMode   : Boolean;
    FInvertMode : Boolean;
  private
}

    procedure   TModKleeBit.LoadPreset;
    var
      i : Integer;
    begin
      FPattern := 0;
      for i := i_pre1 to i_pre8
      do begin
        FPattern := FPattern shr 1;
        if SignalToLogic( FInputs[ i])
        then FPattern := FPattern or 128;
      end;
    end;

    procedure   TModKleeBit.Step;
    var
      InBit   : Boolean;
      Carry   : Boolean;
      i       : Integer;
      Bit     : Boolean;
      MainLev : TSignal;
    begin
      FInvertMode := SignalToLogic( FInputs[ i_invertmode]) xor SignalToLogic( FInputs[ i_invertinvert]);
      FLoopMode   := SignalToLogic( FInputs[ i_loopmode  ]) xor SignalToLogic( FInputs[ i_loopinvert  ]);

      Carry := ( FPattern and 128 <> 0) xor FInvertMode;
      if FLoopMode
      then InBit := Carry
      else InBit := SignalToLogic( FInputs[ i_inbit]);

      FPattern := FPattern shl 1;
      if Inbit
      then FPattern := FPattern or 1;

      FBus1   := SignalToLogic( FInputs[ i_inbus1   ]);
      FBus2   := SignalToLogic( FInputs[ i_inbus2   ]);
      FBus3   := SignalToLogic( FInputs[ i_inbus3   ]);
      FAnalog :=                FInputs[ i_inanalog ];
      MainLev :=                FInputs[ i_mainlevel] * FInputs[ i_mainlevelmod];

      for i := 0 to 7
      do begin
        Bit := FPattern and ( 1 shl i) <> 0;
        if Bit
        then FAnalog := FAnalog + FInputs[ i_level1 + i] * MainLev;

        case Round( FInputs[ i_bussel1 + i]) of
          0 : FBus1 := FBus1 or Bit;
          1 : FBus2 := FBus2 or Bit;
          2 : FBus3 := FBus3 or Bit;
        end;
      end;

      FOutputs[ o_outbit   ] := LogicToSignal( Carry);
      FOutputs[ o_outanalog] := FAnalog;
      FOutputs[ o_outbus1  ] := LogicToSignal( FBus1      );
      FOutputs[ o_outbus2  ] := LogicToSignal( FBus2      );
      FOutputs[ o_outbus3  ] := LogicToSignal( FBus3      );
      FOutputs[ o_outloop  ] := LogicToSignal( FLoopMode  );
      FOutputs[ o_outinvert] := LogicToSignal( FInvertMode);
    end;

//  public

    constructor TModKleeBit.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModKleeBit.SetDefaults; // override;
    begin
      FInputs[ i_inanalog    ] := 0.0;
      FInputs[ i_inbus1      ] := 0.0;
      FInputs[ i_inbus2      ] := 0.0;
      FInputs[ i_inbus3      ] := 0.0;
      FInputs[ i_instep      ] := 0.0;
      FInputs[ i_inload      ] := 0.0;
      FInputs[ i_loopinvert  ] := 0.0;
      FInputs[ i_invertinvert] := 0.0;
      FInputs[ i_mainlevelmod] := 1.0;
      LoadPreset;
    end;

    procedure   TModKleeBit.DoSlowTick; // override;
    var
      aValue : Boolean;
    begin
      if ResetFlag
      then SetDefaults;
      ResetFlag := False;

      aValue := SignalToLogic( FInputs[ i_inload]);
      if not FOldLoad and aValue
      then LoadPreset;
      FOldLoad := aValue;

      aValue := SignalToLogic( FInputs[ i_instep]);
      if not FOldStep and aValue
      then Step;
      FOldStep := aValue;
    end;

    procedure   TModKleeBit.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/step1/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   1 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step2/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   2 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step3/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   4 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step4/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   8 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step5/%f' , [ aPrefix, Name, LogicToSignal( FPattern and  16 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step6/%f' , [ aPrefix, Name, LogicToSignal( FPattern and  32 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step7/%f' , [ aPrefix, Name, LogicToSignal( FPattern and  64 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step8/%f' , [ aPrefix, Name, LogicToSignal( FPattern and 128 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/loop/%f'  , [ aPrefix, Name, LogicToSignal( FLoopMode            )], AppLocale));
      aCallback( Self, Format( '%s/%s/invert/%f', [ aPrefix, Name, LogicToSignal( FInvertMode          )], AppLocale));
      aCallback( Self, Format( '%s/%s/bus1/%f'  , [ aPrefix, Name, FOutputs[ o_outbus1]                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/bus2/%f'  , [ aPrefix, Name, FOutputs[ o_outbus2]                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/bus3/%f'  , [ aPrefix, Name, FOutputs[ o_outbus3]                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/outbit/%f', [ aPrefix, Name, FOutputs[ o_outbit ]                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%f'  , [ aPrefix, Name, FInputs [ i_instep ]                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/load/%f'  , [ aPrefix, Name, FInputs [ i_inload ]                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/inbit/%f' , [ aPrefix, Name, FInputs [ i_inbit  ]                 ], AppLocale));
    end;

    function    TModKleeBit.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_instep       : Result := 'instep';
        i_inload       : Result := 'inload';
        i_loopmode     : Result := 'loopmode';
        i_inanalog     : Result := 'inanalog';
        i_inbus1       : Result := 'inbus1';
        i_inbus2       : Result := 'inbus2';
        i_inbus3       : Result := 'inbus3';
        i_pre1         : Result := 'pre1';
        i_pre2         : Result := 'pre2';
        i_pre3         : Result := 'pre3';
        i_pre4         : Result := 'pre4';
        i_pre5         : Result := 'pre5';
        i_pre6         : Result := 'pre6';
        i_pre7         : Result := 'pre7';
        i_pre8         : Result := 'pre8';
        i_level1       : Result := 'level1';
        i_level2       : Result := 'level2';
        i_level3       : Result := 'level3';
        i_level4       : Result := 'level4';
        i_level5       : Result := 'level5';
        i_level6       : Result := 'level6';
        i_level7       : Result := 'level7';
        i_level8       : Result := 'level8';
        i_bussel1      : Result := 'bussel1';
        i_bussel2      : Result := 'bussel2';
        i_bussel3      : Result := 'bussel3';
        i_bussel4      : Result := 'bussel4';
        i_bussel5      : Result := 'bussel5';
        i_bussel6      : Result := 'bussel6';
        i_bussel7      : Result := 'bussel7';
        i_bussel8      : Result := 'bussel8';
        i_inbit        : Result := 'inbit';
        i_invertmode   : Result := 'invertmode';
        i_mainlevel    : Result := 'mainlevel';
        i_loopinvert   : Result := 'loopinvert';
        i_invertinvert : Result := 'invertinvert';
        i_mainlevelmod : Result := 'mainlevelmod';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModKleeBit.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outanalog : Result := 'outanalog';
        o_outbus1   : Result := 'outbus1';
        o_outbus2   : Result := 'outbus2';
        o_outbus3   : Result := 'outbus3';
        o_outbit    : Result := 'outbit';
        o_outloop   : Result := 'outloop';
        o_outinvert : Result := 'outinvert';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSeqRandom = class( TMod)
  private
    FValues       : array[ Byte] of TScaleNote;
    FPattern      : array[ Byte] of TScaleNote;
    FCount        : Integer;
    FScale        : Integer;
    FBaseNote     : TScaleNote;
    FLowNote      : TScaleNote;
    FHighNote     : TScaleNote;
    FNoteCount    : Integer;
    FScaleNotes   : array[ TScaleNote] of TScaleNote;
    FOldTrig      : Boolean;
    FOldOne       : Boolean;
    FPosition     : Byte;
    FDistribution : TSignal;
  private
}

    procedure   TModSeqRandom.MakeScale;
    var
      aScale : TScaleNoteSet;
      i      : TScaleNote;
    begin
      aScale := TransPoseAndCompleteScale( SystemScales.ScaleNotes[ FScale], FLowNote, FHighNote, FBaseNote, 12);

      FNoteCount := 0;
      for i := Low( i) to High( i)
      do begin
        if i in aScale
        then begin
          FScaleNotes[ FNoteCount] := i;
          Inc( FNoteCount);
        end;
      end;
    end;

    function    TModSeqRandom.RandomNote( aPosition: Byte; aDistribution: TSignal): TScaleNote;
    var
      aRandomValue : TScaleNote;
      anOldValue   : TScaleNote;
      aNewValue    : TScaleNote;
    begin
      aRandomValue := Random( FNoteCount);
      anOldValue   := FValues[ aPosition];
      aNewValue    := Round( aDistribution * aRandomValue + ( 1 - aDistribution) * anOldValue);
      Result       := FScaleNotes[ aNewValue];
      FValues[ aPosition] := aNewValue;
    end;

    procedure   TModSeqRandom.ChangeScale( aScale: Integer);
    begin
      FScale := aScale;
      MakeScale;
      ChangeAllNotes;
    end;

    procedure   TModSeqRandom.ChangeOneNote( aPosition: Byte);
    begin
      FPattern[ FPosition] := RandomNote( aPosition, FDistribution); // correlated random selection
    end;

    procedure   TModSeqRandom.ChangeAllNotes;
    var
      i : Integer;
    begin
      for i := Low( FPattern) to High( FPattern)
      do FPattern[ i] := RandomNote( i, 1.0);       // totally random selection
    end;

    procedure   TModSeqRandom.Step;
    begin
      if FPosition < FCount - 1
      then Inc( FPosition)
      else FPosition := Low( FPosition);
    end;

//  public

    constructor TModSeqRandom.Create( aParent: TSynthPatch; const aName: string); //override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqRandom.SetDefaults; // override;
    begin
      ChangeScale( 1);
      ChangeScale( 0);
    end;

    procedure   TModSeqRandom.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'count')
      then FCount := Clip( Ceil( aValue), 1, 256)
      else inherited;
    end;

    function    TModSeqRandom.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'count')
      then Result := FCount
      else Result := inherited;
    end;

    procedure   TModSeqRandom.DoSlowTick; // override;
    var
      aBool   : Boolean;
      anInt   : Integer;
      aChange : Boolean;
      aScale  : Integer;
    begin
      if ResetFlag
      then FPosition := 0;
      ResetFlag := False;

      FDistribution := Clip( FInputs[ i_dist] + FInputs[ i_distmod] * FInputs[ i_distmodamt], 0.0, 1.0);

      aBool := False;
      anInt := Round( 12 * FInputs[ i_basenote]);
      if anInt <> FBaseNote
      then begin
        aBool     := True;
        FBaseNote := anInt;
      end;
      anInt := Round( UnitsToNoteNumber( FInputs[ i_lownote]));
      if anInt <> FLowNote
      then begin
        aBool    := True;
        FLowNote := anInt;
      end;
      anInt := Round( UnitsToNoteNumber( FInputs[ i_highnote]));
      if anInt <> FHighNote
      then begin
        aBool     := True;
        FHighNote := anInt;
      end;

      aScale := Round( FInputs[ i_scale]);
      if aBool or ( aScale <> FScale)
      then ChangeScale( aSCale);


      aBool := SignalToLogic( FInputs[ i_trig]);
      if not FOldTrig and aBool
      then begin
        aChange := SignalToLogic( FInputs[ i_changeone]);
        if not FOldOne and aChange
        then ChangeOneNote( FPosition);
        FOldOne := aChange;
        Step;
      end;
      FOldTrig := aBool;

      FOutputs[ o_out] := NoteNumberToUnits( FPattern[ FPosition] - MiddleNote) + FInputs[ i_chain];
    end;

    procedure   TModSeqRandom.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'     , [ aPrefix, Name, FInputs[ i_trig     ]], AppLocale));
      aCallback( Self, Format( '%s/%s/changeone/%f', [ aPrefix, Name, FInputs[ i_changeone]], AppLocale));
    end;

    function    TModSeqRandom.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig       : Result := 'trig';
        i_changeone  : Result := 'changeone';
        i_scale      : Result := 'scale';
        i_basenote   : Result := 'basenote';
        i_lownote    : Result := 'lownote';
        i_highnote   : Result := 'highnote';
        i_distmod    : Result := 'distmod';
        i_distmodamt : Result := 'distmodamt';
        i_dist       : Result := 'dist';
        i_chain      : Result := 'chain';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModSeqRandom.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;

    function    TModSeqRandom.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_count : Result := 'count';
        else      Result := NO_INTERNAL;
      end;
    end;


{ ========
  TMorseSymbol = (
    msDot,
    msDash,
    msInterSymbol,
    msShortGap,
    msMediumGap
  );

  TMorseSymbols = array of TMorseSymbol;

  TModMorse = class( TMod)
  private
    FMessage     : string;
    FSymbols     : string;
    FSymbolIndex : Integer;
    FOldReset    : Boolean;
    FOldClock    : Boolean;
  public
    property    Message: string read FMessage write SetMessage;
  public
}

    function    TModMorse.CharToSymbols( aChar: Char): string;
    // Using ITU encoding
    begin
      case UpCase( aChar) of
        'A' : Result := '_ ___   '                  ;     'B' : Result := '___ _ _ _   '        ;
        'C' : Result := '___ _ ___ _   '            ;     'D' : Result := '___ _ _   '          ;
        'E' : Result := '_   '                      ;     'F' : Result := '_ _ ___ _   '        ;
        'G' : Result := '___ ___ _   '              ;     'H' : Result := '_ _ _ _   '          ;
        'I' : Result := '_ _   '                    ;     'J' : Result := '_ ___ ___ ___   '    ;
        'K' : Result := '___ _ ___   '              ;     'L' : Result := '_ ___ _ _   '        ;
        'M' : Result := '___ ___   '                ;     'N' : Result := '___ _   '            ;
        'O' : Result := '___ ___ ___   '            ;     'P' : Result := '_ ___ ____   '       ;
        'Q' : Result := '___ ___ _ ___   '          ;     'R' : Result := '_ ___ _   '          ;
        'S' : Result := '_ _ _   '                  ;     'T' : Result := '___   '              ;
        'U' : Result := '_ _ ___   '                ;     'V' : Result := '_ _ _ ___   '        ;
        'W' : Result := '_ ___ ___   '              ;     'X' : Result := '___ _ _ ___   '      ;
        'Y' : Result := '___ _ ___ ___   '          ;     'Z' : Result := '___ ___ _ _   '      ;
        '0' : Result := '___ ___ ___ ___ ___   '    ;     '1' : Result := '_ ___ ___ ___ ___   ';
        '2' : Result := '_ _ ___ ___ ___   '        ;     '3' : Result := '_ _ _ ___ ___   '    ;
        '4' : Result := '_ _ _ ____   '             ;     '5' : Result := '_ _ _ _ _   '        ;
        '6' : Result := '____ _ _ _   '             ;     '7' : Result := '___ ___ _ _ _   '    ;
        '8' : Result := '___ ___ ___ _ _   '        ;     '9' : Result := '___ ___ ___ ___ _   ';
        '*' : Result := '_ _ _ ___ ___ ___ _ _ _   ';  // SOS symbol
        else  Result := '       ';                     // Inter word delay, for all unknown chars as well as for spaces.
      end;
    end;

    procedure   TModMorse.MessageToSymbols;
    var
      i : Integer;
    begin
      FSymbols := '';
      for i := 1 to Length( FMessage)
      do FSymbols := FSymbols + CharToSymbols( FMessage[ i]);
    end;

    procedure   TModMorse.SetMessage( const aValue: string);
    begin
      if aValue <> FMessage
      then begin
        FMessage := aValue;
        MessageToSymbols;
      end;
    end;

//  public

    constructor TModMorse.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMorse.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'text')
      then Message := aValue
      else inherited
    end;

    function    TModMorse.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'text')
      then result := Message
      else Result := inherited;
    end;

    procedure   TModMorse.DoSlowTick; // override;
    var
      NewReset : Boolean;
      NewClock : Boolean;
    begin
      NewReset := SignalToLogic( FInputs[ i_reset]);
      NewClock := SignalToLogic( FInputs[ i_clock]);

      if ( not FOldReset and NewReset) or ResetFlag or (  FSymbolIndex > Length( FSymbols))
      then begin
        ResetFlag        := False;
        FSymbolIndex     :=  1;
        FOutputs[ o_out] := -1;
      end;

      if not FOldClock and NewClock
      then begin
        if ( FSymbols = '') or ( FSymbols[ FSymbolIndex] = ' ')
        then FOutputs[ o_out] := -1
        else FOutputs[ o_out] :=  1;
        Inc( FSymbolIndex);
      end;

      FOldReset := NewReset;
      FOldClock := NewClock;
    end;

    procedure   TModMorse.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, LogicToSignal( FOldReset       )], AppLocale));
      aCallback( Self, Format( '%s/%s/clock/%f', [ aPrefix, Name, LogicToSignal( FOldClock       )], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'  , [ aPrefix, Name,                FOutputs[ o_out] ], AppLocale));
    end;

    function    TModMorse.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_reset : Result := 'reset';
        i_clock : Result := 'clock';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModMorse.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModGateSeq = class( TMod)
  private
    FOutValue   : TSignal;
    FSteps      : Integer;
    FOldRes     : Boolean;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FForwards   : Boolean;
    FReset      : Boolean;
  private
}

    procedure   TModGateSeq.SampleInputs;
    begin
      FInState  := seqClkHigh;
      FForwards := SignalToLogic( FInputs[ i_dir]);
      FReset    := SignalToLogic( FInputs[ i_res]);
    end;

//  public

    constructor TModGateSeq.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModGateSeq.SetDefaults; // override;
    begin
      FOutValue           := 0;
      FInputs [ i_chain ] := 0; // Chain input to zero
      FInputs [ i_active] := 1; // Sequencer active by default
      FInputs [ i_dir   ] := 1; // Forwards by default
      SampleInputs;
      if FForwards
      then FCounter := 0
      else FCounter := FSteps - 1;
    end;

    procedure   TModGateSeq.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 16)
      else inherited;
    end;

    function    TModGateSeq.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModGateSeq.DoSlowTick; // override;
    var
      Mode : Integer;
    begin
      if ResetFlag
      then begin
        SampleInputs;
        if FForwards
        then FCounter := 0
        else FCounter := FSteps - 1;
      end
      else begin
        case FInState of
          seqClkLow :
            begin
              if SignalToLogic( FInputs[ i_trig])
              then begin
                SampleInputs;
                if ResetFlag or ( not FOldRes and FReset)
                then begin
                  if FForwards
                  then FCounter := 0
                  else FCounter := FSteps - 1;
                end
                else begin
                  if FSteps > 0
                  then begin
                    SampleInputs;
                    if FForwards
                    then FCounter := ( FCounter + 1) mod FSteps
                    else begin
                      if FCounter <= 0
                      then FCounter := FSteps - 1
                      else Dec( FCounter);
                    end;
                  end;
                end;
                Mode := Round( FInputs[ i_step1 + FCounter]);
                if ( Mode in [ 1, 2]) and SignalToLogic( FInputs[ i_active])
                then FOutValue := 1.0
                else FOutValue := FInputs[ i_chain];
              end;
            end;
          seqClkHigh :
            begin
              if not SignalToLogic( FInputs[ i_trig])
              then begin
                FInState := seqClkLow;
                Mode := Round( FInputs[ i_step1 + FCounter]);
                if ( Mode = 2) and SignalToLogic( FInputs[ i_active])
                then FOutValue := 1.0
                else FOutValue := FInputs[ i_chain];
              end;
            end;
        end; // case
      end;
      ResetFlag := False;
      FOldRes   := FReset;
      FOutputs[ o_out    ] := FOutValue;
      FOutputs[ o_outres ] := FInputs[ i_res ];
      FOutputs[ o_outtrig] := FInputs[ i_trig];
      FOutputs[ o_outdir ] := FInputs[ i_dir ];
    end;

    procedure   TModGateSeq.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, FInputs [ i_trig      ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f' , [ aPrefix, Name, LogicToSignal( FOldRes)], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, FInputs [ i_active    ]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'   , [ aPrefix, Name, FInputs [ i_dir       ]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'   , [ aPrefix, Name, FOutputs[ o_out       ]], AppLocale));
      for i := i_step1 to i_step16
      do begin
        if ( i = FCounter + i_step1) and SignalToLogic( FInputs[ i_active])
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - i_step1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - i_step1 + 1], AppLocale));
      end;
    end;

    function    TModGateSeq.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig   : Result := 'trig';
        i_step1  : Result := 'step1';
        i_step2  : Result := 'step2';
        i_step3  : Result := 'step3';
        i_step4  : Result := 'step4';
        i_step5  : Result := 'step5';
        i_step6  : Result := 'step6';
        i_step7  : Result := 'step7';
        i_step8  : Result := 'step8';
        i_step9  : Result := 'step9';
        i_step10 : Result := 'step10';
        i_step11 : Result := 'step11';
        i_step12 : Result := 'step12';
        i_step13 : Result := 'step13';
        i_step14 : Result := 'step14';
        i_step15 : Result := 'step15';
        i_step16 : Result := 'step16';
        i_chain  : Result := 'chain';
        i_res    : Result := 'res';
        i_active : Result := 'active';
        i_dir    : Result := 'dir';
        else       Result := NO_INPUT;
      end;
    end;

    function    TModGateSeq.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out     : Result := 'out';
        o_outres  : Result := 'outres';
        o_outtrig : Result := 'outtrig';
        o_outdir  : Result := 'outdir';
        else        Result := NO_OUTPUT;
      end;
    end;

    function    TModGateSeq.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_steps : Result := 'steps';
        else      Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModMemory = class( TMod)
  private
    FMemoryL : TSignalArrayWrapper;
    FMemoryR : TSignalArrayWrapper;
  public
}

    constructor TModMemory.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FMemoryL := TSignalArrayWrapper.Create( 0);
      FMemoryR := TSignalArrayWrapper.Create( 0);
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModMemory.Destroy; // override;
    begin
      FreeAndNil( FMemoryL);
      FreeAndNil( FMemoryR);
      inherited;
    end;


{ ========
  TModPitchShift = class( TMod)
  private
    FDelayLine : TSignalArrayWrapper;
    FDelay     : TSignal;
    FLength    : Integer;
    FInPointer : Integer;
    FDelta     : TSignal;
    FPhase     : TSignal;
  private
}

    procedure   TModPitchShift.SetDelay( aValue: TSignal);
    var
      NewDelay : TSignalArrayWrapper;
      OldDelay : TSignalArrayWrapper;
      aLength  : Integer;
    begin
      Assert( aValue > 0);
      aLength      := Ceil( TimeToSampleCount( aValue));
      NewDelay     := TSignalArrayWrapper.Create( aLength);
      FLength      := 0;
      OldDelay     := AtomicExchange( Pointer( FDelayLine), Pointer( NewDelay));
      FLength      := aLength;
      FPhase       := 0.0;
      FDelay       := aValue;
      FreeAndNil( OldDelay);
    end;

//  public

    constructor TModPitchShift.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      SetDelay( 50e-3);
    end;

    destructor  TModPitchShift.Destroy; // override;
    begin
      FreeAndNil( FDelayLine);
      inherited;
    end;

    procedure   TModPitchShift.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'delay')
      then begin
        if aValue < 1e-6
        then aValue := 1e-6;
        if aValue <> FDelay
        then SetDelay( aValue);
      end
      else inherited;
    end;

    function    TModPitchShift.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'delay')
      then Result := FDelay
      else Result := inherited;
    end;

    procedure   TModPitchShift.DoSlowTick; // override;
    var
      ControlValue : TSignal;
      BaseFreq     : TSignal;
      TargetFreq   : TSignal;
      TargetLength : TSignal;
    begin
      if Assigned( FDelayLine) and ( FLength > 0)
      then begin
        ControlValue := FInputs[ i_shift] + FInputs[ i_shiftmod] * FInputs[ i_shiftmodamt] + FInputs[ i_shiftfine];
        BaseFreq     := System_Rate / FLength;
        TargetFreq   := BaseFreq * Power( 2, NOTE_SCALING * ControlValue * ONE_TWELFTH);
        TargetLength := System_Rate / TargetFreq;
        FDelta       := 1 / TargetLength;
      end;
    end;

    procedure   TModPitchShift.DoTick; // override;
    var
      Outp1   : TSignal;
      Outp2   : TSignal;
      Mix1    : TSignal;
      Mix2    : TSignal;
      Phase2  : TSignal;
      InPhase : TSignal;
      Trunced : Integer;
      Input   : TSignal;
      Output  : TSignal;
    begin
      if Assigned( FDelayLine) and ( FLength > 0)
      then begin
        InPhase := FInPointer / FLength;

        if FPhase <= 0.5
        then Phase2 := FPhase + 0.5
        else Phase2 := FPhase - 0.5;

        Outp1 := FPhase * FLength;
        Outp2 := Phase2 * FLength;

        Mix1 := 2 * ( FPhase - InPhase);
        if Mix1 < 0
        then Mix1 := Mix1 + 2;
        if Mix1 > 1
        then Mix1 := 2 - Mix1;

        Mix2 := 2 * ( Phase2 - InPhase);
        if Mix2 < 0
        then Mix2 := Mix2 + 2;
        if Mix2 > 1
        then Mix2 := 2 - Mix2;

        Output := FDelayLine.Interpolate( Outp1) * Mix1 + FDelayLine.Interpolate( Outp2) * Mix2;
        FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Output;

        Input := Clip( Normalize( FInputs[ i_in]), -16, 16);
        FDelayLine.Signal[ FInPointer] := Input;
        if FLength > 0
        then FInPointer := ( FInPointer + 1) mod FLength;

        FPhase      := FPhase + FDelta;
        Trunced     := Trunc( FPhase);
        FPhase      := FPhase - Trunced;
        if FPhase < 0
        then FPhase := FPhase + 1;
      end;
    end;

    procedure   TModPitchShift.SampleRateChanged; // override;
    begin
      SetDelay( FDelay); // recalculate line lengths for new sample rate
    end;

    function    TModPitchShift.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in          : Result := 'in';
        i_shift       : Result := 'shift';
        i_shiftfine   : Result := 'shiftfine';
        i_shiftmod    : Result := 'shiftmod';
        i_shiftmodamt : Result := 'shiftmodamt';
        i_mute        : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModPitchShift.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;

    function    TModPitchShift.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_delay :  Result := 'length';
        else       Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModDelay = class( TModMemory)
  private
    FLength    : TSignal;
    FInPointer : Integer;
  public
}

    procedure   TModDelay.SetDefaults; // override;
    begin
      FInputs[ i_delay] := 1.0; // Set maximum delay time if nothing connected
    end;

    procedure   TModDelay.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      NewMem: TSignalArrayWrapper;
      OldMem: TSignalArrayWrapper;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 1e-6
        then aValue := 1e-6;
        if aValue <> FMemLength
        then begin
          NewMem     := TSignalArrayWrapper.Create( Ceil( TimeToSampleCount( aValue)));
          FMemLength := 0;
          OldMem     := AtomicExchange( Pointer( FMemoryL), Pointer( NewMem));
          FMemLength := aValue;
          FreeAndNil( OldMem);
        end;
      end
      else inherited;
    end;

    function    TModDelay.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FMemLength
      else Result := inherited;
    end;

    procedure   TModDelay.DoTick; // override;
    var
      OutPointerVar : TSignal;
      SampCount     : Integer;
    begin
      if Assigned( FMemoryL) and ( FMemoryL.Count > 0)
      then begin
        SampCount                    := FMemoryL.Count;
        FInPointer                   := FInPointer mod SampCount;
        FOutputs[ o_outfixed]        := FMemoryL.Signal[ FInPointer];
        FMemoryL.Signal[ FInPointer] := Clip( Normalize( FInputs[ i_in]), -16, 16);
        OutPointerVar                := MathFloatMod( FInPointer - Clip( FInputs[ i_delay], 0.0, 1.0) * ( TimeToSampleCount( FMemLength) - 1), SampCount);
        FOutputs[ o_out]             := FMemoryL.Interpolate( OutPointerVar);
        FInPointer                   := ( FInPointer + 1) mod SampCount;
      end;
    end;

    function    TModDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in    : Result := 'in';
        i_delay : Result := 'delay';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out      : Result := 'out';
        o_outfixed : Result := 'outfixed';
        else         Result := NO_OUTPUT;
      end;
    end;

    function    TModDelay.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_length :  Result := 'length';
        else        Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModDelay8 = class( TModMemory)
  private
    FLength    : TSignal;
    FInPointer : Integer;
  public
}

    procedure   TModDelay8.SetDefaults; // override;
    begin
      FInputs[ i_delay1] := 1.0; // Set maximum delay time if nothing connected
      FInputs[ i_delay2] := 1.0;
      FInputs[ i_delay3] := 1.0;
      FInputs[ i_delay4] := 1.0;
      FInputs[ i_delay5] := 1.0;
      FInputs[ i_delay6] := 1.0;
      FInputs[ i_delay7] := 1.0;
      FInputs[ i_delay8] := 1.0;
    end;

    procedure   TModDelay8.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      NewMem: TSignalArrayWrapper;
      OldMem: TSignalArrayWrapper;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 1e-6
        then aValue := 1e-6;
        if aValue <> FMemLength
        then begin
          NewMem     := TSignalArrayWrapper.Create( Ceil( TimeToSampleCount( aValue)));
          FMemLength := 0;
          OldMem     := AtomicExchange( Pointer( FMemoryL), Pointer( NewMem));
          FMemLength := aValue;
          FreeAndNil( OldMem);
        end;
      end
      else inherited;
    end;

    function    TModDelay8.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FMemLength
      else Result := inherited;
    end;

    procedure   TModDelay8.DoTick; // override;
    var
      OutPointerVar1 : TSignal;
      OutPointerVar2 : TSignal;
      OutPointerVar3 : TSignal;
      OutPointerVar4 : TSignal;
      OutPointerVar5 : TSignal;
      OutPointerVar6 : TSignal;
      OutPointerVar7 : TSignal;
      OutPointerVar8 : TSignal;
      SampCount      : Integer;
      SampCount_1    : TSignal;
    begin
      if Assigned( FMemoryL) and ( FMemoryL.Count > 0)
      then begin
        SampCount                    := FMemoryL.Count;
        SampCount_1                  := TimeToSampleCount( FMemLength) - 1;
        FInPointer                   := FInPointer mod SampCount;
        FOutputs[ o_outfixed]        := FMemoryL.Signal[ FInPointer];
        FMemoryL.Signal[ FInPointer] := Clip( Normalize( FInputs[ i_in]), -16, 16);

        OutPointerVar1               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay1], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out1]            := FMemoryL.Interpolate( OutPointerVar1);
        OutPointerVar2               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay2], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out2]            := FMemoryL.Interpolate( OutPointerVar2);
        OutPointerVar3               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay3], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out3]            := FMemoryL.Interpolate( OutPointerVar3);
        OutPointerVar4               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay4], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out4]            := FMemoryL.Interpolate( OutPointerVar4);
        OutPointerVar5               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay5], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out5]            := FMemoryL.Interpolate( OutPointerVar5);
        OutPointerVar6               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay6], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out6]            := FMemoryL.Interpolate( OutPointerVar6);
        OutPointerVar7               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay7], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out7]            := FMemoryL.Interpolate( OutPointerVar7);
        OutPointerVar8               := MathFloatMod( FInPointer - Clip( FInputs[ i_delay8], 0.0, 1.0) * SampCount_1, SampCount);
        FOutputs[ o_out8]            := FMemoryL.Interpolate( OutPointerVar8);

        FInPointer                   := ( FInPointer + 1) mod SampCount;
      end;
    end;

    function    TModDelay8.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in     : Result := 'in';
        i_delay1 : Result := 'delay1';
        i_delay2 : Result := 'delay2';
        i_delay3 : Result := 'delay3';
        i_delay4 : Result := 'delay4';
        i_delay5 : Result := 'delay5';
        i_delay6 : Result := 'delay6';
        i_delay7 : Result := 'delay7';
        i_delay8 : Result := 'delay8';
        else       Result := NO_INPUT;
      end;
    end;

    function    TModDelay8.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outfixed : Result := 'outfixed';
        o_out1     : Result := 'out1';
        o_out2     : Result := 'out2';
        o_out3     : Result := 'out3';
        o_out4     : Result := 'out4';
        o_out5     : Result := 'out5';
        o_out6     : Result := 'out6';
        o_out7     : Result := 'out7';
        o_out8     : Result := 'out8';
        else         Result := NO_OUTPUT;
      end;
    end;

    function    TModDelay8.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_length :  Result := 'length';
        else        Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModClockedDelay = class( TModMemory)
  private
    FLength    : TSignal;
    FInPointer : Integer;
  public
}

    constructor TModClockedDelay.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModClockedDelay.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      NewMem: TSignalArrayWrapper;
      OldMem: TSignalArrayWrapper;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 1
        then aValue := 1;
        if aValue <> FMemLength
        then begin
          NewMem     := TSignalArrayWrapper.Create( Round( aValue));
          FMemLength := 0;
          OldMem     := AtomicExchange( Pointer( FMemoryL), Pointer( NewMem));
          FMemLength := aValue;
          FreeAndNil( OldMem);
        end;
      end
      else inherited;
    end;

    function    TModClockedDelay.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FMemLength
      else Result := inherited;
    end;

    procedure   TModClockedDelay.DoSlowTick; // override;
    var
      SampCount : Integer;
      Clock     : Boolean;
    begin
      if ResetFlag
      then begin
        if Assigned( FMemoryL)
        then FMemoryL.Erase;
        ResetFlag := False;
      end;
      if Assigned( FMemoryL) and ( FMemoryL.Count > 0)
      then begin
        Clock := SignalToLogic( FInputs[ i_clock]);
        if not FOldClock and Clock
        then begin
          SampCount                    := FMemoryL.Count;
          FInPointer                   := FInPointer mod SampCount;
          FOutputs[ o_out]             := FMemoryL.Signal[ FInPointer];
          FMemoryL.Signal[ FInPointer] := FInputs[ i_in];
          FInPointer                   := ( FInPointer + 1) mod SampCount;
        end;
        FOldClock := Clock;
      end;
    end;

    procedure   TModClockedDelay.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/clock/%f', [ aPrefix, Name, LogicToSignal( FOldClock)], AppLocale));
    end;

    function    TModClockedDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in    : Result := 'in';
        i_clock : Result := 'clock';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModClockedDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;

    function    TModClockedDelay.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_length :  Result := 'length';
        else        Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModWave = class( TModMemory)
  private
    FFileName    : string;
    FDuration    : string;
    FIsStereo    : Boolean;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

    procedure   TModWave.SetDuration( aValue: Cardinal);
    begin
      if aValue < 1000
      then FDuration := Format( '%.3f ms', [ aValue * 1.0], AppLocale)
      else if aValue >= 10000
      then FDuration := Format( '%.2d:%.2d', [ Round( aValue * 1e-3) div 60, Round( aValue * 1e-3) mod 60])
      else FDuration := Format( '%.3f s', [ aValue * 1e-3], AppLocale);
    end;

    procedure   TModWave.SwapMemoryFor( aMemory: TSignalArray);
    var
      aNewMemoryL  : TSignalArrayWrapper;
      anOldMemoryL : TSignalArrayWrapper;
      aNewMemoryR  : TSignalArrayWrapper;
      anOldMemoryR : TSignalArrayWrapper;
      i            : Integer;
    begin
      if FIsStereo
      then begin
        aNewMemoryL  := TSignalArrayWrapper.Create( Length( aMemory) shr 1);
        aNewMemoryR  := TSignalArrayWrapper.Create( Length( aMemory) shr 1);
        for i := 0 to aNewMemoryL.Count - 1
        do begin
          aNewMemoryL.Signal[ i] := aMemory[ 2 * i    ];
          aNewMemoryR.Signal[ i] := aMemory[ 2 * i + 1];
        end;

        anOldMemoryL := FMemoryL;
        anOldMemoryR := FMemoryR;
        FMemoryL     := aNewMemoryL;
        FMemoryR     := aNewMemoryR;

        FreeAndNil( anOldMemoryL);
        FreeAndNil( anOldMemoryR);
      end
      else begin
        aNewMemoryL  := TSignalArrayWrapper.Create( Length( aMemory));
        for i := 0 to Length( aMemory) - 1
        do aNewMemoryL.Signal[ i] := aMemory[ i];

        anOldMemoryL := FMemoryL;
        FMemoryL     := aNewMemoryL;

        FreeAndNil( anOldMemoryL);
      end;
    end;

    procedure   TModWave.LoadFile;
    const
      BufSize = 8192;
    type
      TSample = SmallInt;
      TBuffer = array[ 0 .. BufSize - 1] of TSample;

      procedure SetSamples16bit44100Hz( var aMemory: TSignalArray; aBuf: TBuffer; aSampleCount: Cardinal);
      const
        MaxVal    = High( TSample) * 0.5;
        MaxValRec = 1.0 / MaxVal;
      var
        i        : Cardinal;
        Position : Cardinal;
      begin
        Position := Length( aMemory);
        SetLength( aMemory, Position + aSampleCount);
        for i := 0 to aSampleCount - 1
        do begin
          aMemory[ Position] := aBuf[ i] * MaxValRec;
          Inc( Position);
        end;
      end;

    var
      aMemory    : TSignalArray;
      aWaveFile  : TWaveFile;
      aByteCount : Cardinal;
      aBuf       : TBuffer;
      aDuration  : Cardinal;
      aFormat    : TPCMFormat;
    begin
      SetLength( aMemory, 0);
      aDuration := 0;
      if FileExists( FileName) // done: do not limit the file size :: and FileSize < 44100 * 2 * 2 * 301 // 44k1, stereo, 16 bit and duration < 301 s.
      then begin
        aWaveFile := TWaveFile.Create( FileName, fmOpenRead);
        try
          if aWaveFile.Valid
          then begin
            aFormat := aWaveFile.PCMFormat;
            if ( aFormat = Stereo16bit44100Hz) or ( aFormat = Mono16bit44100Hz)
            then begin
              aDuration := aWaveFile.Length;
              aWaveFile.Position := 0;
              if aWaveFile.BeginRead
              then begin
                try
                  while Cardinal( aWaveFile.Position) < aWaveFile.DataSize
                  do begin
                    aByteCount := aWaveFile.Read( aBuf, BufSize * SizeOf( TSample));
                    case aFormat of
                      Stereo16bit44100Hz : begin FIsStereo := True ; SetSamples16bit44100Hz( aMemory, aBuf, aByteCount div SizeOf( TSample)); end;
                      Mono16bit44100Hz   : begin FIsStereo := False; SetSamples16bit44100Hz( aMemory, aBuf, aByteCount div SizeOf( TSample)); end;
                    end;
                  end;
                finally
                  aWaveFile.EndRead;
                end;
              end;
            end;
          end;
        finally
          aWaveFile.DisposeOf;
        end;
      end;
      SwapMemoryFor( aMemory);
      SetDuration( aDuration);
    end;

    procedure   TModWave.SetFileName( const aValue: string);
    begin
      if  aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

//  public

    procedure   TModWave.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModWave.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModWave.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    const
      MonoText : array[ Boolean] of string = ( 'M', 'S');
    begin
      aCallback( Self, Format( '%s/%s/duration/%s', [ aPrefix, Name, FDuration]           , AppLocale));
      aCallback( Self, Format( '%s/%s/mono/%s'    , [ aPrefix, Name, MonoText[ FIsStereo]], AppLocale));
    end;

    function    TModWave.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl :  Result := 'out';        // Left  output
        o_outr :  Result := 'outr';       // Right output
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
}

    procedure   TModWavePlayer.SetDefaults; // override;
    begin
      FInputs[ i_speed]    := 0.0; // Set zero speed
      FInputs[ i_position] := 0.0; // Set start position
      FInputs[ i_res]      := 0.0; // Reset to zero
      FInputs[ i_fm]       := 0.0; // FrequencyToUnits( ReferenceA); // Set FM amount to neutral to make speed work ...
      SetDuration( 0);
    end;

    procedure   TModWavePlayer.DoTick; // override;
    var
      SampCount : Integer;
      Increment : TSignal;
      Position  : TSignal;
      MakeMono  : Boolean;
    begin
      if ResetFlag
      then begin
        FPlayPointer := 0.0;
        ResetFlag    := False;
      end;

      if ( not FOldRes) and SignalToLogic( FInputs[ i_res])
      then FPlayPointer := 0.0;
      FOldRes := SignalToLogic( FInputs[ i_res]);

      if FIsStereo
      then begin
        MakeMono := Round( FInputs[ i_mode]) = 0;
        if Assigned( FMemoryL) and Assigned( FMemoryR)
        then begin
          SampCount := FMemoryL.Count;
          if SampCount > 0
          then begin
            Increment    := Clip( FInputs[ i_speed] * LookupFrequency( FInputs[ i_fm]) / LookupFrequency( 0), - 16, 16);
            Position     := ( 32 + Clip( FInputs[ i_position],  -16, 16)) * SampCount;
            FPlayPointer := MathFloatMod( FPlayPointer + Increment, SampCount);
            Position     := MathFloatMod( Position + FPlayPointer, SampCount);
            if MakeMono
            then begin
              FOutputs[ o_outl] := ( FMemoryL.Interpolate( Position) + FMemoryR.Interpolate( Position)) * 0.5;
              FOutputs[ o_outr] := FOutputs[ o_outl];
            end
            else begin
              FOutputs[ o_outl] := FMemoryL.Interpolate( Position);
              FOutputs[ o_outr] := FMemoryR.Interpolate( Position);
            end;
          end
          else begin
            FOutputs[ o_outl] := 0;
            FOutputs[ o_outr] := 0;
          end;
        end
        else begin
          FOutputs[ o_outl] := 0;
          FOutputs[ o_outr] := 0;
        end;
      end
      else begin
        if Assigned( FMemoryL)
        then begin
          SampCount := FMemoryL.Count;
          if SampCount > 0
          then begin
            Increment     := Clip( FInputs[ i_speed] * LookupFrequency( FInputs[ i_fm]) / LookupFrequency( 0), - 16, 16);
            Position      := ( 32 + Clip( FInputs[ i_position],  -16, 16)) * SampCount;
            FPlayPointer  := MathFloatMod( FPlayPointer + Increment, SampCount);
            Position      := MathFloatMod( Position + FPlayPointer, SampCount);
            FOutputs[ o_outl] := FMemoryL.Interpolate( Position);
          end
          else FOutputs[ o_outl] := 0;
        end
        else FOutputs[ o_outl] := 0;
        FOutputs[ o_outr] := FOutputs[ o_outl];
      end;
    end;

    procedure   TModWavePlayer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f', [ aPrefix, Name, FInputs[ i_res]], AppLocale));
    end;

    function    TModWavePlayer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_speed    : Result := 'speed';
        i_position : Result := 'position';
        i_res      : Result := 'res';
        i_mode     : Result := 'mode';            // 0 = mono, 1 = stereo
        i_fm       : Result := 'fm';
        else         Result := NO_INPUT;
      end;
    end;


{ ========
  TModRndGranulator = class( TMod)
  private
    FGranulator       : TRndGranulator;
    FLocked           : Boolean;
    FDelayTime        : TSignal;
    FGrainCount       : Integer;
    FInterOnset       : TSignal;
    FInterOnsetMod    : TSignal;
    FGrainDuration    : TSignal;
    FGrainDurationMod : TSignal;
    FVolume           : TSignal;
    FVolumeMod        : TSignal;
    FPan              : TSignal;
    FPanMod           : TSignal;
    FRecording        : TSignal;
  public
    property    DelayTime  : TSignal read FDelayTime  write SetDelayTime;
    property    GrainCount : Integer read FGrainCount write SetGrainCount;
  public
}

    procedure   TModRndGranulator.Recreate;
    var
      OldGranulator : TRndGranulator;
    begin
      FLocked := True;
      OldGranulator :=
        TRndGranulator(
          AtomicExchange(
            Pointer( FGranulator),
            Pointer( TRndGranulator.Create( System_Rate, DelayTime, GrainCount))
          )
        );
      FreeAndNil( OldGranulator);
      FLocked := False;
    end;

    procedure   TModRndGranulator.SetDelayTime( aValue: TSignal);
    begin
      if aValue <> FDelayTime
      then begin
        FDelayTime := aValue;
        Recreate;
      end;
    end;

    procedure   TModRndGranulator.SetGrainCount( aValue: Integer);
    begin
      if aValue <> FGrainCount
      then begin
        FGrainCount := aValue;
        Recreate;
      end;
    end;

//  public

    constructor TModRndGranulator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := True;
    end;

    destructor  TModRndGranulator.Destroy; // override;
    var
      OldGranulator : TRndGranulator;
    begin
      FLocked := True;
      OldGranulator := TRndGranulator( AtomicExchange( Pointer( FGranulator), nil));
      FreeAndNil( OldGranulator);
      inherited;
    end;

    procedure   TModRndGranulator.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'graincount')
      then GrainCount := Round( aValue)
      else if SameText( aName, 'delaytime')
      then DelayTime := aValue
      else inherited;
    end;

    function    TModRndGranulator.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'graincount')
      then Result := GrainCount
      else if SameText( aName, 'delaytime')
      then Result := DelayTime
      else Result := inherited;
    end;

    procedure   TModRndGranulator.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute               ,
        i_interonset         ,
        i_interonsetmodamt   ,
        i_grainduration      ,
        i_graindurationmodamt,
        i_volume             ,
        i_volumemodamt       ,
        i_pan                ,
        i_panmodamt          ,
        i_record
      ];
      FInputs[ i_delay] := 1.0;
      FInputs[ i_rec  ] := 1.0;
    end;

    procedure   TModRndGranulator.DoSlowTick; // override;
    begin
      if Assigned( FGranulator) and not FLocked
      then begin
        FRecording := SignalToMute( FInputs[ i_record]) * Clip( FInputs[ i_rec], 0.0, 1.0);

        FInterOnSet    := LookupEnvTime( erFast, Finputs[ i_interonset   ]);
        FGrainDuration := LookupEnvTime( erFast, Finputs[ i_grainduration]);
        FVolume        :=                        Finputs[ i_volume       ] ;
        FPan           :=                        Finputs[ i_pan          ] ;

        FInterOnsetMod    := Clip( FInputs[ i_interonsetmod   ] + FInputs[ i_interonsetmodamt   ], 0.0, 1.0) * FInterOnset   ;
        FGrainDurationMod := Clip( FInputs[ i_graindurationmod] + FInputs[ i_graindurationmodamt], 0.0, 1.0) * FGrainDuration;
        FVolumeMod        := Clip( FInputs[ i_volumemod       ] + FInputs[ i_volumemodamt       ], 0.0, 1.0) * FVolume       ;
        FPanMod           := Clip( FInputs[ i_panmod          ] + FInputs[ i_panmodamt          ], 0.0, 1.0)                 ;

        FGranulator.GrainOffset       := FInputs[ i_grainoffset];
        FGranulator.Mute              :=        FInputs[ i_mute   ] ;
        FGranulator.CurrentEnvelope   := Round( FInputs[ i_envtype]);
        FGranulator.DelayModulation   :=        FInputs[ i_delay  ] ;

        FGranulator.MinInterOnsetTime := FInterOnset    - FInterOnsetMod   ;
        FGranulator.MaxInterOnsetTime := FInterOnset    + FInterOnsetMod   ;
        FGranulator.MinGrainDuration  := FGrainDuration - FGrainDurationMod;
        FGranulator.MaxGrainDuration  := FGrainDuration + FGrainDurationMod;

        FGranulator.MinVolume         := Clip( FVolume - FVolumeMod,  0.0, 4.0);
        FGranulator.MaxVolume         := Clip( FVolume + FVolumeMod,  0.0, 4.0);
        FGranulator.MinPan            := Clip( FPan    - FPanMod   , -1.0, 1.0);
        FGranulator.MaxPan            := Clip( FPan    + FPanMod   , -1.0, 1.0);

        FGranulator.Recording         := FRecording;
      end;
    end;

    procedure   TModRndGranulator.DoTick; // override;
    begin
      if Assigned( FGranulator) and not FLocked
      then FGranulator.Tick( FInputs[ i_in], FOutputs[ o_outl], FOutputs[ o_outr]);
    end;

    procedure   TModRndGranulator.SampleRateChanged; // override;
    begin
      Recreate;
    end;

    function    TModRndGranulator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in                  : Result := 'in';
        i_mute                : Result := 'mute';
        i_envtype             : Result := 'envtype';
        i_interonset          : Result := 'interonset';
        i_interonsetmod       : Result := 'interonsetmod';
        i_interonsetmodamt    : Result := 'interonsetmodamt';
        i_grainduration       : Result := 'grainduration';
        i_graindurationmod    : Result := 'graindurationmod';
        i_graindurationmodamt : Result := 'graindurationmodamt';
        i_volume              : Result := 'volume';
        i_volumemod           : Result := 'volumemod';
        i_volumemodamt        : Result := 'volumemodamt';
        i_pan                 : Result := 'pan';
        i_panmod              : Result := 'panmod';
        i_panmodamt           : Result := 'panmodamt';
        i_record              : Result := 'record';
        i_rec                 : Result := 'rec';
        i_delay               : Result := 'delay';
        i_grainoffset         : Result := 'grainoffset';
        else                    Result := NO_INPUT;
      end;
    end;

    function    TModRndGranulator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outr : Result := 'outr';
        else     Result := NO_OUTPUT;
      end;
    end;

    function    TModRndGranulator.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_delaytime  : Result := 'delaytime';
        s_graincount : Result := 'grainccount';
        else           Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModGranulator = class( TMod)
  private
    FGranulator    : TGranulator;
    FLocked        : Boolean;
    FDelayTime     : TSignal;
    FGrainCount    : Integer;
  public
    property    DelayTime  : TSignal read FDelayTime  write SetDelayTime;
    property    GrainCount : Integer read FGrainCount write SetGrainCount;
  public
}

    procedure   TModGranulator.Recreate;
    var
      OldGranulator : TGranulator;
    begin
      FLocked := True;
      OldGranulator :=
        TGranulator(
          AtomicExchange(
            Pointer( FGranulator),
            Pointer( TGranulator.Create( System_Rate, DelayTime, GrainCount))
          )
        );
      FreeAndNil( OldGranulator);
      FLocked := False;
    end;

    procedure   TModGranulator.SetDelayTime( aValue: TSignal);
    begin
      if aValue <> FDelayTime
      then begin
        FDelayTime := aValue;
        Recreate;
      end;
    end;

    procedure   TModGranulator.SetGrainCount( aValue: Integer);
    begin
      if aValue <> FGrainCount
      then begin
        FGrainCount := aValue;
        Recreate;
      end;
    end;

//  public

    constructor TModGranulator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := True;
    end;

    destructor  TModGranulator.Destroy; // override;
    var
      OldGranulator : TGranulator;
    begin
      FLocked := True;
      OldGranulator := TGranulator( AtomicExchange( Pointer( FGranulator), nil));
      FreeAndNil( OldGranulator);
      inherited;
    end;

    procedure   TModGranulator.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'graincount')
      then GrainCount := Round( aValue)
      else if SameText( aName, 'delaytime')
      then DelayTime := aValue
      else inherited;
    end;

    function    TModGranulator.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'graincount')
      then Result := GrainCount
      else if SameText( aName, 'delaytime')
      then Result := DelayTime
      else Result := inherited;
    end;

    procedure   TModGranulator.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute               ,
        i_interonset         ,
        i_interonsetmodamt   ,
        i_grainduration      ,
        i_graindurationmodamt,
        i_volume             ,
        i_volumemodamt       ,
        i_pan                ,
        i_panmodamt          ,
        i_record
      ];
      FInputs[ i_delay] := 1.0;
      FInputs[ i_rec  ] := 1.0;
    end;

    procedure   TModGranulator.DoSlowTick; // override;
    begin
      if Assigned( FGranulator) and not FLocked
      then begin
        FGranulator.Mute            := FInputs[ i_mute];
        FGranulator.CurrentEnvelope := Round( FInputs[ i_envtype]);
        FGranulator.DelayModulation := FInputs[ i_delay];
        FGranulator.Recording       := SignalToMute( FInputs[ i_record]) * Clip( FInputs[ i_rec], 0.0, 1.0);
        FGranulator.InterOnsetTime  := LookupEnvTime( erFast, Finputs[ i_interonset   ] + FInputs[ i_interonsetmod   ] * FInputs[ i_interonsetmodamt   ]);
        FGranulator.GrainDuration   := LookupEnvTime( erFast, Finputs[ i_grainduration] + FInputs[ i_graindurationmod] * FInputs[ i_graindurationmodamt]);
        FGranulator.GrainOffset     := FInputs[ i_grainoffset];
        FGranulator.GrainVolume     := Finputs[ i_volume     ] + FInputs[ i_volumemod       ] * FInputs[ i_volumemodamt       ];
        FGranulator.Pan             := Finputs[ i_pan        ] + FInputs[ i_panmod          ] * FInputs[ i_panmodamt          ];
      end;
    end;

    procedure   TModGranulator.DoTick; // override;
    begin
      if Assigned( FGranulator) and not FLocked
      then FGranulator.Tick( FInputs[ i_in], FOutputs[ o_outl], FOutputs[ o_outr]);
    end;

    procedure   TModGranulator.SampleRateChanged; // override;
    begin
      Recreate;
    end;

    function    TModGranulator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in                  : Result := 'in';
        i_mute                : Result := 'mute';
        i_envtype             : Result := 'envtype';
        i_interonset          : Result := 'interonset';
        i_interonsetmod       : Result := 'interonsetmod';
        i_interonsetmodamt    : Result := 'interonsetmodamt';
        i_grainduration       : Result := 'grainduration';
        i_graindurationmod    : Result := 'graindurationmod';
        i_graindurationmodamt : Result := 'graindurationmodamt';
        i_volume              : Result := 'volume';
        i_volumemod           : Result := 'volumemod';
        i_volumemodamt        : Result := 'volumemodamt';
        i_pan                 : Result := 'pan';
        i_panmod              : Result := 'panmod';
        i_panmodamt           : Result := 'panmodamt';
        i_record              : Result := 'record';
        i_rec                 : Result := 'rec';
        i_delay               : Result := 'delay';
        i_grainoffset         : Result := 'grainoffset';
        else                    Result := NO_INPUT;
      end;
    end;

    function    TModGranulator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outr : Result := 'outr';
        else     Result := NO_OUTPUT;
      end;
    end;

    function    TModGranulator.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_delaytime  : Result := 'delaytime';
        s_graincount : Result := 'grainccount';
        else           Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModSimpleGranulator = class( TMod)
  private
    FGranulator : TSimpleGranulator;
    FLocked     : Boolean;
    FDelayTime  : TSignal;
    FGrainCount : Integer;
    FOldTrig    : Boolean;
  public
    property    DelayTime  : TSignal read FDelayTime  write SetDelayTime;
    property    GrainCount : Integer read FGrainCount write SetGrainCount;
  private
}

    procedure   TModSimpleGranulator.Recreate;
    var
      OldGranulator : TSimpleGranulator;
    begin
      FLocked := True;
      OldGranulator :=
        TSimpleGranulator(
          AtomicExchange(
            Pointer( FGranulator),
            Pointer( TSimpleGranulator.Create( System_Rate, DelayTime, GrainCount))
          )
        );
      FreeAndNil( OldGranulator);
      FLocked := False;
    end;

    procedure   TModSimpleGranulator.SetDelayTime ( aValue: TSignal);
    begin
      if aValue <> FDelayTime
      then begin
        FDelayTime := aValue;
        Recreate;
      end;
    end;

    procedure   TModSimpleGranulator.SetGrainCount( aValue: Integer);
    begin
      if aValue <> FGrainCount
      then begin
        FGrainCount := aValue;
        Recreate;
      end;
    end;

//  public

    constructor TModSimpleGranulator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := True;
    end;

    destructor  TModSimpleGranulator.Destroy; // override;
    var
      OldGranulator : TSimpleGranulator;
    begin
      FLocked := True;
      OldGranulator := TSimpleGranulator( AtomicExchange( Pointer( FGranulator), nil));
      FreeAndNil( OldGranulator);
      inherited;
    end;

    procedure   TModSimpleGranulator.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'graincount')
      then GrainCount := Round( aValue)
      else if SameText( aName, 'delaytime')
      then DelayTime := aValue
      else inherited;
    end;

    function    TModSimpleGranulator.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'graincount')
      then Result := GrainCount
      else if SameText( aName, 'delaytime')
      then Result := DelayTime
      else Result := inherited;
    end;

    procedure   TModSimpleGranulator.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute               ,
        i_grainduration      ,
        i_graindurationmodamt,
        i_record
      ];
      FInputs[ i_delay] := 1.0;
      FInputs[ i_rec  ] := 1.0;
    end;

    procedure   TModSimpleGranulator.DoSlowTick; // override;
    begin
      if Assigned( FGranulator) and not FLocked
      then begin
        FGranulator.Mute            := FInputs[ i_mute];
        FGranulator.CurrentEnvelope := Round( FInputs[ i_envtype]);
        FGranulator.DelayModulation := FInputs[ i_delay];
        FGranulator.Recording       := SignalToMute( FInputs[ i_record]) * Clip( FInputs[ i_rec], 0.0, 1.0);
        FGranulator.GrainDuration   := LookupEnvTime( erFast, Finputs[ i_grainduration] + FInputs[ i_graindurationmod] * FInputs[ i_graindurationmodamt]);
        FGranulator.GrainOffset     := FInputs[ i_grainoffset];
      end;
    end;

    procedure   TModSimpleGranulator.DoTick; // override;
    var
      Trig : Boolean;
    begin
      if Assigned( FGranulator) and not FLocked
      then begin
        Trig := SignalToLogic( FInputs[ i_trig]);

        if not FOldTrig and Trig
        then FGranulator.StartNewGrain;

        FGranulator.Tick( FInputs[ i_in], FOutputs[ o_out]);
        FOldTrig := Trig;
      end;
    end;

    procedure   TModSimpleGranulator.SampleRateChanged; // override;
    begin
      Recreate;
    end;

    function    TModSimpleGranulator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in                  : Result := 'in';
        i_mute                : Result := 'mute';
        i_envtype             : Result := 'envtype';
        i_grainduration       : Result := 'grainduration';
        i_graindurationmod    : Result := 'graindurationmod';
        i_graindurationmodamt : Result := 'graindurationmodamt';
        i_record              : Result := 'record';
        i_rec                 : Result := 'rec';
        i_delay               : Result := 'delay';
        i_grainoffset         : Result := 'grainoffset';
        i_trig                : Result := 'trig';
        else                    Result := NO_INPUT;
      end;
    end;

    function    TModSimpleGranulator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;

    function    TModSimpleGranulator.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_delaytime  : Result := 'delaytime';
        s_graincount : Result := 'grainccount';
        else           Result := NO_INTERNAL;
      end;
    end;



{ ========
  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
}

    constructor TModSampleAndHold.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSampleAndHold.DoSlowTick; // override;
    var
      Trig : Boolean;
    begin
      Trig := SignalToLogic( FInputs[ i_trig]);

      if not FLastTrig and Trig
      then FOutputs[ o_out] := FInputs[ i_in];

      FLastTrig := Trig;
    end;

    procedure   TModSampleAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModSampleAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig : Result := 'trig';
        i_in   : Result := 'in';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModSampleAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMultiSAndH = class( TMod)
  private
    FLastTrig: Boolean;
  public
}

    constructor TModMultiSAndH.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMultiSAndH.DoSlowTick; // override;
    var
      Trig : Boolean;
      i    : Integer;
    begin
      Trig := SignalToLogic( FInputs[ i_trig]);

      if not FLastTrig and Trig
      then begin
        for i := o_out1 to o_out12
        do FOutputs[ i] := FInputs[ i + i_in1 - o_out1];
      end;

      FLastTrig := Trig;
    end;

    procedure   TModMultiSAndH.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModMultiSAndH.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig : Result := 'trig';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        i_in5  : Result := 'in5';
        i_in6  : Result := 'in6';
        i_in7  : Result := 'in7';
        i_in8  : Result := 'in8';
        i_in9  : Result := 'in9';
        i_in10 : Result := 'in10';
        i_in11 : Result := 'in11';
        i_in12 : Result := 'in12';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModMultiSAndH.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1  : Result := 'out1';
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        o_out12 : Result := 'out12';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMux = class( TMod)
  private
    FState : Integer;
  public
}

    constructor TModMux.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMux.DoSlowTick; // override;
    var
      XFade : TSignal;
      State : TSignal;
      Delta : TSignal;
      Next  : Integer;
      Curr  : TSignal;
    begin
      XFade  := FInputs[ i_xfade];
      State  := MathFloatMod( 12 * Clip( FInputs[ i_control], 0, 1), 12);
      FState := Trunc( State);
      Next   := ( FState + 1) mod 12;
      Delta  := XFade * ( State - FState);
      Curr   := FInputs[ FState];
      FOutputs[ o_out] := Curr + Delta * ( FInputs[ Next] - Curr);
    end;

    procedure   TModMux.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to 11
      do aCallback( Self, Format( '%s/%s/%d/%f', [ aPrefix, Name, i, LogicToSignal( FState = i)], AppLocale));
    end;

    function    TModMux.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in0     : Result := 'in0';
        i_in1     : Result := 'in1';
        i_in2     : Result := 'in2';
        i_in3     : Result := 'in3';
        i_in4     : Result := 'in4';
        i_in5     : Result := 'in5';
        i_in6     : Result := 'in6';
        i_in7     : Result := 'in7';
        i_in8     : Result := 'in8';
        i_in9     : Result := 'in9';
        i_in10    : Result := 'in10';
        i_in11    : Result := 'in11';
        i_control : Result := 'control';
        i_xfade   : Result := 'xfade';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModMux.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDemux = class( TMod)
  private
    FState : Integer;
  public
}

    constructor TModDemux.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModDemux.DoSlowTick; // override;
    var
      XFade : TSignal;
      State : TSignal;
      Delta : TSignal;
      Next  : Integer;
      Input : TSignal;
    begin
      XFade  := FInputs[ i_xfade];
      State  := MathFloatMod( 12 * Clip( FInputs[ i_control], 0, 1), 12);
      FState := Trunc( State);
      Next   := ( FState + 1) mod 12;
      Delta  := XFade * ( State - FState);
      Input  := FInputs[ i_in];

      FOutputs[ FState] :=       Delta  * Input;
      FOutputs[ Next  ] := ( 1 - Delta) * Input;
    end;

    procedure   TModDemux.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to 11
      do aCallback( Self, Format( '%s/%s/%d/%f', [ aPrefix, Name, i, LogicToSignal( FState = i)], AppLocale));
    end;

    function    TModDemux.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in      : Result := 'in';
        i_control : Result := 'control';
        i_xfade   : Result := 'xfade';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModDemux.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out0  : Result := 'out0';
        o_out1  : Result := 'out1';
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDigiMux = class( TMod)
  private
    FState : Integer;
  public
}

    constructor TModDigiMux.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModDigiMux.DoSlowTick; // override;
    const
      State : array[ Boolean] of array[ Boolean] of Array[ Boolean] of Integer = ((( 0, 1), ( 2, 3)), (( 4, 5), ( 6, 7)));
    var
      i : Integer;
    begin
      FState := State[ SignalToLogic( FInputs[ i_sel2]), SignalToLogic( FInputs[ i_sel1]), SignalToLogic( FInputs[ i_sel0])];
      for i := i_in0 to i_in7
      do begin
        if i - i_in0 = FState
        then FOutputs[ o_out] := FInputs[ i]
      end;
    end;

    procedure   TModDigiMux.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to 7
      do aCallback( Self, Format( '%s/%s/%d/%f', [ aPrefix, Name, i, LogicToSignal( FState = i)], AppLocale));
    end;

    function    TModDigiMux.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in0  : Result := 'in0';
        i_in1  : Result := 'in1';
        i_in2  : Result := 'in2';
        i_in3  : Result := 'in3';
        i_in4  : Result := 'in4';
        i_in5  : Result := 'in5';
        i_in6  : Result := 'in6';
        i_in7  : Result := 'in7';
        i_sel0 : Result := 'sel0';
        i_sel1 : Result := 'sel1';
        i_sel2 : Result := 'sel2';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModDigiMux.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModDigiDemux = class( TMod)
  private
    FState : Integer;
  public
}

    constructor TModDigiDemux.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModDigiDemux.DoSlowTick; // override;
    const
      State : array[ Boolean] of array[ Boolean] of Array[ Boolean] of Integer = ((( 0, 1), ( 2, 3)), (( 4, 5), ( 6, 7)));
    var
      i : Integer;
    begin
      FState := State[ SignalToLogic( FInputs[ i_sel2]), SignalToLogic( FInputs[ i_sel1]), SignalToLogic( FInputs[ i_sel0])];
      for i := o_out0 to o_out7
      do begin
        if i - o_out0 = FState
        then FOutputs[ i] := FInputs[ i_in]
        else begin
          if not SignalToLogic( FInputs[ i_hold])
          then FOutputs[ i] := 0;
        end;
      end;
    end;

    procedure   TModDigiDemux.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to 7
      do aCallback( Self, Format( '%s/%s/%d/%f', [ aPrefix, Name, i, LogicToSignal( FState = i)], AppLocale));
    end;

    function    TModDigiDemux.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in   : Result := 'in';
        i_sel0 : Result := 'sel0';
        i_sel1 : Result := 'sel1';
        i_sel2 : Result := 'sel2';
        i_hold : Result := 'hold';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModDigiDemux.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out0 : Result := 'out0';
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        o_out5 : Result := 'out5';
        o_out6 : Result := 'out6';
        o_out7 : Result := 'out7';
        else     Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModTrackAndHold = class( TMod)
  public
}

    constructor TModTrackAndHold.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModTrackAndHold.DoSlowTick; // override;
    begin
      if SignalToLogic( FInputs[ i_trig])
      then FOutputs[ o_out] := FInputs[ i_in];
    end;

    procedure   TModTrackAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ i_trig]], AppLocale));
    end;

    function    TModTrackAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig : Result := 'trig';
        i_in   : Result := 'in';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModTrackAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSwitch2to1 = class( TMod)
  public
}

    constructor TModSwitch2to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSwitch2to1.DoSlowTick; // override;
    begin
      FOn := SignalToLogic( FInputs[ i_control]);
      if FOn
      then begin
        FOutputs[ o_out ] := FInputs[ i_in2];
        FOutputs[ o_out2] := FInputs[ i_in1];
      end
      else begin
        FOutputs[ o_out ] := FInputs[ i_in1];
        FOutputs[ o_out2] := FInputs[ i_in2];
      end;
    end;

    procedure   TModSwitch2to1.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/control/%f', [ aPrefix, Name, LogicToSignal( FOn)], AppLocale));
    end;

    function    TModSwitch2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_control : Result := 'control';
        i_in1     : Result := 'in1';
        i_in2     : Result := 'in2';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModSwitch2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out  : Result := 'out';
        o_out2 : Result := 'out2';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModXFade = class( TMod)
  public
}

    constructor TModXFade.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModXFade.SetDefaults; // override;
    begin
      FInputs[ i_control ] := 0; // Equal amounts of in1 and in2
      FInputs[ i_position] := 0;
      FDezipperMap := [
        i_position,
        i_posmod
      ];
    end;

    procedure   TModXFade.DoSlowTick; // override;
    var
      Control : TSignal;
      InType  : Integer;
    begin
      InType  := Round( FInputs[ i_intype]);
      Control := FInputs[ i_position] + SignalForInputType( FInputs[ i_control], InType) * FInputs[ i_posmod];
      FOutputs[ o_out] := ( FInputs[ i_in1] + FInputs[ i_in2] + Control * ( FInputs[ i_in2] - FInputs[ i_in1])) * 0.5;
    end;

    function    TModXFade.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_control  : Result := 'control';
        i_in1      : Result := 'in1';
        i_in2      : Result := 'in2';
        i_position : Result := 'position';
        i_posmod   : Result := 'posmod';
        i_intype   : Result := 'intype';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModXFade.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPan = class( TMod)
  public
}

    constructor TModPan.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPan.SetDefaults; // override;
    begin
      FInputs[ i_control ] := 0; // Mid position
      FDezipperMap := [
        i_position,
        i_posmod
      ];
    end;

    procedure   TModPan.DoSlowTick; // override;
    var
      Control : TSignal;
      Inp     : TSignal;
      InType  : Integer;
    begin
      InType            := Round( FInputs[ i_intype]);
      Inp               := 0.5 * FInputs[ i_in];
      Control           := Normalize( Inp * ( FInputs[ i_position] + SignalForInputType( FInputs[ i_control], InType) * FInputs[ i_posmod]));
      FOutputs[ o_out1] := Inp - Control;
      FOutputs[ o_out2] := Inp + Control;
    end;

    function    TModPan.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_control  : Result := 'control';
        i_in       : Result := 'in';
        i_position : Result := 'position';
        i_posmod   : Result := 'posmod';
        i_intype   : Result := 'intype';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModPan.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixM4to1 = class( TMod)
  public
}

    constructor TModMixM4to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMixM4to1.SetDefaults; // override;
    begin
      FInputs[ i_in ]     := 0; // Unused inputs should be zero
      FInputs[ i_in1]     := 0;
      FInputs[ i_in2]     := 0;
      FInputs[ i_in3]     := 0;
      FInputs[ i_in4]     := 0;
      FInputs[ i_levmod1] := 1;
      FInputs[ i_levmod2] := 1;
      FInputs[ i_levmod3] := 1;
      FInputs[ i_levmod4] := 1;
      FDezipperMap := [
        i_mute1,
        i_mute2,
        i_mute3,
        i_mute4,
        i_level1,
        i_level2,
        i_level3,
        i_level4
      ];
    end;

    procedure   TModMixM4to1.DoSlowTick; // override;
    begin
      FOutputs[ o_out] :=
        Normalize(
          FInputs[ i_in ] +
          FInputs[ i_in1] * Normalize( FInputs[ i_level1] * FInputs[ i_levmod1] * SignalToMute( FInputs[ i_mute1])) +
          FInputs[ i_in2] * Normalize( FInputs[ i_level2] * FInputs[ i_levmod2] * SignalToMute( FInputs[ i_mute2])) +
          FInputs[ i_in3] * Normalize( FInputs[ i_level3] * FInputs[ i_levmod3] * SignalToMute( FInputs[ i_mute3])) +
          FInputs[ i_in4] * Normalize( FInputs[ i_level4] * FInputs[ i_levmod4] * SignalToMute( FInputs[ i_mute4]))
        );
    end;

    function    TModMixM4to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in      : Result := 'in';
        i_in1     : Result := 'in1';
        i_in2     : Result := 'in2';
        i_in3     : Result := 'in3';
        i_in4     : Result := 'in4';
        i_mute1   : Result := 'mute1';
        i_mute2   : Result := 'mute2';
        i_mute3   : Result := 'mute3';
        i_mute4   : Result := 'mute4';
        i_level1  : Result := 'level1';
        i_level2  : Result := 'level2';
        i_level3  : Result := 'level3';
        i_level4  : Result := 'level4';
        i_levmod1 : Result := 'levmod1';
        i_levmod2 : Result := 'levmod2';
        i_levmod3 : Result := 'levmod3';
        i_levmod4 : Result := 'levmod4';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModMixM4to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixM16to1 = class( TMod)
  public
}

    constructor TModMixM16to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMixM16to1.SetDefaults; // override;
    begin
      FInputs[ i_in  ] := 0; // Unused inputs should be zero
      FInputs[ i_in1 ] := 0;
      FInputs[ i_in2 ] := 0;
      FInputs[ i_in3 ] := 0;
      FInputs[ i_in4 ] := 0;
      FInputs[ i_in5 ] := 0;
      FInputs[ i_in6 ] := 0;
      FInputs[ i_in7 ] := 0;
      FInputs[ i_in8 ] := 0;
      FInputs[ i_in9 ] := 0;
      FInputs[ i_in10] := 0;
      FInputs[ i_in11] := 0;
      FInputs[ i_in12] := 0;
      FInputs[ i_in13] := 0;
      FInputs[ i_in14] := 0;
      FInputs[ i_in15] := 0;
      FInputs[ i_in16] := 0;
      FDezipperMap := [
        i_mute,
        i_lev1,
        i_lev2,
        i_lev3,
        i_lev4,
        i_lev5,
        i_lev6,
        i_lev7,
        i_lev8,
        i_lev9,
        i_lev10,
        i_lev11,
        i_lev12,
        i_lev13,
        i_lev14,
        i_lev15,
        i_lev16
      ];
    end;

    procedure   TModMixM16to1.DoSlowTick; // override;
    begin
      FOutputs[ o_out] :=
        Normalize(
          SignalToMute( FInputs[ i_mute]) *
          (
            FInputs[ i_in ]                      +
            FInputs[ i_in1 ] * FInputs[ i_lev1 ] +
            FInputs[ i_in2 ] * FInputs[ i_lev2 ] +
            FInputs[ i_in3 ] * FInputs[ i_lev3 ] +
            FInputs[ i_in4 ] * FInputs[ i_lev4 ] +
            FInputs[ i_in5 ] * FInputs[ i_lev5 ] +
            FInputs[ i_in6 ] * FInputs[ i_lev6 ] +
            FInputs[ i_in7 ] * FInputs[ i_lev7 ] +
            FInputs[ i_in8 ] * FInputs[ i_lev8 ] +
            FInputs[ i_in9 ] * FInputs[ i_lev9 ] +
            FInputs[ i_in10] * FInputs[ i_lev10] +
            FInputs[ i_in11] * FInputs[ i_lev11] +
            FInputs[ i_in12] * FInputs[ i_lev12] +
            FInputs[ i_in13] * FInputs[ i_lev13] +
            FInputs[ i_in14] * FInputs[ i_lev14] +
            FInputs[ i_in15] * FInputs[ i_lev15] +
            FInputs[ i_in16] * FInputs[ i_lev16]
          )
        );
    end;

    function    TModMixM16to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in    : Result := 'in';
        i_in1   : Result := 'in1';
        i_in2   : Result := 'in2';
        i_in3   : Result := 'in3';
        i_in4   : Result := 'in4';
        i_in5   : Result := 'in5';
        i_in6   : Result := 'in6';
        i_in7   : Result := 'in7';
        i_in8   : Result := 'in8';
        i_in9   : Result := 'in9';
        i_in10  : Result := 'in10';
        i_in11  : Result := 'in11';
        i_in12  : Result := 'in12';
        i_in13  : Result := 'in13';
        i_in14  : Result := 'in14';
        i_in15  : Result := 'in15';
        i_in16  : Result := 'in16';
        i_lev1  : Result := 'lev1';
        i_lev2  : Result := 'lev2';
        i_lev3  : Result := 'lev3';
        i_lev4  : Result := 'lev4';
        i_lev5  : Result := 'lev5';
        i_lev6  : Result := 'lev6';
        i_lev7  : Result := 'lev7';
        i_lev8  : Result := 'lev8';
        i_lev9  : Result := 'lev9';
        i_lev10 : Result := 'lev10';
        i_lev11 : Result := 'lev11';
        i_lev12 : Result := 'lev12';
        i_lev13 : Result := 'lev13';
        i_lev14 : Result := 'lev14';
        i_lev15 : Result := 'lev15';
        i_lev16 : Result := 'lev16';
        i_mute  : Result := 'mute';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModMixM16to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixS2to1 = class( TMod)
  public
}

    constructor TModMixS2to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMixS2to1.SetDefaults; // override;
    begin
      FInputs[ i_inl    ] := 0; // Unused inputs should be zero
      FInputs[ i_inr    ] := 0;
      FInputs[ i_inl1   ] := 0;
      FInputs[ i_inr1   ] := 0;
      FInputs[ i_inl2   ] := 0;
      FInputs[ i_inr2   ] := 0;
      FInputs[ i_levmod1] := 1;
      FInputs[ i_levmod2] := 1;
      FDezipperMap := [
        i_mute1,
        i_mute2,
        i_level1,
        i_level2
      ];
    end;

    procedure   TModMixS2to1.DoSlowTick; // override;
    var
      lev1 : TSignal;
      lev2 : TSignal;
    begin
      lev1 := Normalize( FInputs[ i_levmod1] * FInputs[ i_level1] * SignalToMute( FInputs[ i_mute1]));
      lev2 := Normalize( FInputs[ i_levmod2] * FInputs[ i_level2] * SignalToMute( FInputs[ i_mute2]));

      FOutputs[ o_outl] :=
        Normalize(
          FInputs[ i_inl]         +
          FInputs[ i_inl1] * lev1 +
          FInputs[ i_inl2] * lev2
        );
      FOutputs[ o_outr] :=
        Normalize(
          FInputs[ i_inr]         +
          FInputs[ i_inr1] * lev1 +
          FInputs[ i_inr2] * lev2
        );
    end;

    function    TModMixS2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inl     : Result := 'inl';
        i_inr     : Result := 'inr';
        i_inl1    : Result := 'inl1';
        i_inr1    : Result := 'inr1';
        i_inl2    : Result := 'inl2';
        i_inr2    : Result := 'inr2';
        i_mute1   : Result := 'mute1';
        i_mute2   : Result := 'mute2';
        i_level1  : Result := 'level1';
        i_level2  : Result := 'level2';
        i_levmod1 : Result := 'levmod1';
        i_levmod2 : Result := 'levmod2';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModMixS2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outr : Result := 'outr';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixT1to1 = class( TMod)
  public
}

    constructor TModMixT1to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMixT1to1.SetDefaults; // override;
    begin
      FInputs[ i_inl    ] := 0; // Unused inputs should be zero
      FInputs[ i_inm    ] := 0;
      FInputs[ i_inr    ] := 0;
      FInputs[ i_inl1   ] := 0;
      FInputs[ i_inm1   ] := 0;
      FInputs[ i_inr1   ] := 0;
      FInputs[ i_levmod1] := 1;
      FDezipperMap := [
        i_mute1,
        i_level1
      ];
    end;

    procedure   TModMixT1to1.DoSlowTick; // override;
    var
      lev1 : TSignal;
    begin
      lev1 := Normalize( FInputs[ i_levmod1] * FInputs[ i_level1] * SignalToMute( FInputs[ i_mute1]));

      FOutputs[ o_outl] := Normalize( FInputs[ i_inl] + FInputs[ i_inl1] * lev1);
      FOutputs[ o_outm] := Normalize( FInputs[ i_inm] + FInputs[ i_inm1] * lev1);
      FOutputs[ o_outr] := Normalize( FInputs[ i_inr] + FInputs[ i_inr1] * lev1);
    end;

    function    TModMixT1to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inl     : Result := 'inl';
        i_inm     : Result := 'inm';
        i_inr     : Result := 'inr';
        i_inl1    : Result := 'inl1';
        i_inm1    : Result := 'inm1';
        i_inr1    : Result := 'inr1';
        i_mute1   : Result := 'mute1';
        i_level1  : Result := 'level1';
        i_levmod1 : Result := 'levmod1';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModMixT1to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outm : Result := 'outm';
        o_outr : Result := 'outr';
        else     Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModMixQ1to1 = class( TMod)
  public
}

    constructor TModMixQ1to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMixQ1to1.SetDefaults; // override;
    begin
      FInputs[ i_inll   ] := 0; // Unused inputs should be zero
      FInputs[ i_inlr   ] := 0;
      FInputs[ i_inrl   ] := 0;
      FInputs[ i_inrr   ] := 0;
      FInputs[ i_inll1  ] := 0;
      FInputs[ i_inlr1  ] := 0;
      FInputs[ i_inrl1  ] := 0;
      FInputs[ i_inrr1  ] := 0;
      FInputs[ i_levmod1] := 1;
      FDezipperMap := [
        i_mute1,
        i_level1
      ];
    end;

    procedure   TModMixQ1to1.DoSlowTick; // override;
    var
      lev1 : TSignal;
    begin
      lev1 := FInputs[ i_levmod1] * FInputs[ i_level1] * SignalToMute( FInputs[ i_mute1]);

      FOutputs[ o_outll] := Normalize( FInputs[ i_inll] + FInputs[ i_inll1] * lev1);
      FOutputs[ o_outlr] := Normalize( FInputs[ i_inlr] + FInputs[ i_inlr1] * lev1);
      FOutputs[ o_outrl] := Normalize( FInputs[ i_inrl] + FInputs[ i_inrl1] * lev1);
      FOutputs[ o_outrr] := Normalize( FInputs[ i_inrr] + FInputs[ i_inrr1] * lev1);
    end;

    function    TModMixQ1to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inll    : Result := 'inll';
        i_inlr    : Result := 'inlr';
        i_inrl    : Result := 'inrl';
        i_inrr    : Result := 'inrr';
        i_inll1   : Result := 'inll1';
        i_inlr1   : Result := 'inlr1';
        i_inrl1   : Result := 'inrl1';
        i_inrr1   : Result := 'inrr1';
        i_mute1   : Result := 'mute1';
        i_level1  : Result := 'level1';
        i_levmod1 : Result := 'levmod1';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModMixQ1to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outll : Result := 'outll';
        o_outlr : Result := 'outlr';
        o_outrl : Result := 'outrl';
        o_outrr : Result := 'outrr';
        else      Result := NO_OUTPUT;
      end;
    end;



{ ========
 TModMixM6to6 = class( TMod)
  public
}

    constructor TModMixM6to6.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixM6to6.SetDefaults; // override;
    begin
      FInputs[ i_inchcol1] := 0.0; // Zero chain col inputs
      FInputs[ i_inchcol2] := 0.0;
      FInputs[ i_inchcol3] := 0.0;
      FInputs[ i_inchcol4] := 0.0;
      FInputs[ i_inchcol5] := 0.0;
      FInputs[ i_inchcol6] := 0.0;
      FDezipperMap := [
        i_mute1     .. i_mute6,
        i_levelr1c1 .. i_levelr6c6
      ];
    end;

    procedure   TModMixM6to6.DoTick; // override;
    var
      In1 : TSignal;
      In2 : TSignal;
      In3 : TSignal;
      In4 : TSignal;
      In5 : TSignal;
      In6 : TSignal;
    begin
      In1 := FInputs[ i_in1] * SignalToMute( FInputs[ i_mute1]);
      In2 := FInputs[ i_in2] * SignalToMute( FInputs[ i_mute2]);
      In3 := FInputs[ i_in3] * SignalToMute( FInputs[ i_mute3]);
      In4 := FInputs[ i_in4] * SignalToMute( FInputs[ i_mute4]);
      In5 := FInputs[ i_in5] * SignalToMute( FInputs[ i_mute5]);
      In6 := FInputs[ i_in6] * SignalToMute( FInputs[ i_mute6]);

      FOutputs[ o_out1] := Normalize( In1 * ( FInputs[ i_levelr1c1] + FInputs[ i_inr1c1]) + In2 * ( FInputs[ i_levelr2c1] + FInputs[ i_inr2c1]) + In3 * ( FInputs[ i_levelr3c1] + FInputs[ i_inr3c1]) + In4 * ( FInputs[ i_levelr4c1] + FInputs[ i_inr4c1]) + In5 * ( FInputs[ i_levelr5c1] + FInputs[ i_inr5c1]) + In6 * ( FInputs[ i_levelr6c1] + FInputs[ i_inr6c1]) + FInputs[ i_inchcol1]);
      FOutputs[ o_out2] := Normalize( In1 * ( FInputs[ i_levelr1c2] + FInputs[ i_inr1c2]) + In2 * ( FInputs[ i_levelr2c2] + FInputs[ i_inr2c2]) + In3 * ( FInputs[ i_levelr3c2] + FInputs[ i_inr3c2]) + In4 * ( FInputs[ i_levelr4c2] + FInputs[ i_inr4c2]) + In5 * ( FInputs[ i_levelr5c2] + FInputs[ i_inr5c2]) + In6 * ( FInputs[ i_levelr6c2] + FInputs[ i_inr6c2]) + FInputs[ i_inchcol2]);
      FOutputs[ o_out3] := Normalize( In1 * ( FInputs[ i_levelr1c3] + FInputs[ i_inr1c3]) + In2 * ( FInputs[ i_levelr2c3] + FInputs[ i_inr2c3]) + In3 * ( FInputs[ i_levelr3c3] + FInputs[ i_inr3c3]) + In4 * ( FInputs[ i_levelr4c3] + FInputs[ i_inr4c3]) + In5 * ( FInputs[ i_levelr5c3] + FInputs[ i_inr5c3]) + In6 * ( FInputs[ i_levelr6c3] + FInputs[ i_inr6c3]) + FInputs[ i_inchcol3]);
      FOutputs[ o_out4] := Normalize( In1 * ( FInputs[ i_levelr1c4] + FInputs[ i_inr1c4]) + In2 * ( FInputs[ i_levelr2c4] + FInputs[ i_inr2c4]) + In3 * ( FInputs[ i_levelr3c4] + FInputs[ i_inr3c4]) + In4 * ( FInputs[ i_levelr4c4] + FInputs[ i_inr4c4]) + In5 * ( FInputs[ i_levelr5c4] + FInputs[ i_inr5c4]) + In6 * ( FInputs[ i_levelr6c4] + FInputs[ i_inr6c4]) + FInputs[ i_inchcol4]);
      FOutputs[ o_out5] := Normalize( In1 * ( FInputs[ i_levelr1c5] + FInputs[ i_inr1c5]) + In2 * ( FInputs[ i_levelr2c5] + FInputs[ i_inr2c5]) + In3 * ( FInputs[ i_levelr3c5] + FInputs[ i_inr3c5]) + In4 * ( FInputs[ i_levelr4c5] + FInputs[ i_inr4c5]) + In5 * ( FInputs[ i_levelr5c5] + FInputs[ i_inr5c5]) + In6 * ( FInputs[ i_levelr6c5] + FInputs[ i_inr6c5]) + FInputs[ i_inchcol5]);
      FOutputs[ o_out6] := Normalize( In1 * ( FInputs[ i_levelr1c6] + FInputs[ i_inr1c6]) + In2 * ( FInputs[ i_levelr2c6] + FInputs[ i_inr2c6]) + In3 * ( FInputs[ i_levelr3c6] + FInputs[ i_inr3c6]) + In4 * ( FInputs[ i_levelr4c6] + FInputs[ i_inr4c6]) + In5 * ( FInputs[ i_levelr5c6] + FInputs[ i_inr5c6]) + In6 * ( FInputs[ i_levelr6c6] + FInputs[ i_inr6c6]) + FInputs[ i_inchcol6]);

      FOutputs[ o_outchrow1] := Normalize( In1);
      FOutputs[ o_outchrow2] := Normalize( In2);
      FOutputs[ o_outchrow3] := Normalize( In3);
      FOutputs[ o_outchrow4] := Normalize( In4);
      FOutputs[ o_outchrow5] := Normalize( In5);
      FOutputs[ o_outchrow6] := Normalize( In6);
    end;

    function    TModMixM6to6.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1       : Result := 'in1';
        i_in2       : Result := 'in2';
        i_in3       : Result := 'in3';
        i_in4       : Result := 'in4';
        i_in5       : Result := 'in5';
        i_in6       : Result := 'in6';
        i_mute1     : Result := 'mute1';
        i_mute2     : Result := 'mute2';
        i_mute3     : Result := 'mute3';
        i_mute4     : Result := 'mute4';
        i_mute5     : Result := 'mute5';
        i_mute6     : Result := 'mute6';
        i_inchcol1  : Result := 'inchcol1';
        i_inchcol2  : Result := 'inchcol2';
        i_inchcol3  : Result := 'inchcol3';
        i_inchcol4  : Result := 'inchcol4';
        i_inchcol5  : Result := 'inchcol5';
        i_inchcol6  : Result := 'inchcol6';
        i_levelr1c1 : Result := 'levelr1c1';
        i_levelr1c2 : Result := 'levelr1c2';
        i_levelr1c3 : Result := 'levelr1c3';
        i_levelr1c4 : Result := 'levelr1c4';
        i_levelr1c5 : Result := 'levelr1c5';
        i_levelr1c6 : Result := 'levelr1c6';
        i_levelr2c1 : Result := 'levelr2c1';
        i_levelr2c2 : Result := 'levelr2c2';
        i_levelr2c3 : Result := 'levelr2c3';
        i_levelr2c4 : Result := 'levelr2c4';
        i_levelr2c5 : Result := 'levelr2c5';
        i_levelr2c6 : Result := 'levelr2c6';
        i_levelr3c1 : Result := 'levelr3c1';
        i_levelr3c2 : Result := 'levelr3c2';
        i_levelr3c3 : Result := 'levelr3c3';
        i_levelr3c4 : Result := 'levelr3c4';
        i_levelr3c5 : Result := 'levelr3c5';
        i_levelr3c6 : Result := 'levelr3c6';
        i_levelr4c1 : Result := 'levelr4c1';
        i_levelr4c2 : Result := 'levelr4c2';
        i_levelr4c3 : Result := 'levelr4c3';
        i_levelr4c4 : Result := 'levelr4c4';
        i_levelr4c5 : Result := 'levelr4c5';
        i_levelr4c6 : Result := 'levelr4c6';
        i_levelr5c1 : Result := 'levelr5c1';
        i_levelr5c2 : Result := 'levelr5c2';
        i_levelr5c3 : Result := 'levelr5c3';
        i_levelr5c4 : Result := 'levelr5c4';
        i_levelr5c5 : Result := 'levelr5c5';
        i_levelr5c6 : Result := 'levelr5c6';
        i_levelr6c1 : Result := 'levelr6c1';
        i_levelr6c2 : Result := 'levelr6c2';
        i_levelr6c3 : Result := 'levelr6c3';
        i_levelr6c4 : Result := 'levelr6c4';
        i_levelr6c5 : Result := 'levelr6c5';
        i_levelr6c6 : Result := 'levelr6c6';
        i_inr1c1    : Result := 'inr1c1';
        i_inr1c2    : Result := 'inr1c2';
        i_inr1c3    : Result := 'inr1c3';
        i_inr1c4    : Result := 'inr1c4';
        i_inr1c5    : Result := 'inr1c5';
        i_inr1c6    : Result := 'inr1c6';
        i_inr2c1    : Result := 'inr2c1';
        i_inr2c2    : Result := 'inr2c2';
        i_inr2c3    : Result := 'inr2c3';
        i_inr2c4    : Result := 'inr2c4';
        i_inr2c5    : Result := 'inr2c5';
        i_inr2c6    : Result := 'inr2c6';
        i_inr3c1    : Result := 'inr3c1';
        i_inr3c2    : Result := 'inr3c2';
        i_inr3c3    : Result := 'inr3c3';
        i_inr3c4    : Result := 'inr3c4';
        i_inr3c5    : Result := 'inr3c5';
        i_inr3c6    : Result := 'inr3c6';
        i_inr4c1    : Result := 'inr4c1';
        i_inr4c2    : Result := 'inr4c2';
        i_inr4c3    : Result := 'inr4c3';
        i_inr4c4    : Result := 'inr4c4';
        i_inr4c5    : Result := 'inr4c5';
        i_inr4c6    : Result := 'inr4c6';
        i_inr5c1    : Result := 'inr5c1';
        i_inr5c2    : Result := 'inr5c2';
        i_inr5c3    : Result := 'inr5c3';
        i_inr5c4    : Result := 'inr5c4';
        i_inr5c5    : Result := 'inr5c5';
        i_inr5c6    : Result := 'inr5c6';
        i_inr6c1    : Result := 'inr6c1';
        i_inr6c2    : Result := 'inr6c2';
        i_inr6c3    : Result := 'inr6c3';
        i_inr6c4    : Result := 'inr6c4';
        i_inr6c5    : Result := 'inr6c5';
        i_inr6c6    : Result := 'inr6c6';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMixM6to6.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1      : Result := 'out1';
        o_out2      : Result := 'out2';
        o_out3      : Result := 'out3';
        o_out4      : Result := 'out4';
        o_out5      : Result := 'out5';
        o_out6      : Result := 'out6';
        o_outchrow1 : Result := 'outchrow1';
        o_outchrow2 : Result := 'outchrow2';
        o_outchrow3 : Result := 'outchrow3';
        o_outchrow4 : Result := 'outchrow4';
        o_outchrow5 : Result := 'outchrow5';
        o_outchrow6 : Result := 'outchrow6';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixAudio = class( TMod)
  public
}

    constructor TModMixAudio.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixAudio.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_pan1 .. i_mute8
      ];
    end;

    procedure   TModMixAudio.DoTick; // override;
    var
      Inp      : TSignal;
      Control  : TSignal;
      Signal1L : TSignal;
      Signal1R : TSignal;
      Signal2L : TSignal;
      Signal2R : TSignal;
    begin
      if SignalToLogic( FInputs[ i_insolo])
      then begin
        Signal1L := FInputs[ i_inchain1l];
        Signal1R := FInputs[ i_inchain1r];
        Signal2L := FInputs[ i_inchain2l];
        Signal2R := FInputs[ i_inchain2r];
        FOutputs[ o_outsolo] := LogicToSignal( True);
      end
      else begin
        if SignalToLogic( FInputs[ i_solo1])
        then begin
          Inp      := FInputs[ i_in1] * SignalToMute( FInputs[ i_mute1]);
          Control  := FInputs[ i_pan1];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol11];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol11];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol21];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol21];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo2])
        then begin
          Inp      := FInputs[ i_in2] * SignalToMute( FInputs[ i_mute2]);
          Control  := FInputs[ i_pan2];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol12];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol12];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol22];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol22];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo3])
        then begin
          Inp      := FInputs[ i_in3] * SignalToMute( FInputs[ i_mute3]);
          Control  := FInputs[ i_pan3];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol13];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol13];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol23];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol23];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo4])
        then begin
          Inp      := FInputs[ i_in4] * SignalToMute( FInputs[ i_mute4]);
          Control  := FInputs[ i_pan4];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol14];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol14];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol24];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol24];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo5])
        then begin
          Inp      := FInputs[ i_in5] * SignalToMute( FInputs[ i_mute5]);
          Control  := FInputs[ i_pan5];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol15];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol15];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol25];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol25];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo6])
        then begin
          Inp      := FInputs[ i_in6] * SignalToMute( FInputs[ i_mute6]);
          Control  := FInputs[ i_pan6];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol16];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol16];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol26];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol26];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo7])
        then begin
          Inp      := FInputs[ i_in7] * SignalToMute( FInputs[ i_mute7]);
          Control  := FInputs[ i_pan7];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol17];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol17];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol27];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol27];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ i_solo8])
        then begin
          Inp      := FInputs[ i_in8] * SignalToMute( FInputs[ i_mute8]);
          Control  := FInputs[ i_pan8];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol18];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol18];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol28];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol28];
          FOutputs[ o_outsolo] := LogicToSignal( True);
        end
        else begin
          Signal1L := FInputs[ i_inchain1l];
          Signal1R := FInputs[ i_inchain1r];
          Signal2L := FInputs[ i_inchain2l];
          Signal2R := FInputs[ i_inchain2r];

          Inp      := FInputs[ i_in1] * SignalToMute( FInputs[ i_mute1]);
          Control  := FInputs[ i_pan1];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol11];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol11];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol21];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol21];

          Inp      := FInputs[ i_in2] * SignalToMute( FInputs[ i_mute2]);
          Control  := FInputs[ i_pan2];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol12];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol12];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol22];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol22];

          Inp      := FInputs[ i_in3] * SignalToMute( FInputs[ i_mute3]);
          Control  := FInputs[ i_pan3];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol13];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol13];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol23];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol23];

          Inp      := FInputs[ i_in4] * SignalToMute( FInputs[ i_mute4]);
          Control  := FInputs[ i_pan4];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol14];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol14];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol24];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol24];

          Inp      := FInputs[ i_in5] * SignalToMute( FInputs[ i_mute5]);
          Control  := FInputs[ i_pan5];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol15];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol15];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol25];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol25];

          Inp      := FInputs[ i_in6] * SignalToMute( FInputs[ i_mute6]);
          Control  := FInputs[ i_pan6];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol16];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol16];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol26];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol26];

          Inp      := FInputs[ i_in7] * SignalToMute( FInputs[ i_mute7]);
          Control  := FInputs[ i_pan7];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol17];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol17];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol27];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol27];

          Inp      := FInputs[ i_in8] * SignalToMute( FInputs[ i_mute8]);
          Control  := FInputs[ i_pan8];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol18];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol18];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ i_vol28];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ i_vol28];

          FOutputs[ o_outsolo] := LogicToSignal( False);
        end;
      end;
      FOutputs[ o_out1l] := Normalize( Signal1L);
      FOutputs[ o_out1r] := Normalize( Signal1R);
      FOutputs[ o_out2l] := Normalize( Signal2L);
      FOutputs[ o_out2r] := Normalize( Signal2R);
    end;

    procedure   TModMixAudio.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/solo/%f', [ aPrefix, Name, FOutputs[ o_outsolo]], AppLocale));
    end;

    function    TModMixAudio.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_insolo    : Result := 'insolo';
        i_inchain1l : Result := 'inchain1l';
        i_inchain1r : Result := 'inchain1r';
        i_inchain2l : Result := 'inchain2l';
        i_inchain2r : Result := 'inchain2r';
        i_in1       : Result := 'in1';
        i_in2       : Result := 'in2';
        i_in3       : Result := 'in3';
        i_in4       : Result := 'in4';
        i_in5       : Result := 'in5';
        i_in6       : Result := 'in6';
        i_in7       : Result := 'in7';
        i_in8       : Result := 'in8';
        i_pan1      : Result := 'pan1';
        i_pan2      : Result := 'pan2';
        i_pan3      : Result := 'pan3';
        i_pan4      : Result := 'pan4';
        i_pan5      : Result := 'pan5';
        i_pan6      : Result := 'pan6';
        i_pan7      : Result := 'pan7';
        i_pan8      : Result := 'pan8';
        i_vol11     : Result := 'vol11';
        i_vol12     : Result := 'vol12';
        i_vol13     : Result := 'vol13';
        i_vol14     : Result := 'vol14';
        i_vol15     : Result := 'vol15';
        i_vol16     : Result := 'vol16';
        i_vol17     : Result := 'vol17';
        i_vol18     : Result := 'vol18';
        i_vol21     : Result := 'vol21';
        i_vol22     : Result := 'vol22';
        i_vol23     : Result := 'vol23';
        i_vol24     : Result := 'vol24';
        i_vol25     : Result := 'vol25';
        i_vol26     : Result := 'vol26';
        i_vol27     : Result := 'vol27';
        i_vol28     : Result := 'vol28';
        i_mute1     : Result := 'mute1';
        i_mute2     : Result := 'mute2';
        i_mute3     : Result := 'mute3';
        i_mute4     : Result := 'mute4';
        i_mute5     : Result := 'mute5';
        i_mute6     : Result := 'mute6';
        i_mute7     : Result := 'mute7';
        i_mute8     : Result := 'mute8';
        i_solo1     : Result := 'solo1';
        i_solo2     : Result := 'solo2';
        i_solo3     : Result := 'solo3';
        i_solo4     : Result := 'solo4';
        i_solo5     : Result := 'solo5';
        i_solo6     : Result := 'solo6';
        i_solo7     : Result := 'solo7';
        i_solo8     : Result := 'solo8';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMixAudio.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outsolo : Result := 'outsolo';
        o_out1l   : Result := 'out1l';
        o_out1r   : Result := 'out1r';
        o_out2l   : Result := 'out2l';
        o_out2r   : Result := 'out2r';
        else        Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModQuantize = class( TMod)
  public
}

    constructor TModQuantize.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModQuantize.DoSlowTick; // override;
    var
      Steps: Integer;
    begin
      Steps := Max( 1, Round( FInputs[ i_qsteps]));
      FOutputs[ o_out] := Round( Steps * FInputs[ i_level] * FInputs[ i_in] * SignalToMute( FInputs[ i_mute] )) / Steps;
    end;

    function    TModQuantize.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_qsteps : Result := 'qsteps';
        i_in     : Result := 'in';
        i_mute   : Result := 'mute';
        i_level  : Result := 'level';
        else       Result := NO_INPUT;
      end;
    end;

    function    TModQuantize.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModHardClip = class( TMod)
  public
}

    constructor TModHardClip.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModHardClip.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute,
        i_lowlevel,
        i_highlevel,
        i_outamp
      ];
    end;

    procedure   TModHardClip.DoSlowTick; // override;
    begin
      FOutputs[ o_out] := FInputs[ i_outamp] * SignalToMute( FInputs[ i_mute]) * Clip( FInputs[ i_in], FInputs[ i_lowlevel], FInputs[ i_highlevel]);
    end;

    function    TModHardClip.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in        : Result := 'in';
        i_mute      : Result := 'mute';
        i_lowlevel  : Result := 'lowlevel';
        i_highlevel : Result := 'highlevel';
        i_outamp    : Result := 'outamp';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModHardClip.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTransInv = class( TMod)
  public
}

    constructor TModTransInv.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModTransInv.DoSlowTick; // override;
    // transpose :: T( amt, x) -> ( amt + x) mod modulo // amt and x being integer
    // invert    := I( piv, x) -> ( piv - x) mod modulo // piv and x being integer
    var
      Mode      : Integer;
      Modulo    : Integer;
      AmtPivot  : TSignal;
      InValue1  : TSignal;
      InValue2  : TSignal;
      InValue3  : TSignal;
      InValue4  : TSignal;
    begin
      Mode     := Round( FInputs[ i_mode  ]);
      Modulo   := Clip( Round( FInputs[ i_modulo]), 1, 128);
      AmtPivot := UnitsToNoteNumber( FInputs[ i_amtpivot]) - MiddleNote;
      InValue1 := UnitsToNoteNumber( FInputs[ i_in1]);
      InValue2 := UnitsToNoteNumber( FInputs[ i_in2]);
      InValue3 := UnitsToNoteNumber( FInputs[ i_in3]);
      InValue4 := UnitsToNoteNumber( FInputs[ i_in4]);

      case Mode of
        m_transpose : begin
          FOutputs[ o_out1] := NoteNumberToUnits( MathFloatMod( AmtPivot + InValue1, Modulo));
          FOutputs[ o_out2] := NoteNumberToUnits( MathFloatMod( AmtPivot + InValue2, Modulo));
          FOutputs[ o_out3] := NoteNumberToUnits( MathFloatMod( AmtPivot + InValue3, Modulo));
          FOutputs[ o_out4] := NoteNumberToUnits( MathFloatMod( AmtPivot + InValue4, Modulo));
        end;
        else begin
          FOutputs[ o_out1] := NoteNumberToUnits( MathFloatMod( AmtPivot - InValue1, Modulo));
          FOutputs[ o_out2] := NoteNumberToUnits( MathFloatMod( AmtPivot - InValue2, Modulo));
          FOutputs[ o_out3] := NoteNumberToUnits( MathFloatMod( AmtPivot - InValue3, Modulo));
          FOutputs[ o_out4] := NoteNumberToUnits( MathFloatMod( AmtPivot - InValue4, Modulo));
        end;
      end;
    end;

    function    TModTransInv.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1      : Result := 'in1';
        i_in2      : Result := 'in2';
        i_in3      : Result := 'in3';
        i_in4      : Result := 'in4';
        i_amtpivot : Result := 'amtpivot';
        i_modulo   : Result := 'modulo';
        i_mode     : Result := 'mode';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModTransInv.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModScala = class( TMod)
  private
    FMappingFileName : string;
    FScaleFileName   : string;
    FScale           : TScalaScale;
    FLastNote        : Integer;
    FlastOut         : TSignal;
  public
    property    MappingFileName : string read FMappingFileName write SetMappingFileName;
    property    ScaleFileName   : string read FScaleFileName   write SetScaleFileName;
  private
}

    procedure   TModScala.SetMappingFileName( const aValue: string);
    begin
      if aValue <> FMappingFileName
      then begin
        FMappingFileName      := aValue;
        FScale.KeyMapFileName := FMappingFileName;
      end;
    end;

    procedure   TModScala.SetScaleFileName  ( const aValue: string);
    begin
      if aValue <> FScaleFileName
      then begin
        FScaleFileName  := aValue;
        FScale.FileName := FScaleFileName;
      end;
    end;

    function    TModScala.Lookup( aValue: TSignal): TSignal;
    var
      aNote : Integer;
    begin
      aNote  := Trunc( UnitsToNoteNumber( aValue));
      if aNote = FLastNote
      then Result := FLastOut
      else begin
        Result := FScale.KeyToNote( aNote);

        FLastNote := aNote;
        FLastOut  := Result;
      end;
    end;

//  public

    constructor TModScala.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
      FScale  := TScalaScale.Create;
    end;

    destructor  TModScala.Destroy; // override;
    begin
      FreeAndNil( FScale);
      inherited;
    end;

    procedure   TModScala.SetDefaults; // override;
    begin
      SlowTick;
    end;

    procedure   TModScala.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'mapping')
      then MappingFileName := aValue
      else if SameText( aName, 'scale')
      then ScaleFileName := aValue
      else inherited;
    end;

    function    TModScala.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'mapping')
      then Result := FMappingFileName
      else if SameText( aName, 'scale')
      then Result := FScaleFileName
      else Result := inherited;
    end;

    procedure   TModScala.DoSlowTick; // override;
    begin
      FOutputs[ i_out] := Lookup( FInputs[ i_in]);
    end;

    function    TModScala.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in : Result := 'in';
        else   Result := NO_INPUT;
      end;
    end;

    function    TModScala.OutputName( anIndex: Integer): string; // override;
    begin
     case anIndex of
        i_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModTerraGen = class( TMod)
  public
}

    constructor TModTerragen.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast   := False;
      FIsSlow   := True;
    end;

    procedure   TModTerraGen.SetDefaults; // override;
    begin
      inherited;
      FDezipperMap := [
        i_mute
      ];
      FInputs[ i_x] := 1.0;        // Set unused input to 1
      FInputs[ i_y] := 1.0;        // Set unused input to 1
    end;

    procedure   TModTerraGen.DoSlowTick; // override;
    var
      x, y  : TSignal;
      algo  : Integer;
      prod  : TSignal;
    begin
      x    := Clip( FInputs[ i_x], -1.0, 1.0);
      y    := Clip( FInputs[ i_y], -1.0, 1.0);
      prod := x * y;
      algo := Round( FInputs[ i_algo]);
      case algo of
        m_off          : FOutputs[ o_out] := 0;
        m_inx          : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * x;
        m_iny          : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * y;
        m_xy           : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * prod;
        m_curtis_roads : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( ( x - y) * ( x - 1) * ( x + 1) * ( y - 1) * ( y + 1));
        m_bass         : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( Sqrt( Abs( x * x * x * x + y * y * y * y * y - x * x - y * y - x * prod * y)));
        m_lead1        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( ( x - x * x * x / 12 - y * y / 4 + 0.5));
        m_hollow_lead  : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( ( -5 * x / ( x * x + y * y + 1)));
        m_pulsar       : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( Sqr( Sin( x * x + y * y)));
        m_flat_growl   : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( Sqrt( Abs( ( 1 - x * x) * (( 1 - y) + y * cos( 1 / ( x * x + 0.01))))) - 1);
        m_atan         : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( ArcTan( 15 * prod) / HALF_PI);
        m_softclip     : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( prod / 1 + Abs( prod));
        m_third        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( 1.5 * prod + 0.5 * prod * prod * prod);
        m_cheb2        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize( 2 * prod * prod - 1);
        m_cheb3        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize(( 4 * prod * prod - 3) * prod);
        m_cheb4        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize(( 8 * prod * prod - 8) * prod * prod + 1);
        m_cheb5        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize((( 16 * prod * prod - 20) * prod * prod + 5) * prod);
        m_cheb6        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize((( 32 * prod * prod - 48) * prod * prod + 18) * prod * prod - 1);
        m_cheb7        : FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * Normalize(((( 64 * prod * prod - 112) * prod * prod + 56) * prod * prod - 7) * prod);
        else             FOutputs[ o_out] := 0;
      end;
    end;

    function    TModTerraGen.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_x    : Result := 'x';
        i_y    : Result := 'y';
        i_mute : Result := 'mute';
        i_algo : Result := 'algo';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModTerraGen.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModChebGen = class( TMod)
  strict private
  const
    i_in    =  0;
  const
  public
}

    constructor TModChebGen.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast   := False;
      FIsSlow   := True;
    end;

    procedure   TModChebGen.DoSlowTick; // override;
    var
      Tn_2 : TSignal;   // Tn-2
      Tn_1 : TSignal;   // Tn-1
      Tn   : TSignal;   // Tn-0
      x    : TSignal;
    begin
      x := Normalize( FInputs[ i_in]);

      case Round( FInputs[ i_wmode]) of
        m_tri : begin
          FOutputs[ o_out2 ] := abs( 2 * x) - 1;
          FOutputs[ o_out3 ] := abs( abs( 3 * x - 1) - 2) - 1;
          FOutputs[ o_out4 ] := abs( abs( 4 * x) - 2) - 1;
          FOutputs[ o_out5 ] := abs( abs( abs( abs( 5 * x - 1 ) - 2 ) - 2 ) - 2 ) - 1;
          FOutputs[ o_out6 ] := abs( abs( 3 * ( abs( 2 * x) - 1) - 1) - 2) - 1;
          FOutputs[ o_out7 ] := abs( abs( abs( abs( abs( abs( 7 * x - 1 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 1;
          FOutputs[ o_out8 ] := abs( abs( abs( 8 * x) - 4) - 2) - 1;
          FOutputs[ o_out9 ] := abs( abs( abs( abs( abs( abs( abs( abs( 9 * x - 1 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 1;
          FOutputs[ o_out10] := abs( abs( abs( abs( 5 * ( abs( 2 * x) - 1) - 1 ) - 2 ) - 2 ) - 2 ) - 1;
          FOutputs[ o_out11] := abs( abs( abs( abs( abs( abs( abs( abs( abs( abs( 11 * x - 1 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 1;
          FOutputs[ o_out12] := abs( abs( 4 * ( abs( abs( 3 * x - 1) - 2) - 1)) - 1) - 1;
          FOutputs[ o_out13] := abs( abs( abs( abs( abs( abs( abs( abs( abs( abs( abs( abs( 13 * x - 1 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 2 ) - 1;
        end;
        else begin
          Tn_2 := 1;
          Tn_1 := x;

          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out2 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out3 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out4 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out5 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out6 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out7 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out8 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out9 ] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out10] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out11] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2; Tn_2 := Tn_1; Tn_1 := Tn; FOutputs[ o_out12] := Tn;
          Tn := 2 * x * Tn_1 - Tn_2;                           FOutputs[ o_out13] := Tn;
        end;
      end;
    end;

    function    TModChebGen.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in    : Result := 'in';
        i_wmode : Result := 'wmode';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModChebGen.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        o_out12 : Result := 'out12';
        o_out13 : Result := 'out13';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPoetry = class( TMod)
  strict private
  const
    i_in         = 0;
    i_newinput   = 1;
    i_newoutput  = 2;
  const
    o_out        = 0;
  const
    s_history    = 0;
    s_seedlen    = 1;
  private
    FLocked       : Boolean;
    FGenerator    : TPoetryGenerator;
    FHistory      : Integer;
    FSeedLen      : Integer;
    FEpsilon      : TSignal;
    FOldNewInput  : Boolean;
    FOldNewOutput : Boolean;
  private
    property    History : Integer read FHistory write SetHistory;
    property    SeedLen : Integer read FSeedLen write SetSeedLen;
  private
}

    procedure   TModPoetry.Recreate;
    var
      anOld : TPoetryGenerator;
    begin
      FLocked := True;
      anOld :=
        AtomicExchange(
          Pointer( FGenerator),
          Pointer(
            TPoetryGenerator.Create(
              FHistory,
              FSeedLen,
              FEpsilon
            )
          )
        );
      FreeAndNil( anOld);
      FLocked := False;
    end;

    procedure   TModPoetry.SetHistory( aValue: Integer);
    begin
      if aValue <> FHistory
      then begin
        FHistory := aValue;
        Recreate;
      end;
    end;

    procedure   TModPoetry.SetSeedLen( aValue: Integer);
    begin
      if aValue <> FSeedLen
      then begin
        FSeedLen := aValue;
        Recreate;
      end;
    end;

//  public

    constructor TModPoetry.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsSlow   := True;
      FIsFast   := False;
      FHistory  := 256;
      FSeedLen  :=  32;
      FEpsilon  := 1e-3;
      Recreate;
    end;

    procedure   TModPoetry.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'history')
      then History := Round( aValue)
      else if SameText( aName, 'seedlen')
      then SeedLen := Round( aValue)
      else inherited;
    end;

    function    TModPoetry.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'history')
      then Result := History
      else if SameText( aName, 'seedlen')
      then Result := SeedLen
      else Result := inherited;
    end;

    procedure   TModPoetry.DoSlowTick; // override;
    var
      NewInput  : Boolean;
      NewOutput : Boolean;
    begin
      if Assigned( FGenerator) and not FLocked
      then begin
        NewInput  := SignalToLogic( FInputs[ i_newinput ]);
        NewOutput := SignalToLogic( FInputs[ i_newoutput]);
        if not FOldNewInput and NewInput
        then FGenerator.AddHistory( FInputs[ i_in]);
        if not FOldNewOutput and NewOutput
        then FOutputs[ o_out] := FGenerator.Tick;
        FOldNewInput  := NewInput;
        FOldNewOutput := NewOutput;
      end;
    end;

    procedure   TModPoetry.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/newinput/%f' , [ aPrefix, Name, LogicToSignal( FOldNewInput )], AppLocale));
      aCallback( Self, Format( '%s/%s/newoutput/%f', [ aPrefix, Name, LogicToSignal( FOldNewOutput)], AppLocale));
    end;

    function    TModPoetry.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in        : Result := 'in';
        i_newinput  : Result := 'newinput';
        i_newoutput : Result := 'newoutput';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModPoetry.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out  : Result := 'out';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPluck = class( TMod)
  private
    FMemory      : TSignalArrayWrapper;
    FDelayTime   : TSignal;
    FInPointer   : Integer;
    FOutValue    : TSignal;
    FOldFeedback : TSignal;
  public
 }

    constructor TModPluck.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FDelayTime   := 1 / LookupFrequency( 0);
      FMemory      := TSignalArrayWrapper.Create( Ceil( TimeToSampleCount( FDelayTime))); // lowest note about 8.176 Hz - for default tuning
      FOldFeedback := 0.0;
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModPluck.Destroy; // override;
    begin
      FreeAndNil( FMemory);
    end;

    procedure   TModPluck.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_feedback,
        i_brightness,
        i_fmlevel,
        i_frequency,
        i_cents,
        i_mute,
        i_noisemod,
        i_noisedepth
      ];
    end;

    procedure   TModPluck.DoTick; // override;
    var
      OutPointer : TSignal;
      SampCount  : Integer;
      Alfa       : TSignal;
      Delay      : TSignal;
      Feedback   : TSignal;
      Signal     : TSignal;
      Frequency  : TSignal;
      Noise      : TSignal;
    begin
      SampCount                   := FMemory.Count;
      FInPointer                  := FInPointer mod SampCount;

      Noise                       := Clip( FInputs[ i_noise] + FInputs[ i_noisemod] * FInputs[ i_noisedepth], 0, 1);
      Frequency                   := LookupFrequency( FInputs[ i_frequency] + FInputs[ i_freq] + FInputs[ i_fm] * FInputs[ i_fmlevel] + FInputs[ i_cents]);
      Delay                       := 4 / Frequency; // but why 4 / ...
      Alfa                        := Clip( FInputs[ i_brightness] + FInputs[ i_brighctrl] * FInputs[ i_bclevel], 0, 1);
      Feedback                    := FOldFeedback + Alfa * ( FOutValue * FInputs[ i_feedback] - FOldFeedback);
      if Random > Noise
      then Feedback := - Feedback;
      Signal                      := FInputs[ i_pluck] * FInputs[ i_plucklevel] - Feedback;
      FMemory.Signal[ FInPointer] := Clip( Signal, -16, 16);
      OutPointer                  := MathFloatMod( FInPointer - Delay * ( TimeToSampleCount( FDelayTime) - 1), SampCount);
      FOutValue                   := Normalize( FMemory.Interpolate( OutPointer));

      FInPointer                  := ( FInPointer + 1) mod SampCount;
      FOldFeedback                := Feedback;
      FOutputs[ o_out]            := Normalize( FOutValue * SignalToMute( FInputs[ i_mute]));
    end;

    function    TModPluck.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         i_pluck       : Result := 'pluck';
         i_plucklevel  : Result := 'plucklevel';
         i_feedback    : Result := 'feedback';
         i_brightness  : Result := 'brightness';
         i_brighctrl   : Result := 'brighctrl';
         i_bclevel     : Result := 'bclevel';
         i_fm          : Result := 'fm';
         i_fmlevel     : Result := 'fmlevel';
         i_frequency   : Result := 'frequency';
         i_cents       : Result := 'cents';
         i_mute        : Result := 'mute';
         i_noise       : Result := 'noise';
         i_noisemod    : Result := 'noisemod';
         i_noisedepth  : Result := 'noisedepth';
         i_freq        : Result := 'freq';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModPluck.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModDisplay = class( TMod)
  private
    FMax1 : TSignal;
    FMax2 : TSignal;
  public
}

    constructor TModDisplay.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModDisplay.DoTick; // override;
    begin
      if Round( FInputs[ i_mode]) = 0
      then begin
        FMax1 := Max( FMax1, Abs( FInputs[ i_in1]));
        FMax2 := Max( FMax2, Abs( FInputs[ i_in2]));
      end
      else begin
        FMax1 := Max( FMax1, FInputs[ i_in1]);
        FMax2 := Max( FMax2, FInputs[ i_in2]);
      end;
    end;

    procedure   TModDisplay.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    const
      LOWVAL  = -45.0;
      HIGHVAL =  18.0;
      LOWLED  =   0.0;
      HIGHLED =   1.0;
    var
      adB1  : TSignal;
      adB2  : TSignal;
      Speed : TSignal;
    begin
      Speed := 0.707 * ( 0.2 + 0.8 * FInputs[ i_speed]);
      adB1  := RangeMap( Clip( SignalTodB( FMax1 ), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      adB2  := RangeMap( Clip( SignalTodB( FMax2 ), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      aCallback( Self, Format( '%s/%s/level1/%f', [ aPrefix, Name, adB1], AppLocale));
      aCallback( Self, Format( '%s/%s/level2/%f', [ aPrefix, Name, adB2], AppLocale));
      FMax1 := FMax1 * Speed;
      FMax2 := FMax2 * Speed;
    end;

    function    TModDisplay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1   : Result := 'in1';
        i_in2   : Result := 'in2';
        i_mode  : Result := 'mode';
        i_speed : Result := 'speed';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModDisplay.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;


{ ========
  TMod_O_S_C = class( TMod)
  public
}

    constructor TMod_O_S_C.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
      FIsOSC  := True;
    end;


{ ========
  TModSync = class( TMod_O_S_C)
  private
    FOscBytes    : TBytes;
    FSync        : Boolean;
    FOldSync     : Boolean;
    FLedCounter  : Integer;
    FSyncCounter : Integer;
  public
}

    constructor TModSync.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      aMsg : TOSCMessage;
    begin
      inherited;
      aMsg := TOSCMessage.Create( '/sync');

      try
        FOscBytes := aMsg.ToOSCBytes;
      finally
        aMsg.DisposeOf;
      end;
    end;

    procedure   TModSync.AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket); // override;
    var
      aLocMsg : TOSCMessage;
    begin
      aLocMsg := aMsg.MatchAddress( Format( '/%s/sync', [ aSynthName], AppLocale));   // Do not dispose off aLocMsg, it is a copy of aMsg, or a sub packet owned by it

      if Assigned( aLocMsg)
      then FSync := True;
    end;

    procedure   TModSync.DoSlowTick; // override;
    var
      Sync : Boolean;
    begin
      Sync := SignalToLogic( FInputs[ i_sync]);

      if (( not FOldSync) and Sync) or FSync
      then begin
        SendOSCBytes( FOscBytes);
        FLedCounter  := Round( SlowTimeToSampleCount( 0.125));
        FSyncCounter := 4;
      end;

      FSync    := False;
      FOldSync := Sync;

      if FLedCounter > 0
      then Dec( FLedCounter);

      if FSyncCounter > 0
      then begin
        Dec( FSyncCounter);
        FOutputs[ o_sync] := LogicToSignal( True);
      end
      else FOutputs[ o_sync] := LogicToSignal( False);
    end;

    procedure   TModSync.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/syncin/%f' , [ aPrefix, Name, LogicToSignal( FOldSync)        ], AppLocale));
      aCallback( Self, Format( '%s/%s/syncout/%f', [ aPrefix, Name, LogicToSignal( FLedCounter > 0)], AppLocale));
    end;

    function    TModSync.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_sync : Result := 'syncin';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModSync.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_sync : Result := 'syncout';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModOscMessage = class( TMod_O_S_C)
  strict private
  const
    i_trigger = 0;
    i_input   = 1;
  const
    o_out     = 0;
  private
    FAddress    : string;
    FValue      : Single;
    FValuePtr   : PSingle;
    FOscBytes   : TBytes;
    FOldTrigger : Boolean;
    FLedCounter : Integer;
  public
    property    Address : string read FAddress write SetAddress;
    property    Value   : Single read FValue   write SetValue;
  private
}

    procedure   TModOscMessage.Recreate;
    var
      aMsg   : TOSCMessage;
      aValue : TBytes;
    begin
      aMsg := TOSCMessage.Create( '/' + Address);
      aMsg.AddFloat( 0);

      try
        aValue    := aMsg.ToOSCBytes;
        FValuePtr := @ aValue[ Length( aValue) - SizeOf( Single)];
        FOscBytes := aValue;
      finally
        aMsg.DisposeOf;
      end;
    end;

    procedure   TModOscMessage.SetAddress( const aValue: string);
    begin
      if aValue <> FAddress
      then begin
        FAddress := aValue;
        Recreate;
      end;
    end;

    procedure   TModOscMessage.SetValue( const aValue: Single);
    begin
      if aValue <> FValue
      then begin
        FValue     := aValue;
        FValuePtr^ := ReverseFloat( FValue);
      end;
    end;

//  public

    constructor TModOscMessage.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      Address := 'oscaddress';
    end;

    procedure   TModOscMessage.AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket); // override;
    var
      aLocMsg : TOSCMessage;
    begin
      aLocMsg := aMsg.MatchAddress( Format( '/%s/%s', [ aSynthName, Address], AppLocale));   // Do not dispose off aLocMsg, it is a copy of aMsg, or a sub packet owned by it

      if Assigned( aLocMsg)
      then begin
        aLocMsg.Decode;
        if SameText( string( aLocMsg.TypeTag[ 0]), 'f')
        then begin
          Value := aLocMsg.ArgumentAsFloat[ 0];
          FLedCounter := Round( SlowTimeToSampleCount( 0.125));
        end;
      end;
    end;

    procedure   TModOscMessage.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'address')
      then Address := aValue
      else inherited;
    end;

    function    TModOscMessage.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'address')
      then Result := Address
      else Result := inherited;
    end;

    procedure   TModOscMessage.DoSlowTick; // override;
    var
      Trigger : Boolean;
    begin
      Trigger := SignalToLogic( FInputs[ i_trigger]);

      if not FOldTrigger and Trigger
      then begin
        Value := Normalize( FInputs[ i_input]);
        SendOSCBytes( FOscBytes);
      end;

      if FLedCounter > 0
      then Dec( FLedCounter);

      FOldTrigger := Trigger;
      FOutputs[ o_out] := Value;
    end;

    procedure   TModOscMessage.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigger/%f', [ aPrefix, Name, LogicToSignal( FOldTrigger)], AppLocale));
      aCallback( Self, Format( '%s/%s/syncout/%f', [ aPrefix, Name, LogicToSignal( FLedCounter > 0)], AppLocale));
    end;

    function    TModOscMessage.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trigger : Result := 'trigger';
        i_input   : Result := 'input';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModOscMessage.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiIn = class( TMod)
  public
}

    constructor TModMidiIn.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
      FIsMidi := True;
    end;


{ ========
  TModMidiNoteIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FLastNote    : Byte;
    FLastChannel : Byte;
    FOldChannel  : Byte;
    FGate        : Boolean;
    FNote        : TSignal;
    FVelocityOn  : TSignal;
    FVelocityOff : TSignal;
  public
}

    procedure   TModMidiNoteIn.NotesOff;
    begin
      FGate        := False;
      FVelocityOn  := 0;
      FVelocityOff := 0;
    end;

    procedure   TModMidiNoteIn.AcceptMidi( const aMsg: TMidiMessage); // override;
    begin
      if ( FChannel = MIDI_CH_OMNI) or ( aMsg.Channel = FChannel)            // 255 being omni
      then begin
        if Length( aMsg.Data) = 2
        then begin
          if aMSg.Command = CNoteOn
          then begin
            FGate        := True;
            FLastNote    := aMsg.Data[ 0];
            FLastChannel := aMSg.Channel;
            FNote        := NoteNumberToUnits( FLastNote - MiddleNote);
            FVelocityOn  := aMsg.Data[ 1] * MIDI_SCALING_REC;
          end
          else if ( aMsg.Command = CNoteOff) and ( aMsg.Data[ 0] = FLastNote) and ( aMsg.Channel = FLastChannel)
          then begin
            FGate        := False;
            FVelocityOff := aMsg.Data[ 1] * MIDI_SCALING_REC;
          end
          else if ( aMsg.Command = CControlChange) and ( aMsg.Data[ 0] = CCAllNotesOff) and ( aMsg.Data[ 1] = 0) // All notes off
          then NotesOff;
        end;
      end;
    end;

    procedure   TModMidiNoteIn.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        NotesOff;
      end;

      FChannel := Round( Clip( FInputs[ i_ch], 0, 16));
      if FOldChannel <> FChannel
      then NotesOff;
      FOutputs[ o_gate  ] := LogicToSignal( FGate);
      FOutputs[ o_note  ] := FNote;
      FOutputs[ o_velon ] := FVelocityOn;
      FOutputs[ o_veloff] := FVelocityOff;
      FOldChannel                  := FChannel;
    end;

    procedure   TModMidiNoteIn.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiNoteIn.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_ch : Result := 'ch';
        else   Result := NO_INPUT;
      end;
    end;

    function    TModMidiNoteIn.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_gate   : Result := 'gate';
        o_note   : Result := 'note';
        o_velon  : Result := 'velon';
        o_veloff : Result := 'veloff';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiCCIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FCC          : Byte;
    FGate        : Boolean;
    FValue       : TSignal;
    FCounter     : Integer;
    FLedCounter  : Integer;
  public
}

    procedure   TModMidiCCIn.AcceptMidi( const aMsg: TMidiMessage); // override;
    begin
      if ( FChannel = MIDI_CH_OMNI) or ( aMsg.Channel = FChannel)
      then begin
        if Length( aMsg.Data) = 2
        then begin
          if ( aMsg.Command = CControlChange) and ( aMsg.Data[ 0] = FCC)
          then begin
            FValue      := aMsg.Data[ 1];
            FCounter    := 44;                 //   1 ms gate pulse
            FLedCounter := 11025;              // 250 ms LED  pulse
            FGate       := True;
          end;
        end;
      end;
    end;

    procedure   TModMidiCCIn.DoSlowTick; // override;
    begin
      if FLedCounter > 0
      then Dec( FLedCounter);
      FChannel := Round( FInputs[ i_ch]);
      FCC      := Round( FInputs[ i_cc]);
      if FCounter > 0
      then Dec( FCounter)
      else FGate := False;
      FOutputs[ o_gate ] := LogicToSignal( FGate);
      FOutputs[ o_value] := FValue * MIDI_SCALING_REC;
    end;

    procedure   TModMidiCCIn.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FLedCounter <> 0)], AppLocale));
    end;

    function    TModMidiCCIn.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_ch : Result := 'ch';
        i_cc : Result := 'cc';
        else   Result := NO_INPUT;
      end;
    end;

    function    TModMidiCCIn.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_gate  : Result := 'gate';
        o_value : Result := 'value';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiOut = class( TMod)
  public
}

    constructor TModMidiOut.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
      FIsMidi := False;
    end;


{ ========
  TModMidiNoteOut = class( TModMidiOut)
  private
    FLastNote    : Byte;
    FLastChannel : Byte;
    FGate        : Boolean;
    FDidSend     : Boolean;
  public
}

    procedure   TModMidiNoteOut.SetDefaults; // override;
    begin
      FInputs[ i_gate  ] := 0;    // Gate off
      FInputs[ i_note  ] := 0;    // Note 0
      FInputs[ i_velon ] := 1;    // Max velon
      FInputs[ i_veloff] := 1;    // Max veloff
    end;

    procedure   TModMidiNoteOut.DoSlowTick; // override;
    // FNote := ( FLastNote - 69) / NOTE_SCALING;
    var
      NewGate : Boolean;
      aMsg    : TMidiMessage;
    begin
      NewGate := SignalToLogic( FInputs[ i_gate]);
      if NewGate <> FGate
      then begin
        if NewGate
        then begin
          FLastChannel := Round( FInputs[ i_ch]);
          FLastNote    := Round( Clip( FInputs[ i_note] * NOTE_SCALING + MiddleNote, 0, 127));
          SetLength( aMsg.Data, 2);
          aMsg.Command  := CNoteOn;
          aMsg.Channel  := FLastChannel;
          aMsg.Data[ 0] := FlastNote;
          aMsg.Data[ 1] := Round( Clip( FInputs[ i_velon], 0, 1) * 127);
          SendMidi( aMsg);
          FDidSend := True;
        end
        else begin
          if FDidSend
          then begin
            FDidSend := False;
            SetLength( aMsg.Data, 2);
            aMsg.Command  := CNoteOff;
            aMsg.Channel  := FLastChannel;
            aMsg.Data[ 0] := FlastNote;
            aMsg.Data[ 1] := Round( Clip( FInputs[ i_veloff], 0, 1) * 127);
            SendMidi( aMsg);
          end;
        end;
        FGate := NewGate;
      end;
    end;

    procedure   TModMidiNoteOut.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiNoteOut.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_ch     : Result := 'ch';
        i_gate   : Result := 'gate';
        i_note   : Result := 'note';
        i_velon  : Result := 'velon';
        i_veloff : Result := 'veloff';
        else       Result := NO_INPUT;
      end;
    end;


{ ========
  TModMidiCCOut = class( TModMidiOut)
  private
    FGate  : Boolean;
  public
}

    procedure   TModMidiCCOut.DoSlowTick; // override;
    var
      NewGate : Boolean;
      aMsg    : TMidiMessage;
    begin
      NewGate := SignalToLogic( FInputs[ i_gate]);
      if NewGate <> FGate
      then begin
        if NewGate
        then begin
          SetLength( aMsg.Data, 2);
          aMsg.Command  := CControlChange;
          aMsg.Channel  := Round( FInputs[ i_ch]);
          aMsg.Data[ 0] := Round( FInputs[ i_cc]);
          aMsg.Data[ 1] := Round( Clip( FInputs[ i_value], 0, 1) * 127);
          SendMidi( aMsg);
        end;
        FGate := NewGate;
      end;
    end;

    procedure   TModMidiCCOut.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiCCOut.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_ch    : Result := 'ch';
        i_cc    : Result := 'cc';
        i_gate  : Result := 'gate';
        i_value : Result := 'value';
        else      Result := NO_INPUT;
      end;
    end;


{ ========
  TModMidiPlayer = class( TMod)
  private
    FState           : TMidiPlayerState;
    FReader          : TWrenMidiReader;
    FEvent           : TWrenMidiTimedEvent;
    FFileName        : string;
    FReading         : Boolean;
    FCounter         : Integer;
    FChannel         : Byte;
    FOldTrig         : Boolean;
    FOldReset        : Boolean;
    FGateCount       : Integer;
    FNote            : TSignal;
    FVelocityOn      : TSignal;
    FVelocityOff     : TSignal;
    FDoneCounter     : Integer;
    FDoneLedCounter  : Integer;
    FResetLedCounter : Integer;
    FChOffset        : Integer;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

    procedure   TModMidiPlayer.SetFileName( const aValue: string);
    begin
      try
        if aValue <> FFileName
        then begin
          if Assigned( FReader)
          then begin
            if FileExists( aValue)
            then begin
              FReading := True;
              try
                NotesOff;
                FFileName := aValue;
                FReader.ReadFile( FFileName);
                if FReader.Format = 1
                then FChOffset :=  0
                else FChOffset := -1;
                FEvent := GetNextEvent;
              finally
                FReading := False;
              end;
            end
            else begin
              NotesOff;
              FEvent    := nil;
              FFileName := aValue;
            end;
          end;
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    function    TModMidiPlayer.GetNextEvent: TWrenMidiTimedEvent;
    begin
      Result := FReader.NextTimedEvent( FChannel + FChOffset);
    end;

//  public

    procedure   TModMidiPlayer.NotesOff;
    begin
      FGateCount   := 0;
      FVelocityOn  := 0;
      FVelocityOff := 0;
    end;

    constructor TModMidiPlayer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FReader   := TWrenMidiReader.Create;
      FState    := msActive;
      FChOffset := -1;
      FIsFast   := False;
      FIsSlow   := True;
    end;

    destructor  TModMidiPlayer.Destroy; // override;
    begin
      FreeAndNil( FReader);
    end;

    procedure   TModMidiPlayer.DoSlowTick; // override;
    var
      Trig      : Boolean;
      Reset     : Boolean;
      Channel   : Integer;
      Mode      : Integer;
      MidiEvent : TWrenMidiEvent;
    begin
      Reset := SignalToLogic( FInputs[ i_reset]);
      if ResetFlag or ( Reset and not FOldReset)
      then begin
        ResetFlag := False;
        FOldReset := Reset;
        NotesOff;
        FCounter := 0;
        FOldTrig := False;
        if Assigned( FReader)
        then FReader.Reset;
        FEvent := nil;
        FResetLedCounter := 1000;
        FState := msActive;
      end;

      Channel  := Round( FInputs[ i_ch]);
      if Channel <> FChannel
      then NotesOff;
      FChannel := Channel;

      if FReading
      then NotesOff
      else begin
        Trig := SignalToLogic( FInputs[ i_trig]);
        if Trig and not FOldTrig
        then begin

          if FCounter = 0
          then begin
            if Assigned( FEvent)
            then begin

              MidiEvent := FEvent.Event;
              if FEvent.IsAllNotesOffEvent
              then NotesOff
              else if FEvent.IsEndOfTrackEvent
              then begin
                Mode := Round( FInputs[ i_mode]);
                if Mode = m_once
                then FState := msDone;
                FReader.Reset;
                FEvent := nil;
                NotesOff;
                FDoneCounter    :=  100;
                FDoneLedCounter := 1000;
              end
              else if FEvent.IsNoteOffEvent
              then begin
                FVelocityOff := TWrenMidiNoteOffEvent( MidiEvent).Velocity * MIDI_SCALING_REC;
                if FGateCount > 0
                then Dec( FGateCount);
              end
              else if FEvent.IsNoteOnEvent
              then begin
                FNote       := NoteNumberToUnits( TWrenMidiNoteOffEvent( MidiEvent).Key - MiddleNote);
                FVelocityOn := TWrenMidiNoteOffEvent( MidiEvent).Velocity * MIDI_SCALING_REC;
                if FVelocityOn = 0
                then Dec( FGateCount)
                else Inc( FGateCount);
              end;
            end;

            if FState = msActive
            then begin
              FEvent := GetNextEvent;
              if not Assigned( FEvent)
              then begin
                FReader.Reset;
                FEvent:= GetNextEvent;
              end;
              if Assigned( FEvent)
              then FCounter := FEvent.DeltaTime.DeltaTime;
            end
            else begin
              FEvent := nil;
              NotesOff;
            end;
          end
          else Dec( FCounter)

        end;
        FOldTrig := Trig;
      end;

      if FDoneCounter     > 0 then Dec( FDoneCounter    );
      if FDoneLedCounter  > 0 then Dec( FDoneLedCounter );
      if FResetLedCounter > 0 then Dec( FResetLedCounter);

      FOutputs[ o_note  ] := FNote;
      FOutputs[ o_velon ] := FVelocityOn;
      FOutputs[ o_veloff] := FVelocityOff;
      FOutputs[ o_gate  ] := LogicToSignal( FGateCount   > 0);
      FOutputs[ o_done  ] := LogicToSignal( FDoneCounter > 0);
    end;

    procedure   TModMidiPlayer.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModMidiPlayer.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FileName
      else Result := inherited;
    end;

    procedure   TModMidiPlayer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, LogicToSignal( FOldTrig            )], AppLocale));
      aCallback( Self, Format( '%s/%s/gate/%f' , [ aPrefix, Name, LogicToSignal( FGateCount       > 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/done/%f' , [ aPrefix, Name, LogicToSignal( FDoneLedCounter  > 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, LogicToSignal( FResetLedCounter > 0)], AppLocale));
    end;

    function    TModMidiPlayer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig  : Result := 'trig';
        i_ch    : Result := 'ch';
        i_reset : Result := 'reset';
        i_mode  : Result := 'mode';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModMidiPlayer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_note   : Result := 'note';
        o_velon  : Result := 'velon';
        o_veloff : Result := 'veloff';
        o_gate   : Result := 'gate';
        o_done   : Result := 'done';
       else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModReverb = class( TMod)
  private
    FReverb : TReverb;
  public
}

    constructor TModReverb.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FReverb := TReverb.Create( System_Rate);
      FIsFast := True;
      FIsSlow := True;
    end;

    destructor  TModReverb.Destroy; // override;
    begin
      FreeAndNil( FReverb);
      inherited;
    end;

    procedure   TModReverb.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_roomsize,
        i_damp    ,
        i_wet     ,
        i_dry     ,
        i_width   ,
        i_mute
      ];
    end;

    procedure   TModReverb.DoSlowTick; // override;
    begin
      with FReverb
      do begin
        SeRoomSize( FInputs[ i_roomsize]);
        SetDamp    ( FInputs[ i_damp    ]);
        SetWet     ( FInputs[ i_wet     ]);
        SetDry     ( FInputs[ i_dry     ]);
        SetWidth   ( FInputs[ i_width   ]);
        SetMode    ( LogicToSignal( SignalToLogic( FInputs[ i_mode]) xor SignalToLogic( FInputs[ i_freezein])));
      end;
    end;

    procedure   TModReverb.DoTick; // override;
    var
      OutL     : Single;
      OutR     : Single;
      Mute     : TSignal;
      MuteType : Integer;
    begin
      FReverb.ProcessReplace( Normalize( FInputs[ i_inl]), Normalize( FInputs[ i_inr]), OutL, OutR);
      MuteType := Round( FInputs[ i_mute]);
      if MuteType = 2
      then begin
        FOutputs[ o_outl] := FInputs[ i_inl];
        FOutputs[ o_outr] := FInputs[ i_inr];
      end
      else begin
        Mute := SignalToMute( FInputs[ i_mute]);
        FOutputs[ o_outl] := Normalize( OutL) * Mute;
        FOutputs[ o_outr] := Normalize( OutR) * Mute;
      end;
    end;

    procedure   TModReverb.SampleRateChanged; // override;
    var
      OldReverb : TReverb;
      NewReverb : TReverb;
    begin
      NewReverb := TReverb.Create( System_Rate);
      OldReverb := AtomicExChange( Pointer( FReverb), Pointer( NewReverb));
      FreeAndNil( OldReverb);
    end;

    function    TModReverb.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inl      : Result := 'inl';
        i_inr      : Result := 'inr';
        i_roomsize : Result := 'roomsize';
        i_damp     : Result := 'damp';
        i_wet      : Result := 'wet';
        i_dry      : Result := 'dry';
        i_width    : Result := 'width';
        i_mode     : Result := 'mode';
        i_mute     : Result := 'mute';
        i_freezein : Result := 'freezein';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModReverb.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outr : Result := 'outr';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModGVerb = class( TMod)
  // https://github.com/swh/lv2/tree/master/gverb
  private
    FReverb       : TGVerb;
    FRecreateFlag : Boolean;
    FMaxRoomSize  : TSignal;
    FRoomSize     : TSignal;
    FRevTime      : TSignal;
    FDamping      : TSignal;
    FSpread       : TSignal;
    FInputBW      : TSignal;
    FEarlyLevel   : TSignal;
    FTailLevel    : TSignal;
    FDryLevel     : TSignal;
    FInputLevel   : TSignal;
    FMode         : Integer;
  private
}

    procedure   TModGVerb.ReCreateReverb;
    var
      anOld : TGVerb;
    begin
      anOld :=
        AtomicExchange(
          Pointer( FReverb),
          Pointer(
            TGVerb.Create(
              System_Rate,
              FMaxRoomSize,
              FRoomSize,
              FRevTime,
              FDamping,
              FSpread * 100,               // 180 will crash it .. super collider uses 100 as a max it seems, lets do that too then
              FInputBW,
              FEarlyLevel,
              FTailLevel,
              FDryLevel,
              FMode - 1
            )
          )
        );
      FreeAndNil( anOld);
    end;

//  public

    constructor TModGVerb.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast      := True;
      FIsSlow      := True;
      FMaxRoomSize := 302.0;
      FRoomSize    := FMaxRoomSize * 0.103;
      FRevTime     := 3.16;
      FDamping     := 0.5;
      FSpread      := 0.5;
      FInputBW     := 0.702;
      FEarlyLevel  := 0.7;
      FTailLevel   := 0.7;
      FDryLevel    := 0.7;
      FInputLevel  := 0.7;
      FMode        := 0;
      ReCreateReverb;
    end;

    destructor  TModGVerb.Destroy; // override;
    begin
      FreeAndNil( FReverb);
      inherited;
    end;

    procedure   TModGVerb.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_roomsize      ,
        i_roomsizemodlev,
        i_revtime       ,
        i_damping       ,
        i_inputbw       ,
        i_earlylevel    ,
        i_taillevel     ,
        i_drylevel      ,
        i_inputlevel    ,
        i_mute
      ];
      FInputs[ i_revtimemod   ] := 1; // Set all modulation multipliers to 1 to make stuff work with unconnected inputs
      FInputs[ i_dampingmod   ] := 1;
      FInputs[ i_inputbwmod   ] := 1;
      FInputs[ i_earlylevelmod] := 1;
      FInputs[ i_taillevelmod ] := 1;
      FInputs[ i_drylevelmod  ] := 1;
    end;

    procedure   TModGVerb.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'maxroomsize')
      then begin
        FMaxRoomSize := aValue;
        FRoomSize    := Clip( FInputs[ i_roomsizemod] * FInputs[i_roomsizemodlev] + FInputs[ i_roomsize], 0.0, 1.0) * FMaxRoomSize;
        ReCreateReverb;
      end
      else if SameText( aName, 'spread')
      then begin
        FSpread   := aValue;
        FRoomSize := Clip( FInputs[ i_roomsizemod] * FInputs[i_roomsizemodlev] + FInputs[ i_roomsize], 0.0, 1.0) * FMaxRoomSize;
        ReCreateReverb;
      end
      else if SameText( aName, 'mode')
      then begin
        FMode     := Round( aValue);
        FRoomSize := Clip( FInputs[ i_roomsizemod] * FInputs[i_roomsizemodlev] + FInputs[ i_roomsize], 0.0, 1.0) * FMaxRoomSize;
        ReCreateReverb;
      end
      else inherited;
    end;

    function    TModGVerb.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'maxroomsize')
      then Result := FMaxRoomSize
      else if SameText( aName, 'spread')
      then Result := FSpread
      else if SameText( aName, 'mode')
      then Result := FMode
      else Result := inherited;
    end;

    procedure   TModGVerb.DoSlowTick; // override;
    begin
      FRoomSize    := Clip( FInputs[ i_roomsizemod] * FInputs[i_roomsizemodlev] + FInputs[ i_roomsize], 0.0, 1.0) * FMaxRoomSize;
      FRevTime     := FInputs[ i_revtime   ] * Clip( FInputs[ i_revtimemod   ], 0, 1);
      FDamping     := FInputs[ i_damping   ] * Clip( FInputs[ i_dampingmod   ], 0, 1);
      FInputBW     := FInputs[ i_inputbw   ] * Clip( FInputs[ i_inputbwmod   ], 0, 1);
      FEarlyLevel  := FInputs[ i_earlylevel] * Clip( FInputs[ i_earlylevelmod], 0, 1);
      FTailLevel   := FInputs[ i_taillevel ] * Clip( FInputs[ i_taillevelmod ], 0, 1);
      FDryLevel    := FInputs[ i_drylevel  ] * Clip( FInputs[ i_drylevelmod  ], 0, 1);
      FInputLevel  := FInputs[ i_inputlevel];

      FReverb.SetRoomSize      ( FRoomSize  );
      FReverb.SetRevTime       ( FRevTime   );
      FReverb.SetDamping       ( FDamping   );
      FReverb.SetInputBandWidth( FInputBW   );
      FReverb.SetEarlyLevel    ( FEarlyLevel);
      FReverb.SetTailLevel     ( FTailLevel );
      FReverb.SetDryLevel      ( FDryLevel  );

      // this is a kludge, the spread paramater does not get properly initialized unless the knob is moved, or unless
      // the reverb is recreated after we came here ... the lights handler will recreate the reverb when the FRecreate
      // flag is set .. still no idea why this happens ... yuck. search ***1 for more source remarks.
      // o .. and it is best not to recreate stuff here, that'd be memory manager access from the audio thread, bad
      // idea.
      if ResetFlag
      then FRecreateFlag := True;
      ResetFlag := False;
      // end kludge.
    end;

    procedure   TModGVerb.DoTick; // override;
    var
      anOutL   : TSignal;
      anOutR   : TSignal;
      Mute     : TSignal;
      MuteType : Integer;
    begin
      FReverb.Tick(
        FInputs [ i_inl ] * FInputLevel,
        FInputs [ i_inr ] * FInputLevel,
        anOutL,
        anOutR
      );
      MuteType := Round( FInputs[ i_mute]);
      if MuteType = 2
      then begin
        FOutputs[ o_outl] := FInputs[ i_inl];
        FOutputs[ o_outr] := FInputs[ i_inr];
      end
      else begin
        Mute := SignalToMute( FInputs[ i_mute]);
        FOutputs[ o_outl] := Normalize( anOutL) * Mute;
        FOutputs[ o_outr] := Normalize( anOutR) * Mute;
      end;
    end;

    procedure   TModGVerb.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      // This is a kludge ... search ***1 for more source remarks.
      if FRecreateFlag
      then ReCreateReverb;
      FRecreateFlag := False;
      // end kludge
    end;

    procedure   TModGVerb.SampleRateChanged; // override;
    begin
      ReCreateReverb;
    end;

    function    TModGVerb.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inl            : Result := 'inl';
        i_inr            : Result := 'inr';
        i_roomsize       : Result := 'roomsize';
        i_roomsizemod    : Result := 'roomsizemod';
        i_roomsizemodlev : Result := 'roomsizemodlev';
        i_revtime        : Result := 'revtime';
        i_damping        : Result := 'damping';
        i_inputbw        : Result := 'inputbw';
        i_earlylevel     : Result := 'earlylevel';
        i_taillevel      : Result := 'taillevel';
        i_drylevel       : Result := 'drylevel';
        i_inputlevel     : Result := 'inputlevel';
        i_mute           : Result := 'mute';
        i_revtimemod     : Result := 'revtimemod';
        i_dampingmod     : Result := 'dampingmod';
        i_inputbwmod     : Result := 'inputbwmod';
        i_earlylevelmod  : Result := 'earlylevelmod';
        i_taillevelmod   : Result := 'taillevelmod';
        i_drylevelmod    : Result := 'drylevelmod';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModGVerb.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outr : Result := 'outr';
        else     Result := NO_OUTPUT;
      end;
    end;

    function    TModGVerb.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gverb_maxroomsize : Result := 'maxroomsize';
        gverb_spread      : Result := 'spread';
        gverb_mode        : Result := 'mode';
        else                Result := NO_INTERNAL;
      end;
    end;



{ ========
  TModEnvelope = class( TMod)
  private
    FValue     : TSignal;
    FAttack    : TSignal;
    FRelease   : TSignal;
    FPeakValue : TSignal;
    FPeakDecay : TSignal;
  public
}


    constructor TModEnvelope.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEnvelope.SetDefaults; // override;
    begin
      FInputs[ i_attack ] := 1;
      FInputs[ i_release] := 1;
      DoSlowTick;
    end;

    procedure   TModEnvelope.DoSlowTick; // override;
    begin
      FAttack    := Normalize( Exp( - 1 / Max( 1, SlowTimeToSampleCount( FInputs[ i_attack   ]))));
      FRelease   := Normalize( Exp( - 1 / Max( 1, SlowTimeToSampleCount( FInputs[ i_release  ]))));
      FPeakDecay := Normalize( Exp( - 1 / Max( 1, SlowTimeToSampleCount( FInputs[ i_peakdecay]))));
    end;

    procedure   TModEnvelope.DoTick; // override;
    var
      aValue : TSignal;
      aPeak  : TSignal;
    begin
      aValue := Abs( Normalize( FInputs[ i_in]));
      if FValue < aValue
      then begin
        FValue := FValue * FAttack;
        FValue := Normalize( FValue + ( 1 - FAttack) * aValue);
        if FValue > FPeakValue
        then FPeakValue := FValue;
      end
      else begin
        FValue := FValue * FRelease;
        FValue := Normalize( FValue + ( 1 - FRelease) * aValue);
        if SignalToLogic( FInputs[ i_peakdecayon])
        then begin
          aPeak      := FPeakValue;
          FPeakValue := FPeakValue * FPeakDecay;
          FPeakValue := Normalize( FPeakValue + ( 1 - FPeakDecay) * aPeak);
        end;
      end;
      FOutputs[ o_out   ] := FValue;
      FOutputs[ o_invout] := FPeakValue - FValue;
    end;

    function    TModEnvelope.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in          : Result := 'in';
        i_attack      : Result := 'attack';
        i_release     : Result := 'release';
        i_peakdecay   : Result := 'peakdecay';
        i_peakdecayon : Result := 'peakdecayon';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModEnvelope.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out    : Result := 'out';
        o_invout : Result := 'invout';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModCompressor = class( TMod)
  private
    const
      COMP_MAX_GAIN =  64.0;
      COMP_MIN_GAIN = -32.0;
  private
    FEnvelope    : TSignal;
    FAttack      : TSignal;
    FRelease     : TSignal;
    FRefLevel    : TSignal;
    FPreGain     : TSignal;
    FGain        : TSignal;
    FBypass      : Integer;
    FOutL        : TSignal;
    FOutR        : TSignal;
    FLoopAmp     : TSignal;
    FMaxLoopAmp  : TSignal;
    FOldAttack   : TSignal;
    FOldRelease  : TSignal;
    FMinDbVal    : TSignal;
    FMaxDbVal    : TSignal;
  public
}

    constructor TModCompressor.Create( aParent: TSynthPatch; const aName: string); // override;
    const
      ONE_DIV_20 = 1.0 / 20.0;
    begin
      inherited;
      FMaxDbVal := Power( 10, COMP_MAX_GAIN * ONE_DIV_20);
      FMinDbVal := Power( 10, COMP_MIN_GAIN * ONE_DIV_20);
      FIsFast   := True;
      FIsSlow   := True;
    end;

    procedure   TModCompressor.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_reflevel,
        i_gain    ,
        i_attack  ,
        i_release
      ];
      FInputs[ i_attack ] := 1;
      FInputs[ i_release] := 1;
      DoSlowTick;
    end;

    procedure   TModCompressor.DoSlowTick; // override;
    var
      aValue: TSignal;
    begin
      aValue := FInputs[ i_attack];
      if aValue <> FOldAttack
      then begin
        FOldAttack := aValue;
        FAttack    := Exp( - 1 / Max( 1, SlowTimeToSampleCount( aValue)));
      end;

      aValue := FInputs[ i_release];
      if aValue <> FOldRelease
      then begin
        FOldRelease := aValue;
        FRelease    := Exp( - 1 / Max( 1, SlowTimeToSampleCount( aValue)));
      end;

      FRefLevel   := FInputs[ i_reflevel];
      FPreGain    := FInputs[ i_preamp  ];
      FGain       := FInputs[ i_gain    ];
      FBypass     := Round( FInputs[ i_bypass]);
      FMaxLoopAmp := Max( FMaxLoopAmp, FLoopAmp);
    end;

    procedure   TModCompressor.DoTick; // override;
    var
      anInL      : TSignal;
      anInR      : TSignal;
      anOutLevel : TSignal;
      anEnvValue : TSignal;
    begin
      anInL      := FPreGain * FInputs [ i_inl];
      anInR      := FPreGain * FInputs [ i_inr];
      anOutlevel := Max( Abs( FOutL), Abs( FOutR));

      if FEnvelope < anOutLevel
      then anEnvValue := FAttack  * ( FEnvelope - anOutLevel) + anOutLevel
      else anEnvValue := FRelease * ( FEnvelope - anOutLevel) + anOutLevel;

      FEnvelope   := CompressorSmoothing * ( anEnvValue - FEnvelope) + FEnvelope;
      FLoopAmp := Clip( FGain * ( FRefLevel -  FEnvelope), FMinDbVal, FMaxDbVal);
      FOutL    := FLoopAmp * anInL;
      FOutR    := FLoopAmp * anInR;

      case FBypass of
        m_bypass_byp : // outputs are same as inputs
          begin
            FOutputs[ o_outl] := FInputs[ i_inl]; // Do not use anInL - that one has pre-gain applied
            FOutputs[ o_outr] := FInputs[ i_inr]; // Likewise for anInR
          end;
        else begin              // normal operation
          FOutputs[ o_outl] := FOutL;
          FOutputs[ o_outr] := FOutR;
        end;
      end;
      FOutputs[ o_env] := FEnvelope;
    end;

    procedure   TModCompressor.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    const
      LOWLED  =   0.000;
      HIGHLED =   1.000;
      DECAY   =   0.707;
    var
      adB : TSignal;
    begin
      adB := RangeMap( Clip( SignalTodB( FMaxLoopAmp), COMP_MIN_GAIN, COMP_MAX_GAIN), COMP_MIN_GAIN, COMP_MAX_GAIN, LOWLED, HIGHLED);
      aCallback( Self, Format( '%s/%s/level/%f', [ aPrefix, Name, adB], AppLocale));
      FMaxLoopAmp := FMaxLoopAmp * DECAY;
    end;

    function    TModCompressor.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inl       : Result := 'inl';
        i_inr       : Result := 'inr';
        i_reflevel  : Result := 'reflevel';
        i_gain      : Result := 'gain';
        i_attack    : Result := 'attack';
        i_release   : Result := 'release';
        i_bypass    : Result := 'bypass';
        i_preamp    : Result := 'preamp';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModCompressor.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outl : Result := 'outl';
        o_outr : Result := 'outr';
        o_env  : Result := 'env';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFormant = class( TMod)
  private
    FCoeffStore    : array[ 0 .. 10] of TSignalArray;
    FCoeffiecients : array[ 0 .. 10] of TSignal;
    FMemory        : TSignalArray;
    FVowel         : TSignal;
  public
}

    constructor TModFormant.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      i : Integer;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      SetLength( FMemory, 10);
      for i := 0 to 10
      do FCoeffStore[ i] := TSignalArrayWrapper.Create( 5);

      // A

      FCoeffStore[  0].FSignals[ 0] :=    8.11044e-06;
      FCoeffStore[  1].FSignals[ 0] :=    8.943665402;
      FCoeffStore[  2].FSignals[ 0] :=  -36.83889529 ;
      FCoeffStore[  3].FSignals[ 0] :=   92.01697887 ;
      FCoeffStore[  4].FSignals[ 0] := -154.337906   ;
      FCoeffStore[  5].FSignals[ 0] :=  181.6233289  ;
      FCoeffStore[  6].FSignals[ 0] := -151.8651235  ;
      FCoeffStore[  7].FSignals[ 0] :=   89.09614114 ;
      FCoeffStore[  8].FSignals[ 0] :=  -35.10298511 ;
      FCoeffStore[  9].FSignals[ 0] :=    8.388101016;
      FCoeffStore[ 10].FSignals[ 0] :=   -0.923313471;

      // E

      FCoeffStore[  0].FSignals[ 1] :=    4.36215e-06;
      FCoeffStore[  1].FSignals[ 1] :=    8.90438318 ;
      FCoeffStore[  2].FSignals[ 1] :=  -36.55179099 ;
      FCoeffStore[  3].FSignals[ 1] :=   91.05750846 ;
      FCoeffStore[  4].FSignals[ 1] := -152.422234   ;
      FCoeffStore[  5].FSignals[ 1] :=  179.1170248  ;
      FCoeffStore[  6].FSignals[ 1] := -149.6496211  ;
      FCoeffStore[  7].FSignals[ 1] :=   87.78352223 ;
      FCoeffStore[  8].FSignals[ 1] :=  -34.60687431 ;
      FCoeffStore[  9].FSignals[ 1] :=    8.282228154;
      FCoeffStore[ 10].FSignals[ 1] :=   -0.914150747;

      // I

      FCoeffStore[  0].FSignals[ 2] :=    3.33819e-06;
      FCoeffStore[  1].FSignals[ 2] :=    8.893102966;
      FCoeffStore[  2].FSignals[ 2] :=  -36.49532826 ;
      FCoeffStore[  3].FSignals[ 2] :=   90.96543286 ;
      FCoeffStore[  4].FSignals[ 2] := -152.4545478  ;
      FCoeffStore[  5].FSignals[ 2] :=  179.4835618  ;
      FCoeffStore[  6].FSignals[ 2] := -150.315433   ;
      FCoeffStore[  7].FSignals[ 2] :=   88.43409371 ;
      FCoeffStore[  8].FSignals[ 2] :=  -34.98612086 ;
      FCoeffStore[  9].FSignals[ 2] :=    8.407803364;
      FCoeffStore[ 10].FSignals[ 2] :=   -0.932568035;

       // O

      FCoeffStore[  0].FSignals[ 3] :=    1.13572e-06;
      FCoeffStore[  1].FSignals[ 3] :=    8.994734087;
      FCoeffStore[  2].FSignals[ 3] :=  -37.2084849  ;
      FCoeffStore[  3].FSignals[ 3] :=   93.22900521 ;
      FCoeffStore[  4].FSignals[ 3] := -156.6929844  ;
      FCoeffStore[  5].FSignals[ 3] :=  184.596544   ;
      FCoeffStore[  6].FSignals[ 3] := -154.3755513  ;
      FCoeffStore[  7].FSignals[ 3] :=   90.49663749 ;
      FCoeffStore[  8].FSignals[ 3] :=  -35.58964535 ;
      FCoeffStore[  9].FSignals[ 3] :=    8.478996281;
      FCoeffStore[ 10].FSignals[ 3] :=   -0.929252233;

      // U

      FCoeffStore[  0].FSignals[ 4] :=    4.09431e-07;
      FCoeffStore[  1].FSignals[ 4] :=    8.997322763;
      FCoeffStore[  2].FSignals[ 4] :=  -37.20218544 ;
      FCoeffStore[  3].FSignals[ 4] :=   93.11385476 ;
      FCoeffStore[  4].FSignals[ 4] := -156.2530937  ;
      FCoeffStore[  5].FSignals[ 4] :=  183.7080141  ;
      FCoeffStore[  6].FSignals[ 4] := -153.2631681  ;
      FCoeffStore[  7].FSignals[ 4] :=   89.59539726 ;
      FCoeffStore[  8].FSignals[ 4] :=  -35.12454591 ;
      FCoeffStore[  9].FSignals[ 4] :=    8.338655623;
      FCoeffStore[ 10].FSignals[ 4] :=   -0.910251753;
    end;

    destructor  TModFormant.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to 10
      do FCoeffStore[ i].DisposeOf;
      inherited;
    end;

    procedure   TModFormant.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_vowel,
        i_controlmod
      ];
    end;

    procedure   TModFormant.DoSlowTick; // override;
    var
      i : Integer;
    begin
      FVowel := Clip( 4 * ( FInputs[ i_vowel] + FInputs[ i_controlmod] * FInputs[ i_vowelin]), 0.01, 3.99);
      for i := 0 to 10
      do FCoeffiecients[ i] := Normalize( FCoeffStore[ i].Interpolate( FVowel));
    end;

    procedure   TModFormant.DoTick; // override;
    begin
      FOutputs[ o_out] :=
        Normalize(
          FCoeffiecients[  0] * FInputs[ i_in] +
          FCoeffiecients[  1] * FMemory[ 0]    +
          FCoeffiecients[  2] * FMemory[ 1]    +
          FCoeffiecients[  3] * FMemory[ 2]    +
          FCoeffiecients[  4] * FMemory[ 3]    +
          FCoeffiecients[  5] * FMemory[ 4]    +
          FCoeffiecients[  6] * FMemory[ 5]    +
          FCoeffiecients[  7] * FMemory[ 6]    +
          FCoeffiecients[  8] * FMemory[ 7]    +
          FCoeffiecients[  9] * FMemory[ 8]    +
          FCoeffiecients[ 10] * FMemory[ 9]
        );

      FMemory[ 9] := FMemory[ 8];
      FMemory[ 8] := FMemory[ 7];
      FMemory[ 7] := FMemory[ 6];
      FMemory[ 6] := FMemory[ 5];
      FMemory[ 5] := FMemory[ 4];
      FMemory[ 4] := FMemory[ 3];
      FMemory[ 3] := FMemory[ 2];
      FMemory[ 2] := FMemory[ 1];
      FMemory[ 1] := FMemory[ 0];
      FMemory[ 0] := FOutputs[ o_out];
    end;

    function    TModFormant.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in         : Result := 'in';
        i_vowel      : Result := 'vowel';
        i_vowelin    : Result := 'vowelin';
        i_controlmod : Result := 'controlmod';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModFormant.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFlangeChorus = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedBack : TSignal;
    FRate     : TSignal;
    FSweep    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp        : Integer;
    ep1       : Integer;
    ep2       : Integer;
    step      : Integer;
    depth     : Integer;
    idelay    : Integer;
    min_sweep : Integer;
    max_sweep : Integer;
    wlSweep   : TLongWord;
    Buf       : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
  private
}

    procedure   TModFlangeChorus.Setup;
      //  flange_chorus - original comments
      //
      //  Does flanging/chorusing family of effects based on a single
      //  varying delay.
      //
      //  dry_mix     mix of unaffected signal ( -0.999 to 0.999)
      //  wet_mix     mix of affected signal ( -0.999 - 0.999)
      //  feedback    amount of recirculation ( -0.9 - 0.9)
      //  rate        rate of delay change in millisecs per sec
      //  sweep       sweep range in millisecs
      //  delay       fixed additional delay in millisecs
    begin
      // fetch params
      step   := Round( FRate  * 65.536);                 // done: So Rate goes  0 - 1 now
      depth  := Round( FSweep * System_Rate * 1e-3);     // done: so Sweep goes 0 - 1 now
      idelay := Round( FDelay * System_Rate * 1e-3);     // done: so Delay goes 0 - 1 now

      // init/calc some stuff
      max_sweep := DELAY_BUF_SIZE - 2 - idelay;
      min_sweep := max_sweep - depth;

      if min_sweep < 0
      then min_sweep := 0;

      wlSweep.Hi := ( min_sweep + max_sweep) shr 1;
      wlSweep.Lo := 0;

      // init store and read ptrs to known value
      fp  := 0;
      ep1 := 0;
      ep2 := 0;
    end;

//  public

    constructor TModFlangeChorus.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      Setup;
    end;

    procedure   TModFlangeChorus.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute    ,
        i_dry     ,
        i_wet     ,
        i_feedback
      ];
    end;

    procedure   TModFlangeChorus.DoSlowTick; // override;
    var
      OldRate  : TSignal;
      OldSweep : TSignal;
      OldDelay : TSignal;
    begin
      FFrozen   := SignalToLogic( FInputs[ i_freeze  ]) xor SignalToLogic( FInputs[ i_freezein]);
      FDryMix   :=                FInputs[ i_dry     ];
      FWetMix   :=                FInputs[ i_wet     ];
      FFeedBack := Clip         ( FInputs[ i_feedback], -0.999, 0.999);

      OldRate  := FRate;
      OldSweep := FSweep;
      OldDelay := FDelay;

      FRate  := FInputs[ i_rate ];
      FSweep := FInputs[ i_sweep];
      FDelay := FInputs[ i_delay];

      if
        ( FRate  <> OldRate ) or
        ( FSweep <> OldSweep) or
        ( FDelay <> OldDelay)
      then Setup;
    end;

    procedure   TModFlangeChorus.DoTick; // override;
    const
      ODM = 1.0 / 65536.0;
    var
      outval : TSignal;
      aSample: TSignal;
    begin
      aSample := FInputs[ i_in];

      // interpolate from the 2 read values

      outval :=
        ( Buf[ ep1] * wlSweep.Lo +
          Buf[ ep2] * ( 65536.0 - wlSweep.Lo)
        ) * ODM;

      // develop and output final output mix

      outval := Normalize( outval * FWetMix + aSample * FDryMix);
      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * outval;

      // update ptrs

       // store finished input plus feedback - if not frozen
      if not FFrozen
      then Buf[ fp] := aSample + outval * FFeedback;

      fp  := ( fp + 1) and ( DELAY_BUF_SIZE - 1);
      Inc( Integer( wlSweep), step);
      ep1 := ( fp + wlSweep.Hi) and ( DELAY_BUF_SIZE - 1);
      ep2 := ( ep1 - 1) and ( DELAY_BUF_SIZE - 1);
      if wlSweep.Hi > max_sweep       // see if we hit top of sweep
      then step := - step             // reverse sweep
      else if wlSweep.Hi < min_sweep  // or if we hit bottom of sweep
      then step := - step;            // reverse sweep
    end;

    procedure   TModFlangeChorus.SampleRateChanged; // override;
    begin
      Setup;
    end;

    function    TModFlangeChorus.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in       : Result := 'in';
        i_mute     : Result := 'mute';
        i_freeze   : Result := 'freeze';
        i_dry      : Result := 'dry';
        i_wet      : Result := 'wet';
        i_feedback : Result := 'feedback';
        i_rate     : Result := 'rate';
        i_sweep    : Result := 'sweep';
        i_delay    : Result := 'delay';
        i_freezein : Result := 'freezein';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModFlangeChorus.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPitchChange = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedback : TSignal;
    FRate     : TSignal;
    FDepth    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp         : Integer;
    ep1        : Integer;
    ep2        : Integer;
    ep3        : Integer;
    ep4        : Integer;
    idepth     : Integer;
    idelay     : Integer;
    min_sweep  : Integer;
    max_sweep  : Integer;
    step       : Word;
    xfade      : Integer;
    xfade_cnt  : Integer;
    active     : Integer;
    active_cnt : Integer;
    sweep_up   : Boolean;
    chanA      : Boolean;
    blendA     : TSignal;
    blendB     : TSignal;
    fadeA      : PFadeTable;
    fadeB      : PFadeTable;
    wlSweep    : TLongWord;
    Buf        : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
    Fade_Out   : TFadeTable;
    Fade_In    : TFadeTable;
  private
}

    procedure   TModPitchChange.Setup;
      //  pitch_change - original comments
      //
      //  dry_mix     mix of unaffected signal ( -0.999 to 0.999)
      //  wet_mix     mix of affected signal ( -0.999 - 0.999)
      //  feedback    amount of recirculation ( -0.9 - 0.9)
      //  rate        amount of pitch change ( see table below for values)
      //  depth       sweep range in millisecs for generating pitch shift
      //  delay       fixed additional delay

      //  Semitones      Up              Down
      //      1       0.059463        -0.056126
      //      2       0.122462        -0.109101
      //      3       0.189207        -0.159104
      //      4       0.259921        -0.206299
      //      5       0.334840        -0.250846
      //      6       0.414214        -0.292893
      //      7       0.498307        -0.332580
      //      8       0.587401        -0.370039
      //      9       0.681793        -0.405396
      //      10      0.781797        -0.438769
      //      11      0.887749        -0.470268
      //      12      1.000000        -0.500000
    begin
      sweep_up  := FRate > 0;
      idepth    := Round( FDepth * System_Rate * 1e-3);
      idelay    := Round( FDelay * System_Rate * 1e-3);
      max_sweep := DELAY_BUF_SIZE - 2 - idelay;
      min_sweep := max_sweep - idepth;
      if min_sweep < 0
      then min_sweep := 0;
      active    := max_sweep - min_sweep - Round( xfade * FRate) - 2;

      // init store and read ptrs to known value, chanA active 1st

      fp                := 0;
      ep3               := 0;
      ep4               := 0;
      xfade_cnt         := 0;
      Integer( wlSweep) := 0;

      active_cnt := active;
      blendA     := 1.0;
      blendB     := 0.0;
      fadeA      := @ fade_out;
      fadeB      := @ fade_in;
      chanA      := True;
    end;

//   public

    constructor TModPitchChange.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      Setup;
    end;

    procedure   TModPitchChange.SetDefaults; // override;
    var
      i : Integer;
    begin
      FDezipperMap := [
        i_mute    ,
        i_dry     ,
        i_wet     ,
        i_feedback
      ];

      xfade := Round( 12.0 * System_Rate * 0.001);
      if xfade > MAX_XTAB
      // then raise( ECrossfade.Create( 'Can''t do pitch change crossfade at this sample rate'));
      then xfade := MAX_XTAB;

      // build the crossfade lookup tables

      for i := 0 to xfade - 1
      do begin
        fade_in [ i] := cos( i * HALF_PI / xfade);
        fade_out[ i] := sin( i * HALF_PI / xfade);
      end;
    end;

    procedure   TModPitchChange.DoSlowTick; // override;
    var
      OldRate  : TSignal;
      OldDepth : TSignal;
      OldDelay : TSignal;
      rStep    : TSignal;
    begin
      FFrozen   := SignalToLogic( FInputs[ i_freeze  ]) xor SignalToLogic( FInputs[ i_freezein]);
      FDryMix   :=                FInputs[ i_dry     ];
      FWetMix   :=                FInputs[ i_wet     ];
      FFeedBack := Clip         ( FInputs[ i_feedback], -0.999, 0.999);

      OldRate  := FRate;
      OldDepth := FDepth;
      OldDelay := FDelay;

      FRate  := FInputs[ i_rate ];
      FDepth := FInputs[ i_depth];
      FDelay := FInputs[ i_delay];

      if FRate <> OldRate
      then begin
        rStep := FRate * 65536.0;
        while rStep < 0
        do rStep := rStep + 65536.0;
        if rStep > 65535.0
        then rStep := 65535.0;
        step := Round( rStep);
        sweep_up := FRate > 0;

        if sweep_up
        then begin
          ep1 := min_sweep;
          ep2 := min_sweep;
        end
        else begin
          ep1 := max_sweep;
          ep2 := max_sweep;
        end;
      end;

      if
        ( FDepth <> OldDepth) or
        ( FDelay <> OldDelay)
      then Setup;
    end;

    procedure   TModPitchChange.DoTick; // override;
    const
      ODM = 1.0 / 65536.0;
    var
      outval  : TSignal;
      comp    : TSignal;
      aSample : TSignal;
    begin
      aSample := FInputs[ i_in];

      // messy expression to interpolate from both pairs of read ptrs

      comp   := 65536.0 - wlSweep.Lo;
      outval :=
        (( Buf[ ep1] * wlSweep.Lo + Buf[ ep2] * comp) * blendA +
         ( Buf[ ep3] * wlSweep.Lo + Buf[ ep4] * comp) * blendB) * ODM;


      // store finished input plus feedback

      if not FFrozen
      then Buf[ fp] := aSample + outval * FFeedBack;

      // develop final output mix

      outval := Normalize( outval * FWetMix + aSample * FDryMix);
      FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * outval;

      // see if crossfade active

      if xfade_cnt <> 0
      then begin
        Dec( xfade_cnt);
        blendA := fadeA^[ xfade_cnt];
        blendB := fadeB^[ xfade_cnt];
      end;

      // update store ptr
      fp := ( fp + 1) and ( DELAY_BUF_SIZE - 1);

      // see which direction
      if sweep_up
      then begin

        // update sweep
        Inc( Integer( wlSweep), step);

        // always inc at least once
        ep2 := ep1;
        ep1 := ( ep1 + 1) and ( DELAY_BUF_SIZE - 1);
        ep4 := ep3;
        ep3 := ( ep3 + 1) and ( DELAY_BUF_SIZE - 1);

        // if sweep didn't overflow, we're done
        if wlSweep.Hi = 0
        Then Exit;

        // sweep overflowed, inc again
        ep2 := ep1;
        ep1 := ( ep1 + 1) and ( DELAY_BUF_SIZE - 1);
        ep4 := ep3;
        ep3 := ( ep3 + 1) and ( DELAY_BUF_SIZE - 1);
        wlSweep.Hi := 0;

        // see if it's time to switch over to other delay channel
        if active_cnt = 0
        then begin
          xfade_cnt  := xfade;        // initiate crossfade
          active_cnt := active;       // start counter on new channel
          if chanA
          then begin                  // A has been active, go to B
            chanA := False;
            ep3   := ( fp + min_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_out;
            fadeB := @ fade_in;
          end
          else begin
            chanA := True;
            ep1   := ( fp + min_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_in;
            fadeB := @ fade_out;
          end;
        end
        else Dec( active_cnt);
      end
      else begin                              // do downward sweep
        Inc( Integer( wlSweep), step);        // update sweep

        // if sweep didn't overflow, inc ptrs, that's all
        if wlSweep.Hi = 0
        then begin
          ep2 := ep1;
          ep1 := ( ep1 + 1) and ( DELAY_BUF_SIZE - 1);
          ep4 := ep3;
          ep3 := ( ep3 + 1) and ( DELAY_BUF_SIZE - 1);
          Exit;
        End;


        // sweep overflowed, check on stuff but skip ptr inc
        wlSweep.Hi := 0;


        // see if it's time to switch over to other delay channel
        if active_cnt = 0
        then begin
          xfade_cnt  := xfade;
          active_cnt := active;
          if chanA
          then begin                  // A has been active, go to B
            chanA := False;
            ep3   := ( fp + max_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_out;
            fadeB := @ fade_in;
          end
          else begin
            chanA := True;
            ep1   := ( fp + max_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_in;
            fadeB := @ fade_out;
          end;
        end
        else Dec( active_cnt);
      end; // end down sweep
    end;

    procedure   TModPitchChange.SampleRateChanged; // override;
    begin
      Setup;
    end;

    function    TModPitchChange.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in       : Result := 'in';
        i_mute     : Result := 'mute';
        i_freeze   : Result := 'freeze';
        i_dry      : Result := 'dry';
        i_wet      : Result := 'wet';
        i_feedback : Result := 'feedback';
        i_rate     : Result := 'rate';
        i_depth    : Result := 'depth';
        i_delay    : Result := 'delay';
        i_freezein : Result := 'freezein';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModPitchChange.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEuclids = class( TMod)
  private
    FBits     : TScaleNoteSet;
    FSteps    : Integer;
    FPattern  : Integer;
    FOldTrig  : Boolean;
    FPosition : Integer;
  public
}

    constructor TModEuclideanRhythm.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModEuclideanRhythm.SetDefaults; // override;
    var
      aCount : Integer;
    begin
      FSteps    := 16;
      FPattern  := 7;
      FOldTrig  := False;
      FPosition := 0;
      FBits     := EuclideanPattern( FPattern, FSteps, aCount);
    end;

    procedure   TModEuclideanRhythm.DoSlowTick; // override;
    var
      aTrig    : Boolean;
      anOut    : Boolean;
      aSteps   : Integer;
      aPattern : Integer;
      aCount   : Integer;
    begin
      aTrig := SignalToLogic( FInputs[ i_trig]);
      if aTrig <> FOldTrig
      then begin
        FOldTrig := aTrig;
        if aTrig
        then begin
          FPosition := ( FPosition + 1) mod FSteps;
          anOut     := FPosition in FBits;
        end
        else anOut := False;
        FOutputs[ o_out] := LogicToSignal( anOut);
      end;
      aSteps   := Clip( Round( FInputs[ i_steps]), 1, 256);
      aPattern := Clip(( Round( aSteps * ( FInputs[ i_pattern] * FInputs[ i_patternmod] + FInputs[ i_pulses]))), 0, aSteps);
      if ( aSteps <> FSteps) or ( aPattern <> FPattern)
      then begin
        FBits    := EuclideanPattern( aPattern, aSteps, aCount);
        FSteps   := aSteps;
        FPattern := aPattern;
      end;
      FPosition := FPosition mod FSteps;
    end;

    procedure   TModEuclideanRhythm.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutputs[ o_out]        ], AppLocale));
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, LogicToSignal( FOldTrig)], AppLocale));
    end;

    function    TModEuclideanRhythm.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_trig       : Result := 'trig';
        i_pattern    : Result := 'pattern';
        i_patternmod : Result := 'patternmod';
        i_pulses     : Result := 'pulses';
        i_steps      : Result := 'steps';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModEuclideanRhythm.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEuclideanScale = class( TMod)
  private
    FBits       : TScaleNoteSet;
    FSteps      : Integer;
    FPattern    : Integer;
    FScaleNotes : array[ Byte] of Byte;
    FNoteCount  : Integer;
  private
}

    procedure   TModEuclideanScale.MakeScale;
    var
      aScale : TScaleNoteSet;
      i      : TScaleNote;
    begin
      aScale := TransPoseAndCompleteScale( FBits, Low( TScaleNote), High( TScaleNote), - 1, FSteps);

      FNoteCount := 0;
      for i := Low( i) to High( i)
      do begin
        if i in aScale
        then begin
          FScaleNotes[ FNoteCount] := i;
          Inc( FNoteCount);
        end;
      end;
    end;

//  public

    constructor TModEuclideanScale.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModEuclideanScale.SetDefaults; // override;
    begin
      FSteps   := 12;
      FPattern := 7;
      FBits    := EuclideanPattern( FPattern, FSteps, FCount);
      MakeScale;
    end;

    procedure   TModEuclideanScale.DoSlowTick; // override;
    const
      OGT = 1.0 / ( 12.0 * 12.0);
    var
      aSteps   : Integer;
      aPattern : Integer;
      aNoteNr  : Integer;
    begin
      aSteps   := Clip( Round( FInputs[ i_steps]), 1, 256);
      aPattern := Clip(( Round( aSteps * ( FInputs[ i_patternmod] * FInputs[ i_patternmodamt] + FInputs[ i_pattern]))), 0, aSteps);
      if ( aSteps <> FSteps) or ( aPattern <> FPattern)
      then begin
        FBits    := EuclideanPattern( aPattern, aSteps, FCount);
        FSteps   := aSteps;
        FPattern := aPattern;
        MakeScale;
      end;
      aNoteNr := Round( UnitsToNoteNumber( FCount * FSteps * FInputs[ i_in] * OGT));
      FOutputs[ o_out] := NoteNumberToUnits( 12 * FScaleNotes[ aNoteNr] / FSteps);
    end;

    function    TModEuclideanScale.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in            : Result := 'in';
        i_pattern       : Result := 'pattern';
        i_patternmod    : Result := 'patternmod';
        i_patternmodamt : Result := 'patternmodamt';
        i_steps         : Result := 'steps';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModEuclideanScale.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModXYScope = class( TMod)
  private
    FHistory   : Integer;
    FDataQueue : TSignalFifo;
  public
    property    History: Integer read FHistory write SetHistory;
  private
    procedure   SetHistory( aValue: Integer);
  public
}

    procedure   TModXYScope.SetHistory( aValue: Integer);
    var
      OldData : TSignalPairFifo;
      NewData : TSignalPairFifo;
    begin
      if ( aValue <> FHistory) and ( aValue > 0)
      then begin
        NewData := TSignalPairFifo.Create( aValue);
        NewData.CountOverflows := False;
        FHistory   := 0;
        OldData    := AtomicExchange( Pointer( FDataQueue), Pointer( NewData));
        FHistory   := aValue;
        FreeAndNil( OldData);
      end;
    end;

    procedure   TModXYScope.AddPoint( anX, anY: TSignal);
    var
      aPair : TSignalPair;
    begin
      aPair.Left  := anX;
      aPair.Right := anY;
      if Assigned( FDataQueue) and ( FHistory > 0)
      then FDataQueue.Append( aPair);
    end;

//  public

    constructor TModXYScope.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModXYScope.SetDefaults; // override;
    begin
      History := 500;
    end;

    procedure   TModXYScope.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'history')
      then History := Round( aValue)
      else inherited;
    end;

    function    TModXYScope.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'history')
      then Result := History
      else Result := inherited;
    end;

    procedure   TModXYScope.DoTick; // override;
    begin
      AddPoint( FInputs[ i_inx] * FInputs[ i_levx], FInputs[ i_iny] * FInputs[ i_levy]);
    end;

    procedure   TModXYScope.GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler); // override;
    begin
      if ( FHistory > 0) and Assigned( FDataQueue) { and ( FDataQueue.FillCount = FDataQueue.Size) }
      then aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FDataQueue);
    end;

    function    TModXYScope.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inx     : Result := 'inx';
        i_iny     : Result := 'iny';
        i_levx    : Result := 'levx';
        i_levy    : Result := 'levy';
        i_lockxy  : Result := 'lockxy';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModXYScope.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;

    function    TModXYScope.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_history : Result := 'history';
        else        Result := NO_INTERNAL;
      end;
    end;


{ ========
  TFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FCount        : Integer;
    FFTWInBuffer  : Pfftw_real;
    FFTWOutBuffer : Pfftw_complex;
    FFTWPlan      : fftw_plan;
  end;
  TFFTDataArray = array of TFFTData;
}

//  public

    procedure   TFFTData.Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    var
      i : Integer;
    begin
      FWindow       := aWindow;
      FBinCount     := aBinCount;
      FIndex        := aStartIndex;
      FFTWInBuffer  := fftw_malloc( aBinCount * SizeOf( Tfftw_real   ));
      FFTWOutBuffer := fftw_malloc( aBinCount * SizeOf( Tfftw_complex));
      FFTWPlan      := fftw_plan_dft_r2c_1d( aBinCount, FFTWInBuffer, FFTWOutBuffer, FFTW_MEASURE);
      for i := 0 to aBinCount - 1
      do begin
        Pfftw_complex_array( FFTWOutBuffer)^[ i].re.v := 0;
        Pfftw_complex_array( FFTWOutBuffer)^[ i].im.v := 0;
        Pfftw_real_array   ( FFTWInBuffer )^[ i].v    := 0;
      end;
    end;

    procedure   TFFTData.Finalize;
    begin
      if Assigned( FFTWInBuffer)
      then fftw_free( FFTWInBuffer);
      if Assigned( FFTWOutBuffer)
      then fftw_free( FFTWOutBuffer);
      if Assigned( FFTWPlan)
      then fftw_destroy_plan( FFTWPlan);
    end;

    procedure   TFFTData.Tick( anInValue: TSignal; var anOutValue: TComplex);
    begin
      if FIndex >= FBinCount
      then begin
        FIndex := 0;
        fftw_execute( FFTWPlan);
      end;
      if FIndex >= 0
      then begin
        Pfftw_real_array( FFTWInBuffer)[ FIndex].v := anInValue * FWindow[ FIndex];
        anOutValue.Real := anOutValue.Real + Pfftw_complex_array( FFTWOutBuffer)^[ FIndex].re.v;
        anOutValue.Imag := anOutValue.Imag + Pfftw_complex_array( FFTWOutBuffer)^[ FIndex].im.v;
      end;
      Inc( FIndex);
    end;


{ ========
  TIFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FCount        : Integer;
    FFTWInBuffer  : Pfftw_complex;
    FFTWOutBuffer : Pfftw_real;
    FFTWPlan      : fftw_plan;
  end;
  TIFFTDataArray = array of TIFFTData;
}

//  public

    procedure   TIFFTData.Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    var
      i : Integer;
    begin
      FWindow       := aWindow;
      FBinCount     := aBinCount;
      FIndex        := aStartIndex;
      FFTWInBuffer  := fftw_malloc( aBinCount * SizeOf( Tfftw_complex));
      FFTWOutBuffer := fftw_malloc( aBinCount * SizeOf( Tfftw_real   ));
      FFTWPlan      := fftw_plan_dft_c2r_1d( aBinCount, FFTWInBuffer, FFTWOutBuffer, FFTW_MEASURE);
      for i := 0 to aBinCount - 1
      do begin
        Pfftw_complex_array( FFTWInBuffer )^[ i].re.v := 0;
        Pfftw_complex_array( FFTWInBuffer )^[ i].im.v := 0;
        Pfftw_real_array   ( FFTWOutBuffer)^[ i].v    := 0;
      end;
    end;

    procedure   TIFFTData.Finalize;
    begin
      if Assigned( FFTWInBuffer)
      then fftw_free( FFTWInBuffer);
      if Assigned( FFTWOutBuffer)
      then fftw_free( FFTWOutBuffer);
      if Assigned( FFTWPlan)
      then fftw_destroy_plan( FFTWPlan);
    end;

    procedure   TIFFTData.Tick( anInValue: TComplex; var anOutValue: TSignal);
    begin
      if FIndex >= FBinCount
      then begin
        FIndex := 0;
        fftw_execute( FFTWPlan);
      end;
      if FIndex >= 0
      then begin
        Pfftw_complex_array( FFTWInBuffer)[ FIndex].re.v := anInValue.Real;
        Pfftw_complex_array( FFTWInBuffer)[ FIndex].im.v := anInValue.Imag;
        anOutValue := anOutValue  + FWindow[ FIndex] * Pfftw_real_array( FFTWOutBuffer)^[ FIndex].v;
      end;
      Inc( FIndex);
    end;


{ ========
  TFFTHandler = class
  private
    FBinCount : Integer;
    FWindow   : TFFTWindow;
    FData     : TFFTDataArray;
  public
    property    WindowCount: Integer read GetWindowCount;
  private
}

    function    TFFTHandler.GetWindowCount: Integer;
    begin
      Result := Length( FData);
    end;

//  public

    constructor TFFTHandler.Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      i : Integer;
    begin
      Assert( aBinCount    > 0);
      Assert( aWindowCount > 0);
      inherited Create;
      FDestroying := False;
      FBinCount   := aBinCount;
      FWindow     := aWindowType.Create( aBinCount);
      SetLength( FData, aWindowCount);
      for i := 0 to aWindowCount - 1
      do FData[ i].Initialize( FBinCount, - (( i * FBinCount) div aWindowCount), FWindow);
    end;

    destructor  TFFTHandler.Destroy; // override;
    begin
      FDestroying := True;
      while WindowCount > 0
      do begin
        FData[ WindowCount - 1].Finalize;
        SetLength( FData, WindowCount - 1);
      end;
      FreeAndNil( FWindow);
      inherited;
    end;

    procedure   TFFTHandler.Tick( anInValue: TSignal; var anOutValue: TComplex);
    var
      i : Integer;
    begin
      anOutValue.Real := 0;
      anOutValue.Imag := 0;
      if not FDestroying
      then begin
        for i := 0 to WindowCount - 1
        do FData[ i].Tick( anInValue, anOutValue);
        anOutValue := anOutValue / WindowCount;
      end;
    end;


{ ========
  TIFFTHandler = class
  private
    FBinCount : Integer;
    FWindow   : TFFTWindow;
    FData     : TIFFTDataArray;
  public
    property    WindowCount: Integer read GetWindowCount;
  private
}

    function    TIFFTHandler.GetWindowCount: Integer;
    begin
      Result := Length( FData);
    end;

//  public

    constructor TIFFTHandler.Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      i : Integer;
    begin
      Assert( aBinCount    > 0);
      Assert( aWindowCount > 0);
      inherited Create;
      FDestroying := False;
      FBinCount   := aBinCount;
      FWindow     := aWindowType.Create( aBinCount);
      SetLength( FData, aWindowCount);
      for i := 0 to aWindowCount - 1
      do FData[ i].Initialize( FBinCount, ( i * FBinCount) div aWindowCount, FWindow);
    end;

    destructor  TIFFTHandler.Destroy; // override;
    begin
      FDestroying := True;
      while WindowCount > 0
      do begin
        FData[ WindowCount - 1].Finalize;
        SetLength( FData, WindowCount - 1);
      end;
      FreeAndNil( FWindow);
      inherited;
    end;

    procedure   TIFFTHandler.Tick( anInValue: TComplex; var anOutValue: TSignal);
    var
      i : Integer;
    begin
      anOutValue := 0;
      if not FDestroying
      then begin
        for i := 0 to WindowCount - 1
        do FData[ i].Tick( anInValue, anOutValue);
        anOutValue  := anOutValue  / WindowCount;
      end;
    end;


{ ========
  TModVocoder = class( TModBaseFFTIFFT)
  private
    FBinCount         : Integer;
    FWindowCount      : Integer;
    FModulatorHandler : TFFTHandler;
    FCarrierHandler   : TFFTHandler;
    FOutHandler       : TIFFTHandler;
  private
}

    procedure   TModVocoder.SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldModulatorHandler : TFFTHandler;
      NewModulatorHandler : TFFTHandler;
      OldCarrierHandler   : TFFTHandler;
      NewCarrierHandler   : TFFTHandler;
      OldOutHandler       : TIFFTHandler;
      NewOutHandler       : TIFFTHandler;
    begin
      try
        OldModulatorHandler := FModulatorHandler;
        if aBinCount > 0
        then NewModulatorHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewModulatorHandler := nil;
        FModulatorHandler := NewModulatorHandler;
        FreeAndNil( OldModulatorHandler);

        OldCarrierHandler := FCarrierHandler;
        if aBinCount > 0
        then NewCarrierHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewCarrierHandler := nil;
        FCarrierHandler := NewCarrierHandler;
        FreeAndNil( OldCarrierHandler);

        OldOutHandler := FOutHandler;
        if aBinCount > 0
        then NewOutHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewOutHandler := nil;
        FOutHandler := NewOutHandler;
        FreeAndNil( OldOutHandler);
      except
        on E: Exception
        do FftwWarning( 'Vocoder');
      end;
    end;

    procedure   TModVocoder.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandlers( FWindowCount, aValue, TFFTHammingWindow); // The default settings for Hamming actually produce a Hann window.
        FBinCount := aValue;
      end;
    end;

//  public

    constructor TModVocoder.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModVocoder.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute
      ];
      FWindowCount :=    4;
      BinCount     := 4096;
    end;

    procedure   TModVocoder.DoTick; // override;
    var
      aModulatorValue : TComplex;
      aCarrierValue   : TComplex;
      anOutValue      : TSignal;
    begin
      if Assigned( FModulatorHandler) and Assigned( FCarrierHandler) and Assigned( FOutHandler)
      then begin
        FModulatorHandler.Tick( FInputs[ i_modulator], aModulatorValue);
        FCarrierHandler  .Tick( FInputs[ i_carrier  ], aCarrierValue  );
        FOutHandler.Tick( aCarrierValue * aModulatorValue.Norm / BinCount, anOutValue);
        FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * anOutValue;
      end;
    end;

    function    TModVocoder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_modulator : Result := 'modulator';
        i_carrier   : Result := 'carrier';
        i_mute      : Result := 'mute';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModVocoder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModConvoder = class( TModBaseFFTIFFT)
  private
    FSize             : Integer;
    FBinCount         : Integer;
    FWindowCount      : Integer;
    FAttenuation      : TSignal;
    FModulatorHandler : TFFTHandler;
    FCarrierHandler   : TFFTHandler;
    FOutHandler       : TIFFTHandler;
  private
}

    procedure   TModConvoder.SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldModulatorHandler : TFFTHandler;
      NewModulatorHandler : TFFTHandler;
      OldCarrierHandler   : TFFTHandler;
      NewCarrierHandler   : TFFTHandler;
      OldOutHandler       : TIFFTHandler;
      NewOutHandler       : TIFFTHandler;
    begin
      try
        OldModulatorHandler := FModulatorHandler;
        if aBinCount > 0
        then NewModulatorHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewModulatorHandler := nil;
        FModulatorHandler := NewModulatorHandler;
        FreeAndNil( OldModulatorHandler);

        OldCarrierHandler := FCarrierHandler;
        if aBinCount > 0
        then NewCarrierHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewCarrierHandler := nil;
        FCarrierHandler := NewCarrierHandler;
        FreeAndNil( OldCarrierHandler);

        OldOutHandler := FOutHandler;
        if aBinCount > 0
        then NewOutHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewOutHandler := nil;
        FOutHandler := NewOutHandler;
        FreeAndNil( OldOutHandler);
      except
        on E: Exception
        do FftwWarning( 'Convoder');
      end;
    end;

    procedure   TModConvoder.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandlers( FWindowCount, aValue, TFFTTriangularWindow);
     // SetNewHandlers( FWindowCount, aValue, TFFTHammingWindow);
        FBinCount    := aValue;
        FAttenuation := ( 1.0 * BinCount) * sqrt( 1.0 * BinCount);
      end;
    end;

//  public

    constructor TModConvoder.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModConvoder.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_mute
      ];
      FWindowCount :=   2;
      BinCount     := 512;
    end;

    procedure   TModConvoder.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'size')
      then begin
        FSize    := Round( Clip( aValue, Low( c_sizes), High( c_sizes)));
        BInCount := c_sizes[ FSize];
      end
      else inherited;
    end;

    function    TModConvoder.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'size')
      then Result := FSize
      else Result := inherited;
    end;

    procedure   TModConvoder.DoTick; // override;
    var
      aModulatorValue : TComplex;
      aCarrierValue   : TComplex;
      anOutValue      : TSignal;
    begin
      if Assigned( FModulatorHandler) and Assigned( FCarrierHandler) and Assigned( FOutHandler) and ( BinCount > 0)
      then begin
        FModulatorHandler.Tick( FInputs[ i_modulatorlev] * FInputs[ i_modulator], aModulatorValue);
        FCarrierHandler  .Tick( FInputs[ i_carrierlev  ] * FInputs[ i_carrier  ], aCarrierValue  );
        FOutHandler.Tick( aCarrierValue * aModulatorValue / FAttenuation, anOutValue);
        FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * anOutValue;
      end;
    end;

    function    TModConvoder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_modulator    : Result := 'modulator';
        i_carrier      : Result := 'carrier';
        i_mute         : Result := 'mute';
        i_modulatorlev : Result := 'modulatorlev';
        i_carrierlev   : Result := 'carrierlev';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModConvoder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;

    function    TModConvoder.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        s_size : Result := 'size';
        else     Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModFilterBank = class( TMod)
  private
    FBinCount     : Integer;
    FWindowCount  : Integer;
    FInHandler    : TFFTHandler;
    FOutHandler   : TIFFTHandler;
    FCoefficients : TSignalArray;
    FCurrCoeff    : Integer;
    FFrequencies  : TSignalArray;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  private
}

    procedure   TModFilterBank.SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldInHandler  : TFFTHandler;
      NewInHandler  : TFFTHandler;
      OldOutHandler : TIFFTHandler;
      NewOutHandler : TIFFTHandler;
    begin
      try
        OldInHandler := FInHandler;
        if aBinCount > 0
        then NewInHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewInHandler := nil;
        FInHandler := NewInHandler;
        FreeAndNil( OldInHandler);

        OldOutHandler := FOutHandler;
        if aBinCount > 0
        then NewOutHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewOutHandler := nil;
        FOutHandler := NewOutHandler;
        FreeAndNil( OldOutHandler);

        SetLength( FCoefficients, aBinCount);
        FCurrCoeff := 0;
      except
        on E: Exception
        do FftwWarning( 'FilterBank');
      end;
    end;

    procedure   TModFilterBank.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandlers( FWindowCount, 2 * aValue, TFFTHammingWindow); // Hamming defaults to a Hann window, bin size twice as large for zero padding
        FBinCount := aValue;
      end;
    end;

//  public

    constructor TModFilterBank.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      Freq : TSignal;
      i    : Integer;
    begin
      inherited;
      SetLength( FFrequencies, 17);
      Freq := 40 / Sqrt( 2);               // about 28.3 Hz for the lowest frequency
      for i := 0 to 16
      do begin
        FFrequencies[ i] := Freq;
        Freq             := Freq * Sqrt( 2);
      end;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModFilterBank.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_fader1,
        i_fader2,
        i_fader3,
        i_fader4,
        i_fader5,
        i_fader6,
        i_fader7,
        i_fader8,
        i_fader9,
        i_fader10,
        i_fader11,
        i_fader12,
        i_fader13,
        i_fader14,
        i_fader15,
        i_fader16,
        i_fader17
      ];
      FWindowCount := 16;
      BinCount     := 1024;
    end;

    procedure   TModFilterBank.DoSlowTick; // override;
    // Need to smear out the faders over an array of values
    // to be used to multiply the FFT'd input values with to get an output value
    // Use rectamgular smear for now ... also .. the top half of the fft result
    // represents negative frequencies ... so ... the top half of the Coefficients
    // array shuld do that too .. the order is inverted tho ...
    var
      aHop  : TSignal;
      aFreq : TSignal;
      i     : Integer;
      p     : Integer;
    begin
      if BinCount > 0
      then begin
        aHop  := ( 0.5 * System_Rate) / BinCount;
        aFreq := FFrequencies[ 0];
        p     := 0;

        for i := 0 to 16
        do begin
          while ( aFreq <= FFrequencies[ i]) and ( p <= BinCount)
          do begin
            FCoefficients[ p] := FInputs[ i_fader1 + i];
            aFreq := aFreq + aHop;
            Inc( p);
          end;
        end;

        // Fill out possible remainder of first half (array size is BinCount * 2)
        while p <= BinCount shr 1
        do begin
          FCoefficients[ p] := FInputs[ i_fader17];
          Inc( p);
        end;

        // Set up 2nd half as a mirror image of the first half
        while p < BinCount
        do begin
          FCoefficients[ p] := FCoefficients[ BinCount - p - 1];
          Inc( p);
        end;

        // The second half is zero padded
        while p < BinCount
        do begin
          FCoefficients[ p] := 0;
          Inc( p);
        end;
      end;
    end;

    procedure   TModFilterBank.DoTick; // override;
    var
      anInValue  : TComplex;
      anOutValue : TSignal;
    begin
      if Assigned( FInHandler) and Assigned( FOutHandler)
      then begin
        if FCurrCoeff < BinCount
        then FInHandler.Tick( FInputs[ i_in], anInValue)
        else FInHandler.Tick( 0             , anInValue);

        anInValue.Real := anInValue.Real * FCoefficients[ FCurrCoeff];
        anInValue.Imag := anInValue.Imag * FCoefficients[ FCurrCoeff];
        FCurrCoeff     := ( FCurrCoeff + 1) mod ( 2 * BinCount);

        FOutHandler.Tick( anInValue, anOutValue);
        FOutputs[ o_out] := anOutValue;
      end;
    end;

    function    TModFilterBank.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in      : Result := 'in';
        i_fader1  : Result := 'fader1';
        i_fader2  : Result := 'fader2';
        i_fader3  : Result := 'fader3';
        i_fader4  : Result := 'fader4';
        i_fader5  : Result := 'fader5';
        i_fader6  : Result := 'fader6';
        i_fader7  : Result := 'fader7';
        i_fader8  : Result := 'fader8';
        i_fader9  : Result := 'fader9';
        i_fader10 : Result := 'fader10';
        i_fader11 : Result := 'fader11';
        i_fader12 : Result := 'fader12';
        i_fader13 : Result := 'fader13';
        i_fader14 : Result := 'fader14';
        i_fader15 : Result := 'fader15';
        i_fader16 : Result := 'fader16';
        i_fader17 : Result := 'fader17';
        else        Result := NO_INPUT;
      end;
    end;

    function    TModFilterBank.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModTalkie = class( TMod)
  private
    FTalkie     : TTalkie;
    FDestroying : Boolean;
    FStart      : Boolean;
    FOldBankSel : TSignal;
    FBank       : TSignal;
    FPhrase     : TSignal;
    FSampleRate : TSignal;
    FFrameRate  : TSignal;
    FSpeaking   : Boolean;
    FDidRun     : Boolean;
  private
}

    procedure   TModTalkie.TalkStarted( aSender: TObject);
    begin
      FSpeaking := True;
      FDidRun   := True;
    end;

    procedure   TModTalkie.TalkDone( aSender: TObject);
    begin
      FSpeaking := False;
    end;

    procedure   TModTalkie.Say;
    var
      aPhraseSets : TTalkiePhraseSets;
      aBankMax    : Integer;
      aPhraseMax  : Integer;
      TheBank     : Integer;
      ThePhrase   : Integer;
    begin
      if Assigned( FTalkie)
      then begin
        aPhraseSets := TalkiePhraseSets;
        if Assigned( aPhraseSets)
        then begin
          aBankMax   := aPhraseSets.Count - 1;
          TheBank    := Clip( Round( aBankMax * FBank), 0, aBankMax);
          aPhraseMax := aPhraseSets.PhraseSet[ TheBank].Count - 1;
          ThePhrase  := Clip( Round(  aPhraseMax * FPhrase), 0, aPhraseMax);
          FTalkie.Say( FFrameRate, FSampleRate, aPhraseSets.PhraseSet[ TheBank].Data[ ThePhrase]);
        end;
      end;
    end;

    procedure   TModTalkie.NewFramRate( aValue: TSignal);
    begin
      if ( aValue <> FFrameRate) and Assigned( FTalkie)
      then begin
        FFrameRate        := aValue;
        FTalkie.FrameRate := aValue;
      end;
    end;

    procedure   TModTalkie.NewSampleRate( aValue: TSignal);
    begin
      if ( aValue <> FSampleRate) and Assigned( FTalkie)
      then begin
        FSampleRate        := aValue;
        FTalkie.SampleRate := aValue;
      end;
    end;

//  public

    constructor TModTalkie.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FFrameRate  :=   40.0; // Frame  rate in frames/s, processing rate of speech frames
      FSampleRate := 8000.0; // Sample rate in samples/s for speech, Talkie will rate convert this to System_Rate
      FTalkie := TTalkie.Create;
      FTalkie.OnStarted := TalkStarted;
      FTalkie.OnDone    := TalkDone;
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModTalkie.Destroy; // override;
    var
      FOldTalkie : TTalkie;
    begin
      FDestroying := True;
      FOldTalkie  := FTalkie;
      FTalkie     := nil;
      FOldTalkie.Stop;
      FOldTalkie.OnStarted := nil;
      FOldTalkie.OnDone    := nil;
      FreeAndNil( FOldTalkie);
      inherited;
    end;

    procedure   TModTalkie.SetDefaults; // override;
    begin
      FStart  := False;
      FDidRun := False;
      FBank   := -1;
      FPhrase := -1;
      FInputs[ i_fm     ] := 0.0; // Set FM amount to neutral to make speed work ...
      FInputs[ i_banksel] := -1;  // Mark input as unused
    end;

    procedure   TModTalkie.DoTick; // override;
    var
      Start      : Boolean;
      BankSel    : TSignal;
      FrameRate  : TSignal;
      SampleRate : TSignal;
      FmFactor   : TSignal;
    begin
      if not FDestroying and Assigned( FTalkie)
      then begin
        Start      := SignalToLogic(    FInputs[ i_start  ]);
        BankSel    :=                   FInputs[ i_banksel];
        FBank      :=                   FInputs[ i_bank   ];
        FPhrase    :=                   FInputs[ i_phrase ];
        FmFactor   := LookupFrequency(  FInputs[ i_fm]) / LookupFrequency( 0);
        FrameRate  :=      80.0 * Clip( FInputs[ i_fr] + FInputs[ i_frmod] * FInputs[ i_frmodamt], 0, 1);
        SampleRate := Clip( 16000.0 * ( FInputs[ i_sr] + FInputs[ i_srmod] * FInputs[ i_srmodamt]) * FmFactor, 0, System_Rate shr 1);

        if ( BankSel >= 0) and ( BankSel <> FOldBankSel) then FBank := Clip( BankSel, 0, 1);
        if FrameRate  <> FFrameRate                      then NewFramRate  ( FrameRate );
        if SampleRate <> FSampleRate                     then NewSampleRate( SampleRate);

        if ( not FStart and Start)
        then Say;

        FStart      := Start;
        FOldBankSel := BankSel;

        FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]) * FTalkie.Tick;
        FOutputs[ o_rdy] := LogicToSignal( FDidRun and not FSpeaking);
      end
      else begin
        FSpeaking        := False;
        FStart           := False;
        FOutputs[ o_out] := SignalToMute( FInputs[ i_mute]);
        FOutputs[ o_rdy] := LogicToSignal( False);
      end;
      ResetFlag := False;
    end;

    procedure   TModTalkie.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/runs/%f', [ aPrefix, Name, LogicToSignal( FSpeaking)], AppLocale));
      aCallback( Self, Format( '%s/%s/rdy/%f' , [ aPrefix, Name, FOutputs[ o_rdy]         ], AppLocale));
    end;

    function    TModTalkie.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_start    : Result := 'start';
        i_phrase   : Result := 'phrase';
        i_mute     : Result := 'mute';
        i_fr       : Result := 'fr';
        i_frmod    : Result := 'frmod';
        i_frmodamt : Result := 'frmodamt';
        i_sr       : Result := 'sr';
        i_srmod    : Result := 'srmod';
        i_srmodamt : Result := 'srmodamt';
        i_bank     : Result := 'bank';
        i_banksel  : Result := 'banksel';
        i_fm       : Result := 'fm';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModTalkie.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out : Result := 'out';
        o_rdy : Result := 'rdy';
        else    Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAttractor = class( TMod)
  private
    FLorenz  : TLorenzOsc;
    FRossler : TRosslerOsc;
    FX       : TSignal;
    FY       : TSignal;
    FZ       : TSignal;
  public
}

    constructor TModAttractor.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast  := True;
      FIsSlow  := False;
      FLorenz  := TLorenzOsc.Create;
      FRossler := TRosslerOsc.Create;
    end;

    destructor  TModAttractor.Destroy; // override;
    begin
      FLorenz .DisposeOf;
      FRossler.DisposeOf;
      inherited;
    end;

    procedure   TModAttractor.SetDefaults; // override;
    begin
      FDezipperMap := [
        i_frequency,
        i_fm       ,
        i_fmlevel  ,
        i_cents    ,
        i_mute
      ];
    end;

    procedure   TModAttractor.DoTick; // override;
    var
      Mode      : Integer;
      Mute      : TSignal;
      Frequency : TSignal;
    begin
      if Assigned( FLorenz) and Assigned( FRossler)
      then begin
        Mode      := Round( FInputs[ i_mode      ]);
        Mute      := SignalToMute( FInputs[ i_mute      ]);
        Frequency := LookupFrequency( Clip( FInputs[ i_frequency] + FInputs[ i_freq] + FInputs[ i_fm] * FInputs[ i_fmlevel] + FInputs[ i_cents], 0.0, 1.0));

        case Mode of
          m_lorenz : begin FLorenz .SetFreq( Frequency); FLorenz .Tick( FX, FY, FZ); end;
          else       begin FRossler.SetFreq( Frequency); FRossler.Tick( FX, FY, FZ); end;
        end;

        FOutputs[ o_outx] := FX * Mute;
        FOutputs[ o_outy] := FY * Mute;
        FOutputs[ o_outz] := FZ * Mute;
      end;
    end;

    procedure   TModAttractor.SampleRateChanged; // override;
    begin
      if Assigned( FLorenz ) then FLorenz .SetSampleRate( System_Rate);
      if Assigned( FRossler) then FRossler.SetSampleRate( System_Rate);
    end;

    function    TModAttractor.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_frequency : Result := 'frequency';
        i_fm        : Result := 'fm';
        i_fmlevel   : Result := 'fmlevel';
        i_cents     : Result := 'cents';
        i_freq      : Result := 'freq';
        i_mute      : Result := 'mute';
        i_mode      : Result := 'mode';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModAttractor.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_outx : Result := 'outx';
        o_outy : Result := 'outy';
        o_outz : Result := 'outz';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNeuron = class( TMod)
  private
    FBrain      : TFannNetwork;
    FIns        : array of fann_type;
    FOuts       : array of fann_type;
    FFileName   : string;
    FDestroying : Boolean;
    FUpdating   : Boolean;
  private
}

    procedure   TModNeuron.SetFileName( const aValue: string);
    begin
      if aValue <> FFileName
      then begin
        FUpdating := True;
        FFileName := aValue;
        FBrain.LoadFromFile( FFileName);
        FUpdating := False;
      end;
    end;

//  public

    constructor TModNeuron.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FIns , i_in_last  - i_in_first  + 1);
      SetLength( FOuts, o_out_last - o_out_first + 1);
      FBrain  := TFannNetwork.Create;
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModNeuron.Destroy; // override;
    begin
      FDestroying := True;
      FreeAndNil( FBrain);
      inherited;
    end;

    procedure   TModNeuron.DoTick; // override;
    var
      i : Integer;
    begin
      if not FUpdating and not FDestroying and Assigned( FBrain) and FBrain.Built
      then begin
        for i := i_in_first to i_in_last
        do FIns[ i - i_in_first] := FInputs[ i] * FInputs[ i_inlevel];

        FBrain.Run( FIns, FOuts);

        for i := o_out_first to o_out_last
        do FOutputs[ i] := FOuts[ i - o_out_first] * Finputs[ i_outlevel];
      end;
    end;

    function    TModNeuron.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_inlevel  : Result := 'inlevel';
        i_outlevel : Result := 'outlevel';
        i_in1      : Result := 'in1';
        i_in2      : Result := 'in2';
        i_in3      : Result := 'in3';
        i_in4      : Result := 'in4';
        i_in5      : Result := 'in5';
        i_in6      : Result := 'in6';
        i_in7      : Result := 'in7';
        i_in8      : Result := 'in8';
        i_in9      : Result := 'in9';
        i_in10     : Result := 'in10';
        i_in11     : Result := 'in11';
        i_in12     : Result := 'in12';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModNeuron.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1  : Result := 'out1';
        o_out2  : Result := 'out2';
        o_out3  : Result := 'out3';
        o_out4  : Result := 'out4';
        o_out5  : Result := 'out5';
        o_out6  : Result := 'out6';
        o_out7  : Result := 'out7';
        o_out8  : Result := 'out8';
        o_out9  : Result := 'out9';
        o_out10 : Result := 'out10';
        o_out11 : Result := 'out11';
        o_out12 : Result := 'out12';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModInterpreter = class( TMod)
  private
    FForth    : TForth;
    FFileName : string;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

    procedure   TModInterpreter.LoadFile;
    var
      aForth: TForth;
    begin
      if Assigned( FForth) and FileExists( Filename)
      then begin
        aForth := FForth;
        FForth := nil;                     // Set FForth to nil to prevent real time execution during file loading
        aForth.LoadWords( FileName);
        FForth := aForth;                  // Re-enable real time stuff
      end;
    end;

    procedure   TModInterpreter.SetFileName( const aValue: string);
    begin
      if aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

    procedure   TModInterpreter.HandleOutData( const aSender: TObject; const aData: TExternalData);
    begin
      FOutputs[ o_out1] := aData[ 0];
      FOutputs[ o_out2] := aData[ 1];
      FOutputs[ o_out3] := aData[ 2];
      FOutputs[ o_out4] := aData[ 3];
      FOutputs[ o_out5] := aData[ 4];
      FOutputs[ o_out6] := aData[ 5];
      FOutputs[ o_out7] := aData[ 6];
      FOutputs[ o_out8] := aData[ 7];
    end;

    procedure   TModInterpreter.HandleForthLog( const aSender: TObject; const aMsg: string);
    begin
      LogFmt( 'script log: %s', [ aMsg]);
    end;

    procedure   TModInterpreter.HandleForthResponse( const aSender: TObject; const aMsg: string);
    begin
      LogFmt( 'script response: %s', [ aMsg]);
    end;

    procedure   TModInterpreter.HandleForthError( const aSender: TObject; const aMsg: string);
    begin
      ErrorFmt( 'script error: %s', [ aMsg]);
    end;

    procedure   TModInterpreter.SetCallbacks;
    begin
      if Assigned( FForth)
      then begin
        FForth.OnDataOut   := HandleOutData;        // Set a data_out handler for forth
        FForth.OnLog       := HandleForthLog;       // Set a log      handler for Forth
        FForth.OnResponse  := HandleForthResponse;  // Set a response handler for Forth
        FForth.OnError     := HandleForthError;     // Set a error    handler for Forth
      end;
    end;

    procedure   TModInterpreter.ClearCallbacks;
    begin
      if Assigned( FForth)
      then begin
        FForth.OnDataOut   := nil;
        FForth.OnLog       := nil;
        FForth.OnResponse  := nil;
        FForth.OnError     := nil;
      end;
    end;

//  public

    constructor TModInterpreter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;

      FForth := TForth.Create( nil);              // Create the interpreter
      SetCallbacks;                               // Install the forth callback handlers
      FForth.StartupFile := ForthFileName;        // Set a startup file for the Forth interpreter
      FForth.Reset;                               // Initialize the Forth system, this also loads and compiles the startup file
    end;

    destructor  TModInterpreter.Destroy; // override;
    begin
      ClearCallbacks;                             // Clear the callback handlers
      FreeAndNil( FForth);                        // Nils FForth too to prevent any real time execution
      inherited;
    end;

    procedure   TModInterpreter.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModInterpreter.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModInterpreter.DoSlowTick; // override;
    var
      aData : TExternalData;
    begin
      if Assigned( FForth)
      then begin
        aData[ 0] := FInputs[ i_in1];
        aData[ 1] := FInputs[ i_in2];
        aData[ 2] := FInputs[ i_in3];
        aData[ 3] := FInputs[ i_in4];
        aData[ 4] := FInputs[ i_in5];
        aData[ 5] := FInputs[ i_in6];
        aData[ 6] := FInputs[ i_in7];
        aData[ 7] := FInputs[ i_in8];
        FForth.AcceptDataIn( aData);
      end;
    end;

    function    TModInterpreter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1 : Result := 'in1';
        i_in2 : Result := 'in2';
        i_in3 : Result := 'in3';
        i_in4 : Result := 'in4';
        i_in5 : Result := 'in5';
        i_in6 : Result := 'in6';
        i_in7 : Result := 'in7';
        i_in8 : Result := 'in8';
        else    Result := NO_INPUT;
      end;
    end;

    function    TModInterpreter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        o_out5 : Result := 'out5';
        o_out6 : Result := 'out6';
        o_out7 : Result := 'out7';
        o_out8 : Result := 'out8';
        else     Result := NO_OUTPUT;
      end;
    end;


end.
